
#include "annot_proc.hpp"
#include "oper_graph.hpp"
#include "operation_ast.hpp"
#include "machine_ast.hpp"
#include "define_ast.hpp"
#include "manager_ast.hpp"
#include "regalloc_annot.hpp"
#include "osm_exp_ast.hpp"
#include <map>
#include <set>
#include <list>

using namespace mad_ast;
using namespace mad;
using std::vector;
using std::string;
using std::map;
using std::set;
using std::cerr;
using std::endl;
using std::pair;
using std::list;

Symbol SYMBOL_COMP;
Symbol SYMBOL_REG_BANK;
Symbol SYMBOL_SET_NAME;
Symbol SYMBOL_FORMAT;
Symbol SYMBOL_name;
Symbol SYMBOL_src;
Symbol SYMBOL_dst;
Symbol SYMBOL_symbols;
Symbol SYMBOL_data;
Symbol SYMBOL_REG_OVERLAPS;

Symbol SYMBOL_SCHED_RESOURCE;

AnnotType *TYPE_REG_BANK;
AnnotType *TYPE_REG_OVERLAPS;
AnnotType *TYPE_FORMAT;

RegallocSpec *proc_result;

static annot_context global_context;

/* all the registers. */
static map<Symbol, Register *> reg_table;
static map<Symbol, RegisterBank *> regbank_table;

/* all register overlaps. */
static vector<OverlapRegs *> regoverlap_list;
static vector<Oper *> oper_list;

extern const string compute_name(const vector<string>& names);

/* scheduler information */
static map<Symbol, int> resource_type_table;	//useless
static map<Symbol, Symbol> resource_inst_table; //manager name, alias name
static map<Symbol, int> sched_state_table;		//name, cycle
static set<Symbol> resource_alias_set;

/* initialize some variables. */
void extract_regalloc_init()
{
	SYMBOL_COMP     = symbol_find_or_insert("COMP");
	SYMBOL_SET_NAME = symbol_find_or_insert("SET_NAME");
	SYMBOL_FORMAT   = symbol_find_or_insert("FORMAT");
	SYMBOL_name     = symbol_find_or_insert("name");
	SYMBOL_src      = symbol_find_or_insert("src");
	SYMBOL_dst     = symbol_find_or_insert("dst");
	SYMBOL_symbols  = symbol_find_or_insert("symbols");
	SYMBOL_REG_BANK = symbol_find_or_insert("REG_BANK");
	SYMBOL_REG_OVERLAPS = symbol_find_or_insert("REG_OVERLAPS");
	SYMBOL_data     = symbol_find_or_insert("data");

	SYMBOL_SCHED_RESOURCE = symbol_find_or_insert("SCHED_RESOURCE");

	TYPE_REG_BANK = new annot_type_set(new annot_type_str(NULL), NULL);
	vector<annot_type *> *sub_overlaps = new vector<annot_type *>;
	sub_overlaps->push_back(new annot_type_str(NULL));
	sub_overlaps->push_back(new annot_type_str(NULL));
	TYPE_REG_OVERLAPS =
		new annot_type_set(new annot_type_tuple(sub_overlaps, NULL), NULL);

	TYPE_FORMAT = new annot_type_set(new annot_type_str(NULL), NULL);
}

Oper *extract_oper_constraints(vector<const OperationSection *>& opers,
	const annot_context& ctxt)
{
	vector<const OperationSection *>::const_iterator oper_it;
	vector<const annot_statement_relation *> constraints;

	/* find the FORMAT command first. */
	const annot_statement_command *format_cmd = NULL;
	for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++) {

		const vector<annot_paragraph *> *pannots =
			(*oper_it)->get_parsed_annotations();

		if (!pannots) continue;

		vector<annot_paragraph *>::const_iterator an_it;
		for (an_it=pannots->begin(); an_it!=pannots->end(); an_it++) {

			if (!(*an_it)->label_match(SYMBOL_COMP)) continue;

			vector<const annot_statement_relation *> rels =
				(*an_it)->get_relations();
			constraints.insert(constraints.end(), rels.begin(), rels.end());
			format_cmd = (*an_it)->get_command(SYMBOL_FORMAT);
			if (format_cmd) break; //find the format command
		}

		if (format_cmd) break;
	}

	if (format_cmd==NULL) {
		cerr << "Error: FORMAT not found for operation ";
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			cerr << symbol_get_string((*oper_it)->get_name()) << " ";
		cerr << endl;
		return NULL;
	}

	const annot_value *src = format_cmd->get_argument_value(SYMBOL_src);
	const annot_value *dst = format_cmd->get_argument_value(SYMBOL_dst);
	if (!src) {
		cerr << "Error: src not found in FORMAT, at "
			<< *format_cmd->get_loc() << endl;
		return NULL;
	}
	if (!dst) {
		cerr << "Error: dst not found in FORMAT, at "
			<< *format_cmd->get_loc() << endl;
		return NULL;
	}

	annot_context ct1 = ctxt + annot_context(*oper_it, SYMBOL_COMP);
	bool result = ct1.type_check();
	assert(result);

	annot_type *src_type = src->infer_type(&ct1);
	if (!src_type->compatible(TYPE_FORMAT)) {
		cerr << "Error: src in FORMAT of wrong type "
			<< *format_cmd->get_loc() << endl;
		return NULL;
	}

	annot_type *dst_type = dst->infer_type(&ct1);
	if (!dst_type->compatible(TYPE_FORMAT)) {
		cerr << "Error: dst in FORMAT of wrong type "
			<< *format_cmd->get_loc() << endl;
		return NULL;
	}

	delete src_type;
	delete dst_type;

	/* get all variables. */
	const annot_value_set *src_set = static_cast<const annot_value_set *>(src);
	const annot_value_set *dst_set = static_cast<const annot_value_set *>(dst);
	const vector<annot_value *> *src_vals = src_set->get_values();
	const vector<annot_value *> *dst_vals = dst_set->get_values();

	map<Symbol, Variable *> vars;
	vector<annot_value *>::const_iterator vit;
	if (src_vals) {
	for (vit=src_vals->begin(); vit!=src_vals->end(); vit++) {
		if ((*vit)->get_type()!=ANNOT_VAL_IDENT) {
			cerr << "Error: src in FORMAT of wrong type "
				<< *(*vit)->get_loc() << endl;
			return NULL;
		}
		else {
			annot_value_ident *vident = static_cast<annot_value_ident *>(*vit);
			if (vars.find(vident->get_ident())!=vars.end()) {
				cerr << "Error: variable appeared more than once in FORMAT "
					<< *vident->get_loc() << endl;
				return NULL;
			}
			else {
				Variable *new_var = new Variable;
				new_var->name = vident->get_ident();
				new_var->type = SRC_OPERAND;
				vars[vident->get_ident()] = new_var;	
			}
		}
	}
	}

	if (dst_vals) {
	for (vit=dst_vals->begin(); vit!=dst_vals->end(); vit++) {
		if ((*vit)->get_type()!=ANNOT_VAL_IDENT) {
			cerr << "Error: dst in FORMAT of wrong type "
				<< *(*vit)->get_loc() << endl;
			return NULL;
		}
		else {
			annot_value_ident *vident = static_cast<annot_value_ident *>(*vit);
			if (vars.find(vident->get_ident())!=vars.end()) {
				cerr << "Error: variable appeared more than once in FORMAT "
					<< *vident->get_loc() << endl;
				return NULL;
			}
			else {
				Variable *new_var = new Variable;
				new_var->name = vident->get_ident();
				new_var->type = DST_OPERAND;
				vars[vident->get_ident()] = new_var;	
			}
		}
	}
	}

	vector<Constraint *> *constraint_list = new vector<Constraint *>;
	/* now lets classify constraints */
	vector<const annot_statement_relation *>::iterator cst_it;
	for (cst_it=constraints.begin(); cst_it!=constraints.end(); cst_it++) {

		const annot_value *left = (*cst_it)->get_left();
		const annot_value *right = (*cst_it)->get_right();

		switch ((*cst_it)->get_operator()) {

			case ANNOT_REL_EQ:
				/* means assignment if constant on right*/
				if (right->get_type()==ANNOT_VAL_STR) {
					assert(left->get_type()==ANNOT_VAL_IDENT);
					const annot_value_ident *left_id =
						static_cast<const annot_value_ident *>(left);
					Symbol vname = left_id->get_ident();
					if (vars.find(vname)==vars.end()) {
						cerr << "Error : Unknown operand at "
							<< *left->get_loc() << endl;
						continue;
					}

					const annot_value_string *right_str =
						static_cast<const annot_value_string *>(right);
					Symbol rname =
						symbol_find_or_insert(right_str->get_value());
					if (reg_table.find(rname)==reg_table.end()) {
						cerr << "Error : Unknown register at "
							<< *right->get_loc() << endl;
					}
				
					RegassignConstraint *new_cst =
						new RegassignConstraint();	
					new_cst->var = vars[vname];
					new_cst->reg = reg_table[vname];
					Constraint *new_elem = new Constraint();
					new_elem->type = ASSIGN_CONSTR;
					new_elem->constr.rconstr = new_cst;
					constraint_list->push_back(new_elem);
				}
				else { /* otherwise equ */
					assert(left->get_type()==ANNOT_VAL_IDENT);
					const annot_value_ident *left_id =
						static_cast<const annot_value_ident *>(left);
					Symbol vname1 = left_id->get_ident();
					if (vars.find(vname1)==vars.end()) {
						cerr << "Error : Unknown operand at "
							<< *left->get_loc() << endl;
						continue;
					}

					assert(right->get_type()==ANNOT_VAL_IDENT);
					const annot_value_ident *right_id =
						static_cast<const annot_value_ident *>(right);
					Symbol vname2 = right_id->get_ident();
					if (vars.find(vname2)==vars.end()) {
						cerr << "Error : Unknown operand at "
							<< *right->get_loc() << endl;
						continue;
					}

					VarConstraint *r_cst = new VarConstraint();
					r_cst->var1 = vars[vname1];
					r_cst->var2 = vars[vname2];
					Constraint *new_elem = new Constraint();
					new_elem->type = VAR_CONSTR;
					new_elem->constr.vconstr = r_cst;
					constraint_list->push_back(new_elem);
				}
				break;

			case ANNOT_REL_IN:
				/* if left is identifier and right is an identifier-> bank */
				if (right->get_type()==ANNOT_VAL_IDENT &&
					left->get_type()==ANNOT_VAL_IDENT) {

					const annot_value_ident *left_id =
						static_cast<const annot_value_ident *>(left);
					Symbol vname1 = left_id->get_ident();
					if (vars.find(vname1)==vars.end()) {
						cerr << "Error : Unknown operand at "
							<< *left->get_loc() << endl;
						continue;
					}

					assert(right->get_type()==ANNOT_VAL_IDENT);
					const annot_value_ident *right_id =
						static_cast<const annot_value_ident *>(right);
					Symbol vname2 = right_id->get_ident();
					if (regbank_table.find(vname2)==regbank_table.end()) {
						cerr << "Error : Unknown bank "
							<< symbol_get_string(vname2) << " at " 
							<< *right->get_loc() << endl;
						continue;
					}

					BankConstraint *r_cst = new BankConstraint();
					r_cst->var = vars[vname1];
					r_cst->bank = regbank_table[vname2];
					Constraint *new_elem = new Constraint();
					new_elem->type = BANK_CONSTR;
					new_elem->constr.bconstr = r_cst;
					constraint_list->push_back(new_elem);
				}
				else { /* if left is tuple, then option constraint. */
					RegoptionsConstraint *r_cst = new RegoptionsConstraint();
					vector<Variable *> *left_vars = new vector<Variable *>;;

					assert(left->get_type()==ANNOT_VAL_TUPLE);
					const annot_value_tuple *left_tuple =
						static_cast<const annot_value_tuple *>(left);
					const vector<annot_value *> *lvals = left_tuple->get_values();
					vector<annot_value *>::const_iterator lv_it;
					for (lv_it=lvals->begin(); lv_it!=lvals->end(); lv_it++) {
						assert((*lv_it)->get_type()==ANNOT_VAL_IDENT);
						Symbol vname1 = static_cast<const annot_value_ident *>
								(*lv_it)->get_ident();
						if (vars.find(vname1)==vars.end()) {
							cerr << "Error : Unknown operand at "
							<< *left->get_loc() << endl;
						}
						else left_vars->push_back(vars[vname1]);
					}

					/* something is wrong. */
					if (left_vars->size()!=lvals->size()) continue;
					r_cst->vars = left_vars;

					right = right->infer_value(&ctxt);

					vector<RegisterList *> *opt = new vector<RegisterList *>;
					assert(right->get_type()==ANNOT_VAL_SET);
					const annot_value_set *right_set =
						static_cast<const annot_value_set *>(right);
					const vector<annot_value *> *rvals =
						right_set->get_values();

					bool is_fine = true;
					vector<annot_value *>::const_iterator rv_it;
					for (rv_it=rvals->begin(); rv_it!=rvals->end(); rv_it++) {
						assert((*rv_it)->get_type()==ANNOT_VAL_TUPLE);
						const annot_value_tuple *tup =
							static_cast<const annot_value_tuple *>(*rv_it);
						const vector<annot_value *> *vals = tup->get_values();

						vector<Register *> *reglist = new vector<Register *>;
						vector<annot_value *>::const_iterator val_it;
						for (val_it=vals->begin();
							val_it!=vals->end(); val_it++) {
							assert((*val_it)->get_type()==ANNOT_VAL_STR);
							const annot_value_string *val_str = 
								static_cast<annot_value_string *>(*val_it);
							Symbol asym = symbol_find_or_insert
								(val_str->get_value());
							if (reg_table.find(asym)==reg_table.end()) {
								cerr << "Error: Unknow register at "
									<< *val_str->get_loc() << endl;
								is_fine = false;
								break;
							}
							else reglist->push_back(reg_table[asym]);
						}
						if (!is_fine) break;
						opt->push_back(reglist);
					}
				
					if (!is_fine) continue;;
					r_cst->options = opt;
					Constraint *new_elem = new Constraint();
					new_elem->type = OPTIONS_CONSTR;
					new_elem->constr.aconstr = r_cst;
					constraint_list->push_back(new_elem);

				}
				break;
		}
	}
	Oper *noper = new Oper();
	noper->constraints = constraint_list;
	vector<Variable *> *vlist = new vector<Variable *>;
	map<Symbol, Variable *>::iterator var_it;
	for (var_it=vars.begin(); var_it!=vars.end(); var_it++) {
		vlist->push_back((*var_it).second);
	}
	noper->operands = vlist;
	return noper;
}

void extract_regbanks(const AnnotOwner *owner, Symbol bank_name)
{
	const vector<annot_paragraph *> *pannots =
			owner->get_parsed_annotations();

	if (!pannots) return;

	vector<annot_paragraph *>::const_iterator an_it;
	for (an_it=pannots->begin(); an_it!=pannots->end(); an_it++) {

		if (!(*an_it)->label_match(SYMBOL_COMP)) continue;

		const annot_statement_command *bank_cmd =
			(*an_it)->get_command(SYMBOL_REG_BANK);

		if (bank_cmd==NULL) continue;

		const annot_value *bank_val = 
				bank_cmd->get_argument_value(SYMBOL_symbols);

		if (bank_val==NULL) {
			cerr << "Error: no symbols found in REG_BANK" << endl;
			continue;
		}
		else if (!bank_val->is_constant(&global_context)) {
			cerr << "Error: REG_BANK symbols not constant" << endl;
			continue;
		}
		else if (!bank_val->is_wellformed(&global_context)) {
			cerr << "Error: REG_BANK symbols not wellformed" << endl;
			continue;
		}

		/* lets try to get the value */
		annot_type *real_type = bank_val->infer_type(&global_context);
		if (!real_type->compatible(TYPE_REG_BANK)) {
			cerr << "Error: REG_BANK symbols of wrong type" << endl;
			delete real_type;
			continue;
		}

		annot_value *real_val  = bank_val->infer_value(&global_context);
		annot_value_set *real_val_set =
			static_cast<annot_value_set *>(real_val);

		const vector<annot_value *> *vals = real_val_set->get_values();
		vector<annot_value *>::const_iterator val_it;
		vector<Register *> *reg_list = new vector<Register *>;
		for (val_it=vals->begin(); val_it!=vals->end(); val_it++) {

			assert((*val_it)->get_type()==ANNOT_VAL_STR);

			const annot_value_string *vstr =
				static_cast<const annot_value_string *>(*val_it);
			const string reg_name = vstr->get_value();

			Symbol sname = symbol_find_or_insert(reg_name.c_str());
			Register *new_reg = new Register();
			new_reg->name = sname;
			reg_table[sname] = new_reg;
			reg_list->push_back(new_reg);
		}
		bank_name = static_cast<const annot_value_ident *>(bank_val)->get_ident();
	cerr << "new bank " << symbol_get_string(bank_name) << endl;

		RegisterBank *new_bank = new RegisterBank();
		new_bank->name = bank_name;
		new_bank->regs = reg_list;

		regbank_table[bank_name] = new_bank;

		delete real_type;
		delete real_val;
	}

	for (an_it=pannots->begin(); an_it!=pannots->end(); an_it++) {

		if (!(*an_it)->label_match(SYMBOL_COMP)) continue;

		const annot_statement_command *overlap_cmd =
			(*an_it)->get_command(SYMBOL_REG_OVERLAPS);

		if (overlap_cmd==NULL) continue;

		const annot_value *overlap_val = 
				overlap_cmd->get_argument_value(SYMBOL_data);

		if (overlap_val==NULL) {
			cerr << "Error: no data found in REG_OVERLAP" << endl;
			continue;
		}
		else if (!overlap_val->is_constant(&global_context)) {
			cerr << "Error: REG_OVERLAP data not constant" << endl;
			continue;
		}
		else if (!overlap_val->is_wellformed(&global_context)) {
			cerr << "Error: REG_OVERLAP data not wellformed" << endl;
			continue;
		}

		/* lets try to get the value */
		annot_type *real_type = overlap_val->infer_type(&global_context);
		if (!real_type->compatible(TYPE_REG_OVERLAPS)) {
			cerr << "Error: REG_OVERLAP data symbols of wrong type" << endl;
			delete real_type;
			continue;
		}

		const annot_value_set *real_value_set =
			static_cast<const annot_value_set *>(overlap_val);

		/* now real_value_set contains a list of tuples. */
		const vector<annot_value *> *vals = real_value_set->get_values();
		vector<annot_value *>::const_iterator val_it;
		for (val_it=vals->begin(); val_it!=vals->end(); val_it++) {

			assert((*val_it)->get_type()==ANNOT_VAL_TUPLE);
			const annot_value_tuple *tuple_val = 
					static_cast<const annot_value_tuple *>(*val_it);
			const vector<annot_value *> *vals = tuple_val->get_values();
			assert(vals->size()==2);

			assert((*vals)[0]->get_type()==ANNOT_VAL_STR &&
					(*vals)[1]->get_type()==ANNOT_VAL_STR);

			const annot_value_string *str1 =
				static_cast<const annot_value_string *>((*vals)[0]);
			const annot_value_string *str2 =
				static_cast<const annot_value_string *>((*vals)[1]);

			string strv1 = str1->get_value();
			string strv2 = str2->get_value();

			Symbol sym1 = symbol_find_or_insert(strv1);
			Symbol sym2 = symbol_find_or_insert(strv2);

			if (reg_table.find(sym1)==reg_table.end()) {
				cerr << "Unknown register " << strv1 << endl;
				continue;
			}
			else if (reg_table.find(sym2)==reg_table.end()) {
				cerr << "Unknown register " << strv2 << endl;
				continue;
			}
			else {
				OverlapRegs *new_overlap = new OverlapRegs();
				new_overlap->reg1 = reg_table[sym1];
				new_overlap->reg2 = reg_table[sym2];
				regoverlap_list.push_back(new_overlap);
			}
		}
	}
}

template<class T>
void extract_annot_from_owners(const vector<const T *>& owners)
{
	typename vector<const T *>::const_iterator owner_it;
	for (owner_it=owners.begin(); owner_it!=owners.end(); owner_it++) {
		global_context.augment(*owner_it, SYMBOL_COMP);
	}
}

template<class T>
void extract_annot_from_owners(const vector<T *> *owners)
{
	if (owners) {
		typename vector<T *>::const_iterator owner_it;
		for (owner_it=owners->begin(); owner_it!=owners->end(); owner_it++) {
			global_context.augment(*owner_it, SYMBOL_COMP);
		}
	}
}

/* extract register information. */
bool extract_reginfo(Program *prog)
{
	/* first we form the global annotation context. */
	extract_annot_from_owners<ImportSection>(prog->get_import_sections());
	extract_annot_from_owners<UsingSection>(prog->get_using_sections());
	extract_annot_from_owners<AnnotationSection>(prog->get_annot_sections());

	vector<const define_section *> defs = prog->get_define_sections();
	vector<const define_section *>::const_iterator def_it;
	for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {

		const vector<define_clause *> *clss = (*def_it)->get_clauses();
		extract_annot_from_owners<define_clause>(clss);
	}

	vector<const manager_section *> mans = prog->get_manager_sections();
	vector<const manager_section *>::const_iterator man_it;
	for (man_it=mans.begin(); man_it!=mans.end(); man_it++) {

		const vector<manager_class_clause *> *classes =
			(*man_it)->get_classes();
		extract_annot_from_owners<manager_class_clause>(classes);

		const vector<manager_inst_clause *> *insts =
			(*man_it)->get_instances();
		extract_annot_from_owners<manager_inst_clause>(insts);
	}

	/* type check the golbal context */
	bool check_status = global_context.type_check();
	if (!check_status) return false;

	/* defines done. now get regbank info. */
	for (man_it=mans.begin(); man_it!=mans.end(); man_it++) {

		const vector<manager_inst_clause *> *insts = (*man_it)->get_instances();
		if (insts) {
			vector<manager_inst_clause *>::const_iterator cls_it;
			for (cls_it=insts->begin(); cls_it!=insts->end(); cls_it++) {
				extract_regbanks(*cls_it, (*cls_it)->get_name());
			}
		}
	}

	return check_status;
}

const annot_value *
	annot_get_command_arg_val(const AnnotOwner *owner, Symbol label,
		Symbol com_name, Symbol arg_name)
{
	const vector<AnnotParagraph *> *paras = owner->get_parsed_annotations();
	if (!paras) return NULL;

	vector<AnnotParagraph *>::const_iterator para_it;
	for (para_it=paras->begin(); para_it!=paras->end(); para_it++) {

		if ((*para_it)->label_match(label)) {

			/* get the command */
			const annot_statement_command *command =
				(*para_it)->get_command(com_name);

			if (command) return command->get_argument_value(arg_name);
		}
	}
	return NULL;
}

/** Get the name annotation of an operation_section. */
string annot_get_oper_name(const operation_section *oper)
{

	/* get the name argument value */
	const annot_value *name_val = annot_get_command_arg_val
		(oper, SYMBOL_COMP, SYMBOL_SET_NAME, SYMBOL_name);

	if (name_val) {
		if (name_val->get_type()!=ANNOT_VAL_STR) {
			cerr << "Find name argument, but of wrong type in "
				<< symbol_get_string(oper->get_name()) << endl;
			return "";
		}
		else {
			return
				static_cast<const annot_value_string *>(name_val)->get_value();
		}
	}

	return "";
}

void extract_schedresource(const AnnotOwner *owner, Symbol man_name)
{
	const vector<annot_paragraph *> *pannots =
			owner->get_parsed_annotations();

	if (!pannots) return;

	vector<annot_paragraph *>::const_iterator an_it;
	for (an_it=pannots->begin(); an_it!=pannots->end(); an_it++) {

		if (!(*an_it)->label_match(SYMBOL_COMP)) continue;

		vector<const annot_statement_command *> cmds = (*an_it)->get_commands();
		vector<const annot_statement_command *>::iterator cmd_it;
		for (cmd_it=cmds.begin(); cmd_it!=cmds.end(); cmd_it++) {

			Symbol cmd_name = (*cmd_it)->get_name();
			if (cmd_name!=SYMBOL_SCHED_RESOURCE) continue;

			const annot_value *bank_val = 
				(*cmd_it)->get_argument_value(SYMBOL_name);

			if (bank_val==NULL) {
				cerr << "Error: no name found in SCHED_RESOURCE" << endl;
				continue;
			}

			if (bank_val->get_type()==ANNOT_VAL_STR) {

				const annot_value_string *ival =
					static_cast<const annot_value_string *>(bank_val);
				string rname = ival->get_value();

				if (resource_inst_table.find(man_name)!=
					resource_inst_table.end()) {
					cerr << "Error: SCHED_RESOURCE redefined" << endl;
					continue;
				}
				resource_inst_table[man_name] = symbol_find_or_insert(rname);
				resource_alias_set.insert(symbol_find_or_insert(rname));
cerr << symbol_get_string(man_name) << " --> " << rname <<
	endl;
			}
			else {
				cerr << "Error: SCHED_RESOURCE name of wrong type" << endl;
			}
		}
	}
}

void extract_schedinfo(Program *prog)
{
	vector<const manager_section *> mans = prog->get_manager_sections();
	vector<const manager_section *>::const_iterator man_it;
	for (man_it=mans.begin(); man_it!=mans.end(); man_it++) {

		const vector<manager_inst_clause *> *insts = (*man_it)->get_instances();
		if (insts) {
			vector<manager_inst_clause *>::const_iterator cls_it;
			for (cls_it=insts->begin(); cls_it!=insts->end(); cls_it++) {
				extract_schedresource(*cls_it, (*cls_it)->get_name());
			}
		}
	}

	/* then we extract state info. we assume only one machine, 
	   and all states are good states.*/
	vector<const machine_section *> machs = prog->get_machine_sections();
	vector<const machine_section *>::const_iterator mach_it;
	for (mach_it=machs.begin(); mach_it!=machs.end(); mach_it++) {
		Symbol init = (*mach_it)->get_initial_state();
		const set<Symbol> *states = (*mach_it)->get_states();
		set<Symbol>::const_iterator sit;
		for (sit=states->begin(); sit!=states->end(); sit++) {
			if ((*sit)!=init) sched_state_table[*sit] = 1;
		}
	}

	const oper_graph *ograph = prog->get_decoding_graph();
	vector<const oper_graph_and_node *> roots = ograph->get_root_nodes();

	std::cerr << "num of roots " << (roots.size()) << std::endl;

	vector<const oper_tree *> exps;
	vector<const oper_graph_and_node *>::const_iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {
		vector<const oper_tree *> ts = (*rit)->get_expansions();
		exps.insert(exps.end(), ts.begin(), ts.end());
	}

	std::cerr << "num of expansions " << (exps.size()) << std::endl;
	vector<const oper_tree *>::iterator exp_it;
	for (exp_it=exps.begin(); exp_it!=exps.end(); exp_it++) {

		vector<string> names;
		vector<const operation_section *> opers = (*exp_it)->get_child_opers();
		vector<const operation_section *>::iterator oper_it;

		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			cerr << symbol_get_string((*oper_it)->get_name()) << " ";
		cerr << endl;

		// now we have all the edges and osm actions
		const machine_section *mach = opers.front()->get_machine();
		set<Symbol> active_table;	//active state
		map<Symbol, vector<osm_expression *>*> osm_exp_table;
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++) {

			const TransParagraph *tpara = (*oper_it)->get_trans();
			if (tpara==NULL) continue;
			const vector<TransClause *> *clss = tpara->get_clauses();
			vector<TransClause *>::const_iterator cls_it;
			for (cls_it=clss->begin(); cls_it!=clss->end(); cls_it++) {
				const vector<osm_expression *> *exps =
					(*cls_it)->get_osm_exps();
				if (exps) {
					Symbol edge_name = (*cls_it)->get_edge();
					if (osm_exp_table.find(edge_name)==osm_exp_table.end()) {
						vector<osm_expression *> *nexps =
							new vector<osm_expression *>;
						nexps->insert(nexps->end(), exps->begin(), exps->end());
						osm_exp_table[edge_name] = nexps;
					}
					else {
						vector<osm_expression *> *nexps =
							osm_exp_table[edge_name];
						nexps->insert(nexps->end(), exps->begin(), exps->end());
					}
					const edge_clause *edge_cls = mach->get_edge(edge_name);
					active_table.insert(edge_cls->get_to_state());
				}
			}
		}

		map<Symbol, set<Symbol> > resource_usage_table; //state->resources
		map<Symbol, int> latency_table; //state->time

		// get a state graph
		mad_dag *stgraph = const_cast<mad_dag *>(mach->get_state_graph());

		

		list<Symbol> sorted_states;
		stgraph->topological_sort(sorted_states);

		//assume that there is only one path
		int latency = 0;
		list<Symbol>::iterator state_it;
		for (state_it=sorted_states.begin();
			state_it!=sorted_states.end(); state_it++) {
			if (sched_state_table.find(*state_it)!=sched_state_table.end() &&
				active_table.find(*state_it)!=active_table.end()) {
				latency_table[*state_it] = latency++;
			}
		}

		/* first, find all allocations. */
		map<Symbol, Symbol> buf_table; //buffer->resource name

		map<Symbol, vector<osm_expression *>*>::iterator osm_it;
		for (osm_it=osm_exp_table.begin();
			osm_it!=osm_exp_table.end(); osm_it++) {

			Symbol edge_name = (*osm_it).first;
			const edge_clause *edge_cls = mach->get_edge(edge_name);
			Symbol to_state = edge_cls->get_to_state();
			if (latency_table.find(to_state)==latency_table.end()) continue;

			vector<osm_expression *> *exps = (*osm_it).second;
			vector<osm_expression *>::iterator exp_it;
			for (exp_it=exps->begin(); exp_it!=exps->end(); exp_it++) {

				if ((*exp_it)->get_type()==OSM_ASSIGN)  {

					osm_assign_expression *aexp =
						static_cast<osm_assign_expression *>(*exp_it);
					const osm_expression *rhs = aexp->get_rhs();
					const osm_expression *lhs = aexp->get_lhs();


					if (rhs->get_type()==OSM_MAN_REF) {
						const osm_man_ref_expression *mexp =
							static_cast<const osm_man_ref_expression *>(rhs);
						// get manager name
						Symbol res_name = mexp->get_name();

				cerr << "find allocate " << symbol_get_string(res_name) << endl;

						assert(lhs->get_type()==OSM_BUF_REF);
						const osm_buf_ref_expression *bexp =
							static_cast<const osm_buf_ref_expression *>(lhs);
						buf_table[bexp->get_name()] = res_name;
bexp->print(cerr);
cerr << "buf " << symbol_get_string(bexp->get_name()) << " holds " <<
	symbol_get_string(res_name) << endl;

						if (resource_inst_table.find(res_name)!=
							resource_inst_table.end()) {

							resource_usage_table[to_state].insert(res_name);
cerr << "add " << symbol_get_string(res_name) << " to " <<
	symbol_get_string(to_state) << endl;
							map<Symbol, int>::iterator state_it;
							for (state_it=latency_table.begin();
								state_it!=latency_table.end(); state_it++) {

								Symbol st_name = (*state_it).first;
								if (st_name==to_state) continue;
								if (stgraph->reachable(to_state,st_name)) {
cerr << "add " << symbol_get_string(res_name) << " to " <<
	symbol_get_string(st_name) << endl;
									resource_usage_table[st_name].insert(res_name);
								}
							}
						}

					}
				}
			}
		}

		/* then find or discard or releases */
		for (osm_it=osm_exp_table.begin();
			osm_it!=osm_exp_table.end(); osm_it++) {

			Symbol edge_name = (*osm_it).first;
			const edge_clause *edge_cls = mach->get_edge(edge_name);
			Symbol to_state = edge_cls->get_to_state();

			if (latency_table.find(to_state)==latency_table.end()) continue;

			vector<osm_expression *> *exps = (*osm_it).second;
			vector<osm_expression *>::iterator exp_it;
			for (exp_it=exps->begin(); exp_it!=exps->end(); exp_it++) {

				const osm_buf_ref_expression *buf;
				if ((*exp_it)->get_type()==OSM_DISCARD) {
				cerr << "find discard " << endl;
					buf = static_cast<osm_release_expression *>(*exp_it)
						->get_buf();
				}
				else if ((*exp_it)->get_type()==OSM_RELEASE) {
				cerr << "find release " << endl;
					buf = static_cast<osm_discard_expression *>(*exp_it)
						->get_buf();
				}
				else continue;
				// if this buffer holds resource, remove all
				if (buf_table.find(buf->get_name())!=buf_table.end()) {

					Symbol rname = buf_table[buf->get_name()];

					resource_usage_table[to_state].erase(rname);
cerr << "remove " << symbol_get_string(rname) << " from " <<
	symbol_get_string(to_state) << endl;
					map<Symbol, int>::iterator state_it;
					for (state_it=latency_table.begin();
						state_it!=latency_table.end(); state_it++) {

						Symbol st_name = (*state_it).first;
						if (st_name==to_state) continue;
						if (stgraph->reachable(to_state,st_name)) {
							resource_usage_table[st_name].erase(rname);
cerr << "remove " << symbol_get_string(st_name) << " from " <<
	symbol_get_string(to_state) << endl;
						}
					}
				}
			}
		}

		map<Symbol, set<Symbol> >::iterator ru_it;
		for (ru_it=resource_usage_table.begin();
			ru_it!=resource_usage_table.end(); ru_it++) {

			Symbol st_name = (*ru_it).first;
			set<Symbol> ress = (*ru_it).second;
			cerr << symbol_get_string(st_name) << " : ";
			set<Symbol>::iterator set_it;
			for (set_it=ress.begin(); set_it!=ress.end(); set_it++) {
				cerr << symbol_get_string(*set_it) << " ";
			}
			cerr << endl;
		}
	}
}

extern const string compute_name(const vector<string>& names);
void extract_regalloc_info(Program *prog)
{
	/* initialize global variables. */
	extract_regalloc_init();

	/* first we get all the defines in the program */
	if (!extract_reginfo(prog)) return;

	/* operations that have been typechecked */
	set<Symbol> checked_opers;
	set<Symbol> checked_mach;
	set<Symbol> bad_mach;

	const oper_graph *ograph = prog->get_decoding_graph();
	if (ograph==NULL) {
		std::cerr << "No decoding graph found" << endl;
		exit(1);
	}
	vector<const oper_graph_and_node *> roots = ograph->get_root_nodes();

	std::cerr << "num of roots " << (roots.size()) << std::endl;

	vector<const oper_tree *> exps;
	vector<const oper_graph_and_node *>::const_iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {
		vector<const oper_tree *> ts = (*rit)->get_expansions();
		exps.insert(exps.end(), ts.begin(), ts.end());
	}

	std::cerr << "num of expansions " << (exps.size()) << std::endl;
	vector<const oper_tree *>::iterator exp_it;
	for (exp_it=exps.begin(); exp_it!=exps.end(); exp_it++) {

		vector<string> names;
		vector<const operation_section *> opers = (*exp_it)->get_child_opers();

		const machine_section *mach = opers.front()->get_machine();
		annot_context mach_ctxt = global_context +
					annot_context(mach, SYMBOL_COMP);

		mach_ctxt.type_check();
		if (checked_mach.find(mach->get_name())==checked_mach.end()) {
			checked_mach.insert(mach->get_name());
			if (!mach_ctxt.type_check()) bad_mach.insert(mach->get_name());
		}
		if (bad_mach.find(mach->get_name())!=bad_mach.end())
			continue;

		bool oper_good = true;
		vector<const operation_section *>::iterator oper_it;
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++) {

			/* if not checked before. */
			if (checked_opers.find((*oper_it)->get_name())==
				checked_opers.end()) {

				cerr << "check operation "
					<< symbol_get_string((*oper_it)->get_name()) << endl;

				annot_context ct1 = mach_ctxt +
					annot_context(*oper_it, SYMBOL_COMP);

				if (!ct1.type_check()) oper_good = false;
				checked_opers.insert((*oper_it)->get_name());
			}

			string oname = annot_get_oper_name(*oper_it);
			if (oname.size()) names.push_back(oname);
		}

		string oper_name;
		if (names.size()) {
			oper_name = compute_name(names);
			std::cerr << "Find name " << oper_name << " for instruction ";
		}
		else {
			cerr << "Error: Instruction has no name " ;
			oper_good = false;
		}
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			cerr << symbol_get_string((*oper_it)->get_name()) << " ";
		cerr << endl;

		if (oper_good) {
			cerr << "extract format" << endl;

			/* lets get the format and the constraints */
			Oper *noper = extract_oper_constraints(opers, mach_ctxt);
			if (noper) {
				noper->name = symbol_find_or_insert(oper_name);
				oper_list.push_back(noper);
			}
		}
	}

	proc_result = new RegallocSpec();
	RegisterList *nreglist = new RegisterList();

	map<Symbol, Register *>::iterator reg_it;
	for (reg_it=reg_table.begin(); reg_it!=reg_table.end(); reg_it++) {
		nreglist->push_back((*reg_it).second);
	}

	RegisterBankList *nregbanklist = new RegisterBankList();
	map<Symbol, RegisterBank *>::iterator bank_it;
	for (bank_it=regbank_table.begin(); bank_it!=regbank_table.end(); bank_it++) {
		nregbanklist->push_back((*bank_it).second);
	}

	OverlapConstraints *noverlaplist = new OverlapConstraints;
	noverlaplist->insert(noverlaplist->end(), regoverlap_list.begin(),
			regoverlap_list.end());

	OperList *noperlist= new OperList;
	noperlist->insert(noperlist->end(), oper_list.begin(), oper_list.end());

	proc_result->reglist = nreglist;
	proc_result->regbanklist = nregbanklist;
	proc_result->overlaplist = noverlaplist;
	proc_result->operlist = noperlist;
	
}
