
#include "mad_dag.hpp"
#include <cassert>

using std::vector;
using std::map;
using std::set;
using std::list;
using std::ostream;
using std::endl;
using namespace mad;

bool mad_dag_vertex::check_cycle(Symbol& cyclic)
{
	if (flag) {
		cyclic = name;
		return true;
	}
	flag = true;

	vector<mad_dag_edge *>::iterator eit;
	for (eit=outgoing_edges.begin(); eit!=outgoing_edges.end(); eit++)
		if ((*eit)->get_dest_vertex()->check_cycle(cyclic)) return true;
		
	flag = false;
	return false;
}

void mad_dag_vertex::depth_first()
{
	if (flag) return;
	flag = true;
	vector<mad_dag_edge *>::iterator eit;
	for (eit=outgoing_edges.begin(); eit!=outgoing_edges.end(); eit++)
		((*eit)->get_dest_vertex()->depth_first());
}

void mad_dag_vertex::reverse_depth_first()
{
	if (flag) return;
	flag = true;
	vector<mad_dag_edge *>::iterator eit;
	for (eit=incoming_edges.begin(); eit!=incoming_edges.end(); eit++)
		((*eit)->get_src_vertex()->reverse_depth_first());
}

void mad_dag_vertex::topological_sort(list<Symbol>& result)
{
	if (flag) return;

	flag = true;
	vector<mad_dag_edge *>::iterator eit;
	for (eit=outgoing_edges.begin(); eit!=outgoing_edges.end(); eit++)
		((*eit)->get_dest_vertex()->topological_sort(result));

	result.push_front(name);
}

void mad_dag_vertex::calc_reachers()
{
	/* source node, do nothing. */
	if (incoming_edges.size()==0) return;

	set<const mad_dag_vertex *> temps;
	vector<mad_dag_edge *>::iterator eit = incoming_edges.begin();
	reachers = (*eit)->get_src_vertex()->reachers;

	for (; eit!=incoming_edges.end(); eit++) {
		set_union(reachers.begin(), reachers.end(),
			(*eit)->get_src_vertex()->reachers.begin(),
			(*eit)->get_src_vertex()->reachers.end(),
			inserter(temps, temps.begin()));
		reachers = temps;
	}

	reachers.insert(this);
}

void mad_dag_vertex::calc_dominator()
{
	/* source node, do nothing. */
	if (incoming_edges.size()==0) return;

	set<const mad_dag_vertex *> temps;
	vector<mad_dag_edge *>::iterator eit = incoming_edges.begin();
	dominators = (*eit)->get_src_vertex()->dominators;

	for (; eit!=incoming_edges.end(); eit++) {
		set_intersection(dominators.begin(), dominators.end(),
			(*eit)->get_src_vertex()->dominators.begin(),
			(*eit)->get_src_vertex()->dominators.end(),
			inserter(temps, temps.begin()));
		dominators = temps;
	}

	dominators.insert(this);
}


mad_dag::~mad_dag()
{
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++)
		delete (*sit).second;

	map<Symbol, mad_dag_edge *>::iterator eit;
	for (eit=edges.begin(); eit!=edges.end(); eit++)
		delete (*eit).second;
}

void mad_dag::insert_vertex(Symbol name) 
{
	// no vertex is inserted twice.
	assert(vertices.find(name)==vertices.end());

	mad_dag_vertex *nvertex = new mad_dag_vertex(name);
	vertices[name] = nvertex;

	// simply inserting a vertex does not affect reachability and
	// dominator analysis since the vertex is disconnected
}

void mad_dag::insert_edge(Symbol name, Symbol from, Symbol to)
{
	assert(vertices.find(from)!=vertices.end());
	assert(vertices.find(to)!=vertices.end());
	assert(edges.find(name)==edges.end());

	mad_dag_edge *nedge = new mad_dag_edge(name, vertices[from], vertices[to]);
	edges[name] = nedge;

	vertices[from]->insert_out_edge(nedge);
	vertices[to]->insert_in_edge(nedge);

	// invalidate previous analysis
	reachable_analyzed = false;
	dominator_analyzed = false;
}

bool mad_dag::cycle_check(Symbol& cyclic)
{
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++)
		(*sit).second->flag = false;

	for (sit=vertices.begin(); sit!=vertices.end(); sit++)
		if ((*sit).second->check_cycle(cyclic)) return true;
	
	return false;
}

//@@@ could utilize reachability analysis
bool mad_dag::all_reachable(Symbol from, Symbol& unreach)
{
	assert(vertices.find(from)!=vertices.end());

	if (!reachable_analyzed)
		analyze_reachability();

	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++) {
		if ((*sit).second->reachers.find(vertices[from]) ==
			(*sit).second->reachers.end()) {
			unreach = (*sit).first;
			return false;
		}
	}

	return true;
}

bool mad_dag::reverse_all_reachable(Symbol from, Symbol& unreach)
{
	assert(vertices.find(from)!=vertices.end());

	if (!reachable_analyzed)
		analyze_reachability();

	set<mad_dag_vertex *> all_vertices;
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++)
		all_vertices.insert((*sit).second);

	set<mad_dag_vertex *> unreachables;
	set_difference(all_vertices.begin(), all_vertices.end(),
			vertices[from]->reachers.begin(), 
			vertices[from]->reachers.end(), 
			inserter(unreachables, unreachables.begin()));

	if (unreachables.empty()) return true;

	unreach = (*unreachables.begin())->get_name();
	return false;
}

bool mad_dag::reachable(Symbol from, Symbol to)
{
	assert(vertices.find(from)!=vertices.end());
	assert(vertices.find(to)!=vertices.end());

	if (!reachable_analyzed)
		analyze_reachability();

	return vertices[to]->has_reacher(vertices[from]);
}

bool mad_dag::dominate(Symbol from, Symbol to)
{
	assert(vertices.find(from)!=vertices.end());
	assert(vertices.find(to)!=vertices.end());

	if (!dominator_analyzed)
		analyze_dominator();

	return vertices[to]->has_dominator(vertices[from]);
}

void mad_dag::topological_sort(list<Symbol> &sorted)
{
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++)
		(*sit).second->flag = false;

	for (sit=vertices.begin(); sit!=vertices.end(); sit++) {
		if ((*sit).second->incoming_edges.size()==0) {
			(*sit).second->topological_sort(sorted);
		}
	}
}

void mad_dag::analyze_reachability()
{
	/* first topological sort the vertices */
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++) {
		(*sit).second->reachers.clear();
	}

	list<Symbol> sorted_vertices;
	topological_sort(sorted_vertices);

	list<Symbol>::iterator it;
	for (it=sorted_vertices.begin(); it!=sorted_vertices.end(); it++) {
		mad_dag_vertex *v = vertices[*it];
		if (v->incoming_edges.size()==0) {
			v->reachers.insert(v);
		}
	}

	for (it=sorted_vertices.begin(); it!=sorted_vertices.end(); it++)
		vertices[*it]->calc_reachers();

	reachable_analyzed = true;
}

void mad_dag::analyze_dominator()
{
	/* first topological sort the vertices */
	map<Symbol, mad_dag_vertex *>::iterator sit;
	for (sit=vertices.begin(); sit!=vertices.end(); sit++) {
		(*sit).second->dominators.clear();
	}

	list<Symbol> sorted_vertices;
	topological_sort(sorted_vertices);

	list<Symbol>::iterator it;
	for (it=sorted_vertices.begin(); it!=sorted_vertices.end(); it++) {
		mad_dag_vertex *v = vertices[*it];
		if (v->incoming_edges.size()==0) {
			v->dominators.insert(v);
		}
	}

	for (it=sorted_vertices.begin(); it!=sorted_vertices.end(); it++)
		vertices[*it]->calc_dominator();

	dominator_analyzed = true;
}

void mad_dag::print(ostream& os) const
{
	os << "Graph Begin:" << endl;
	map<Symbol, mad_dag_edge *>::const_iterator sit;
	for (sit=edges.begin(); sit!=edges.end(); sit++) {
		os  << symbol_get_string((*sit).second->get_src_vertex()->get_name())
			<< " -> "
			<< symbol_get_string((*sit).second->get_dest_vertex()->get_name())
			<< endl;
	}
	os << "Graph End." << endl;
}

const mad_dag_vertex *mad_dag::get_vertex(Symbol name) const
{
	map<Symbol, mad_dag_vertex *>::const_iterator it = vertices.find(name);
	if (it==vertices.end()) return NULL;
	return (*it).second;
}

const mad_dag_edge *mad_dag::get_edge(Symbol name) const
{
	map<Symbol, mad_dag_edge *>::const_iterator it = edges.find(name);
	if (it==edges.end()) return NULL;
	return (*it).second;
}
