
#ifndef __MAD_DAG_HPP__
#define __MAD_DAG_HPP__

#include <vector>
#include <map>
#include <set>
#include <list>
#include "symbol.hpp"

using mad::Symbol;

namespace mad {

	class mad_dag_edge;

	/** Class for vertices in a directed acyclic graph. */
	typedef class mad_dag_vertex {

		public:
			/** Constructor.
			 *  @param name the name of the vertex.
			 */
			mad_dag_vertex(Symbol name) : name(name), flag(false) {}

			/** Get the name of the edge. */
			Symbol get_name() const {return name;}

			/** Insert an outgoing edge. */
			void insert_out_edge(mad_dag_edge *edge) {
				outgoing_edges.push_back(edge);
			}

			/** Insert an incoming edge. */
			void insert_in_edge(mad_dag_edge *edge) {
				incoming_edges.push_back(edge);
			}

			/** Get the outoing edges. */
			const std::vector<mad_dag_edge *> *get_out_edges() const {
				return &outgoing_edges;
			}

			/** Get the outoing edges. */
			const std::vector<mad_dag_edge *> *get_in_edges() const {
				return &incoming_edges;
			}

		private:
			std::vector<mad_dag_edge *> outgoing_edges;
			std::vector<mad_dag_edge *> incoming_edges;

			Symbol name;

			bool flag;

			std::set<const mad_dag_vertex *> dominators;
			std::set<const mad_dag_vertex *> reachers;

			/** check cycle.
			 *  @param cyclic Return a symbol value if a cyclic node found.
			 *  @return True if cyclic.
			 */
			bool check_cycle(Symbol& cyclic);

			/** depth first traverse and setting flag. */
			void depth_first();

			/** reverse depth first traverse and setting flag. */
			void reverse_depth_first();

			/** Update reaching set. */
			void calc_reachers();

			/** Update dominatee set. */
			void calc_dominator();

			/** Topological sort */
			void topological_sort(std::list<Symbol> &sorted);

			// below two are not public since reachability analysis
			// and dominator analysis need to be triggered from within mad_dag
			/** Test if this is a reacher. */
			bool has_reacher(const mad_dag_vertex *st) const {
				return reachers.find(st)!=reachers.end();
			}

			/** Test if this is a dominator. */
			bool has_dominator(const mad_dag_vertex *st) const {
				return dominators.find(st)!=dominators.end();
			}

		friend class mad_dag;

	} MadDagVertex;

	/** Class for edges in a directed acyclic graph. */
	typedef class mad_dag_edge {

		public:
			/** Constructor.
			 *  @param name The name of the edge.
			 *  @param from The source vertex.
			 *  @param to   The destination vertex.
			 */
			mad_dag_edge(Symbol name,
				mad_dag_vertex *from, mad_dag_vertex *to) :
				from(from), to(to), name(name) {}

			/** Destructor. */
			~mad_dag_edge() {}

			/** Get the source of the edge. */
			mad_dag_vertex *get_src_vertex() const {return from;}

			/** Get the dest of the edge. */
			mad_dag_vertex *get_dest_vertex() const {return to;}

			/** Get the name of the edge. */
			Symbol get_name() const {return name;}

		private:
			mad_dag_vertex *from;
			mad_dag_vertex *to;
			Symbol name;

		friend class mad_dag;

	} MadDagEdge;

	/** Class for directed acyclic graph. */
	typedef class mad_dag {

		public:
			/** Constructor. */
			mad_dag() :
				reachable_analyzed(false), dominator_analyzed(false) {}

			/** Destructor. */
			~mad_dag();

			/** add one vertex to the graph.
			 *  @param name Name of the vertex.
			 */
			void insert_vertex(Symbol name);

			/** Add one edge to the graph.
			 *  @param name Name of the edge.
			 *  @param from The source vertex.
			 *  @param to   The destination vertex.
			 */
			void insert_edge(Symbol name, Symbol from, Symbol to);

			/** Check if the graph is cyclic.
			 *  @param  cyclic A vertex involved in the cycle, as a return val.
			 *  @return True if the graph is cyclic.
			 */
			bool cycle_check(Symbol& cyclic);

			/** Check if all vertex is reachable from a vertex.
			 *  @param from    The origin vertex.
			 *  @param unreach The first unreachable vertex, as a return val.
			 *  @return True if all vertices in the dag is reachable.
			 */
			bool all_reachable(Symbol from, Symbol& unreach);

			/** Check if all vertex is reachable from a vertex reversely.
			 *  @param from The origin vertex.
			 *  @param unreach The first unreachable vertex, as a return val.
			 *  @return True if all vertices in the dag is reachable.
			 */
			bool reverse_all_reachable(Symbol from, Symbol& unreach);

			/** Check if vertex b is reachable from a.
			 *  @param a The from vertex.
			 *  @param b The to vertex.
			 *  @return True if b is reachable from a.
			 */
			bool reachable(Symbol a, Symbol b);

			/** Check if vertex b is dominated by a.
			 *  @param a The from vertex.
			 *  @param b The to vertex.
			 *  @return True if b is dominated by a.
			 */
			bool dominate(Symbol a, Symbol b);

			/** Get the vertex node. 
			 *  @param name The vertex name.
			 */
			const mad_dag_vertex *get_vertex(Symbol name) const;

			/** Get the edge node. 
			 *  @param name The edge name.
			 */
			const mad_dag_edge *get_edge(Symbol name) const;

			/** Print the graph */
			void print(std::ostream& os) const;

			/** Topological sort. Result valid only for true DAG.
			 *  @param sorted The sorted result.
			 */
			void topological_sort(std::list<Symbol> &sorted);

		private:

			/** Perform reachability analysis.
			 *  The result is only good for DAG -- no cycle.
			 */
			void analyze_reachability();

			/** Perform dominator analysis.
			 *  The result is only good for DAG -- no cycle.
			 */
			void analyze_dominator();

			bool reachable_analyzed;
			bool dominator_analyzed;

			std::map<Symbol, mad_dag_vertex *> vertices;
			std::map<Symbol, mad_dag_edge  *> edges;

	} MadDag;
}

#endif
