
#ifndef __OSM_ISS_HPP__
#define __OSM_ISS_HPP__

#include "bittypes.h"
#include <string>
#include <cstring>

/* naming convention (to avoid naming clash with user defined variables)
 * function name:  prefix with double underscore,
 * variable name:  prefix and suffix with underscore,
 * class name:     prefix and suffix with underscore,
 * macro name:     prefix and suffix with underscore, all capital.
 */

/** Base class for the OSM machine.
 *  Derived class could define a enum of state,
 *  a static table of state names, as long as the activate and reset.
 *  Also buffers, variables, etc.
 */
class _iss_machine_ {

	public:

		/** Constructor.
		 *  @param sim    The simulator object.
		 *  @param id     ID number of the machine.
		 */  
		_iss_machine_(_SIMULATOR *sim, _MACH_ID_T id) :
			_sim_(sim), _id_(id) {}

		/** Destructor. */
		virtual ~_iss_machine_() {}

		/** Get the class id of the machine. */
		_MACH_ID_T __get_id() const {return _id_;}

#if 0
		/** Activate the state machine.
		 *  An OSM should remain if either the activation fails to
		 *  advance state or it advances to an non-initial state.
		 *  @return True if the OSM should remain in list (not retiring).
		 */
		virtual void __activate() = 0;
#endif

		/** The simulator object. */
		_SIMULATOR *__get_simulator() const {return _sim_;}

	public:

		_SIMULATOR *_sim_;

		_MACH_ID_T _id_; 	// class id of the machine
};

/** Token member names will not clash with identifiers, so
 *  the naming convention does not apply here.
 */

/** Token class. 
 *  A token represent a resource.
 *  _M_ -- manager type
 */
template <class _M_>
class _token_ {

	public:

		/** Constructor. */
		_token_(_M_ *m) : _tenant_(NULL), _boss_(m) {}

		/** Destructor. */
		~_token_() {}

		/** Test if a token is free.
		 *	@return True if the token is free.
		 */
		bool is_free() {return _tenant_==NULL;}

		/** Get the object holding the token. */
		_iss_machine_ *get_tenant() {return _tenant_;}

		/** Release the token */
		bool test_release(_iss_machine_ *obj) {
			return true;
		}

		/** Release the token */
		void release(_iss_machine_ *obj) {
			_boss_->reclaim_token(this, obj);
		}

		/** Discard the token */
		void discard(_iss_machine_ *obj) {
			return _boss_->discard_token(this, obj);
		}

		/** Get the manager */
		_M_ *get_boss() {return _boss_;}

		/** Set the token as used by an object. Used by Manager only.
		 *	@param obj The object to hold the token.
		 */
		void set_used(_iss_machine_ *obj) {
			_tenant_=obj;
		}

		/** Set an object as unused. Used by manager only. */
		void set_unused() {
			_tenant_=NULL;
		}

	private:

		_iss_machine_ *_tenant_;
		_M_ *_boss_;
};

#endif
