
#include "simulator.hpp"

#include "opers.hpp"
#include "machines.hpp"
#include <vector>
#include <list>
#include <iomanip>
#include <iostream>

#include <cstdio>
#include <csignal>
#include <cstring>
#include <cstdlib>
#include "config.h"
#include "loader.h"
#include "syscall.h"

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

using std::vector;

arm_simulator::arm_simulator(bool verbose) :
	running(false), debuging(false), verbose(verbose), in_fpe(false), src(NULL),
	mem()
{
	/* create the hardware modules */
	biu = new BIU("biu");
	icache = new ICache("icache", *biu);
	itlb = new ITLB("itlb", *biu);
	dcache = new DCache("dcache", *biu);
	dtlb = new DTLB("dtlb", *biu);

	/* Create the managers */
#define MANG_DEF(cname, iname) iname = new cname(#iname);
#include "mang_list.hpp"

	/* conenct the managers. */
	mIF->connect(mRF, icache, itlb, &mem);
	mMemCtrl->connect(dcache, dtlb, mMemRead, mMemWrite, &mem, mMemAddr, mRF);
	mNewPC->connect(mRF);
	
	/* storage for operation/machine */
	machine_pool = new obj_pool<_base_machine, MACH_ID_T, arm_simulator,
								arm_simulator::mach_allocator>(_mach_total);
	operation_pool = new obj_pool<_base_operation, OPER_ID_T, arm_simulator,
								arm_simulator::oper_allocator>(_oper_total);
	

	/* the scheduler */
	sched = new scheduler<arm_simulator>(this);

	/* reset every thing */
	reset();
}

arm_simulator::~arm_simulator()
{
	delete machine_pool;
	delete operation_pool;

#define MANG_DEF(cname, iname) delete iname;
#include "mang_list.hpp"

	delete sched;
}

void arm_simulator::reset()
{
	mem.reset();
	icache->reset();
	dcache->reset();
	itlb->reset();
	dtlb->reset();
	biu->reset();

#define MANG_DEF(cname, iname) iname->reset();
#include "mang_list.hpp"

	/* clear the value of cpsr */
	mCPSR->set_value(0);

	/* clear the value of the registers */
	for (int i=0; i<16; i++) {
		mRF->set_value(i, 0);
	}

	reset_stats();
	//sched->reset();
	//
	//brk_point = 0;
	//mmap_brk_point = MMAP_BASE;
}

void arm_simulator::reset_stats()
{
	cycle_count = 0;
	alloc_count = 0;
	retire_count = 0;
}

void arm_simulator::dump_stats(FILE *fp)
{
	icache->PrintStats(fp);
	itlb->PrintStats(fp);
	dcache->PrintStats(fp);
	dtlb->PrintStats(fp);
	biu->PrintStats(fp);

	fprintf(fp, "Total allocated %lld\n", alloc_count);
	fprintf(fp, "Total retired %lld\n", retire_count);
	fprintf(fp, "Total cycles %lld\n", cycle_count);
}

void arm_simulator::clock_tick()
{
	if (running) {

#ifdef DEBUG
		fprintf(stderr, "--clock %d--\n", cycle_count);
#endif
		sched->step();

		cycle_count++;

		mMemCtrl->update_on_clock();
		mNewPC->update_on_clock();
		mIF->update_on_clock();

		icache->updateOnClockTick();
		itlb->updateOnClockTick();
		dcache->updateOnClockTick();
		dtlb->updateOnClockTick();
		biu->updateOnClockTick();

		if (mRF->pc_modified()) {
			sched->flush(mRF->get_dynid());
#ifdef DEBUG
fprintf(stderr, "--flush %llx--", mRF->get_dynid());
#endif
			mRF->clear_pc_modified();
		}
	}
}

void arm_simulator::run(UInt64 maxCycle)
{
	running = true;

	
#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	struct timeval begin_u, end_u, begin_s, end_s;
	struct rusage usg;
	float user_time, sys_time;
	getrusage(RUSAGE_SELF, &usg);
	begin_u = usg.ru_utime;
	begin_s = usg.ru_stime;
#endif 

	running = true;
	while(running && cycle_count<=maxCycle) {
		clock_tick();
	}

	dump_stats(stderr);

#if HAVE_SYS_RESOURCE_H && HAVE_SYS_TIME_H
	getrusage(RUSAGE_SELF, &usg);
	end_u = usg.ru_utime;
	end_s = usg.ru_stime;									
	user_time = (end_u.tv_sec+end_u.tv_usec/1000000.0)-	
		(begin_u.tv_sec+begin_u.tv_usec/1000000.0);		
	sys_time = (end_s.tv_sec+end_s.tv_usec/1000000.0)-
		(begin_s.tv_sec+begin_s.tv_usec/1000000.0);
	fprintf(stderr,
		"Total user time  : %.3f sec.\n"					  
		"Total system time: %.3f sec.\n"					  
		"Simulation speed : %.3e cyc./sec.\n",			  
		user_time, sys_time, cycle_count/(user_time+sys_time));					   
#endif
}


_base_machine *arm_simulator::allocate_machine(MACH_ID_T id)
{
	_base_machine *mach = machine_pool->allocate(this,id);
	mach->set_dynid(alloc_count++);
	return mach;
}

_base_operation *arm_simulator::allocate_operation(OPER_ID_T id)
{
	return operation_pool->allocate(this,id);
}

void arm_simulator::sink_machine(_base_machine *mach, bool flush)
{
	vector<_base_operation *> *opers = mach->get_opers();
	vector<_base_operation *>::iterator it;
	for (it=opers->begin(); it!=opers->end(); it++)
		free_operation(*it);

	if (!flush) {
		retire_count++;
		mach->retire();
#if 0
		fprintf(stderr, "0x%08x : 0x%08x %lld\n",
			 static_cast<normal *>(mach)->pc.val(),
			 static_cast<normal *>(mach)->iw.val(), mach->get_dynid());
		fprintf(stderr, "%lld retire\n", mach->get_dynid());
#endif
	}
	else {
		mach->flush();
#if 0
		fprintf(stderr, "%lld flush\n", mach->get_dynid());
#endif
	}

	machine_pool->reclaim(mach);
}

_base_machine *arm_simulator::source_machine(bool test_drive)
{
	if (src==NULL) {
		_base_operation *oper = allocate_operation(_oper_any);
		src = allocate_machine(0);
		oper->init(src, NULL, NULL);
	}

	/* test drive */
	if (test_drive) {
		return src;
	}
	else {
		_base_machine *ret = src;
		src = NULL;
		return ret;
	}
}

_base_machine *arm_simulator::mach_allocator(arm_simulator *s, MACH_ID_T id)
{
  switch (id) {
#define MACH_DEF(_a, _b) \
	case _mach_ ## _b: return new MACH_CLASS(_b)(s);
#include "mach_list.hpp"
	default:
      assert(0);
  }
}

_base_operation *arm_simulator::oper_allocator(arm_simulator *s, OPER_ID_T id)
{
  switch (id) {
#define OPER_DEF(_a, _b) \
	case _oper_ ## _b: return new OPER_CLASS(_b)();
#include "oper_list.hpp"
    default:
      assert(0);
  }
}

