
#ifndef __BIT_TYPE_H__
#define __BIT_TYPE_H__

#include "bittypes.h"
#include <sstream>
#include <string>

#define WORD_T   uint32_t
#define DWORD_T  uint64_t
#define SWORD_T  int32_t
#define SDWORD_T int64_t

#define BIT_T(_a, _b) _bit_t<_a, _b>

#define BIT_CTR(_a, _b, _v) _bit_t<_a, _b>(_v)
#define STR_CTR(_a) std::string(_a)

#define BIT2BIT(_t1,_i1,_t2,_i2,_v) bits_cast<_t1,_i1,_t2,_i2>(_v)

#define BIT_XTR1(_a, _i, _t) \
	_bit_t<_t, 1>((_a).extract((unsigned)(_i)))

#define BIT_XTR2(_a, _i1, _i2, _t3, _i3) \
	_bit_t<_t3, _i3>((_a).extract((unsigned)(_i1), (unsigned)(_i2)))

#define BIT_DEP1(_a, _i, _v) \
	(_a).deposit((unsigned)(_i), _v.val())

#define BIT_DEP2(_a, _i1, _i2, _v) \
	(_a).deposit((unsigned)(_i1), (unsigned)(_i2), _v.val())

#define VAL_OF(_a) (_a).val()

/* bit class template, for Len<=64 */
template <class T, unsigned Len> class _bit_t {

	public:

		/* Constructor 1 */
		_bit_t() : value(0) {}

		/* Constructor 2 */
		_bit_t(const T& v) {
			value = (v<<(sizeof(T)*8-Len))>>(sizeof(T)*8-Len);
		}

		/* Assignment operator */
		_bit_t<T,Len>& operator = (const T& v) {
			value = v;
			return *this;
		}

		/* deposit values into field */
		void deposit(unsigned lmb, const T& v) {
			value = (value&~(1<<lmb)) | ((v&1)<<lmb);
		}

		/* deposit values into field */
		void deposit(unsigned lmb, unsigned rmb, const T& v) {
			T val1 = ((T)-1<<(lmb-rmb))<<1; // we shift twice here so as to 
											// avoid things like useless <<32.
			T val2 = ((T)-1)^val1;	// a mask of (lmb-rmb+1) bits
			T val3 = val2&v;
			value = (value&~(val2<<rmb)) | (val3<<rmb);
		}

		/* extract values from field */
		T extract(unsigned lmb) const {
			return (value>>lmb)&1;
		}

		/* extract values from field */
		T extract(unsigned lmb, unsigned rmb) const {
			T val1 = ((T)-1<<(lmb-rmb))<<1;
			T val2 = ((T)-1)^val1;	// a mask of (lmb-rmb+1) bits
			return (value>>rmb)&val2;
		}

		/* Get the value */
		T val() const {return value;}

		std::string hex() {
			std::stringstream sstr;
			sstr << std::hex << std::showbase << value;
			return sstr.str();
		}

		std::string dec() {
			std::stringstream sstr;
			sstr << std::dec << std::showbase << value;
			return sstr.str();
		}

		std::string oct() {
			std::stringstream sstr;
			sstr << std::oct << std::showbase << value;
			return sstr.str();
		}

		std::string bin() {

			char buf[Len+3];
			T v = value;
			int l=Len+3;

			buf[--l] = 0;
			while (v) {
				buf[--l] = (v&1)+'0';
				v = v>>1;
			}
			if (l==Len+2) buf[--l] = '0';
			buf[--l] = 'b';
			buf[--l] = '0';

			return std::string(buf+l);
		}

	private:
		T value;
};

template <class T, unsigned L> 
inline const _bit_t<T,L> operator + (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	return _bit_t<T,L> (x.val() + y.val());
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator - (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() - y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator * (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() * y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator / (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() / y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator | (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() | y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator % (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() % y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator & (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() & y.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<T,L> operator ^ (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>& y) {
	_bit_t<T,L> __result(x.val() ^ y.val());
	return __result;
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator - (const _bit_t<T,L>&x)
{
	_bit_t<T,L> __result(-x.val());
	return __result;
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator ~ (const _bit_t<T,L>&x)
{
	_bit_t<T,L> __result(~x.val());
	return __result;
}

template <class T, unsigned L> 
inline const _bit_t<unsigned, 1> operator == (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() == y.val();
}

template <class T, unsigned L> 
inline const _bit_t<unsigned, 1> operator != (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() != y.val();
}

template <class T, unsigned L> 
inline const _bit_t<unsigned, 1> operator <  (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() < y.val();
}

template <class T, unsigned L>
inline const _bit_t<unsigned, 1> operator >  (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() > y.val();
}

template <class T, unsigned L> 
inline const _bit_t<unsigned, 1> operator <=  (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() <= y.val();
}

template <class T, unsigned L>
inline const _bit_t<unsigned, 1> operator >=  (const _bit_t<T,L>& x,
											const _bit_t<T,L>& y) {
	return x.val() >= y.val();
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator >> (const _bit_t<T,L>& x,
				   					const _bit_t<T,L>&y) {
	_bit_t<T,L> __result(x.val() >> y.val());
	return __result;
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator >> (const _bit_t<T,L>& x, unsigned y) {
	_bit_t<T,L> __result(x.val() >> y);
	return __result;
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator << (const _bit_t<T,L>& x,
			   						const _bit_t<T,L>&y) {
	_bit_t<T,L> __result(x.val() << y.val());
	return __result;
}

template <class T, unsigned L>
inline const _bit_t<T,L> operator << (const _bit_t<T,L>& x, unsigned y) {
	_bit_t<T,L> __result(x.val() << y);
	return __result;
}

template <class T, unsigned L>
std::ostream& operator << (std::ostream& os, const _bit_t<T,L>& x)
{
	return os << x.val();
}

template <class T1, unsigned L1, class T2, unsigned L2>
inline const _bit_t<T2,L2> bits_cast(const _bit_t<T1,L1>& val)
{
	return _bit_t<T2,L2>((T1)(val.val()));
}

#endif
