
#ifndef __OSM_HPP__
#define __OSM_HPP__

#ifndef MAX_FUNC_ON_EDGE
#define MAX_FUNC_ON_EDGE 4
#define USE_DEFAULT_MOOE
#endif

#include "bittypes.h"
#include <vector>
#include <string>
#include <cstring>

/* forward declaration */
class _base_machine;

/** Base operation class. */
class _base_operation {

	public:

		/** Constructor. */
		_base_operation(OPER_ID_T id) : id(id) {};

		/** Destructor, do nothing. */
		virtual ~_base_operation() {}

		/** Initialize the operation.
		 *  Decode subopers, the EVAL section, register edges.
		 *  @param mach The machine associated with the operation.
		 *  @param dtab The decoding table.
		 *  @param cdng The coding.
		 */
		virtual void init(_base_machine *mach,
			OPER_ID_T *dtab, void *cdng) = 0;

		/** Get the class id of the operation. */
		OPER_ID_T get_id() const {return id;}

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual bool test_osm(unsigned int id) = 0;

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual void run_osm_pre(unsigned int id) = 0;

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual void run_osm_post(unsigned int id) = 0;

		/** Run the expressions associated on an edge.
		 *  @param id The edge's index.
		 */
		virtual void run_exp(unsigned int id) = 0;

		/** Get the assembly syntax of the operation. */
		virtual std::string get_syntax() {return "";}

		/** Use void * as an efficient polymorphic type. */
		virtual const void *get_coding() {return NULL;}

	protected:

		OPER_ID_T id;	//class id of the operation

};

/** Edge is a place holder for function pointers. */
class _edge {

	public:

		/** Constructor. */
		_edge(unsigned int id, unsigned int& t, unsigned int& e) :
			 id(id), ntran(t), neval(e) {reset();}

		/** Destructor. */
		~_edge() {}

		/** Reset the edge. */
		void reset() {
			ntran = 0;
			neval = 0;
		}

		/** Fire the transactions on the edge.
		 *  @return True if successful.
		 */
		bool fire() const {

			unsigned int i;

			/* empty edge, not to execute. */
			if (ntran+neval==0) return false;

			/* test the conditions. */
			for (i=0; i<ntran; i++) {
				if (!(tran_oper[i]->test_osm)(id)) return false;
			}
			/* run the OSM expressions that gets resources. */
			for (i=0; i<ntran; i++) (tran_oper[i]->run_osm_pre)(id);

			/* run the expressions. */
			for (i=0; i<neval; i++) (eval_oper[i]->run_exp)(id);

			/* run the OSM expressions that releases resources. */
			for (i=0; i<ntran; i++) (tran_oper[i]->run_osm_post)(id);

			return true;
		}

		/** Register a condition on an edge. */
		void register_cond(_base_operation *oper) {
#ifdef USE_DEFAULT_MOOE
			assert(ntran< MAX_FUNC_ON_EDGE);
#endif
			tran_oper[ntran] = oper;
			ntran++;
		}

		/** Register a function on an edge. */
		void register_eval(_base_operation *oper) {
#ifdef USE_DEFAULT_MOOE
			assert(ntran< MAX_FUNC_ON_EDGE);
#endif
			eval_oper[neval] = oper;
			neval++;
		}

	private:

		/* index of the edge */
		unsigned int id;

		/* the operation that test the condition */
		_base_operation *tran_oper[MAX_FUNC_ON_EDGE];

		/* the operations that evaluates */
		_base_operation *eval_oper[MAX_FUNC_ON_EDGE];

		/* number of transition, eval functions */
		unsigned int& ntran;
		unsigned int& neval;
};

/** Base class for the OSM machine.
 *  Derived class could define a enum of state,
 *  a static table of state names, as long as the activate and reset.
 *  Also buffers, variables, etc.
 */
class _base_machine {

	public:

		/** Constructor.
		 *  @param sim    The simulator object.
		 *  @param id     ID number of the machine.
		 *  @param n_edge Number of edges.
		 */  
		_base_machine(SIMULATOR *sim, MACH_ID_T id, unsigned int n_edge) :
			id(id), state(0), edges(), sim(sim) {

			opers = new std::vector<_base_operation *>;
			opers->reserve(4);

			counters = new unsigned int[2*n_edge];
			for (unsigned int i=0; i<n_edge; i++) 
				edges.push_back(new _edge(i,counters[2*i],counters[2*i+1]));
		}

		/** Destructor. */
		virtual ~_base_machine() {

			for (unsigned int i=0; i<edges.size(); i++)
				delete edges[i];

			delete opers;
			delete [] counters;
		}

		/** Get the class id of the machine. */
		MACH_ID_T get_id() const {return id;}

		/** Activate the state machine.
		 *  @return True if the current state is not initial.
		 */
		virtual bool activate() = 0;

		/** Reset the state machine.
		 *  Discarding all buffers, reset state, edges?
		 *  The derived class should call this at the end of its retire().
		 */
		virtual void retire() {
			opers->clear();
			memset(counters, 0, edges.size()*2*sizeof(unsigned int));
		}

		/** Reset the state machine.
		 *  Discarding all buffers, reset state, edges?
		 *  The derived class should call this at the end of its flush().
		 */
		virtual void flush() {
			opers->clear();
			memset(counters, 0, edges.size()*2*sizeof(unsigned int));
		}

		/** Get the edge by index. */
		_edge *get_edge(unsigned int id) {return edges[id];}

		/** Get the current state. */
		unsigned int get_state() const {return state;}

		/** Get the dynamic id of the operation */
		uint64_t get_dynid() const {return dynid;}

		/** Get the dynamic id of the operation */
		void set_dynid(uint64_t id) {dynid = id;}

		/** Get the operations associated with the machine. */
		std::vector<_base_operation *> *get_opers() {return opers;}

		/** Register an operation on the machine. */
		void register_oper(_base_operation *oper) {
			opers->push_back(oper);
		}

		/** The simulator object. */
		SIMULATOR *get_simulator() const {return sim;}

	protected:

		MACH_ID_T id; 	// class id of the operation

		unsigned int state;	// current state

		std::vector<_edge *> edges;	// edges

		std::vector<_base_operation *> *opers;	// associated operations

		unsigned int *counters;

		uint64_t dynid;

		SIMULATOR *sim;

};

/** Token class. 
 *  A token represent a resource.
 *  M -- manager type
 *  T -- token value type
 */
template <class M, class T>
class _token {

	public:

		/** Constructor. */
		_token(M *m) : tenant(NULL), boss(m) {}

		/** Destructor. */
		~_token() {}

		/** Test if a token is free.
		 *	@return True if the token is free.
		 */
		bool isFree() {return tenant==NULL;}

		/** Get the object holding the token. */
		_base_machine *get_tenant() {return tenant;}

		/** Release the token */
		bool test_release(_base_machine *obj) {
			return boss->token_good_to_reclaim(this, obj);
		}

		/** Release the token */
		void release(_base_machine *obj) {
			boss->reclaim_token(this, obj);
		}

		/** Discard the token */
		void discard(_base_machine *obj) {
			return boss->discard_token(this, obj);
		}

		/** Read the value of the token */
		T read(_base_machine *obj) {
			return boss->read_token(obj, this);
		}

		/** Write value to the token */
		void write(const T& v, _base_machine *obj) {
			return boss->write_token(v, obj, this);
		}

		/** Get the manager */
		M *get_boss() {return boss;}

		/** Set the token as used by an object. Used by Manager only.
		 *	@param obj The object to hold the token.
		 */
		void set_used(_base_machine *obj) {
			tenant=obj;
		}

		/** Set an object as unused. Used by manager only. */
		void set_unused() {
			tenant=NULL;
		}

	private:

		_base_machine *tenant;
		M *boss;
};

#endif
