
#ifndef __MORE_MANAGERS_HPP__
#define __MORE_MANAGERS_HPP__

/* include libosm stuff */
#include "include_osm.hpp"
#include "icache.h"
#include "itlb.h"
#include "dcache.h"
#include "dtlb.h"
#include "emumem.h"
#include <iostream>

/** short hand */
typedef class emulator::memory _MEM;

/** simple resource manager. */
typedef unit_manager<int> simple_resource;

/** The register file. */
class regfile_manager {

	public:

		/** index type. */
		typedef _UINT_T(4)  I;

		/** value type. */
		typedef _UINT_T(32) T;

		/** token type. */
		typedef class _token_<regfile_manager > TOKEN_T;

		/** Constructor.
		 *  @param name Name of the register file
		 */
		regfile_manager(const std::string& name) :
		   	name(name), or_temps(false), pc_flag(false) {

			temp_flags = new bool[16](false);
			tokens = new TOKEN_T[16](this);
			values = new T[16];
		}

		/** Destructor. */
		~regfile_manager() {
			delete [] temp_flags;
			delete [] tokens;
			delete [] values;
		}

		/** See if token is available for allocation.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_alloc(const I& ind, _base_machine_ *obj) {
			return tokens[ind.val()].is_free() 
				/*|| obj->__get_dynid() >
					tokens[ind.val()].get_tenant()->__get_dynid()*/;
		}

		/** See if token is available for inquire.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_inquire(const I& ind, _base_machine_ *obj) {
			return tokens[ind.val()].is_free() ||
				   obj==tokens[ind.val()].get_tenant();
		}

		/** See if it is ok to release token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _base_machine_ *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(const I& ind, _base_machine_ *obj) {
			tokens[ind.val()].set_used(obj);
			return tokens+ind.val();
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool reclaim_token(TOKEN_T *token, _base_machine_ *obj) {
			if (token->get_tenant()==obj)
				token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _base_machine_ *obj) {
			if (token->get_tenant()==obj)
				token->set_unused();
		}

		/** Allocate a token temporarily.
		 *  This is equivalent to an allocation followed by
		 *  a successful release one cycle later.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token_temp(const I& ind, _base_machine_ *obj) {
			temp_flags[ind.val()] = or_temps = true;
			tokens[ind.val()].set_used(obj);
			return tokens+ind.val();
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}


		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _base_machine_ *obj) {
			return values[ind.val()];
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_base_machine_ *obj, TOKEN_T *token) {
			return values[token-tokens];
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index of the token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _base_machine_ *obj) {

			/* Stop updating if the pc is modified. */
			if (pc_flag && dynid<obj->__get_dynid()) {
				return;
			}

			allocate_token_temp(ind, obj);	// prevent double write
			values[ind.val()] = val;

			/* monitor the pc */			
			if (ind.val()==15) {
				pc_flag = true;
				dynid = obj->__get_dynid();
			}
		}													 

		/** Update the value of the token.
		 * This function is called from the token. This corresponds
		 * to *token = val in OSM expression.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 * @param token The token to be updated.
		 */
		void write_token(const T& val,
			_base_machine_ *obj, TOKEN_T *token) {

#if 0
			std::cerr << "updating " << token-tokens << std::endl;
#endif
			/* Stop updating if the pc is modified. */
			if (pc_flag && dynid<obj->__get_dynid()) {
#if 0
				std::cerr << "abort writing" << std::endl;
#endif
				return;
			}

			values[token-tokens] = val;

			/* monitor the pc */
			if (token-tokens==15) {
				pc_flag = true;
				dynid = obj->__get_dynid();
			}
		}


		/** Get the value of the program counter. */
		T get_pc() const {
			return values[15];
		}

		/** Set the value of the program counter. */
		void set_pc(const T& val) {
			values[15] = val;
		}

		/** Set the value of the program counter and set the pc modify flag.
		 *  @param val The new pc value.
		 *  @param id  The dynid of the OSM performing the update
		 */
		void set_pc(const T& val, const uint64_t& id) {
			values[15] = val;
			dynid = id;
			pc_flag = true;
		}

		/** Get the value of a register.
		 *  This is used for value access from the hardware layer.
		 *  @param ind Index of the token.
		 */
		T get_value(int ind) {
			return values[ind];
		}

		/* Set the value of a register.
		 * This is used for value access from the hardware layer. So
		 * this does not monitor the value of pc.
		 * @param ind Index of the token.
		 * @param val Value to write.
		 */
		void set_value(int ind, const T& val) {
			values[ind] = val;
		}

		/** Check if pc is modifier. */
		bool pc_modified() const {return pc_flag;}

		/** Clear the pc_flag. */
		void clear_pc_modified() {pc_flag=false;}

		/** Free those temporarily allocated. */
		void update_on_clock() {
			if (or_temps) {
				for (int i=0; i<16; i++) {
					if (temp_flags[i])  {
						tokens[i].set_unused();
					   	temp_flags[i] = false;
					}
				}
				or_temps = false;
			}
		}

		/** Get the dynid (of the machine that updates pc). */
		uint64_t get_dynid() const {return dynid;}

		/** Resetting states. */
		void reset() {
			pc_flag = or_temps = false;
			for (int i=0; i<16; i++) {
				temp_flags[i] = false;
				tokens[i].set_unused();
				values[i] = 0;
			}
		}

	private:

		std::string name;

		bool or_temps;
		bool *temp_flags;
		TOKEN_T *tokens;
		T *values;

		bool pc_flag;
		uint64_t dynid;
};

/** fetching manager. */
class fetch_manager {

	public:

		/** The token type. */
		typedef class _token_<fetch_manager> TOKEN_T;

		/** The constructor.
		 *  @param name Name of the manager.
		 */
		fetch_manager(const std::string& name) :
			name(name), token(new TOKEN_T(this)), value(0),
	   		fetch_status(0), trigger(false) {}

		/** Destructor. */
		~fetch_manager() {
			delete token;
		}


		/** See if the token is available for allocation.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_base_machine_ *obj) {
			return token->is_free();
		}

		/** See if the token is available for inquire.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_base_machine_ *obj) {
			return token->is_free();
		}

		/** See if the token is available for release.
		 *  @param obj The requesting OSM.
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _base_machine_ *obj);

		/** Allocate a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(_base_machine_ *obj) {
			token->set_used(obj);
			fetch_status = 1;
			trigger = true;
			return token;
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool reclaim_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}

		/** Get the name of the manager. */
		const std::string& get_name() {return name;}

		/** Access the value of the token.
		 * This corresponds to val = *m[] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		_UINT_T(32) read_token(_base_machine_ *obj)
	   		{return value;}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		_UINT_T(32) read_token (_base_machine_ *obj, TOKEN_T *token)
	   		{return value;}

		/** Connect the fetcher with other modules.
		 *  @param mrf  The register file manager, for program counter.
		 *  @param csh  The instruction cache object.
		 *  @param tlb  The instruction tlb object.
		 *  @param imm  The memory to get instrution from.
		 */
		void connect(regfile_manager *mrf, ICache *csh, ITLB *tlb, _MEM *imm) {
		   	mRF = mrf;
			icache = csh;
			itlb = tlb;
			mem = imm;
		}

		/** Regular state update on clock edge. */
		void update_on_clock();

		/** Reset the states/ */
		void reset() {
			trigger = false;
			token->set_unused();
			fetch_status = 0;
		}

	private:

		std::string name;
		TOKEN_T *token;
		_UINT_T(32) value;

		unsigned current_pc;	/* the pc currently fetching */
		int fetch_status;		/* state of fetching. */
		int trigger;			/* start a new request. */

		/* connected components. */
		regfile_manager *mRF;
		ICache *icache;
		ITLB   *itlb;
		_MEM *mem;
};

typedef class value_manager<_UINT_T(32)> mem_addr_port;
typedef class value_manager<_UINT_T(32)> mem_read_port;
typedef class value_manager<_UINT_T(32)> mem_write_port;
typedef class value_manager<_UINT_T(4)> cpsr_manager;

/** The memory control port.
 *  This port does nothing but to synchronize data memory access.
 *  An OSM should inquire this port to see if the memory request finishes.
 */
class mem_ctrl_port {

	public:

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		mem_ctrl_port(const std::string& name) :
			name(name), count(0), status(0), mergeable(false), trigger(false) {}

		/** Destructor. */
		~mem_ctrl_port() {}

		/** See if token is availble for inquire.
		 *  Equivalent to if the memory access has finished.
		 *  @param obj The object performing the memory transaction.
		 */
		bool token_free_for_inquire(_base_machine_ *obj) {
			return status==0;
		}

		/** See if the token is available for allocate, dummy always true. */
		bool token_free_for_alloc(_base_machine_ *obj) {
			return status==0;
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const _UINT_T(32)& val, _base_machine_ *obj) {
			isRead = 1 - (val.val() & 1);
			size = (val.val()>>1)&7;
			count = val.val()>>4;
			trigger = true;
			dynid = obj->__get_dynid();
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Connect the module with other modules.
		 *  @param csh  The data cache object.
		 *  @param tlb  The data tlb object.
		 *  @param rpt  The memory read  port.
		 *  @param wpt  The memory write port.
		 *  @param dmm  The data memory containing real data.
		 *  @param map  The memory address port.
		 *  @param reg  The register file.
		 */
		void connect(DCache *csh, DTLB *tlb,
			mem_read_port *rpt, mem_write_port *wpt, _MEM *dmm,
			mem_addr_port *map, regfile_manager *reg) {
			dcache = csh;
			dtlb = tlb;
			mMemRead = rpt;
			mMemWrite = wpt;
			mem = dmm;
			mMemAddr = map;
			mRF = reg;
		}

		/** Regular update on clock edge. */
		void update_on_clock();

		/** reset states. */
		void reset() {
			status = 0;
		}

	private:

		std::string name;

		unsigned count;
		unsigned status;
		unsigned addr;
		unsigned size;
		bool mergeable;
		bool isRead;
		bool trigger;

		DCache *dcache;
		DTLB   *dtlb;
		mem_read_port *mMemRead;
		mem_write_port *mMemWrite;
		mem_addr_port *mMemAddr;
		regfile_manager *mRF;
		_MEM *mem;

		uint64_t dynid;//requesting obj id

		friend class mem_mode_port;
};

#if 0
/** The memory mode port. */
class mem_mode_port {

	public:

		/** Constructor.
		 *  @param name Name of the port.
		 */
		mem_mode_port(const std::string& name) :
			name(name), trigger(false), count(0) {}

		/** Destructor. */
		~mem_mode_port() {}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Connect the module with other modules.
		 *  @param mc The memory control manager.
		 *  @param ma The memory address manager.
		 */
		void connect(mem_ctrl_port *mc, mem_addr_port *ma, regfile_manager *r) {
			mMemCtrl = mc;
			mMemAddr = ma;
			mRF = r;
		}

		/** Regular update on clock edge. */
		void update_on_clock();

		/** Reset states. */
		void reset() {
			trigger = false;
			count = 0;
		}

	private:
		std::string name;
		unsigned isRead;
		unsigned size;
		unsigned count;
		bool trigger;	//true if access starts.

		mem_ctrl_port *mMemCtrl;
		mem_addr_port *mMemAddr;
		regfile_manager *mRF;
};

// @@@ consider merge mem_mode_port and mem_ctrl port @@@
#endif



/** The new program counter manager.
 *  Branch instruction write result to this.
 */
class new_pc {

	public:

		/** Token value type. */
		typedef _TUPLE_T(_UINT_T(32),_UINT_T(1)) TVAL;

		/** Token type. */
		typedef class _token_<new_pc> TOKEN_T;

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		new_pc(const std::string& name) :
			name(name), token(new TOKEN_T(this)) {
			token->set_unused();
		}

		/** Destructor. */
		~new_pc() {
			delete token;
		}

		/** See if the token is good for allocation, dummy, always true.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_base_machine_ *obj) {
			return true;
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const TVAL& val, _base_machine_ *obj) {

			/* second is the predicate bit */
			if (val._second_.val()>0) {
				npc = val._first_.val();
				dynid = obj->__get_dynid();
				trigger = true;
			}
		}

		/** Connect the fetcher with other modules.
		 *  @param mrf  The register file manager, for program counter.
		 */
		void connect(regfile_manager *mrf) {
		   	mRF = mrf;
		}

		/** Regular update on clock edge. */
		void update_on_clock();

		/** Reset states. */
		void reset() {
			trigger = false;
		}

	private:

		std::string name;
		TOKEN_T *token;
		bool trigger;
		_WORD_T npc;
		uint64_t dynid;

		regfile_manager *mRF;
};

#endif
