
#include "simulator.hpp"

#include "opers.hpp"
#include "machines.hpp"
#include <vector>
#include <iomanip>
#include <iostream>

#include <cstdio>
#include <csignal>
#include <cstring>
#include <cstdlib>
#include "config.h"
#include "loader.h"
#include "syscall.h"

using std::vector;
using std::list;

arm_simulator::arm_simulator(bool verbose) :
	running(false), debuging(false), verbose(verbose), in_fpe(false), src(NULL),
	mem()
{
	/* create the hardware modules */
	biu = new BIU("biu");
	icache = new ICache("icache", *biu);
	itlb = new ITLB("itlb", *biu);
	dcache = new DCache("dcache", *biu);
	dtlb = new DTLB("dtlb", *biu);

	/* Create the managers */
#define MANG_DEF(cname, iname) iname = new cname(#iname);
#include "mang_list.hpp"

	/* conenct the managers. */
	mIF->connect(mRF, icache, itlb, &mem);
	mMemCtrl->connect(dcache, dtlb, mMemRead, mMemWrite, &mem, mMemAddr, mRF);
	mNewPC->connect(mRF);
	
	/* storage for operation/machine */
	machine_pool = new obj_pool<_base_machine_, _MACH_ID_T, arm_simulator,
								arm_simulator::mach_allocator>(_mach_total);
	operation_pool = new obj_pool<_base_operation_, _OPER_ID_T, arm_simulator,
								arm_simulator::oper_allocator>(_oper_total);
	

	/* reset every thing */
	reset();
}

arm_simulator::~arm_simulator()
{
	delete machine_pool;
	delete operation_pool;

#define MANG_DEF(cname, iname) delete iname;
#include "mang_list.hpp"

}

void arm_simulator::reset()
{
	mem.reset();
	icache->reset();
	dcache->reset();
	itlb->reset();
	dtlb->reset();
	biu->reset();

#define MANG_DEF(cname, iname) iname->reset();
#include "mang_list.hpp"

	/* clear the value of cpsr */
	mCPSR->set_value(0);

	/* clear the value of the registers */
	for (int i=0; i<16; i++) {
		mRF->set_value(i, 0);
	}

	reset_stats();
	//
	//brk_point = 0;
	//mmap_brk_point = MMAP_BASE;
}

void arm_simulator::reset_stats()
{
	cycle_count = 0;
	alloc_count = 0;
	retire_count = 0;
}

void arm_simulator::dump_stats(FILE *fp)
{
	icache->PrintStats(fp);
	itlb->PrintStats(fp);
	dcache->PrintStats(fp);
	dtlb->PrintStats(fp);
	biu->PrintStats(fp);

	fprintf(fp, "Total allocated %lld\n", alloc_count);
	fprintf(fp, "Total retired %lld\n", retire_count);
	fprintf(fp, "Total cycles %lld\n", cycle_count);
}

void arm_simulator::flush(uint64_t ind)
{
	list<_base_machine_ *>::iterator it;
	for (it=work_list.begin(); it!=work_list.end();) {
		if ((*it)->__get_dynid()>ind && (*it)->__get_id()!=_mach_fetcher) {
			sink_machine(*it, true);
			it = work_list.erase(it);
		}
		else it++;
	}
}

void arm_simulator::clock_tick()
{
	if (running) {

#ifdef DEBUG
		fprintf(stderr, "--clock %d--\n", cycle_count);
#endif

		/** Activate all OSMs in order. */
		for (work_it=work_list.begin(); work_it!=work_list.end();) {

			ins_it = work_it;
			ins_it++;
#ifdef DEBUG
			int oldstate = (*work_it)->__get_state();
#endif
			bool ret = (*work_it)->__activate();
#ifdef DEBUG
			std::cerr << (*work_it)->__get_dynid() << " " << oldstate << "->";
			std::cerr << (*work_it)->__get_state() << " " << ret << std::endl;
#endif

			/* Test if the OSM is back to the initial state. */
			if (ret) { /* no */
				work_it++;
			}
			else { /* yes */
				/* send retired ones to the sink. */
				sink_machine(*work_it, false);
				work_it = work_list.erase(work_it);
			}
		}

#if 0
		/* get new OSMs from the source. */
		bool change = true;
		while (change) {
			change = false;
			_base_machine_ *src = source_machine(true);
			if (src->activate()) {
				work_list.push_back(src);
				source_machine(false);
			}
		}
#endif

		cycle_count++;

		mMemCtrl->update_on_clock();
		mNewPC->update_on_clock();
		mIF->update_on_clock();

		icache->updateOnClockTick();
		itlb->updateOnClockTick();
		dcache->updateOnClockTick();
		dtlb->updateOnClockTick();
		biu->updateOnClockTick();

		if (mRF->pc_modified()) {
			flush(mRF->get_dynid());
#ifdef DEBUG
fprintf(stderr, "--flush %llx--", mRF->get_dynid());
#endif
			mRF->clear_pc_modified();
		}
	}
}

void arm_simulator::run(UInt64 maxCycle)
{
	_base_machine_ *src = source_machine(false);
	work_list.push_back(src);

	running = true;
	while(running && cycle_count<=maxCycle) {
		clock_tick();
	}
}


void arm_simulator::activate_operation(_base_operation_ *oper,
	_OPER_ID_T *ids, void *cdg)
{
	_base_machine_ *mach = allocate_machine(oper->__get_mach_id());
	oper->__init(mach, ids, cdg);

	work_list.insert(ins_it, mach);
}

_base_machine_ *arm_simulator::allocate_machine(_MACH_ID_T id)
{
	_base_machine_ *mach = machine_pool->allocate(this,id);
	mach->__set_dynid(alloc_count++);
	return mach;
}

_base_operation_ *arm_simulator::allocate_operation(_OPER_ID_T id)
{
	return operation_pool->allocate(this,id);
}

void arm_simulator::sink_machine(_base_machine_ *mach, bool flush)
{
	vector<_base_operation_ *> *opers = mach->__get_opers();
	vector<_base_operation_ *>::iterator it;
	for (it=opers->begin(); it!=opers->end(); it++)
		free_operation(*it);

	if (!flush) {
		retire_count++;
		mach->__retire();
#if 0
		fprintf(stderr, "%lld retire\n", mach->get_dynid());
#endif
	}
	else {
		mach->__flush();
#if 0
		fprintf(stderr, "%lld flush\n", mach->get_dynid());
#endif
	}

	machine_pool->reclaim(mach);
}

_base_machine_ *arm_simulator::source_machine(bool test_drive)
{
	if (src==NULL) {
		_base_operation_ *oper = allocate_operation(_oper_fetch);
		src = allocate_machine(_mach_fetcher);
		oper->__init(src, NULL, NULL);
	}

	/* test drive */
	if (test_drive) {
		return src;
	}
	else {
		_base_machine_ *ret = src;
		src = NULL;
		return ret;
	}
}

_base_machine_ *arm_simulator::mach_allocator(arm_simulator *s, _MACH_ID_T id)
{
  switch (id) {
#define _MACH_DEF(_a, _b) \
	case _mach_ ## _b: return new _MACH_CLASS(_b)(s);
#include "mach_list.hpp"
	default:
      assert(0);
  }
}

_base_operation_ *arm_simulator::oper_allocator(arm_simulator *s, _OPER_ID_T id)
{
  switch (id) {
#define _OPER_DEF(_a, _b, _c) \
	case _oper_ ## _b: return new _OPER_CLASS(_b)();
#include "oper_list.hpp"
    default:
      assert(0);
  }
}

