
#ifndef __ARMSIMUL_H__
#define __ARMSIMUL_H__

#include <list>
#include "emumem.h"
#include "icache.h"
#include "itlb.h"
#include "dcache.h"
#include "dtlb.h"

#include "more_managers.hpp"
#include "object_pool.hpp"

class arm_simulator {

	enum _machine_id {
		_mach_dummy,
		#define _MACH_DEF(_a,_b) _mach_ ## _b,
		#include "mach_list.hpp"
		_mach_total,
	};

	enum _operation_id {
		_oper_dummy,
		#define _OPER_DEF(_a,_b,_c) _oper_ ## _b,
		#include "oper_list.hpp"
		_oper_total,
	};

	public:

		/** Constructor.
		 *  @param verbose Verbose mode
		 */
		arm_simulator(bool verbose);

		/** Destructor. */
		~arm_simulator();

		/** simulate for n cycles.
		 *  @param n The maximum number of cycles to simulate (may exit).
		 */
		void run(UInt64 n);

		/** simulate one cycle. */
		void clock_tick();

		/** stop running. 
		 *  Called by the signal handler or syscall interpreter.
		 */
		void stop() {running = false;}

		/** See if the simulator is running.
		 */
		bool is_running() const {return running;}

		/** Stop debuging, useless here.
		 *  Kept only to satisfy syscall.cpp.
		 */
		void stop_debug() {debuging = false;}

		/** See if in debugging mode, useless here.
		 *  Kept only to satisfy syscall.cpp.
		 */
		bool is_debugging() {return debuging;}

		/** stats output. */
		void dump_stats(FILE *fp);

		/** stats resetting. */
		void reset_stats();

		/* set brk point for syscall interpretation. */
		void set_brk(arm_addr_t addr) {brk_point = addr;}
		void set_mmap_brk(arm_addr_t addr) {mmap_brk_point = addr;}

		/** load the program into memory. */
		void load_program(const char *, int argc, char *argv[], char *env[]);

		/** load fpe emulation library into memory. */
		void load_fpe(const char *);

		/** reset */
		void reset();

		/** Get the value of a register. For debugging use. */
		word_t read_gpr(int ind) const {
			return mRF->get_value(ind).val();
		}

		/** These two hooks for the syscall interpretor. */
		void write_gpr(int ind, word_t val) {
			mRF->set_value(ind, val);
		}

		void write_gpr2(int ind, word_t val) {
			mRF->set_value(ind, val);
		}

		uint64_t get_cycle_count() const {return cycle_count;}
#if 0
		word_t read_cpsr() {return my_regs.cpsr;}
		word_t read_spsr() {return my_regs.spsr;}

		void write_cpsr(word_t val) {my_regs.cpsr=val;}
		void write_spsr(word_t val) {my_regs.spsr=val;}
#endif

	private:

		void flush(uint64_t);

		/* status flags */
		bool running;
		bool debuging;

		uint64_t cycle_count;	/* number of simulation cycles*/

		std::list<_base_machine_ *> work_list;
		std::list<_base_machine_ *>::iterator work_it, ins_it;

	public:
		/* syscall data */
		bool verbose;
		bool emulate_syscall;
		target_addr_t brk_point;
		target_addr_t mmap_brk_point;

		/* emulating fpe */
		bool in_fpe;


		/* memory management related members */

	public:

		uint64_t alloc_count;	/* number of machine allocations. */
		uint64_t retire_count;	/* number of machine retires. */

		/* Allocate a machine of given id type. */
		_base_machine_ *allocate_machine(_MACH_ID_T id);

		/* Allocate an operation of given id type. */
		_base_operation_ *allocate_operation(_OPER_ID_T id);

		/* Free a machine, flush is true if this is flushed.
		 * This is the sink of OSMs. */
		void sink_machine(_base_machine_ *mach, bool flush);

		/* Get a new machine.
		 * This is the source of OSMs.
		 * @param commit True if just test drive the source.
		 */
		_base_machine_ *source_machine(bool test_drive);

		/** Activate a new operation.
		 *  @param oper The activated operation.
		 *  @param ids The ids of the machine.
		 *  @param cdg The coding of the new operation.
		 */
		void activate_operation(_base_operation_ * oper,
				_OPER_ID_T *ids, void *cdg);

	private:

		/* Free an operation object. */
		void free_operation(_base_operation_ *oper) {
			operation_pool->reclaim(oper);
		}

		static _base_machine_ *mach_allocator(arm_simulator *, _MACH_ID_T id);
		static _base_operation_ *oper_allocator(arm_simulator *, _OPER_ID_T id);

		/* storage for machine objects. */
		obj_pool<_base_machine_, _MACH_ID_T, arm_simulator,
	   		arm_simulator::mach_allocator> *machine_pool;

		/* storage for operation objects. */
		obj_pool<_base_operation_, _OPER_ID_T, arm_simulator,
	  	 	arm_simulator::oper_allocator> *operation_pool;

		_base_machine_ *src;
		/* memory management related members end/ */

		/*hardware units */

	public:

		/* caches, TLBs & bus interface unit, modeling timing only */
		ICache *icache;
		DCache *dcache;
		ITLB *itlb;
		DTLB *dtlb;
		BIU *biu;

		/* the token managers */
#define MANG_DEF(cname, iname) cname *iname;
#include "mang_list.hpp"

	public:

		/* memory saving the real data value */
		emulator::memory mem;

};

#endif
