
#ifndef __OSM_HPP__
#define __OSM_HPP__

#ifndef _MAX_FUNC_ON_EDGE
#define _MAX_FUNC_ON_EDGE 4
#define _USE_DEFAULT_MODE
#endif

#include "bittypes.h"
#include <vector>
#include <string>
#include <cstring>

#if _MAX_FUNC_ON_EDGE < 256
#define _ECOUNT_TYPE uint8_t
#elif _MAX_FUNC_ON_EDGE < 65536
#define _ECOUNT_TYPE uint16_t
#else
#define _ECOUNT_TYPE uint32_t
#endif

/* naming convention (to avoid naming clash with user defined variables)
 * function name:  prefix with double underscore,
 * variable name:  prefix and suffix with underscore,
 * class name:     prefix and suffix with underscore,
 * macro name:     prefix and suffix with underscore, all capital.
 */

/* forward declaration */
class _base_machine_;

/** Base operation class. */
class _base_operation_ {

	public:

		/** Constructor. */
		_base_operation_(_OPER_ID_T _id, _MACH_ID_T _mid) :
			_id_(_id), _mid_(_mid) {};

		/** Destructor, do nothing. */
		virtual ~_base_operation_() {}

		/** Initialize the operation.
		 *  Decode subopers, the EVAL section, register edges.
		 *  @param mach The machine associated with the operation.
		 *  @param dtab The decoding table.
		 *  @param cdng The coding.
		 */
		virtual void __init(_base_machine_ *mach,
			_OPER_ID_T *dtab, void *cdng) = 0;

		/** Get the class id of the operation. */
		_OPER_ID_T __get_id() const {return _id_;}

		/** Get the class id of the machine. */
		_MACH_ID_T __get_mach_id() const {return _mid_;}

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual bool __test_osm(unsigned int id) = 0;

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual void __run_osm_pre(unsigned int id) = 0;

		/** Run the OSM expressions on an edge.
		 *  @param id  The edge's index.
		 *  @return True if the OSM expressions are successful.
		 */
		virtual void __run_osm_post(unsigned int id) = 0;

		/** Run the expressions associated on an edge.
		 *  @param id The edge's index.
		 */
		virtual void __run_exp(unsigned int id) = 0;

		/** Run everything associated on the edge.
		 *  This is the combination of above four.
		 *  @param id The edge's index.
		 */
		virtual bool __run_all(unsigned int id) = 0;

		/** Get the assembly syntax of the operation. */
		virtual std::string __get_syntax() {return "";}

		/** Use void * as an efficient polymorphic type. */
		virtual const void *__get_coding() {return NULL;}

	protected:

		_OPER_ID_T _id_;		//class id of the operation
		_MACH_ID_T _mid_;		//class id of the required machine

};

/** Edge is a place holder for function pointers. */
class _edge_ {

	public:

		/** Constructor. */
		_edge_(unsigned int id, _ECOUNT_TYPE & t, _ECOUNT_TYPE & e) :
			 _id_(id), _ntran_(t), _neval_(e) {
			 __reset();
		 }

		/** Destructor. */
		~_edge_() {}

		/** Reset the edge. */
		void __reset() {
			_ntran_ = 0;
			_neval_ = 0;
		}

		/** Fire the transactions on the edge.
		 *  @return True if successful.
		 */
		bool __fire() {

			unsigned int i;

			/* empty edge, not to execute. */
			if (_ntran_+_neval_==0) return false;

			if (_run_all_) {
				return _run_all_->__run_all(_id_);
			}

			/* test the conditions. */
			for (i=0; i<_ntran_; i++) {
				if (!(_tran_oper_[i]->__test_osm)(_id_)) return false;
			}
			/* run the OSM expressions that gets resources. */
			for (i=0; i<_ntran_; i++) (_tran_oper_[i]->__run_osm_pre)(_id_);

			/* run the expressions. */
			for (i=0; i<_neval_; i++) (_eval_oper_[i]->__run_exp)(_id_);

			/* run the OSM expressions that releases resources. */
			for (i=0; i<_ntran_; i++) (_tran_oper_[i]->__run_osm_post)(_id_);

			return true;
		}

		/** Register a condition on an edge. */
		void __register_cond(_base_operation_ *oper) {
#ifdef _USE_DEFAULT_MODE
			assert(_ntran_< _MAX_FUNC_ON_EDGE);
#endif
			_tran_oper_[_ntran_] = oper;
			_ntran_++;

			_run_all_ = NULL;
			if (_ntran_==1) {
				if (_neval_==0) 
					_run_all_ = oper;
				else if (_neval_==1 && _eval_oper_[0]==oper)
					_run_all_ = oper;
			}
		}

		/** Register a function on an edge. */
		void __register_eval(_base_operation_ *oper) {
#ifdef _USE_DEFAULT_MODE
			assert(_ntran_< _MAX_FUNC_ON_EDGE);
#endif
			_eval_oper_[_neval_] = oper;
			_neval_++;

			_run_all_ = false;
			if (_neval_==1) {
				if (_ntran_==0) 
					_run_all_ = oper;
				else if (_ntran_==1 || _tran_oper_[0]==oper)
					_run_all_ = oper;
			}
		}

	private:

		_base_operation_ *_run_all_;

		/* index of the edge */
		unsigned int _id_;

		/* the operation that test the condition */
		_base_operation_ *_tran_oper_[_MAX_FUNC_ON_EDGE];

		/* the operations that evaluates */
		_base_operation_ *_eval_oper_[_MAX_FUNC_ON_EDGE];

		/* number of transition, eval functions */
		_ECOUNT_TYPE& _ntran_;
		_ECOUNT_TYPE& _neval_;
};

/** Base class for the OSM machine.
 *  Derived class could define a enum of state,
 *  a static table of state names, as long as the activate and reset.
 *  Also buffers, variables, etc.
 */
class _base_machine_ {

	public:

		/** Constructor.
		 *  @param sim    The simulator object.
		 *  @param id     ID number of the machine.
		 *  @param n_edge Number of edges.
		 */  
		_base_machine_(_SIMULATOR *sim, _MACH_ID_T id, unsigned int n_edge) :
			_id_(id), _edges_(), _sim_(sim) {

			_opers_ = new std::vector<_base_operation_ *>;
			_opers_->reserve(4);

			_counters_ = new _ECOUNT_TYPE[2*n_edge];
			for (unsigned int i=0; i<n_edge; i++) 
				_edges_.push_back(new
						_edge_(i,_counters_[2*i],_counters_[2*i+1]));
		}

		/** Destructor. */
		virtual ~_base_machine_() {

			for (unsigned int i=0; i<_edges_.size(); i++)
				delete _edges_[i];

			delete _opers_;
			delete [] _counters_;
		}

		/** Get the class id of the machine. */
		_MACH_ID_T __get_id() const {return _id_;}

		/** Activate the state machine.
		 *  An OSM should remain if either the activation fails to
		 *  advance state or it advances to an non-initial state.
		 *  @return True if the OSM should remain in list (not retiring).
		 */
		virtual bool __activate() = 0;

		/** Reset the state machine.
		 *  Discarding all buffers, reset state, edges?
		 *  The derived class should call this at the end of its retire().
		 */
		virtual void __retire() {
			_opers_->clear();
			memset(_counters_, 0, _edges_.size()*2*sizeof(_ECOUNT_TYPE));
		}

		/** Reset the state machine.
		 *  Discarding all buffers, reset state, edges?
		 *  The derived class should call this at the end of its flush().
		 */
		virtual void __flush() {
			_opers_->clear();
			memset(_counters_, 0, _edges_.size()*2*sizeof(_ECOUNT_TYPE));
		}

		/** Get the edge by index. */
		_edge_ *__get_edge(unsigned int id) {return _edges_[id];}

		/** Get the current state. */
		unsigned int __get_state() const {return _state_;}

		/** Get the dynamic id of the operation */
		uint64_t __get_dynid() const {return _dynid_;}

		/** Get the dynamic id of the operation */
		void __set_dynid(uint64_t id) {_dynid_ = id;}

		/** Get the operations associated with the machine. */
		std::vector<_base_operation_ *> *__get_opers() {return _opers_;}

		/** Register an operation on the machine. */
		void __register_oper(_base_operation_ *oper) {
			_opers_->push_back(oper);
		}

		/** The simulator object. */
		_SIMULATOR *__get_simulator() const {return _sim_;}

	protected:

		_MACH_ID_T _id_; 	// class id of the operation

		unsigned int _state_;	// current _state

		std::vector<_edge_ *> _edges_;	// edges

		std::vector<_base_operation_ *> *_opers_;	// associated operations

		_ECOUNT_TYPE *_counters_;

		uint64_t _dynid_;

		_SIMULATOR *_sim_;

};

/** Token member names will not clash with identifiers, so
 *  the naming convention does not apply here.
 */

/** Token class. 
 *  A token represent a resource.
 *  _M_ -- manager type
 *  _T_ -- token value type
 */
template <class _M_, class _T_>
class _token_ {

	public:

		/** Constructor. */
		_token_(_M_ *m) : _tenant_(NULL), _boss_(m) {}

		/** Destructor. */
		~_token_() {}

		/** Test if a token is free.
		 *	@return True if the token is free.
		 */
		bool is_free() {return _tenant_==NULL;}

		/** Get the object holding the token. */
		_base_machine_ *get_tenant() {return _tenant_;}

		/** Release the token */
		bool test_release(_base_machine_ *obj) {
			return _boss_->token_good_to_reclaim(this, obj);
		}

		/** Release the token */
		void release(_base_machine_ *obj) {
			_boss_->reclaim_token(this, obj);
		}

		/** Discard the token */
		void discard(_base_machine_ *obj) {
			return _boss_->discard_token(this, obj);
		}

		/** Read the value of the token */
		_T_ read(_base_machine_ *obj) {
			return _boss_->read_token(obj, this);
		}

		/** Write value to the token */
		void write(const _T_& v, _base_machine_ *obj) {
			return _boss_->write_token(v, obj, this);
		}

		/** Get the manager */
		_M_ *get_boss() {return boss;}

		/** Set the token as used by an object. Used by Manager only.
		 *	@param obj The object to hold the token.
		 */
		void set_used(_base_machine_ *obj) {
			_tenant_=obj;
		}

		/** Set an object as unused. Used by manager only. */
		void set_unused() {
			_tenant_=NULL;
		}

	private:

		_base_machine_ *_tenant_;
		_M_ *_boss_;
};

#endif
