
#ifndef __DEVICE_EMULATE_HPP_
#define __DEVICE_EMULATE_HPP_

#include <map>

typedef unsigned uint32_t;

typedef uint32_t dev_id_t;
typedef uint32_t dev_addr_t;
typedef uint32_t dev_data_t;

/** Base class for device.
 *  This class corresponds to a device driver. */
class sima_device {

	public:
		/** Constructor. */
		sima_device() {}

		/** Virtual destructor of base class. */
		virtual ~sima_device() {}

		/** Write a value to the device.
		 *  @param addr The address to write to.
		 *  @param data The data to write.
		 *  @return False to block.
		 */
		virtual bool write(dev_data_t data, dev_addr_t addr) = 0;

		/** Read a value from a device.
		 *  @param addr The address to write to.
		 *  @param data The data to write.
		 *  @return False to block.
		 */
		virtual bool read(dev_data_t& data, dev_addr_t addr) = 0;
};


/** Manager of all devices.
 *  This corresponds to the device table in OS.
 */
class device_emulator {


	public:
		/** Constructor. */
		device_emulator();

		/** Destructor. */
		~device_emulator();

		/** Register a device.
		 *  @param id  The id of the device.
		 *  @param dev A pointer to the device.
		 *  @return True if register successful, false if id already in use.
		 */
		bool register_device(dev_id_t id, sima_device *dev);

		/** Unregister a device.
		 *  @param id The id of the device.
		 *  @return True if successfully unregistered, false if no such device.
		 */
		bool unregister_device(dev_id_t id);

		/** Send a value to a device.
		 *  @param id   The id of the device.
		 *  @param val  The value.
		 *  @param addr The address within the device.
		 *  @return False to block.
		 */
		bool send(dev_id_t id, dev_data_t val, dev_addr_t addr=0);

		/** Receive a value from a device.
		 *  @param id   The id of the device.
		 *  @param val  Reference to the value to be read.
		 *  @param addr The address within the device.
		 *  @return False to block.
		 */
		bool receive(dev_id_t id, dev_data_t &val, dev_addr_t addr=0);

	private:

		/* The last device accessed */
		dev_id_t last_id;
		sima_device *last_dev;

		/* registered devices. */
		std::map<dev_id_t, sima_device *> devs;

};

class arm_simulator;

/* to initialize all devices and to register them */
void init_devices(arm_simulator *);

/* to close all devices */
void close_devices(arm_simulator *);

#endif
