
#ifndef __EMU_MEMORY_H__
#define __EMU_MEMORY_H__

#include "misc.h"
#include <cstring>

namespace emulator {

const unsigned int C54X_MEMORY_SIZE = 65536;

class memory
{
	public:

		/* constructor */
		memory();

		/* destructor */
		~memory();

		/* free all memory pages */
		void reset();

		/* read/write operations */
		byte_t read_byte(target_addr_t addr) {
			return data[addr];
		}

		void write_byte(target_addr_t addr, byte_t value) {
			data[addr] = value;
		}

		word_t read_word(target_addr_t addr) {
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			return swap_word(*((word_t *)(data+addr)));
#else
			return *((word_t *)(data+addr));
#endif
		}

		void write_word(target_addr_t addr, word_t value) {
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			*((word_t *)(data+addr)) = swap_word(value);
#else
			*((word_t *)(data+addr)) = value;
#endif
		}

		/* write operations */
		dword_t read_dword(target_addr_t addr) {
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			return swap_dword(*((dword_t *)(data+addr)));
#else
			return *((dword_t *)(data+addr));
#endif
		}

		void write_dword(target_addr_t addr, dword_t value) {
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			*((dword_t *)(data+addr)) = swap_dword(value);
#else
			*((dword_t *)(data+addr)) = value;
#endif
		}

		/* block read, write */
		void set_block(target_addr_t addr, byte_t value, unsigned int size);
		void read_block(void *buf, target_addr_t addr, unsigned int size);
		void write_block(target_addr_t, void *buf, unsigned int size);

	protected:

		static word_t swap_word(word_t val) {
			return (val>>8) | (val<<8);
		}

		static dword_t swap_dword(dword_t val) {
			return (val>>24) | ((val>>8)&0xFF00) |
				   ((val&0xFF00)<<8) | (val<<24) ;
		}

		byte_t *data;
};

class data_memory : public memory 
{

	public:

	typedef struct __attribute__((packed))
	{
		byte_t byte0;
		byte_t byte1;
		byte_t byte2;
		byte_t byte3;
		byte_t byte4;

	} GP_Reg;

	typedef struct __attribute__((packed))
	{
		// CPU Memory-Mapped Registers
		word_t IMR; // 0 - Interrupt Mask Register
		word_t IFR; // 1 - Interrupt Flag Register
		word_t reserved1; // 2
		word_t reserved2; // 3
		word_t reserved3; // 4
		word_t reserved4; // 5
		word_t ST0; // 6 - Status Register 0  see pg 90, 4-2 spru131g.pdf
				  // 15-13 : ARP
				  // 12	: TC
				  // 11	: C
				  // 10	: OVA
				  //  9	: OVB
				  //  8-0  : DP
		word_t ST1; // 7 - Status Register 1  see pg 92 4-4 spru131g.pdf
				  // 15	: BRAF
				  // 14	: CPL
				  // 13	: XF
				  // 12	: HM
				  // 11	: INTM
				  // 10	: 0
				  //  9	: OVM
				  //  8	: SXM
				  //  7	: C16
				  //  6	: FRCT
				  //  5	: CMPT
				  //  4-0  : ASM

		GP_Reg A; // 8 - Accumulator A Low
				  // 9 - Accumulator A High
				  // A - Accumulator A Guard Bits
		char left_over_guard_bits_A; // fitting 40 bit A into a 48 bit slot
		GP_Reg B;
		char left_over_guard_bits_B; // fitting 40 bit B into a 48 bit slot
		word_t T;	//E
		word_t TRN; //F
		// DAGEN registers begin
		word_t ar0;
		word_t ar1;
		word_t ar2;
		word_t ar3;
		word_t ar4;
		word_t ar5;
		word_t ar6;
		word_t ar7;
		word_t SP;
		word_t BK; // 19 - Circular Buffer Size Register
		// DAGEN registers end
		word_t BRC; // 1A - Block Repeat Counter
		word_t RSA; // 1B - Block Repeat Start Address
		word_t REA; // 1C - Block Repeat End Address
		word_t PMST; // 1D - Processor Mode Status Register
				   // 15-7 : IPTR
				   //  6   : MP/MC'
				   //  5   : OVLY
				   //  4   : AVIS
				   //  3   : DROM
				   //  2   : CLKOFF
				   //  1   : SMUL
				   //  0   : SST
	
		word_t XPC; // 1E - Program counter extension;
		word_t reserved5; // 1F
	} MMR;

	data_memory() : memory() {
		mmr = reinterpret_cast<MMR *>(data);
	}

	public:
		MMR *mmr;
};

typedef memory ProgMemory;
typedef data_memory DataMemory;

} /* namespace */

#endif





