
#include <cstdio>
#include <cstdlib>
#include "simulator.hpp"
#include "loader.hpp"

CoffHeader *read_coff_header(FILE *fp)
{
	CoffHeader *hdr = (CoffHeader *)malloc(sizeof(CoffHeader));

	if(fseek(fp, 0, SEEK_SET) != 0) {
		free(hdr);
		return NULL;
	}

	if(fread(hdr, sizeof(CoffHeader), 1, fp) != 1) {
	    free(hdr);
	    return NULL;
	}

#if DEBUG
	printf("Program header read.\n");
	printf("version_id = 0x%x\n", hdr->version_id);
	printf("flags = 0x%x\n", hdr->flags);
	printf("target_id = 0x%x\n", hdr->target_id);
#endif

	return hdr;
}

CoffOptHeader *read_optional_header(FILE *fp, uint16_t size)
{
	assert(size!=0);

	if(fseek(fp, sizeof(CoffHeader), SEEK_SET) != 0) {
		return NULL;
	}

    if (size != sizeof(CoffOptHeader)) {
		return NULL;
	}

	CoffOptHeader *opt_hdr = (CoffOptHeader *)malloc(sizeof(CoffOptHeader));

	if (fread(opt_hdr, size, 1, fp) != 1) {
		free(opt_hdr);
		return NULL;
	}

#if DEBUG
	printf("Program optional header read.\n");
#endif
	return opt_hdr;
}



void c54x_simulator::load_program(const char *filename)
{
	FILE *fp;

	CoffHeader *header;
	CoffOptHeader *opt_header;

	SecHeader *section_header=NULL;


	bool binutil; // Flag telling whether file is generated from binutil or ti

	/* open the file for read */
	fp = fopen(filename, "rb");
	if (fp == NULL) {
		fprintf(stderr, "Cannot open executable: %s\n", filename);
		exit(1);
	}

	/* read COFF header */
	header = read_coff_header(fp);
	if(header == NULL) {
		fprintf(stderr, "Could not read COFF header: %s\n", filename);
		exit(1);
	}

	/* see if executable */
	if ((header->flags&0x02)==0) {
		fprintf(stderr, "File not executable: %s\n", filename);
		exit(1);
	}

	/* check target */
	if (header->target_id!=0x98) {
		fprintf(stderr, "Target not C54x: %s\n", filename);
		exit(1);
	}

	/* get the size of optional header */
	if (header->num_bytes_opt) {
		opt_header = read_optional_header(fp, header->num_bytes_opt);
		if (opt_header==NULL) {
			fprintf(stderr, "Could not read COFF opt header: %s\n", filename);
			exit(1);
		}

		prog_base = opt_header->start_address;
#if DEBUG
		fprintf(stderr, "Program entry at 0x%x\n", prog_base);
#endif
	}
	else
	{
		fprintf(stderr, "Could not read COFF opt header: %s\n", filename);
		exit(1);
	}

	binutil = header->version_id == 0xc1;

	size_t read_size = binutil?(sizeof(SecHeader)-8):sizeof(SecHeader);

	section_header = (SecHeader *)malloc(sizeof(SecHeader)*
			header->num_sections);

	int k;
	for (k=0; k<header->num_sections; k++) {

		if (fread(section_header+k, read_size, 1, fp) != 1) {
			fprintf(stderr, "Reading section header error: %s\n", filename);
			exit(1);
		}
	}

	for (k=0; k<header->num_sections; k++) {

		unsigned size;
		SecHeader *sec_ptr = section_header + k;

		size = sec_ptr->s_size;

		if (size && (sec_ptr->s_paddr || sec_ptr->s_name[0]))
		{

			if (sec_ptr->s_scnptr )
			{
#if DEBUG
				fprintf(stderr, "Loading section %s at %s::0x%x"
					" from position 0x%x"
					" for length 0x%x\n",
				   	sec_ptr->s_name, sec_ptr->s_flags&0x20?"pmem":"dmem",
					sec_ptr->s_paddr, sec_ptr->s_scnptr, size);
#endif
				// Set position in file to start reading
				if (fseek(fp, sec_ptr->s_scnptr, SEEK_SET)) {
					fprintf(stderr, "Locating section error: %s\n", filename);
					exit(1);
				}

				void *buffer = malloc(size);
				if (fread(buffer, size, 1, fp)!=1) {
					fprintf(stderr, "Reading section error: %s\n", filename);
					exit(1);
				}

				pmem.write_block(sec_ptr->s_paddr, buffer, size);
				dmem.write_block(sec_ptr->s_paddr, buffer, size);

				free(buffer);
			}
			else {
#if DEBUG
				fprintf(stderr, "Initializing section %s at dmem::0x%x"
					" for length 0x%x\n",
					sec_ptr->s_name, sec_ptr->s_paddr, size);
#endif
				// Undefined data, just allocate in blocks of max MX_BUFF
				dmem.set_block(sec_ptr->s_paddr, 0, size);
				pmem.set_block(sec_ptr->s_paddr, 0, size);
			}
	    }
	}
}
