
#ifndef __MORE_MANAGERS_HPP__
#define __MORE_MANAGERS_HPP__

/* include libosm stuff */
#include <iostream>

typedef unit_manager<int> simple_resource;

class branch_manager {

	public:
		/** Token value type. */
		typedef _UINT_T(16) TVAL;

		/**
		 *  @param name Name of the manager.
		 */
		branch_manager(const std::string& name) :
			name(name), trigger(false) {}

		/** Destructor. */
		~branch_manager() {}

		/** See if the token is good for allocation, dummy, always true.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_base_machine_ *obj) {
			return trigger;
		}

		/** See if the token is good for allocation, dummy, always true.	
		 *  Called by writers.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_base_machine_ *obj) {
			return true;
		}

		/** Get the name of the manager. */				   
		const std::string& get_name() {
			return name;
		}

		/** Update the value of the token.					
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const TVAL& val, _base_machine_ *obj) {
			trigger = true;
			new_pc = val;
		}

		void inquire_token(_base_machine_ *obj) {
		}

		/** query from the ifetcher */						
		bool branch_taken() const {return trigger;}

		TVAL get_target() const {return new_pc;}

		/** Regular update on clock edge. */				  
		void update_on_clock() {
			reset();
		}

		/** Reset states. */								  
		void reset() {
			trigger = false;
		}

	private:												  

		std::string name;									 
		bool trigger;
		TVAL new_pc;

};


class fetch_manager {

	public:

		/** The token type. */
		typedef class _token_<fetch_manager> TOKEN_T;

		/** Token value type. */
		typedef _TUPLE_T(_UINT_T(16), _UINT_T(16)) TVAL;

		/** The constructor.
		 *  @param name Name of the manager.
		 */
		fetch_manager(const std::string& name) :
			name(name), token(new TOKEN_T(this)) {
			reset();
		}

		/** Destructor. */
		~fetch_manager() {
			delete token;
		}

		/** See if the token is available for allocation.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_base_machine_ *obj) {
			return token->is_free();
		}

		/** See if the token is available for inquire.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_base_machine_ *obj) {
			return token->is_free();
		}

		/** See if the token is available for release.
		 *  @param obj The requesting OSM.
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _base_machine_ *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(_base_machine_ *obj) {
			token->set_used(obj);
			return token;
		}

		void inquire_token(_base_machine_ *obj) {
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}

		/** Get the name of the manager. */
		const std::string& get_name() {return name;}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		TVAL read_token(_base_machine_ *obj, TOKEN_T *token) {
			_UINT_T(16) iw = get_instruction();
	//		fprintf(stderr, "pc = %d, iw = %x\n", current_pc.val(), iw.val());
			TVAL ret = _TUPLE_CTR(_UINT_T(16), _UINT_T(16), current_pc, iw);
			current_pc = current_pc + _UINT_CTR(16,2u);
			return ret;
		}

		/** Connect the fetcher with other modules.
		 *  @param mbr  The branch manager, for program counter.
		 */
		void connect(branch_manager *mbr) {
			mBR = mbr;
		}

		/** Regular state update on clock edge. */
		void update_on_clock() {
			if (mBR->branch_taken()) {
				current_pc = mBR->get_target();
	//			fprintf(stderr, "branch to %d\n", current_pc.val());
			}
		}

		/** Reset the states/ */
		void reset() {
			token->set_unused();
			current_pc = 0;
		}

		_UINT_T(16) get_instruction(); 

	private:

		std::string name;
		TOKEN_T *token;

		_UINT_T(16) current_pc;	/* the pc currently fetching */

		branch_manager *mBR;
};


/** The register file. */
class register_file {

	public:

		/** index type. */
		typedef _UINT_T(2)  I;

		/** value type. */
		typedef _UINT_T(128) T;

		/** token type. */
		typedef class _token_<register_file> TOKEN_T;

		/** Constructor.
		 *  @param name Name of the register file
		 */
		register_file(const std::string& name) : name(name) {

			tokens = new TOKEN_T[4](this);
			values = new T[4];
			reset();
		}

		/** Destructor. */
		~register_file() {
			delete [] tokens;
			delete [] values;
		}

		/** See if token is available for allocation.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_alloc(const I& ind, _base_machine_ *obj) {
			return true;
		}

		/** See if token is available for inquire.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_inquire(const I& ind, _base_machine_ *obj) {
			return true;
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		void inquire_token(const I& ind, _base_machine_ *obj) {
		}

		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _base_machine_ *obj) {
			return values[ind.val()];
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index of the token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _base_machine_ *obj) {
			int index = ind.val();
			if (index>0) values[index] = val;
	//		fprintf(stderr, "r%d = %lld\n", index, val.val());
		}													 


		/** Resetting states. */
		void reset() {
			for (int i=0; i<4; i++) tokens[i].set_unused();
			values[0] = 0;
			values[1] = 0;
			values[2] = 0;
			values[3] = 1;
		}

		/* called bu the simulator */
		T get_value(int index) {
			return values[index];
		}

	private:

		std::string name;

		TOKEN_T *tokens;
		T *values;
};

#endif
