
#ifndef __MORE_MANAGERS_HPP__
#define __MORE_MANAGERS_HPP__

/* include libosm stuff */
#include <iostream>

/** The register file. */
class shift_reg_manager {

	public:

		/** index type. */
		typedef _UINT_T(2)  I;

		/** value type. */
		typedef _UINT_T(64) T;

		/** token type. */
		typedef class _token_<shift_reg_manager > TOKEN_T;

		/** Constructor.
		 *  @param name Name of the register file
		 */
		shift_reg_manager(const std::string& name) : name(name) {

			tokens = new TOKEN_T[3](this);
			values = new T[3];
			reset();
		}

		/** Destructor. */
		~shift_reg_manager() {
			delete [] tokens;
			delete [] values;
		}

		/** See if token is available for allocation.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_alloc(const I& ind, _base_machine_ *obj) {
			return true;
		}

		/** See if token is available for inquire.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_inquire(const I& ind, _base_machine_ *obj) {
			return true;
		}

#if 0 //These are no needed for this manager.
		/** See if it is ok to release token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _base_machine_ *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(const I& ind, _base_machine_ *obj) {
			assert(ind.val()<3);
			return tokens+ind.val();
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _base_machine_ *obj) {
			token->set_unused();
		}
#endif

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}


		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _base_machine_ *obj) {
			assert(ind.val()<3);
			int index = ind.val()+shift_ind;
			return values[index>2?index-3:index];
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index of the token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _base_machine_ *obj) {

			int index = ind.val()+shift_ind;
			values[index>2?index-3:index] = val;
			/* shift the registers on write */

			shift_ind++;
			if (shift_ind>2) shift_ind=0;
		}													 

		/** Get the ultimate result */
		_UINT_T(64) get_result() {
			return values[shift_ind];
		}


		/** Resetting states. */
		void reset() {
			for (int i=0; i<3; i++) {
				tokens[i].set_unused();
			}
			values[0] = 0;
			values[1] = 1;
			values[2] = 0;
			shift_ind = 0;
		}

	private:

		std::string name;

		TOKEN_T *tokens;
		T *values;
		unsigned shift_ind;

};

#endif
