
#include "simulator.hpp"

#include "opers.hpp"
#include "machines.hpp"
#include <vector>
#include <iomanip>
#include <iostream>

#include <csignal>
#include <cstring>
#include <cstdlib>
#include "config.h"

using std::vector;
using std::list;

simple_simulator::simple_simulator(bool verbose) :
	running(false), debuging(false)
{

	/* Create the managers */
#define MANG_DEF(cname, iname) iname = new cname(#iname);
#include "mang_list.hpp"

	/* storage for operation/machine */
	machine_pool = new obj_pool<_base_machine_, _MACH_ID_T, simple_simulator,
								simple_simulator::mach_allocator>(_mach_total);
	operation_pool = new obj_pool<_base_operation_, _OPER_ID_T, simple_simulator,
								simple_simulator::oper_allocator>(_oper_total);
	
	/* reset every thing */
	reset();
}

simple_simulator::~simple_simulator()
{
	delete machine_pool;
	delete operation_pool;

#define MANG_DEF(cname, iname) delete iname;
#include "mang_list.hpp"

}

void simple_simulator::reset()
{
#define MANG_DEF(cname, iname) iname->reset();
#include "mang_list.hpp"

	reset_stats();
}

void simple_simulator::reset_stats()
{
	cycle_count = 0;
	alloc_count = 0;
	retire_count = 0;
}

void simple_simulator::dump_stats(FILE *fp)
{
	fprintf(fp, "Total allocated OSMs : %lld\n", alloc_count);
	fprintf(fp, "Total retired OSMs   : %lld\n", retire_count);

	fprintf(fp, "Total cycles         : %lld\n", cycle_count);
	fprintf(fp, "Result : %lld\n", mSHIFT->get_result().val());

}


void simple_simulator::clock_tick()
{
	if (running) {

#undef DEBUG 
#undef DEBUG_CYCLE
#ifdef DEBUG_CYCLE
		fprintf(stderr, "--clock %d--\n", cycle_count);
#endif

		/** Activate all OSMs in order. */
		for (work_it=work_list.begin(); work_it!=work_list.end();) {

			ins_it = work_it;
			ins_it++;
#if defined DEBUG_CYCLE || defined DEBUG
			int oldstate = (*work_it)->__get_state();
#endif
			bool ret = (*work_it)->__activate();
#ifdef DEBUG_CYCLE
			std::cerr << (*work_it)->__get_dynid() << " " << oldstate << "->";
			std::cerr << (*work_it)->__get_state() << " " << std::endl;
#endif

			/* Test if the OSM is back to the initial state. */
			if (ret) { /* no */
				work_it++;
			}
			else { /* yes */

#ifdef DEBUG
				//fprintf(stderr, "-%d-", oldstate);
				if (oldstate>3 && oldstate!=8)
					fprintf(stderr, "0x%08x 0x%08x 0x%08x\n",
						static_cast<MC_normal *>(*work_it)->G_iw.val(),
						static_cast<MC_normal *>(*work_it)->G_v_iflag.val(),
						static_cast<MC_normal *>(*work_it)->G_v_oflag.val());
#endif

				/* send retired ones to the sink. */
				sink_machine(*work_it);
				work_it = work_list.erase(work_it);
			}
		}

		cycle_count++;
	}
}

void simple_simulator::run(uint64_t maxCycle)
{
	_base_machine_ *src = source_machine();
	work_list.push_back(src);

	running = true;
	while(running && cycle_count<=maxCycle) {
		clock_tick();
	}
}

void simple_simulator::activate_operation(_base_operation_ *oper,
	_OPER_ID_T *ids, void *cdg)
{
	_base_machine_ *mach = allocate_machine(oper->__get_mach_id());
	oper->__init(mach, ids, cdg);

	work_list.insert(ins_it, mach);
}

_base_machine_ *simple_simulator::allocate_machine(_MACH_ID_T id)
{
	_base_machine_ *mach = machine_pool->allocate(this,id);
	mach->__set_dynid(alloc_count++);
	return mach;
}

#if 0
_base_operation_ *simple_simulator::allocate_operation(_OPER_ID_T id)
{
	return operation_pool->allocate(this,id);
}
#endif

void simple_simulator::sink_machine(_base_machine_ *mach)
{
	_base_operation_ **opers;
	unsigned ocount = mach->__get_opers(opers);

	for (unsigned i=0; i<ocount; i++)
		free_operation(opers[i]);

	retire_count++;
	mach->__retire();
	machine_pool->reclaim(mach);
}

_base_machine_ *simple_simulator::source_machine()
{
	_base_operation_ *oper = allocate_operation(_oper_any);
	_base_machine_ *src = allocate_machine(_mach_normal);
	oper->__init(src, NULL, NULL);

	return src;
}

_base_machine_ *simple_simulator::mach_allocator(simple_simulator *s, _MACH_ID_T id)
{
  switch (id) {
#define _MACH_DEF(_a, _b) \
	case _mach_ ## _b: return new _MACH_CLASS(_b)(s);
#include "mach_list.hpp"
	default:
      assert(0);
  }
}

_base_operation_ *simple_simulator::oper_allocator(simple_simulator *s, _OPER_ID_T id)
{
  switch (id) {
#define _OPER_DEF(_a, _b, _c) \
	case _oper_ ## _b: return new _OPER_CLASS(_b)();
#include "oper_list.hpp"
    default:
      assert(0);
  }
}

