
#ifndef __MORE_MANAGERS_HPP__
#define __MORE_MANAGERS_HPP__

/* include libosm stuff */
#include "include_osm.hpp"
#include "emumem.h"
#include <string>

/** short hand */

/** The register file + forwarding path manager. */
template <class I, class T, unsigned N>
class regval_manager {

	public:

		/** token type. */
		typedef class _token_<regval_manager<I, T, N> > TOKEN_T;

		/** Constructor.
		 *  @param name Name of the register file
		 */
		regval_manager(const std::string& name) : name(name) {

			tokens = new TOKEN_T[N](this);
			values = new T[N];
		}

		/** Destructor. */
		~regval_manager() {
			delete [] tokens;
			delete [] values;
		}

		/** See if token is available for allocation.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_alloc(const I& ind, _BASE_MACHINE *obj) {
			return tokens[ind.val()].is_free() 
				/*|| obj->__get_dynid() >
					tokens[ind.val()].get_tenant()->__get_dynid()*/;
		}

		/** See if token is available for inquire.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_inquire(const I& ind, _BASE_MACHINE *obj) {
			return tokens[ind.val()].is_free() ||
				   obj==tokens[ind.val()].get_tenant();
		}

		/** See if it is ok to release token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(const I& ind, _BASE_MACHINE *obj) {
			tokens[ind.val()].set_used(obj);
			return tokens+ind.val();
		}

		/** Inquire a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		void inquire_token(const I& ind, _BASE_MACHINE *obj) {
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {
			if (token->get_tenant()==obj)
				token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {
			if (token->get_tenant()==obj)
				token->set_unused();
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}


		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _BASE_MACHINE *obj) {
			return values[ind.val()];
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {
			return values[token-tokens];
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index of the token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _BASE_MACHINE *obj) {
			values[ind.val()] = val;
		}													 

		/** Update the value of the token.
		 * This function is called from the token. This corresponds
		 * to *token = val in OSM expression.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 * @param token The token to be updated.
		 */
		void write_token(const T& val, _BASE_MACHINE *obj, TOKEN_T *token) {
			values[token-tokens] = val;
		}


		/** Get the value of a register.
		 *  This is used for value access from the hardware layer.
		 *  @param ind Index of the token.
		 */
		T get_value(int ind) {
			return values[ind];
		}

		/* Set the value of a register.
		 * This is used for value access from the hardware layer. So
		 * this does not monitor the value of pc.
		 * @param ind Index of the token.
		 * @param val Value to write.
		 */
		void set_value(int ind, const T& val) {
			values[ind] = val;
		}

		/** Free those temporarily allocated. */
		void update_on_clock() {}

		/** Resetting states. */
		void reset() {
			for (unsigned i=0; i<N; i++) {
				tokens[i].set_unused();
				values[i] = 0;
			}
		}

	private:

		std::string name;

		TOKEN_T *tokens;
		T *values;

};

class newpc_manager {

	public:

		/** Token value type. */
		typedef _UINT_T(16) TVAL;

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		newpc_manager(const std::string& name) :
			name(name), trigger(false) {}

		/** Destructor. */
		~newpc_manager() {}

		/** See if the token is good for allocation, dummy, always true.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {
			return trigger;
		}

		/** See if the token is good for allocation, dummy, always true.
		 *  Called by writers.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return true;
		}

		/** Inquire a token.
		 *  @param obj The OSM sending the request.			
		 */
		void inquire_token(_BASE_MACHINE *obj) {
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const TVAL& val, _BASE_MACHINE *obj) {
			target = (uint32_t)(val.val());
			trigger = true;
#ifdef DEBUG
			fprintf(stderr, "branch target: 0x%08x\n", target);
#endif
		}

		/** queries from the ifetcher */
		bool branch_taken() const {return trigger;}
		uint32_t get_target() const {return target;}


		/** Regular update on clock edge. */
		void update_on_clock() {
			reset();
		}

		/** Reset states. */
		void reset() {
			trigger = false;
		}

	private:

		std::string name;
		bool trigger;
		uint32_t target;

};

/** fetching manager. */
class fetch_manager {

	public:

		/** The token type. */
		typedef class _token_<fetch_manager> TOKEN_T;

		typedef _TUPLE_T(_UINT_T(16), _UINT_T(16)) T;

		/** The constructor.
		 *  @param name Name of the manager.
		 */
		fetch_manager(const std::string& name) :
			name(name), token(new TOKEN_T(this)),
	   		fetch_status(0), trigger(false) {}

		/** Destructor. */
		~fetch_manager() {
			delete token;
		}


#if 0
		/** See if the token is available for allocation.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return token->is_free();
		}
#endif

		/** See if the token is available for inquire.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {
			return trigger == false;
		}

		/** Inquire a token.
		 *  @param obj The OSM sending the request.			
		 */
		void inquire_token(_BASE_MACHINE *obj) {
		}

#if 0
		/** See if the token is available for release.
		 *  @param obj The requesting OSM.
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return fetch_status==0;
		}

		/** Allocate a token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(_BASE_MACHINE *obj) {
			token->set_used(obj);
			fetch_status = 1;
			trigger = true;
			return token;
		}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {
			token->set_unused();
		}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {
			token->set_unused();
		}
#endif

		/** Get the name of the manager. */
		const std::string& get_name() {return name;}

		/** Access the value of the token.
		 * This corresponds to val = *m[] in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj) {

			trigger = true;
			uint16_t iw = mem->read_word(pc<<1);
#undef DEBUG
#ifdef DEBUG
			fprintf(stderr, "0x%04x : 0x%04x\n", pc, iw);
#endif
			return _TUPLE_CTR(_UINT_T(16), _UINT_T(16), pc, iw);
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {

			trigger = true;
			uint16_t iw = mem->read_word(pc<<1);
#ifdef DEBUG
			fprintf(stderr, "0x%04x : 0x%04x\n", pc, iw);
#endif
			return _TUPLE_CTR(_UINT_T(16), _UINT_T(16), pc, iw);
		}

		/** Connect the fetcher with other modules.
		 *  @param mrf  The register file manager, for program counter.
		 *  @param csh  The instruction cache object.
		 *  @param tlb  The instruction tlb object.
		 *  @param imm  The memory to get instrution from.
		 */
		void connect(newpc_manager *mrs, ProgMemory *imm) {
			mRS = mrs;
			mem = imm;
		}

		/** Set the program counter. */
		void set_pc(uint16_t npc) {
			pc = npc;
		}

		uint16_t get_pc() const {return pc;}

		/** Regular state update on clock edge. */
		void update_on_clock();

		/** Reset the states/ */
		void reset() {
			trigger = false;
			token->set_unused();
			fetch_status = 0;
		}

	private:

		std::string name;
		TOKEN_T *token;

		unsigned pc;	/* the pc currently fetching */
		int trigger;	/* start a new request. */
		int fetch_status;

		/* connected components. */
		newpc_manager *mRS;
		ProgMemory *mem;
};

typedef class unit_value_manager<_UINT_T(16)> mem_addr;
typedef class unit_value_manager<_UINT_T(16)> mem_data;

/** The memory control port.
 *  This port does nothing but to synchronize data memory access.
 *  An OSM should inquire this port to see if the memory request finishes.
 */
class mem_ctrl {

	public:

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		mem_ctrl(const std::string& name) :
			name(name), status(0), trigger(false) {}

		/** Destructor. */
		~mem_ctrl() {}

#if 0
		/** See if token is availble for inquire.
		 *  Equivalent to if the memory access has finished.
		 *  @param obj The object performing the memory transaction.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {
			return status==0;
		}
#endif

		/** See if the token is available for allocate, dummy always true. */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return true;
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const _UINT_T(16)& val, _BASE_MACHINE *obj) {
			isRead = val.val();
			size = 16;
			trigger = true;
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Connect the module with other modules.
		 *  @param csh  The data cache object.
		 *  @param tlb  The data tlb object.
		 *  @param rpt  The memory read  port.
		 *  @param wpt  The memory write port.
		 *  @param dmm  The data memory containing real data.
		 *  @param map  The memory address port.
		 */
		void connect(mem_data *rpt, DataMemory *dmm, mem_addr *map) {
			mMemData = rpt;
			mem = dmm;
			mMemAddr = map;
		}

		/** Regular update on clock edge. */
		void update_on_clock();

		/** reset states. */
		void reset() {
			status = 0;
			trigger = false;
		}

	private:

		std::string name;

		unsigned status;
		unsigned addr;
		unsigned size;
		bool isRead;
		bool trigger;

		mem_data *mMemData;
		mem_addr *mMemAddr;
		DataMemory *mem;

};

typedef unit_manager<int> simple_resource;
typedef regval_manager<_UINT_T(3), _UINT_T(16), 8> ireg_manager;
typedef regval_manager<_UINT_T(3), _UINT_T(16), 8> addr_manager;

/* manager controlling several non-exclusive resource.
 * Suitable be used as a data port.
 * I -- token index type
 * T -- token value type
 */
class longreg_manager {

	public:

		typedef _UINT_T(2)  I;
		typedef _UINT_T(32) T;

		/** type of the token */
		typedef class _token_<longreg_manager> TOKEN_T;

		/** constructor.
		 * @param name Name of the manager.
		 * @param n    Number of tokens.
		 */
		longreg_manager(const std::string& name) : nTotal(4), name(name) {
			tokens = new TOKEN_T[4](this);
		}

		/**  destructor. */
		~longreg_manager() {
			delete [] tokens;
		}

		/** non-exclusive resource -> always available for allocation. */
		bool token_free_for_alloc(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> always available for inquire. */
		bool token_free_for_inquire(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> always ok to release. */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> no need to set tenant field. */
		TOKEN_T *allocate_token(const I& ind, _BASE_MACHINE *obj) {
			return tokens;
		}

		/** non-exclusive resource -> no need to set tenant field. */
		TOKEN_T *allocate_token_temp(const I& ind, _BASE_MACHINE *obj) {
			return tokens;
		}

		/** Inquire a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		void inquire_token(const I& ind, _BASE_MACHINE *obj) {
		}

		/** release the token, do nothing. */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {
		}

		/** discard the token, do nothing. */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {
		}


		/** get the name of the token. */
		const std::string& get_name() {
			return name;
		}

		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index to token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _BASE_MACHINE *obj) {
			_UINT_T(16) lo = ireg->get_value(ind.val()*2);
			_UINT_T(16) hi = ireg->get_value(ind.val()*2+1);
			return (hi.val()<<16) | lo.val();
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index to token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _BASE_MACHINE *obj) {
			ireg->set_value(ind.val()*2, val.val());
			ireg->set_value(ind.val()*2+1, val.val()>>16);
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {
			unsigned ind = token-tokens;
			_UINT_T(16) lo = ireg->get_value(ind*2);
			_UINT_T(16) hi = ireg->get_value(ind*2+1);
			return (hi.val()<<16) | lo.val();
		}

		/** Update the value of the token.
		 * This function is called from the token. This corresponds
		 * to *token = val in OSM expression.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 * @param token The token to be updated.
		 */
		void write_token(const T& val, _BASE_MACHINE *obj, TOKEN_T *token) {
			unsigned ind = token-tokens;
			ireg->set_value(ind*2, val.val());
			ireg->set_value(ind*2+1, val.val()>>16);
		}

		/** connect with the integer manager. */
		void connect(ireg_manager *ir) {
			ireg = ir;
		}

		T get_value(int ind) {
			_UINT_T(16) lo = ireg->get_value(ind*2);
			_UINT_T(16) hi = ireg->get_value(ind*2+1);
			return (hi.val()<<16) | lo.val();
		}

		/** Do nothing. */
		void update_on_clock() {}

		/** reset the manager. */
		void reset() {
			for (int i=0; i<nTotal; i++)
				tokens[i].set_unused();
		}

	private:

		const int nTotal;
		std::string name;
		TOKEN_T *tokens;
		ireg_manager *ireg;
};

/* manager controlling several non-exclusive resource.
 * Suitable be used as a data port.
 * I -- token index type
 * T -- token value type
 */
class acc_manager {

	public:

		typedef _UINT_T(1)  I;
		typedef _UINT_T(32) T;

		/** type of the token */
		typedef class _token_<acc_manager> TOKEN_T;

		/** constructor.
		 * @param name Name of the manager.
		 * @param n    Number of tokens.
		 */
		acc_manager(const std::string& name) : 
				nTotal(2), name(name) {
			tokens = new TOKEN_T[2](this);
		}

		/**  destructor. */
		~acc_manager() {
			delete [] tokens;
		}

		/** non-exclusive resource -> always available for allocation. */
		bool token_free_for_alloc(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> always available for inquire. */
		bool token_free_for_inquire(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> always ok to release. */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return true;
		}

		/** non-exclusive resource -> no need to set tenant field. */
		TOKEN_T *allocate_token(const I& ind, _BASE_MACHINE *obj) {
			return tokens;
		}

		/** non-exclusive resource -> no need to set tenant field. */
		TOKEN_T *allocate_token_temp(const I& ind, _BASE_MACHINE *obj) {
			return tokens;
		}

		/** Inquire a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		void inquire_token(const I& ind, _BASE_MACHINE *obj) {
		}

		/** release the token, do nothing. */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {
		}

		/** discard the token, do nothing. */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {
		}


		/** get the name of the token. */
		const std::string& get_name() {
			return name;
		}

		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index to token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _BASE_MACHINE *obj) {
			_UINT_T(16) lo = ireg->get_value(ind.val()*2+4);
			_UINT_T(16) hi = ireg->get_value(ind.val()*2+5);
			return (hi.val()<<16) | lo.val();
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index to token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _BASE_MACHINE *obj) {
			ireg->set_value(ind.val()*2+4, val.val());
			ireg->set_value(ind.val()*2+5, val.val()>>16);
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {
			unsigned ind = token-tokens;
			_UINT_T(16) lo = ireg->get_value(ind*2+4);
			_UINT_T(16) hi = ireg->get_value(ind*2+5);
			return (hi.val()<<16) | lo.val();
		}

		/** Update the value of the token.
		 * This function is called from the token. This corresponds
		 * to *token = val in OSM expression.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 * @param token The token to be updated.
		 */
		void write_token(const T& val, _BASE_MACHINE *obj, TOKEN_T *token) {
			unsigned ind = token-tokens;
			ireg->set_value(ind*2+4, val.val());
			ireg->set_value(ind*2+5, val.val()>>16);
		}

		/** connect with the integer manager. */
		void connect(ireg_manager *ir) {
			ireg = ir;
		}

		T get_value(int ind) {
			_UINT_T(16) lo = ireg->get_value(ind*2+4);
			_UINT_T(16) hi = ireg->get_value(ind*2+5);
			return (hi.val()<<16) | lo.val();
		}
		/** Do nothing. */
		void update_on_clock() {}

		/** reset the manager. */
		void reset() {
			for (int i=0; i<nTotal; i++)
				tokens[i].set_unused();
		}

	private:

		const int nTotal;
		std::string name;
		TOKEN_T *tokens;
		ireg_manager *ireg;
};


//typedef regval_manager<_UINT_T(2), _UINT_T(32), 8> longreg_manager;
//typedef regval_manager<_UINT_T(1), _UINT_T(32), 8> acc_manager;

typedef unit_value_manager<_UINT_T(16)> mac16_manager;
typedef unit_value_manager<_UINT_T(1)> flag_manager;
typedef unit_value_manager<_UINT_T(1)> flag_manager;
typedef unit_value_manager<_UINT_T(1)> flag_manager;


#endif
