
#include "simulator.hpp"

#include "machines.hpp"
#include <vector>
#include <iostream>

#include <cstdio>
#include <csignal>
#include <cstring>
#include <cstdlib>
#include "config.h"

using std::vector;
using std::list;
using std::cerr;
using std::endl;

hip_simulator::hip_simulator(bool verbose) :
	verbose(verbose), running(false), pmem(), dmem()
{
	/* Create the managers */
#define MANG_DEF(cname, iname) iname = new cname(#iname);
#include "mang_list.hpp"

	/* conenct the managers. */
	
	/* storage for operation/machine */
	machine_pool = new obj_pool<_opt_machine_, _MACH_ID_T, hip_simulator,
								hip_simulator::mach_allocator>(_mach_total);
	

	mFetch->connect(mBranch, &pmem);
	mMemCtrl1->connect(mMemIData1, &dmem, mMemAddr1);
	mMemCtrl2->connect(mMemIData2, &dmem, mMemAddr2);
	mLReg->connect(mIReg);
	mAccReg->connect(mIReg);

	/* reset every thing */
	reset();
}

hip_simulator::~hip_simulator()
{
	delete machine_pool;

#define MANG_DEF(cname, iname) delete iname;
#include "mang_list.hpp"

}

void hip_simulator::reset()
{
	pmem.reset();
	dmem.reset();

#define MANG_DEF(cname, iname) iname->reset();
#include "mang_list.hpp"

	reset_stats();
}

void hip_simulator::reset_stats()
{
	cycle_count = 0;
	alloc_count = 0;
	retire_count = 0;
}

void hip_simulator::dump_stats(FILE *fp)
{
	fprintf(fp, "Total allocated OSMs : %lld\n", alloc_count);
	fprintf(fp, "Total retired OSMs   : %lld\n", retire_count);

	fprintf(fp, "Total cycles         : ");
	dump_int64(cycle_count, fp);
	fprintf(stderr, "\n");
}


void hip_simulator::clock_tick()
{

#undef DEBUG_CYCLE
#ifdef DEBUG_CYCLE
		fprintf(stderr, "--clock %d--\n", cycle_count);
#endif

		/** Activate all OSMs in order. */
		for (work_it=work_list.begin(); work_it!=work_list.end();) {

			ins_it = work_it;
			ins_it++;
#if defined DEBUG_CYCLE 
			int oldstate = (*work_it)->__get_state();
#endif
			bool ret = (*work_it)->__activate();
#ifdef DEBUG_CYCLE
			std::cerr << (*work_it)->__get_dynid() << " " << oldstate << "->";
			std::cerr << (*work_it)->__get_state() << " " << ret << std::endl;
#endif

			/* Test if the OSM is back to the initial state. */
			if (ret) { /* no */
				work_it++;
			}
			else { /* yes */

#if 0 
				//fprintf(stderr, "-%d-", oldstate);
				if (oldstate>3 && oldstate!=8)
					fprintf(stderr, "0x%08x 0x%08x 0x%08x\n",
						static_cast<MC_normal *>(*work_it)->G_iw.val(),
						static_cast<MC_normal *>(*work_it)->G_v_iflag.val(),
						static_cast<MC_normal *>(*work_it)->G_v_oflag.val());
#endif

				/* send retired ones to the sink. */
				sink_machine(*work_it);
				work_it = work_list.erase(work_it);
			}
		}

		cycle_count++;

#define MANG_DEF(cname, iname) iname->update_on_clock();
#include "mang_list.hpp"

#if 0
	std::cerr << "NEG = " << mN->get_value().hex() << " ";
	std::cerr << "ZERO = " << mZ->get_value().hex() << " ";
	std::cerr << "CARRY = " << mC->get_value().hex() << std::endl;
	std::cerr << "SP = " << mSP->get_value().hex() << std::endl;

	static char *reg_names[] = {"A0", "A1", "B0", "B1", "C0", "C1", "D0", "D1"};
	static char *lreg_names[] = {"AX", "BX", "CX", "DX"};
	static char *accreg_names[] = {"CX", "DX"};

	for (unsigned ii=0; ii<8; ii++) {
		std::cerr << reg_names[ii] << ":" << mIReg->get_value(ii).hex() << " ";
	}
	std::cerr << std::endl;

	for (unsigned ii=0; ii<4; ii++) {
		std::cerr << lreg_names[ii] << ":" << mLReg->get_value(ii).hex() << " ";
	}
	std::cerr << std::endl;

	for (unsigned ii=0; ii<2; ii++) {
		std::cerr << accreg_names[ii] << ":" << mAccReg->get_value(ii).hex() << " ";
	}
	std::cerr << std::endl;

	for (unsigned ii=0; ii<8; ii++) {
		std::cerr << "X" << ii << ":" << mAddrReg->get_value(ii).hex()
			 << " ";
	}
	std::cerr << std::endl;
#endif
}

void hip_simulator::run(UInt64 maxCycle)
{
	_opt_machine_ *src = source_machine();
	work_list.push_back(src);

	running = true;
	while(running && cycle_count<=maxCycle) {
		clock_tick();
	}
}

void hip_simulator::activate_machine(_MACH_ID_T id)
{
	_opt_machine_ *mach = allocate_machine(id);
	mach->__init();

	work_list.insert(ins_it, mach);
}

_opt_machine_ *hip_simulator::allocate_machine(_MACH_ID_T id)
{
	_opt_machine_ *mach = machine_pool->allocate(this,id);
	mach->__set_dynid(alloc_count++);
	return mach;
}

#if 0
_base_operation_ *hip_simulator::allocate_operation(_OPER_ID_T id)
{
	return operation_pool->allocate(this,id);
}
#endif

void hip_simulator::sink_machine(_opt_machine_ *mach)
{
	retire_count++;
	mach->__retire();
#if 0
	fprintf(stderr, "0x%08x 0x%08x\n",
		 static_cast<MC_normal *>(mach)->G_iw.val(),
		 static_cast<MC_normal *>(mach)->G_v_iflag.val());
	fprintf(stderr, "0x%08x : 0x%08x %lld\n",
		 static_cast<normal *>(mach)->pc.val(),
		 static_cast<normal *>(mach)->iw.val(), mach->get_dynid());
	fprintf(stderr, "%lld retire\n", mach->get_dynid());
#endif

	machine_pool->reclaim(mach);
}

_opt_machine_ *hip_simulator::source_machine()
{
	_opt_machine_ *src = allocate_machine(_mach_fetcher_normal);
	src->__init();

	return src;
}

_opt_machine_ *hip_simulator::mach_allocator(hip_simulator *s, _MACH_ID_T id)
{
  switch (id) {
#define _MACH_DEF(_a, _b) \
	case _mach_ ## _b: return new _MACH_CLASS(_b)(s);
#include "mach_list.hpp"
	default:
      assert(0);
  }
}

