
#ifndef __EXTERNAL_DEV_H_
#define __EXTERNAL_DEV_H_

#include <queue>
#include "emu_device.hpp"

using emulator::device_master;
using emulator::emu_device;

/* stub functions. */
void init_devices(device_master *dm);
void close_devices(device_master *dm);

/* This is an example device. To define your own device,
   1. derive you device class from emu_device.
   2. define your own write/read functions.
   3. create device instances and register them in 
      external_dev_stub.cpp::init_devices.
   4. unregister devices and destroy then instances
      in external_dev_stub.cpp::close_devices.

   Note: The device can block the execution of the processor core
      by returning a false in its write/read methods. For devices
      featuring non-blocking read/write, a true value should always
      be returned.

   When programming your code to simulate, you can use the functions
   defined in device/arm_coproc.h to access coprocessors.
   The function implementation is in device/arm_coproc.s. So remember to
   assemble and link the file with your application.
   A sample file device/arm_test.c is given.
*/


namespace emulator {
  
  typedef std::pair <unsigned, UInt32>  ns_data;

  class arm_if_device : public emu_device {


	public:
		arm_if_device() : emu_device() {
                  input_queue = new std::queue<unsigned>;
                  output_queue = new std::queue<ns_data *>;
                }

		~arm_if_device() {
                  delete input_queue;
                  delete output_queue;
                }

                

		bool write(dev_data_t data, dev_addr_t addr) {
                 
                    ns_data *mydata = new ns_data(addr, data);
                    (output_queue)->push(mydata);
                    return true;
                  
                }
                
		bool read(dev_data_t& data, dev_addr_t addr) {
                  
                    if( (input_queue)->empty()){
                    return false;
                    }else{
                        data = (input_queue)->front();
                        input_queue->pop();
			return true;
                   }
		}
                
                bool get_output_queue(ns_data& data){
                  if( (output_queue)->empty()){
                    return false;
                  }else{
                        data = *((output_queue)->front());
                        output_queue->pop();
			return true;
                  }
                }
                
                bool put_input_queue(ns_data data){
                   
                   input_queue->push(data.second);
		   return true;
                }


                bool get_output_queue_status(){
                 return !((output_queue)->empty());
                  
                }
                
               
	private: 
		std::queue<unsigned> *input_queue;
                std::queue<ns_data *> *output_queue;
  };



class arm_id_device : public emu_device {


	public:
		arm_id_device() : emu_device() {
                  
                }

		~arm_id_device() {
                  
                  
                }

                void set_address(unsigned addr){
                  _addr = addr;
                  return; 
                }
             	bool write(dev_data_t data, dev_addr_t addr) {
                
                    return true;
                  
                }
                	
		bool read(dev_data_t& data, dev_addr_t addr) {
                  data = _addr;
				  return true;
		}
                
                
	private: 
                unsigned _addr;
  };

}


#endif

