
#ifndef __EMU_MEMORY_H__
#define __EMU_MEMORY_H__

#include "misc.h"
#include <cstring>

#ifdef COSIM_STUB
#include <map>
#include "external_mem.h"
#endif

namespace emulator {

const unsigned int MEMORY_PAGE_SIZE = 4096;
const unsigned int PRIMARY_MEMORY_HASH_TABLE_SIZE = 4096;
const unsigned int SECONDARY_MEMORY_HASH_TABLE_SIZE = 16;

typedef struct memory_page_table_entry_t
{
	target_addr_t addr;
	struct memory_page_table_entry_t *next;
	byte_t *storage;
} memory_page_table_entry_t;

typedef struct
{
	memory_page_table_entry_t *pte[SECONDARY_MEMORY_HASH_TABLE_SIZE];
} secondary_memory_hash_table_t;

class memory
{

#ifdef COSIM_STUB
		std::map<target_addr_t, int> external_decoded_addr;
#endif

	public:

		/* constructor */
		memory();

		/* destructor */
		~memory();

		/* copy constructor, not implemented */
		memory(const memory&);

		/* free all memory pages */
		void reset();

		/* read/write operations */
		halfword_t read_half_word(target_addr_t addr);
		dword_t read_dword(target_addr_t addr);

		byte_t read_byte(target_addr_t addr)
		{
			target_addr_t offset = addr % MEMORY_PAGE_SIZE;
			memory_page_table_entry_t *pte = get_page(addr);

#ifdef COSIM_STUB   
			if (external_decoded_addr.find(addr) != external_decoded_addr.end()) {
				return ext_read_byte(external_decoded_addr[addr]);
			}                                                   
#endif
	
			return pte->storage[offset];
		}

		void write_byte(target_addr_t addr, byte_t value)
		{
			target_addr_t offset = addr % MEMORY_PAGE_SIZE;
			memory_page_table_entry_t *pte = get_page(addr);
#ifdef COSIM_STUB
			if (external_decoded_addr.find(addr) != external_decoded_addr.end()) {
				ext_write_byte(external_decoded_addr[addr], value);
				return;
			}
#endif
			pte->storage[offset] = value;
		}

		word_t read_word(target_addr_t addr)
		{
			target_addr_t offset = addr % MEMORY_PAGE_SIZE;
			memory_page_table_entry_t *pte = get_page(addr);

#ifdef COSIM_STUB
			if (external_decoded_addr.find(addr) != external_decoded_addr.end()) {
				return ext_read_word(external_decoded_addr[addr]);
			}
#endif
	
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			return swap_word(*((word_t *)(pte->storage+offset)));
#else
			return *((word_t *)(pte->storage+offset));
#endif
		}

		void write_word(target_addr_t addr, word_t value)
		{
			target_addr_t offset = addr % MEMORY_PAGE_SIZE;
			memory_page_table_entry_t *pte = get_page(addr);
#ifdef COSIM_STUB
			if (external_decoded_addr.find(addr) != external_decoded_addr.end()) {
				ext_write_word(external_decoded_addr[addr], value);
				return;
			}
#endif

#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
			*((word_t *)(pte->storage+offset)) = swap_word(value);
#else
			*((word_t *)(pte->storage+offset)) = value;
#endif
		}

		void read_block(void *buf, target_addr_t addr, unsigned int size);

		/* write operations */
		void write_half_word(target_addr_t addr, halfword_t value);
		void write_dword(target_addr_t addr, dword_t value);

		void set_block(target_addr_t addr, byte_t value, unsigned int size);
		void write_block(target_addr_t, void *buf, unsigned int size);

		/* some usefule utilities */
		target_addr_t align_to_page_boundary(target_addr_t addr);

#ifdef TEST_MEMORY
		/* memory test , true->succeed */
		bool test();
#endif

		/* if the fist read is to address 0, this will crash, so beware */
		word_t read_word_fast(target_addr_t addr) {

 			memory_page_table_entry_t *pte; 
			target_addr_t offset = addr % MEMORY_PAGE_SIZE; 

		    if (cached_addr==addr-offset) pte = cached_pte; 
		    else {
				cached_pte = pte = get_page(addr);
				cached_addr = addr-offset;
			}

#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
		    return swap_word(*((word_t *)(pte->storage+offset)));
#else
		    return *((word_t *)(pte->storage+offset));
#endif
		}                               

		/* return the number of pages allocated */
		unsigned int get_page_count() {
			return page_count;
		}

#ifdef COSIM_STUB                                   
		int register_addr(target_addr_t addr);                      
#endif
			   
	private:

		void _reset();

		static halfword_t swap_half_word(halfword_t val) {
			return (val>>8) | (val<<8);
		}

		static word_t swap_word(word_t val) {
			return (val>>24) | ((val>>8)&0xFF00) |
				   ((val&0xFF00)<<8) |  (val<<24) ;
		}

		static dword_t swap_dword(dword_t val) {
			return ((dword_t)swap_word((word_t)val)<<32) |
				    (dword_t)swap_word((word_t)(val>>32));
		}

		static UInt32 hash1(target_addr_t addr)
		{
			return (addr / MEMORY_PAGE_SIZE) % PRIMARY_MEMORY_HASH_TABLE_SIZE;
		}

		static UInt32 hash2(target_addr_t addr)
		{
			return (addr / MEMORY_PAGE_SIZE / PRIMARY_MEMORY_HASH_TABLE_SIZE) %
			   SECONDARY_MEMORY_HASH_TABLE_SIZE;
		}

		memory_page_table_entry_t *allocate_page(target_addr_t addr);

		memory_page_table_entry_t *search_page(target_addr_t addr)
		{
			UInt32 h1;
			UInt32 h2;
			secondary_memory_hash_table_t *secondary_hash_table;
	
			addr = addr - (addr % MEMORY_PAGE_SIZE);
	
			h1 = hash1(addr);
			secondary_hash_table = primary_hash_table[h1];
	
			if(secondary_hash_table)
			{
				memory_page_table_entry_t *pte;
				h2 = hash2(addr);
				pte = secondary_hash_table->pte[h2];
		
				while(pte)
				{
					if(pte->addr == addr) return pte;
					pte = pte->next;
				}
			}
			return 0;
		}

		memory_page_table_entry_t *get_page(target_addr_t addr)
		{
			memory_page_table_entry_t *pte;
	
			addr = addr - (addr % MEMORY_PAGE_SIZE);
			pte = search_page(addr);
			if(!pte)
				pte = allocate_page(addr);

			return pte;
		}

		secondary_memory_hash_table_t 
			*primary_hash_table[PRIMARY_MEMORY_HASH_TABLE_SIZE];

		unsigned int page_count;	/*stats information */

		/* for possibly sequential reads by read_word_fast */
		target_addr_t cached_addr;
		memory_page_table_entry_t *cached_pte;

};

} /* namespace */

#endif





