
#ifndef __ARMSIMUL_H__
#define __ARMSIMUL_H__

#include <list>
#include "bittypes.h"
#include "object_pool.hpp"

class _simulator {

	enum _machine_id {
		_mach_dummy,
		#define _MACH_DEF(_a,_b) _mach_ ## _b,
		#include "ns_mach_list.hpp"
		_mach_total,
	};

	enum _operation_id {
		_oper_dummy,
		#define _OPER_DEF(_a,_b,_c) _oper_ ## _b,
		#include "ns_oper_list.hpp"
		_oper_total,
	};

	public:

		/** Constructor.
		 *  @param verbose Verbose mode
		 */
		_simulator(bool verbose);

		/** Destructor. */
		~_simulator();

		/** simulate for n cycles.
		 *  @param n The maximum number of cycles to simulate (may exit).
		 */
		void run(uint64_t n);

		/** simulate one cycle. */
		void clock_tick();

		/** stop running. 
		 *  Called by the signal handler or syscall interpreter.
		 */
		void stop() {running = false;}

		/** See if the simulator is running.
		 */
		bool is_running() const {return running;}

		/** reset */
		void reset();

		uint64_t get_cycle_count() const {return cycle_count;}

	private:

		/* status flags */
		bool running;

		uint64_t cycle_count;	/* number of simulation cycles*/

		std::list<_base_machine_ *> work_list;
		std::list<_base_machine_ *>::iterator work_it, ins_it;

	public:
		/* syscall data */
		bool verbose;

		/* memory management related members */

	public:

		uint64_t alloc_count;	/* number of machine allocations. */

		uint64_t retire_count;	/* number of machine retires. */

                uint64_t max_osm_count; /* max number of machines each cycle */

		/* Allocate a machine of given id type. */
		_base_machine_ *allocate_machine(_MACH_ID_T id);

		/* Allocate an operation of given id type. */
		_base_operation_ *allocate_operation(_OPER_ID_T id);

		/* Free a machine, flush is true if this is flushed.
		 * This is the sink of OSMs. */
		void sink_machine(_base_machine_ *mach, bool flush);

		/* This sets the initial state of the simulator
		 */
		void initialize();

		/** Activate a new operation.
		 *  @param oper The activated operation.
		 *  @param ids The ids of the machine.
		 *  @param cdg The coding of the new operation.
		 */
		void activate_operation(_base_operation_ * oper,
				_OPER_ID_T *ids, void *cdg);

		void dump_stats();

		uint64_t get_cycle_count() {return cycle_count;}

	private:

		/* Free an operation object. */
		void free_operation(_base_operation_ *oper) {
			operation_pool->reclaim(oper);
		}

		static _base_machine_ *mach_allocator(_simulator *, _MACH_ID_T id);
		static _base_operation_ *oper_allocator(_simulator *, _OPER_ID_T id);

		/* storage for machine objects. */
		obj_pool<_base_machine_, _MACH_ID_T, _simulator,
	   		_simulator::mach_allocator> *machine_pool;

		/* storage for operation objects. */
		obj_pool<_base_operation_, _OPER_ID_T, _simulator,
	  	 	_simulator::oper_allocator> *operation_pool;

		/*hardware units */

	public:

		/* the token managers */
#define MANG_DEF(cname, iname) cname *iname;
#include "ns_mang_list.hpp"
         

};

#endif
