
#include "simulator.hpp"

#include "opers.hpp"
#include "machines.hpp"
#include <vector>
#include <iostream>

#include <cstdio>
#include <csignal>
#include <cstring>
#include <cstdlib>
#include <string>
#include <fstream>
#include "config.h"

using std::vector;
using std::list;
using std::cerr;
using std::endl;
using std::ostream;
using std::ofstream;
using std::string;

plx_simulator::plx_simulator(bool verbose) :
	verbose(verbose), running(false), pmem(), dmem()
{
	/* create the hardware modules */
	biu = new BIU("biu");
	icache = new ICache("icache", *biu);
	itlb = new ITLB("itlb", *biu);
	dcache = new DCache("dcache", *biu);
	dtlb = new DTLB("dtlb", *biu);

	/* Create the managers */
#define MANG_DEF(cname, iname) iname = new cname(#iname);
#include "mang_list.hpp"

	/* conenct the managers. */
	mIF->connect(mBranch, icache, itlb, &pmem);
	mMemCtrl->connect(dcache, dtlb, mMemRead, mMemWrite, &dmem, mMemAddr);
	
	/* storage for operation/machine */
	machine_pool = new obj_pool<_base_machine_, _MACH_ID_T, plx_simulator,
								plx_simulator::mach_allocator>(_mach_total);
	operation_pool = new obj_pool<_base_operation_, _OPER_ID_T, plx_simulator,
								plx_simulator::oper_allocator>(_oper_total);
	

	/* reset every thing */
	reset();
}

plx_simulator::~plx_simulator()
{
	delete machine_pool;
	delete operation_pool;

#define MANG_DEF(cname, iname) delete iname;
#include "mang_list.hpp"

}

void plx_simulator::reset()
{
	pmem.reset();
	dmem.reset();
	icache->reset();
	dcache->reset();
	itlb->reset();
	dtlb->reset();
	biu->reset();

#define MANG_DEF(cname, iname) iname->reset();
#include "mang_list.hpp"

	reset_stats();
}

void plx_simulator::reset_stats()
{
	cycle_count = 0;
	alloc_count = 0;
	retire_count = 0;
}

void plx_simulator::dump_stats(FILE *fp)
{
	icache->PrintStats(fp);
	itlb->PrintStats(fp);
	dcache->PrintStats(fp);
	dtlb->PrintStats(fp);
	biu->PrintStats(fp);

	fprintf(fp, "Total allocated OSMs : %lld\n", alloc_count);
	fprintf(fp, "Total retired OSMs   : %lld\n", retire_count);

	fprintf(fp, "Total cycles         : ");
	dump_int64(cycle_count, fp);
	fprintf(stderr, "\n");
	fprintf(stderr, "Total data memory pages allocated : %d\n",
		 dmem.get_page_count());
	fprintf(stderr, "Total inst. memory pages allocated : %d\n",
		 pmem.get_page_count());

	fprintf(stderr, "pc = 0x%08x\n", mIF->get_pc());
}

void plx_simulator::dump_iregs(ostream& os)
{
	for (unsigned i=0; i<16; i++) {
		for (unsigned j=2*i; j<2*i+2; j++) {
			string rval = mRF->get_value(j).hex();
			rval.resize(34,' ');
			os << "r" << j;
		   	if (j<10) os << ' '; 
			os	<< " = " << rval;
		}
		os << endl;
	}
}

void plx_simulator::dump_fregs(ostream& os)
{
	for (unsigned i=0; i<16; i++) {
		for (unsigned j=2*i; j<2*i+2; j++) {
			string rval = mFRF->get_value(j).hex();
			rval.resize(34,' ');
			os << "f" << j;
		   	if (j<10) os << ' '; 
			os	<< " = " << rval;
		}
		os << endl;
	}
}

void plx_simulator::dump_preds(ostream& os)
{
	os << "Predicates: ";
	for (unsigned i=0; i<8; i++) {
		os << mPF->get_value(i).val() << " ";
	}
	os << endl;
}

void plx_simulator::dump_pmem(ostream&os, uint32_t addr, uint32_t size)
{
	pmem.dump(os, addr, size);
}

void plx_simulator::dump_dmem(ostream&os, uint32_t addr, uint32_t size)
{
	dmem.dump(os, addr, size);
}

void plx_simulator::dump_core(ostream& os)
{
	ofstream  ofs;
	ofs.open("core.plx");

	dmem.dump(ofs);
	ofs.close();
}

void plx_simulator::clock_tick()
{

#ifdef DEBUG_CYCLE
		fprintf(stderr, "--clock %d--\n", cycle_count);
#endif

		/** Activate all OSMs in order. */
		for (work_it=work_list.begin(); work_it!=work_list.end();) {

			ins_it = work_it;
			ins_it++;
#if defined DEBUG_CYCLE 
			int oldstate = (*work_it)->__get_state();
#endif
			bool ret = (*work_it)->__activate();
#ifdef DEBUG_CYCLE
			std::cerr << (*work_it)->__get_dynid() << " " << oldstate << "->";
			std::cerr << (*work_it)->__get_state() << " " << ret << std::endl;
#endif

			/* Test if the OSM is back to the initial state. */
			if (ret) { /* no */
				work_it++;
			}
			else { /* yes */

#if 0 
				//fprintf(stderr, "-%d-", oldstate);
				if (oldstate>3 && oldstate!=8)
					fprintf(stderr, "0x%08x 0x%08x 0x%08x\n",
						static_cast<MC_normal *>(*work_it)->G_iw.val(),
						static_cast<MC_normal *>(*work_it)->G_v_iflag.val(),
						static_cast<MC_normal *>(*work_it)->G_v_oflag.val());
#endif

				/* send retired ones to the sink. */
				sink_machine(*work_it);
				work_it = work_list.erase(work_it);
			}
		}

		cycle_count++;

#define MANG_DEF(cname, iname) iname->update_on_clock();
#include "mang_list.hpp"

		icache->updateOnClockTick();
		itlb->updateOnClockTick();
		dcache->updateOnClockTick();
		dtlb->updateOnClockTick();
		biu->updateOnClockTick();
}

void plx_simulator::run(UInt64 maxCycle)
{
	_base_machine_ *src = source_machine();
	work_list.push_back(src);

	running = true;
	while(running && cycle_count<=maxCycle) {
		clock_tick();
	}
}

void plx_simulator::activate_operation(_base_operation_ *oper,
	_OPER_ID_T *ids, void *cdg)
{
	_base_machine_ *mach = allocate_machine(oper->__get_mach_id());
	oper->__init(mach, ids, cdg);

	work_list.insert(ins_it, mach);
}

_base_machine_ *plx_simulator::allocate_machine(_MACH_ID_T id)
{
	_base_machine_ *mach = machine_pool->allocate(this,id);
	mach->__set_dynid(alloc_count++);
	return mach;
}

#if 0
_base_operation_ *plx_simulator::allocate_operation(_OPER_ID_T id)
{
	return operation_pool->allocate(this,id);
}
#endif

void plx_simulator::sink_machine(_base_machine_ *mach)
{
	_base_operation_ **opers;
	unsigned ocount = mach->__get_opers(opers);

	for (unsigned i=0; i<ocount; i++)
		free_operation(opers[i]);

	retire_count++;
	mach->__retire();
#if 0
	fprintf(stderr, "0x%08x 0x%08x\n",
		 static_cast<MC_normal *>(mach)->G_iw.val(),
		 static_cast<MC_normal *>(mach)->G_v_iflag.val());
	fprintf(stderr, "0x%08x : 0x%08x %lld\n",
		 static_cast<normal *>(mach)->pc.val(),
		 static_cast<normal *>(mach)->iw.val(), mach->get_dynid());
	fprintf(stderr, "%lld retire\n", mach->get_dynid());
#endif

	machine_pool->reclaim(mach);
}

_base_machine_ *plx_simulator::source_machine()
{
	_base_operation_ *oper = allocate_operation(_oper_fetch);
	_base_machine_ *src = allocate_machine(_mach_normal);
	oper->__init(src, NULL, NULL);

	return src;
}

_base_machine_ *plx_simulator::mach_allocator(plx_simulator *s, _MACH_ID_T id)
{
  switch (id) {
#define _MACH_DEF(_a, _b) \
	case _mach_ ## _b: return new _MACH_CLASS(_b)(s);
#include "mach_list.hpp"
	default:
      assert(0);
  }
}

_base_operation_ *plx_simulator::oper_allocator(plx_simulator *s, _OPER_ID_T id)
{
  switch (id) {
#define _OPER_DEF(_a, _b, _c) \
	case _oper_ ## _b: return new _OPER_CLASS(_b)();
#include "oper_list.hpp"
    default:
      assert(0);
  }
}

