
#ifndef __PLXSIMUL_H__
#define __PLXSIMUL_H__

#include <list>
#include <iostream>
#include "emumem.h"
#include "icache.h"
#include "itlb.h"
#include "dcache.h"
#include "dtlb.h"

#include "more_managers.hpp"
#include "object_pool.hpp"

class plx_simulator {

	enum _machine_id {
		_mach_dummy,
		#define _MACH_DEF(_a,_b) _mach_ ## _b,
		#include "mach_list.hpp"
		_mach_total,
	};

	enum _operation_id {
		_oper_dummy,
		#define _OPER_DEF(_a,_b,_c) _oper_ ## _b,
		#include "oper_list.hpp"
		_oper_total,
	};

	public:

		/** Constructor.
		 *  @param verbose Verbose mode
		 */
		plx_simulator(bool verbose);

		/** Destructor. */
		~plx_simulator();

		/** simulate for n cycles.
		 *  @param n The maximum number of cycles to simulate (may exit).
		 */
		void run(UInt64 n);

		/** simulate one cycle. */
		void clock_tick();

		/** stop running. 
		 *  Called by the signal handler or syscall interpreter.
		 */
		void stop() {running = false;}

		/** See if the simulator is running.
		 */
		bool is_running() const {return running;}

		/** stats output. */
		void dump_stats(FILE *fp);

		/** dump integer registers. */
		void dump_iregs(std::ostream& os);

		/** dump floating-point registers. */
		void dump_fregs(std::ostream& os);

		/** dump predicate registers. */
		void dump_preds(std::ostream& os);

		/** dump prog memory . */
		void dump_pmem(std::ostream& os, uint32_t start_addr, uint32_t length);

		/** dump data memory . */
		void dump_dmem(std::ostream& os, uint32_t start_addr, uint32_t length);

		/** dump all data memory . */
		void dump_core(std::ostream& os);

		/** stats resetting. */
		void reset_stats();

		/** load the program into memory. */
		void load_program(const char *);

		/** reset */
		void reset();

#if 0
		/** Get the value of a register. For debugging use. */
		word_t read_gpr(int ind) const {
			return mRF->get_value(ind).val();
		}

		/** These two hooks for the syscall interpretor. */
		void write_gpr(int ind, word_t val) {
			mRF->set_value(ind, val);
		}
#endif

		uint64_t get_cycle_count() const {return cycle_count;}


	private:
		bool verbose;

		/* status flags */
		bool running;

		uint64_t cycle_count;	/* number of simulation cycles*/

		friend void trap_emul(plx_simulator *,
						const _UINT_T(32)& pc, const _UINT_T(23)& num);

		std::list<_base_machine_ *> work_list;
		std::list<_base_machine_ *>::iterator work_it, ins_it;

	public:

		uint64_t alloc_count;	/* number of machine allocations. */
		uint64_t retire_count;	/* number of machine retires. */


		/* Allocate a machine of given id type. */
		_base_machine_ *allocate_machine(_MACH_ID_T id);

		/* Allocate an operation of given id type. */
		_base_operation_ *allocate_operation(_OPER_ID_T id) {
			return operation_pool->allocate(this,id);
		}

		 /* This is the sink of OSMs. */
		void sink_machine(_base_machine_ *mach);

		/* Get a new machine.
		 * This is the source of OSMs.
		 */
		_base_machine_ *source_machine();

		/** Activate a new operation.
		 *  @param oper The activated operation.
		 *  @param ids The ids of the machine.
		 *  @param cdg The coding of the new operation.
		 */
		void activate_operation(_base_operation_ * oper,
				_OPER_ID_T *ids, void *cdg);

	private:

		/* Free an operation object. */
		void free_operation(_base_operation_ *oper) {
			operation_pool->reclaim(oper);
		}

		static _base_machine_ *mach_allocator(plx_simulator *, _MACH_ID_T id);
		static _base_operation_ *oper_allocator(plx_simulator *, _OPER_ID_T id);

		/* storage for machine objects. */
		obj_pool<_base_machine_, _MACH_ID_T, plx_simulator,
	   		plx_simulator::mach_allocator> *machine_pool;

		/* storage for operation objects. */
		obj_pool<_base_operation_, _OPER_ID_T, plx_simulator,
	  	 	plx_simulator::oper_allocator> *operation_pool;

		/* memory management related members end/ */

		/*hardware units */

	public:

		/* caches, TLBs & bus interface unit, modeling timing only */
		ICache *icache;
		DCache *dcache;
		ITLB *itlb;
		DTLB *dtlb;
		BIU *biu;

		/* the token managers */
#define MANG_DEF(cname, iname) cname *iname;
#include "mang_list.hpp"

	public:

		/* memory saving the real data value */
		emulator::memory pmem, dmem;

};

#endif
