
#ifndef __MORE_MANAGERS_HPP__
#define __MORE_MANAGERS_HPP__

/* include libosm stuff */
#include "include_osm.hpp"
#include "icache.h"
#include "itlb.h"
#include "dcache.h"
#include "dtlb.h"
#include "emumem.h"
#include <string>

/** short hand */
typedef class emulator::memory _MEM;

/** The register file + forwarding path manager. */
template <class I, class T, unsigned N>
class regval_manager {

	public:

		/** token type. */
		typedef class _token_<regval_manager<I, T, N> > TOKEN_T;

		/** Constructor.
		 *  @param name Name of the register file
		 */
		regval_manager(const std::string& name) : name(name) {

			tokens = new TOKEN_T[N](this);
			values = new T[N];
		}

		/** Destructor. */
		~regval_manager() {
			delete [] tokens;
			delete [] values;
		}

		/** See if token is available for allocation.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_alloc(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** See if token is available for inquire.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_free_for_inquire(const I& ind, _BASE_MACHINE *obj) {
			return true;
		}

		/** See if it is ok to release token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param ind The index of the token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(const I& ind, _BASE_MACHINE *obj) {
			return tokens+ind.val();
		}

		/** Inquire a token, do nothing for now. */
		void inquire_token(const I& ind, _BASE_MACHINE *obj) {}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}


		/** Access the value of the token.
		 * This corresponds to val = *m[ind] in OSM expression.
		 * @param ind The index of the token.
		 * @param obj The OSM reading the value.
		 */
		T read_token(const I& ind, _BASE_MACHINE *obj) {
			return values[ind.val()];
		}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {
			return values[token-tokens];
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[ind] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param ind The index of the token.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const I& ind, const T& val, _BASE_MACHINE *obj) {
			if (ind.val()) values[ind.val()] = val;
		}													 

		/** Update the value of the token.
		 * This function is called from the token. This corresponds
		 * to *token = val in OSM expression.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 * @param token The token to be updated.
		 */
		void write_token(const T& val, _BASE_MACHINE *obj, TOKEN_T *token) {
			if (token!=tokens) values[token-tokens] = val;
		}


		/** Get the value of a register.
		 *  This is used for value access from the hardware layer.
		 *  @param ind Index of the token.
		 */
		T get_value(int ind) {
			return values[ind];
		}

		/* Set the value of a register.
		 * This is used for value access from the hardware layer. So
		 * this does not monitor the value of pc.
		 * @param ind Index of the token.
		 * @param val Value to write.
		 */
		void set_value(int ind, const T& val) {
			values[ind] = val;
		}

		/** Resetting states. */
		void reset() {
			for (unsigned i=0; i<N; i++) {
				tokens[i].set_unused();
				values[i] = 0;
			}
		}

	private:

		std::string name;

		TOKEN_T *tokens;
		T *values;

};

class branch_manager {

	public:

		/** Token value type. */
		typedef _UINT_T(32) TVAL;

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		branch_manager(const std::string& name) :
			name(name), trigger(false) {}

		/** Destructor. */
		~branch_manager() {}

		/** See if the token is good for allocation, dummy, always true.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {
			return false;
		}

		/** See if the token is good for allocation, dummy, always true.
		 *  Called by writers.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return true;
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Inquire a token, do nothing for now. */
		void inquire_token(_BASE_MACHINE *obj) {}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const TVAL& val, _BASE_MACHINE *obj) {
			target = (uint32_t)(val.val());
			trigger = true;
#ifdef DEBUG
			fprintf(stderr, "branch target: 0x%08x\n", target);
#endif
		}

		/** queries from the ifetcher */
		bool branch_taken() const {return trigger;}
		uint32_t get_target() const {return target;}


		/** Reset states. */
		void reset() {
			trigger = false;
		}

	private:

		std::string name;
		bool trigger;
		uint32_t target;

};

/** fetching manager. */
class fetch_manager {

	public:

		/** The token type. */
		typedef class _token_<fetch_manager> TOKEN_T;

		typedef _TUPLE_T(_UINT_T(32), _UINT_T(32)) T;

		/** The constructor.
		 *  @param name Name of the manager.
		 */
		fetch_manager(const std::string& name) :
			name(name), token(new TOKEN_T(this)) {}

		/** Destructor. */
		~fetch_manager() {
			delete token;
		}


		/** See if the token is available for allocation.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return true;
		}

		/** See if the token is available for inquire.
		 *  @param obj The requesting OSM.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {
			return true;
		}

		/** See if the token is available for release.
		 *  @param obj The requesting OSM.
		 */
		bool token_good_to_reclaim(TOKEN_T *token, _BASE_MACHINE *obj) {
			return true;
		}

		/** Allocate a token.
		 *  @param obj The OSM sending the request.			
		 */
		TOKEN_T *allocate_token(_BASE_MACHINE *obj) {
			return token;
		}

		/** Inquire a token, do nothing for now. */
		void inquire_token(_BASE_MACHINE *obj) {}

		/** Release a token.
		 *  @param token The token to release.
		 *  @param obj The OSM sending the request.			
		 */
		void reclaim_token(TOKEN_T *token, _BASE_MACHINE *obj) {}

		/** Discard a token.
		 *  @param token The token to discard.
		 *  @param obj The OSM sending the request.			
		 */
		void discard_token(TOKEN_T *token, _BASE_MACHINE *obj) {}

		/** Get the name of the manager. */
		const std::string& get_name() {return name;}

		/** Access the value of the token.
		 * This corresponds to val = *token in OSM expression.
		 * @param obj The OSM reading the value.
		 */
		T read_token(_BASE_MACHINE *obj, TOKEN_T *token) {

			if (mBM->branch_taken()) {
				current_pc = mBM->get_target();
				mBM->reset();
			}
			else {
				current_pc += 4;
			}

			unsigned iw = mem->read_word_fast(current_pc);
#ifdef DEBUG
			fprintf(stderr, "0x%08x : 0x%08x\n", current_pc, iw);
#endif
			return _TUPLE_CTR(_UINT_T(32), _UINT_T(32), current_pc, iw);
		}

		/** Connect the fetcher with other modules.
		 *  @param mrf  The register file manager, for program counter.
		 *  @param csh  The instruction cache object.
		 *  @param tlb  The instruction tlb object.
		 *  @param imm  The memory to get instrution from.
		 */
		void connect(branch_manager *mrs, _MEM *imm) {
			mBM = mrs;
			mem = imm;
		}

		/** Set the program counter. */
		void set_pc(uint32_t npc) {
			current_pc = npc;
		}

		uint32_t get_pc() const {return current_pc;}

		/** Reset the states/ */
		void reset() {
			token->set_unused();
		}

	private:

		std::string name;
		TOKEN_T *token;

		unsigned current_pc;	/* the pc currently fetching */

		/* connected components. */
		branch_manager *mBM;
		_MEM *mem;
};

typedef class unit_value_manager<_UINT_T(128)> mem_addr_port;
typedef class unit_value_manager<_UINT_T(128)> mem_read_port;
typedef class unit_value_manager<_UINT_T(128)> mem_write_port;

/** The memory control port.
 *  This port does nothing but to synchronize data memory access.
 *  An OSM should inquire this port to see if the memory request finishes.
 */
class mem_ctrl_port {

	public:

		/** Constructor.
		 *  @param name Name of the manager.
		 */
		mem_ctrl_port(const std::string& name) :
			name(name), status(0), trigger(false) {}

		/** Destructor. */
		~mem_ctrl_port() {}

		/** See if token is availble for inquire.
		 *  Equivalent to if the memory access has finished.
		 *  @param obj The object performing the memory transaction.
		 */
		bool token_free_for_inquire(_BASE_MACHINE *obj) {

			if (trigger) {
				addr = mMemAddr->get_value().val();
			}

			if (isRead) {
				if (size==16) {
					_UINT_T(128) result;
					dword_t vlo = mem->read_dword(addr);
					dword_t vhi = mem->read_dword(addr+8);
#if TARGET_LITTLE_ENDIAN //vlo has low order bits
					result = vhi;
					result = (result << 64) | _UINT_CTR(128, vlo);
#else
					result = vlo;
					result = (result << 64) | _UINT_CTR(128, vhi);
#endif
					mMemRead->set_value(result);
				}
				else if (size==8) {
					mMemRead->set_value(mem->read_dword(addr));
				}
				else if (size==4) {
					mMemRead->set_value(mem->read_word(addr));
				}
				else if (size==2) {
					mMemRead->set_value(mem->read_half_word(addr));
				}
				else if (size==1) {
					mMemRead->set_value(mem->read_byte(addr));
				}
				else {
					std::cerr << "unimplemented memory read size "
						  	<<  size << std::endl;
					abort();
				}
			}
			else {
				if (size==16) {
					_UINT_T(128) result = mMemWrite->get_value();
					dword_t vlo = (dword_t)result.val();
					dword_t vhi = (dword_t)(result>>64).val();
#if TARGET_LITTLE_ENDIAN //vlo has low order bits
					mem->write_dword(addr, vlo);
					mem->write_dword(addr+8, vhi);
#else
					mem->write_dword(addr, vhi);
					mem->write_dword(addr+8, vlo);
#endif
				}
				else if (size==8) {
					mem->write_dword(addr,
						(uint64_t)mMemWrite->get_value().val());
				}
				else if (size==4) {
					mem->write_word(addr,
						(uint32_t)mMemWrite->get_value().val());
				}
				else if (size==2) {
					mem->write_half_word(addr, 
						(uint16_t)(uint32_t)mMemWrite->get_value().val());
				}
				else if (size==1) {
					mem->write_byte(addr,
						(uint8_t)(uint32_t)mMemWrite->get_value().val());
				}
				else {
					std::cerr << "unimplemented memory write size "
					   	<< size << std::endl;
					abort();
				}
			}
		}

		/** See if the token is available for allocate, dummy always true. */
		bool token_free_for_alloc(_BASE_MACHINE *obj) {
			return true;
		}

		/** Inquire a token, do nothing for now. */
		void inquire_token(_BASE_MACHINE *obj) {
			addr += 4;
			trigger = false;
		}

		/** Update the value of the token.
		 * This function is called from the OSM. This
		 * corresponds to *m[] = val in OSM expression.
		 * To prevent double write in the same cycle, we allocate_temp first.
		 * @param val The updated value.
		 * @param obj The OSM doing the update.
		 */
		void write_token(const _UINT_T(32)& val, _BASE_MACHINE *obj) {
			isRead = 1 - (val.val() & 1);
			size = (val.val()>>1)&31;//16 bytes at most
			trigger = true;
		}

		/** Get the name of the manager. */
		const std::string& get_name() {
			return name;
		}

		/** Connect the module with other modules.
		 *  @param csh  The data cache object.
		 *  @param tlb  The data tlb object.
		 *  @param rpt  The memory read  port.
		 *  @param wpt  The memory write port.
		 *  @param dmm  The data memory containing real data.
		 *  @param map  The memory address port.
		 */
		void connect(mem_read_port *rpt, mem_write_port *wpt, _MEM *dmm,
			mem_addr_port *map) {
			mMemRead = rpt;
			mMemWrite = wpt;
			mem = dmm;
			mMemAddr = map;
		}

		/** reset states. */
		void reset() {
			status = 0;
		}

	private:

		std::string name;

		unsigned status;
		unsigned addr;
		unsigned size;
		bool isRead;
		bool trigger;

		mem_read_port *mMemRead;
		mem_write_port *mMemWrite;
		mem_addr_port *mMemAddr;
		_MEM *mem;

};

typedef unit_manager<int> simple_resource;
typedef latency_manager<_UINT_T(32)> delay_manager;
typedef regval_manager<_UINT_T(5), _UINT_T(128), 32> regfile_manager;
typedef regval_manager<_UINT_T(3), _UINT_T(1), 8> pred_manager;


#endif
