
#include "emumem.h"

#include <ios>
#include <cstring>

using namespace emulator;
using std::ostream;
using std::endl;
using std::ios_base;

memory::memory()
{
	memset(primary_hash_table, 0, sizeof(primary_hash_table));
	page_count = 0;

	cached_pte = get_page(0);
	cached_addr = 0;
}

memory::~memory()
{
	reset();
}

memory::memory(const memory& mem)
{
	unsigned int i, j;

	memset(primary_hash_table, 0, sizeof(primary_hash_table));
	page_count = 0;

	for(i = 0; i < PRIMARY_MEMORY_HASH_TABLE_SIZE; i++)
	{
		const secondary_memory_hash_table_t *secondary_hash_table =
		   mem.primary_hash_table[i];
		
		if(secondary_hash_table)
		{
			secondary_memory_hash_table_t *secondary_hash_table_cpy = 
				new secondary_memory_hash_table_t;
			memset(secondary_hash_table_cpy, 0,
				sizeof(secondary_memory_hash_table_t));
			primary_hash_table[i] = secondary_hash_table_cpy;

			for(j = 0; j < SECONDARY_MEMORY_HASH_TABLE_SIZE; j++)
			{
				const memory_page_table_entry_t *pte = 
					secondary_hash_table->pte[j];

				while (pte)
				{
					memory_page_table_entry_t *pte_cpy =
						new memory_page_table_entry_t;
					pte_cpy->addr = pte->addr;
					pte_cpy->storage = new byte_t[MEMORY_PAGE_SIZE];
					memcpy(pte_cpy->storage, pte->storage, MEMORY_PAGE_SIZE);
					pte_cpy->next = secondary_hash_table_cpy->pte[j];
					secondary_hash_table_cpy->pte[j] = pte_cpy;
					page_count++;

					pte = pte->next;
				}
			}
		}
	}
	cached_addr = mem.cached_addr;
	cached_pte = get_page(cached_addr);
}

void memory::reset()
{
	unsigned int i, j;
	
	for(i = 0; i < PRIMARY_MEMORY_HASH_TABLE_SIZE; i++)
	{
		secondary_memory_hash_table_t *secondary_hash_table =
		   primary_hash_table[i];
		
		if(secondary_hash_table)
		{
			for(j = 0; j < SECONDARY_MEMORY_HASH_TABLE_SIZE; j++)
			{
				memory_page_table_entry_t *pte = secondary_hash_table->pte[j];
				memory_page_table_entry_t *nextpte;

				while (pte)
				{
					nextpte = pte->next;
					delete [] pte->storage;
					delete pte;
					pte = nextpte;
				}
			}
			delete secondary_hash_table;
		}
	}
	memset(primary_hash_table, 0, sizeof(primary_hash_table));
	page_count = 0;

	cached_pte = get_page(0);
	cached_addr = 0;
}


memory_page_table_entry_t *memory::allocate_page(target_addr_t addr)
{
	UInt32 h1, h2;
	memory_page_table_entry_t *pte;
	secondary_memory_hash_table_t *secondary_hash_table;
		
	h1 = hash1(addr);
	secondary_hash_table = primary_hash_table[h1];
		
	if(!secondary_hash_table)
	{
		secondary_hash_table = new secondary_memory_hash_table_t;
		memset(secondary_hash_table, 0,
			sizeof(secondary_memory_hash_table_t));
		primary_hash_table[h1] = secondary_hash_table;
	}
		
	h2 = hash2(addr);
	pte = new memory_page_table_entry_t;
	pte->addr = addr;
	pte->storage = new byte_t[MEMORY_PAGE_SIZE];
	memset(pte->storage, 0, MEMORY_PAGE_SIZE);
	pte->next = secondary_hash_table->pte[h2];
	secondary_hash_table->pte[h2] = pte;
	page_count++;
	return pte;
}

halfword_t memory::read_half_word(target_addr_t addr)
{
	target_addr_t offset = addr % MEMORY_PAGE_SIZE;
	memory_page_table_entry_t *pte = get_page(addr);
	
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
	return swap_half_word(*((halfword_t *)(pte->storage+offset)));
#else
	return *((halfword_t *)(pte->storage+offset));
#endif

}

void memory::write_half_word(target_addr_t addr, halfword_t value)
{
	target_addr_t offset = addr % MEMORY_PAGE_SIZE;
	memory_page_table_entry_t *pte = get_page(addr);

#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
	*((halfword_t *)(pte->storage+offset)) = swap_half_word(value);
#else
	*((halfword_t *)(pte->storage+offset)) = value;
#endif
}

dword_t memory::read_dword(target_addr_t addr)
{
	target_addr_t offset = addr % MEMORY_PAGE_SIZE;
	memory_page_table_entry_t *pte = get_page(addr);
	
#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
	return swap_dword(*((dword_t *)(pte->storage+offset)));
#else
	return *((dword_t *)(pte->storage+offset));
#endif
}

void memory::write_dword(target_addr_t addr, dword_t value)
{
	target_addr_t offset = addr % MEMORY_PAGE_SIZE;
	memory_page_table_entry_t *pte = get_page(addr);

#if WORDS_BIGENDIAN==TARGET_LITTLE_ENDIAN
	*((dword_t *)(pte->storage+offset)) = swap_dword(value);
#else
	*((dword_t *)(pte->storage+offset)) = value;
#endif
}


void memory::read_block(void *buf, target_addr_t addr, unsigned int size)
{
	if(size > 0)
	{
		UInt32 offset = addr % MEMORY_PAGE_SIZE;
		memory_page_table_entry_t *pte = get_page(addr);
		UInt32 sz = MEMORY_PAGE_SIZE - offset;	
		if(size > sz)
		{
			memcpy(buf, pte->storage + offset, sz);
			size -= sz;
			addr += sz;
			buf = (byte_t *) buf + sz;
			
			if(size >= MEMORY_PAGE_SIZE)
			{
				do
				{
					pte = get_page(addr);
					memcpy(buf, pte->storage, MEMORY_PAGE_SIZE);
					size -= MEMORY_PAGE_SIZE;
					addr += MEMORY_PAGE_SIZE;
					buf = (byte_t *) buf + MEMORY_PAGE_SIZE;
				} while(size >= MEMORY_PAGE_SIZE);
			}
		
			if(size > 0)
			{
				pte = get_page(addr);
				memcpy(buf, pte->storage, size);
			}
		}
		else
		{
			memcpy(buf, pte->storage + offset, size);
		}
	}
}

void memory::write_block(target_addr_t addr, void *buf, unsigned int size)
{
	if(size > 0)
	{
		UInt32 offset = addr % MEMORY_PAGE_SIZE;
		memory_page_table_entry_t *pte = get_page(addr);
		UInt32 sz = MEMORY_PAGE_SIZE - offset;	
		if(size > sz)
		{
			memcpy(pte->storage + offset, buf, sz);
			size -= sz;
			addr += sz;
			buf = (byte_t *) buf + sz;
			
			if(size >= MEMORY_PAGE_SIZE)
			{
				do
				{
					pte = get_page(addr);
					memcpy(pte->storage, buf, MEMORY_PAGE_SIZE);
					size -= MEMORY_PAGE_SIZE;
					addr += MEMORY_PAGE_SIZE;
					buf = (byte_t *) buf + MEMORY_PAGE_SIZE;
				} while(size >= MEMORY_PAGE_SIZE);
			}
		
			if(size > 0)
			{
				pte = get_page(addr);
				memcpy(pte->storage, buf, size);
			}
		}
		else
		{
			memcpy(pte->storage + offset, buf, size);
		}
	}
}

void memory::set_block(target_addr_t addr, byte_t value, unsigned int size)
{
	if(size > 0)
	{
		UInt32 offset = addr % MEMORY_PAGE_SIZE;
		memory_page_table_entry_t *pte = get_page(addr);
		UInt32 sz = MEMORY_PAGE_SIZE - offset;	
		if(size > sz)
		{
			memset(pte->storage + offset, value, sz);
			size -= sz;
			addr += sz;
			
			if(size >= MEMORY_PAGE_SIZE)
			{
				do
				{
					pte = get_page(addr);
					memset(pte->storage, value, MEMORY_PAGE_SIZE);
					size -= MEMORY_PAGE_SIZE;
					addr += MEMORY_PAGE_SIZE;
				} while(size >= MEMORY_PAGE_SIZE);
			}
		
			if(size > 0)
			{
				pte = get_page(addr);
				memset(pte->storage, value, size);
			}
		}
		else
		{
			memset(pte->storage + offset, value, size);
		}
	}
}

target_addr_t memory::align_to_page_boundary(target_addr_t addr)
{
	return addr - (addr % MEMORY_PAGE_SIZE);
}

void memory::dump(ostream& os, target_addr_t addr, unsigned int size)
{
	ios_base::fmtflags old_flag = os.setf(ios_base::hex);
	int old_width = os.width();

	/* round size, addr */
	size = ((size + (addr&15) + 15) >> 4) << 4;
	addr = (addr >> 4) << 4;

	/* truncate the address to word boundary */
	for (unsigned int i = 0; i < size; i+=16) {

		target_addr_t page_addr = (addr+i) - ((addr+i) % MEMORY_PAGE_SIZE);

		os << "0x";
		os.width(8);
		os.fill('0');
	    os << (addr+i) << ":  ";

		if (search_page(page_addr)==NULL) {
			for (unsigned int j=0; j<16; j++) os << "00 ";
		}
		else {
			for (unsigned int j=0; j<16; j++) {
				os.width(2);
				os.fill('0');
			   	os << (word_t)read_byte(addr+i+j) << ' ';
			}
		}
		os << endl;
	}
	os.setf(old_flag);
	os.width(old_width);
}

void memory::dump(ostream& os)
{
	unsigned int i, j;

	for(i = 0; i < PRIMARY_MEMORY_HASH_TABLE_SIZE; i++)
	{
		const secondary_memory_hash_table_t *secondary_hash_table =
		   primary_hash_table[i];
		
		if(secondary_hash_table)
		{
			for(j = 0; j < SECONDARY_MEMORY_HASH_TABLE_SIZE; j++)
			{
				const memory_page_table_entry_t *pte = 
					secondary_hash_table->pte[j];

				while (pte)
				{
					dump(os, pte->addr, MEMORY_PAGE_SIZE);

					pte = pte->next;
					os << endl;
				}
			}
		}
	}
}


#ifdef TEST_MEMORY
bool memory::test()
{
	unsigned int i;
	bool ret = true;

	/* test half_word */
	halfword_t hw;
	for (i=0; i<1024;i+=2)
		write_half_word(i, (halfword_t)i);

	for (i=0; i<1024;i+=2)
		if ((hw=read_half_word(i))!=(halfword_t)i ||
			read_byte(i)!=(byte_t)i ||
			read_byte(i+1)!=(byte_t)(i>>8)) {
			ret = false;
			std::cerr << "halfword test failed at " << i << ", get " << hw
				  << std::endl;
		}

	/* test word */
	word_t w;
	for (i=0; i<1024;i+=4)
		write_word(i, (word_t)i);

	for (i=0; i<1024;i+=4)
		if ((w=read_word(i))!=i ||
			read_byte(i)!=(byte_t)i ||
			read_byte(i+3)!=(byte_t)(i>>24)) {
			ret = false;
			std::cerr << "word test failed at " << i << ", get " << w
				  << std::endl;
		}

	/* test dword */
	dword_t dw;
	for (i=0; i<1024;i+=8)
		write_dword(i, ((dword_t)i<<32)|i);

	for (i=0; i<1024;i+=8)
		if ((dw=read_dword(i))!=(((dword_t)i<<32)|i) ||
			read_byte(i)!=(byte_t)i ||
			read_byte(i+7)!=(byte_t)(i>>24)) {
			ret = false;
			std::cerr << "dword test failed at " << i << ", get " << dw 
				  << std::endl;
		}


	/* test block */
	char hello[] = "hello world!\n";
	char bufff[sizeof(hello)];
	

	write_block(0xffff, hello, sizeof(hello));
	read_block(bufff, 0xffff, sizeof(hello));

	for (i=0; i<sizeof(hello);i+=8)
		if (bufff[i]!=hello[i]) {
			ret = false;
			std::cerr << "block test failed at " << i << ", get " << bufff[i]
				  << std::endl;
		}

	/* test copy constructor using word */
	for (i=0; i<1024;i+=4)
		write_word(i, (word_t)i);

	memory mem(*this);
	for (i=0; i<1024;i+=4)
		if ((w=mem.read_word(i))!=i ||
			mem.read_byte(i)!=(byte_t)i ||
			mem.read_byte(i+3)!=(byte_t)(i>>24)) {
			ret = false;
			std::cerr << "copy test failed at " << i << ", get " << w
				  << std::endl;
		}

	if (mem.page_count != page_count)
		std::cerr << "page count mismatch after copying" << std::endl;

	return ret;
}
#endif

