
#include "annotation.hpp"
#include "mad_dag.hpp"
#include <sstream>

using std::vector;
using std::map;
using std::set;
using std::stringstream;
using std::cerr;
using std::endl;
using std::ostream;
using std::string;
using namespace mad;
using namespace mad_ast;

annot_type_tuple::~annot_type_tuple()
{
	delete_object_list(child_types);
}

annot_value_set::~annot_value_set()
{
	delete_object_list(val);
}

annot_value_tuple::~annot_value_tuple()
{
	delete_object_list(val);
}

annot_statement_command::~annot_statement_command()
{
	delete_object_list(args);
}

annot_paragraph::~annot_paragraph()
{
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		delete (*st_it);
	}
}

void annot_paragraph::append_statement(annot_statement *stm)
{
	statements.push_back(stm);
}

vector<const annot_statement *> annot_paragraph::get_statements() const
{
	return const_vector_transform<annot_statement>(statements);
}

vector<const annot_statement_command *> annot_paragraph::get_commands() const
{
	vector<const annot_statement_command *> result;
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		if ((*st_it)->get_type()==ANNOT_STM_CMD)
			result.push_back(static_cast<const annot_statement_command *>(*st_it));
	}
	return result;
}

vector<const annot_statement_define *> annot_paragraph::get_defines() const
{
	vector<const annot_statement_define *> result;
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		if ((*st_it)->get_type()==ANNOT_STM_DEF)
			result.push_back(static_cast<const annot_statement_define *>(*st_it));
	}
	return result;
}

vector<const annot_statement_relation *> annot_paragraph::get_relations() const
{
	vector<const annot_statement_relation *> result;
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		if ((*st_it)->get_type()==ANNOT_STM_REL)
			result.push_back(static_cast<const annot_statement_relation *>(*st_it));
	}
	return result;
}

vector<const annot_statement_var *> annot_paragraph::get_vars() const
{
	vector<const annot_statement_var *> result;
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		if ((*st_it)->get_type()==ANNOT_STM_VAR)
			result.push_back(static_cast<const annot_statement_var *>(*st_it));
	}
	return result;
}

const annot_statement_command *annot_paragraph::get_command(Symbol name) const
{
	vector<annot_statement *>::const_iterator st_it;
	for (st_it=statements.begin(); st_it!=statements.end(); st_it++) {
		if ((*st_it)->get_type()==ANNOT_STM_CMD) {
			const annot_statement_command *com = 
				static_cast<const annot_statement_command *>(*st_it);
			if (com->get_name()==name) return com;
		}
	}
	return NULL;
}

const annot_argument *annot_statement_command::get_argument(Symbol aname) const
{
	vector<annot_argument *>::const_iterator ait;
	for (ait=args->begin(); ait!=args->end(); ait++) {
		if ((*ait)->get_name()==aname) return *ait;
	}
	return NULL;
}

const annot_value *
	annot_statement_command::get_argument_value(Symbol aname) const
{
	const annot_argument *arg = this->get_argument(aname);
	if (arg) return arg->get_value();
	return NULL;
}

ostream& annot_type_str::print(ostream& os) const 
{
	os << "string";
	return os;
}

ostream& annot_type_int::print(ostream& os) const 
{
	if (sign)
		os << "int<" << width << ">";
	else
		os << "uint<" << width << ">";
	return os;
}

ostream& annot_type_tuple::print(ostream& os) const 
{
	// tuple cannot be empty
	assert(child_types);
	assert(child_types->size()>0);

	os << "(";
	print_object_list(os, ", ", child_types);
	os << ")";
	return os;
}

ostream& annot_type_set::print(ostream& os) const 
{
	os << "{";
	child_type->print(os);
	os << "}";
	return os;
}

ostream& annot_value_ident::print(ostream& os) const 
{
	os << symbol_get_string(ident);
	return os;
}

ostream& annot_value_integer::print(ostream& os) const 
{
	os << val;
	return os;
}

ostream& annot_value_string::print(ostream& os) const 
{
	os << val;
	return os;
}

ostream& annot_value_set::print(ostream& os) const 
{
	os << "{";
	print_object_list(os, ", ", val);
	os << "}";
	return os;
}

ostream& annot_value_tuple::print(ostream& os) const
{
	os << "(";
	print_object_list(os, ", ", val);
	os << ")";
	return os;
}

ostream& annot_argument::print(ostream& os) const 
{
	os << symbol_get_string(name);
	os << "=";
	value->print(os);
	return os;
}

ostream& annot_statement_command::print(ostream& os) const 
{
	os << symbol_get_string(name);
	if (args) {
		os << "(";
		print_object_list(os, ", ", args);
		os << ")";
	}
	return os;
}

ostream& annot_statement_relation::print(ostream& os) const 
{
	left->print(os);
	switch (op) {
		case ANNOT_REL_EQ: os << "=="; break;
		case ANNOT_REL_IN: os << "<:"; break;
	}
	right->print(os);
	return os;
}

ostream& annot_statement_var::print(ostream& os) const 
{
	os << "VAR " << symbol_get_string(name) << ":";
	vtype->print(os);
	return os;
}

ostream& annot_statement_define::print(ostream& os) const
{
	os << "DEFINE " << symbol_get_string(name) << " ";
	val->print(os);
	return os;
}

ostream& annot_paragraph::print(ostream& os) const 
{
	os << ":" << symbol_get_string(label) << ":";
	print_object_list(os, ";\n", &statements);
	return os;
}

ostream& annotation_t::print(ostream& os) const
{
	os << "$$" << str << "$$";
	return os;
}

bool annot_type_str::compatible(const annot_type *op1) const
{
	return op1->get_type()==ANNOT_TYP_STR;
}

bool annot_type_int::compatible(const annot_type *op1) const
{
	return op1->get_type()==ANNOT_TYP_INT;
}

bool annot_type_tuple::compatible(const annot_type *op1) const
{
	if (op1->get_type()==ANNOT_TYP_TUPLE) {

		const annot_type_tuple *tp1 = 
			static_cast<const annot_type_tuple *>(op1);

		if (child_types->size()!=tp1->child_types->size()) return false;

		for (unsigned ii=0; ii<child_types->size(); ii++)
			if (!(*child_types)[ii]->compatible((*tp1->child_types)[ii]))
				return false;

		return true;
	}
	return false;
}

bool annot_type_set::compatible(const annot_type *op1) const
{
	if (op1->get_type()==ANNOT_TYP_SET) {

		const annot_type_set *tp1 = static_cast<const annot_type_set *>(op1);

		/* empty set type is compatible with anything. */
		if (child_type==NULL || tp1->child_type==NULL) return true;

		return child_type->compatible(tp1->child_type);
	}
	return false;
}

annot_type *annot_value_ident::infer_type(const annot_context *ctxt) const
{
	return ctxt->get_symbol_type(ident);
}

annot_value *annot_value_ident::infer_value(const annot_context *ctxt) const
{
	return ctxt->get_define_value(ident);
}

bool annot_value_ident::is_constant(const annot_context *ctxt) const 
{
	return ctxt->has_define(ident);
}

bool annot_value_ident::is_wellformed(const annot_context *ctxt) const
{
	return ctxt->has_variable(ident) || ctxt->has_define(ident);
}

annot_type *annot_value_integer::infer_type(const annot_context *ctxt) const
{
	int width = val->get_width_inbase(2);
	return new annot_type_int(val->is_negative(), width, NULL);
}

annot_value *annot_value_integer::infer_value(const annot_context *ctxt) const
{
	return new annot_value_integer(val, NULL);
}

bool annot_value_integer::is_constant(const annot_context *ctxt) const 
{
	return true;
}

bool annot_value_integer::is_wellformed(const annot_context *ctxt) const
{
	return true;
}

annot_type *annot_value_string::infer_type(const annot_context *ctxt) const
{
	return new annot_type_str(NULL);
}

annot_value *annot_value_string::infer_value(const annot_context *ctxt) const
{
	return new annot_value_string(val, NULL);
}

bool annot_value_string::is_constant(const annot_context *ctxt) const 
{
	return true;
}

bool annot_value_string::is_wellformed(const annot_context *ctxt) const
{
	return true;
}

annot_type *annot_value_set::infer_type(const annot_context *ctxt) const
{
	/* empty set */
	if (val==NULL || val->size()==0) {
		return new annot_type_set(NULL, NULL);
	}

	/* get the types of all its elements */
	vector<annot_type *> inferred_types;
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		inferred_types.push_back((*vit)->infer_type(ctxt));
	}

	if (inferred_types.size()==1) {
		return new annot_type_set(inferred_types.front(), NULL);
	}
	else {
		annot_type *first = inferred_types.front();
		annot_type *typ1, *typ2 = first;
		vector<annot_type *>::iterator tit = inferred_types.begin();
		for (tit++; tit!=inferred_types.end(); tit++) {
			typ1 = common_super_type(typ2, *tit);
			delete *tit;
			delete typ2;
			typ2 = typ1;
		}
		return new annot_type_set(typ2, NULL);
	}
}

annot_value *annot_value_set::infer_value(const annot_context *ctxt) const
{
	vector<annot_value *> *nvals = new vector<annot_value *>;

	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		nvals->push_back((*vit)->infer_value(ctxt));
	}
	return new annot_value_set(nvals, NULL);
}

bool annot_value_set::is_constant(const annot_context *ctxt) const 
{
	if (val==NULL) return true;
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		if (!(*vit)->is_constant(ctxt)) return false;
	}
	return true;
}

bool annot_value_set::is_wellformed(const annot_context *ctxt) const
{
	/* empty set is good. */
	if (val==NULL) return true;

	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		if (!(*vit)->is_wellformed(ctxt)) return false;
	}

	/* get the types of all its elements */
	vector<annot_type *> inferred_types;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		inferred_types.push_back((*vit)->infer_type(ctxt));
	}

	/* check if all the types are mutually compatible. */
	if (inferred_types.size()) {
		bool ret = true;
		vector<annot_type *>::iterator tit = inferred_types.begin();
		for (; tit!=inferred_types.end(); tit++) {

			vector<annot_type *>::iterator sit = tit;
			for (sit++; sit!=inferred_types.end(); sit++) {
				if (!(*tit)->compatible(*sit))
					ret = false;
			}
		}

		for (tit=inferred_types.begin(); tit!=inferred_types.end(); tit++) {
			delete (*tit);
		}
		return ret;
	}

	return true;
}

annot_type *annot_value_tuple::infer_type(const annot_context *ctxt) const
{
	vector<annot_type *> *ntype = new vector<annot_type *>;
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		ntype->push_back((*vit)->infer_type(ctxt));
	}
	return new annot_type_tuple(ntype, NULL);
}

annot_value *annot_value_tuple::infer_value(const annot_context *ctxt) const
{
	vector<annot_value *> *nval = new vector<annot_value *>;
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		nval->push_back((*vit)->infer_value(ctxt));
	}
	return new annot_value_tuple(nval, NULL);
}

bool annot_value_tuple::is_constant(const annot_context *ctxt) const 
{
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		if (!(*vit)->is_constant(ctxt)) return false;
	}
	return true;
}

bool annot_value_tuple::is_wellformed(const annot_context *ctxt) const
{
	vector<annot_value *>::const_iterator vit;
	for (vit=val->begin(); vit!=val->end(); vit++) {
		if (!(*vit)->is_wellformed(ctxt)) return false;
	}
	return true;
}

void annot_value_ident::get_used_symbols(set<Symbol> &syms) const
{
	syms.insert(ident);
	return;
} 

void annot_value_integer::get_used_symbols(set<Symbol> &syms) const
{
	return;
} 

void annot_value_string::get_used_symbols(set<Symbol> &syms) const
{
	return;
}

void annot_value_set::get_used_symbols(set<Symbol> &syms) const
{
	vector<annot_value *>::const_iterator val_it;
	for (val_it=val->begin(); val_it!=val->end(); val_it++) {
		(*val_it)->get_used_symbols(syms);
	}
	return;
}

void annot_value_tuple::get_used_symbols(set<Symbol> &syms) const
{
	vector<annot_value *>::const_iterator val_it;
	for (val_it=val->begin(); val_it!=val->end(); val_it++) {
		(*val_it)->get_used_symbols(syms);
	}
	return;
}

annot_type *mad_ast::common_super_type(const annot_type *op1,
										const annot_type *op2)
{
	if (!op1->compatible(op2)) return NULL;

	if (op1->get_type()==ANNOT_TYP_STR) {
		return new annot_type_str(NULL);
	}
	else if (op1->get_type()==ANNOT_TYP_INT) {
		const annot_type_int *tp1 = static_cast<const annot_type_int *>(op1);
		const annot_type_int *tp2 = static_cast<const annot_type_int *>(op2);
		return new annot_type_int(tp1->is_signed()&&tp2->is_signed(),
			tp1->get_width()>tp2->get_width()?tp1->get_width():tp2->get_width(),
			NULL);
	}
	else if (op1->get_type()==ANNOT_TYP_TUPLE) {
		const annot_type_tuple *tp1 = 
			static_cast<const annot_type_tuple *>(op1);
		const annot_type_tuple *tp2 = 
			static_cast<const annot_type_tuple *>(op2);

		const vector<annot_type *> *typs1 = tp1->get_child_types();
		const vector<annot_type *> *typs2 = tp2->get_child_types();

		vector<annot_type *> *com_types = new vector<annot_type *>();
		for (unsigned ii=0; ii<typs1->size(); ii++) {
			com_types->push_back(common_super_type((*typs1)[ii], (*typs2)[ii]));
		}
		return new annot_type_tuple(com_types, NULL);

	}
	else { //set

		const annot_type_set *tp1 = static_cast<const annot_type_set *>(op1);
		const annot_type_set *tp2 = static_cast<const annot_type_set *>(op2);

		const annot_type *typ1 = tp1->get_child_type();
		const annot_type *typ2 = tp2->get_child_type();
		if (typ1==NULL && typ2==NULL) {
			return new annot_type_set(NULL, NULL);
		}
		else if (typ1==NULL) {
			return new annot_type_set(common_super_type(typ2, typ2), NULL);
		}
		else if (typ2==NULL) {
			return new annot_type_set(common_super_type(typ1, typ1), NULL);
		}
		else {
			return new annot_type_set(common_super_type(typ1, typ2), NULL);
		}
	}
}

annot_context::annot_context(const annot_paragraph *para)
{
	add_paragraph(para);
}

annot_context::annot_context(const annot_owner *owner, Symbol label)
{
	const vector<AnnotParagraph *> *annots = owner->get_parsed_annotations();
	if (annots) {
		vector<AnnotParagraph *>::const_iterator an_it;
		for (an_it=annots->begin(); an_it!=annots->end(); an_it++) {
			if ((*an_it)->label_match(label))
				add_paragraph(*an_it);
		}
	}
}

void annot_context::add_paragraph(const annot_paragraph *para)
{
	if (para==NULL) return;

	vector<const annot_statement_command *> lcmds = para->get_commands();
	vector<const annot_statement_command *>::iterator cmd_it;
	for (cmd_it=lcmds.begin(); cmd_it!=lcmds.end(); cmd_it++) {
		coms.push_back(*cmd_it);
	}

	vector<const annot_statement_relation *> lrels = para->get_relations();
	vector<const annot_statement_relation *>::iterator rel_it;
	for (rel_it=lrels.begin(); rel_it!=lrels.end(); rel_it++) {
		rels.push_back(*rel_it);
	}

	vector<const annot_statement_var *> lvars = para->get_vars();
	vector<const annot_statement_var *>::iterator var_it;
	for (var_it=lvars.begin(); var_it!=lvars.end(); var_it++) {
		vars.push_back(*var_it);
	}

	vector<const annot_statement_define *> ldefines = para->get_defines();
	vector<const annot_statement_define *>::iterator def_it;
	for (def_it=ldefines.begin(); def_it!=ldefines.end(); def_it++) {
		defs.push_back(*def_it);
	}

	type_checked = false;
}

bool annot_statement_command::type_check(const annot_context *ctxt) const
{
	bool status = true;
	vector<annot_argument *>::const_iterator arg_it;
	for (arg_it=args->begin(); arg_it!=args->end(); arg_it++) {
		if (!(*arg_it)->get_value()->is_wellformed(ctxt)) {
			stringstream sstr;
			sstr << "Error: argument value not well-formed"
				<< " at " << *(*arg_it)->get_loc() << endl;
			annot_error(sstr.str());
			status = false;
		}
	}
	return status;
}

bool annot_statement_relation::type_check(const annot_context *ctxt) const 
{
	bool status = true;
	if (!left->is_wellformed(ctxt)) {
		stringstream sstr;
		sstr << "Error: left-hand value not well-formed"
			<< " at " << *left->get_loc() << endl;
		annot_error(sstr.str());
		status = false;
	}
	if (!right->is_wellformed(ctxt)) {
		stringstream sstr;
		sstr << "Error: right-hand value not well-formed"
			<< " at " << *right->get_loc() << endl;
		annot_error(sstr.str());
		status = false;
	}

	if (left->is_constant(ctxt) && right->is_constant(ctxt)) {
		stringstream sstr;
		sstr << "Error: constant on both side of operator "
			<< " at " << *this->get_loc() << endl;
		annot_error(sstr.str());
		status = false;
	}

	if (!status) return status;

	annot_type *left_type = left->infer_type(ctxt);
	annot_type *right_type = right->infer_type(ctxt);

	switch (op) {
		case ANNOT_REL_EQ: 
			if (!left_type->compatible(right_type)) {
				stringstream sstr;
				sstr << "Error: type mismatch between two operands "
					<< " at " << *this->get_loc() << endl;
				annot_error(sstr.str());
				status = false;
			}
			break;
		case ANNOT_REL_IN:
			if (right_type->get_type()!=ANNOT_TYP_SET) {
				stringstream sstr;
				sstr << "Error: right-hand value is not of set type "
					<< " at " << *right->get_loc() << endl;
				annot_error(sstr.str());
				status = false;
			}
			else {
				annot_type_set *right_type_set = static_cast<annot_type_set *>
					(right_type);
				if (!left_type->compatible(right_type_set->get_child_type())) {
					stringstream sstr;
					sstr << "Error: type mismatch between two operands "
						<< " at " << *this->get_loc() << endl;
					annot_error(sstr.str());
					status = false;
				}
			}
			break;
	}

	delete left_type;
	delete right_type;
	return status;
}

bool annot_context::type_check()
{
	if (type_checked) return check_status;

	type_checked = true;
	check_status = true;

	/* first we clear the tables. */
	def_table.clear();
	var_table.clear();

	/* then we set up the table. */
	vector<const annot_statement_var *>::iterator var_it;
	for (var_it=vars.begin(); var_it!=vars.end(); var_it++) {
		add_var(*var_it);
	}

	vector<const annot_statement_define *>::iterator def_it;
	for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {
		add_def(*def_it);
	}

	/* After this point, has_variable and has_define can be called. 
	   get_symbol_type can be called for variables.
	 */

	/* we try to see if defines are well formed -- no cyclic reference,
	   refer to other macros only. */
	mad_dag ref_dag;
	Symbol temp_name = 1;
	for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {
		ref_dag.insert_vertex((*def_it)->get_name());
	}

	for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {

		set<Symbol> used_syms;
		(*def_it)->get_value()->get_used_symbols(used_syms);

		set<Symbol>::iterator set_it;
		for (set_it=used_syms.begin(); set_it!=used_syms.end(); set_it++) {
			ref_dag.insert_edge(temp_name++, (*def_it)->get_name(), *set_it);
		}

	}

	
	if (ref_dag.cycle_check(temp_name)) {
		stringstream sstr;
		sstr << "Error: cyclic reference involving macro "
			 << symbol_get_string(temp_name)
			<< " at " << *(def_table[temp_name]->get_loc()) << endl;
		annot_error(sstr.str());
		check_status = false;
	}

	/* if every thing is fine, then we check constness and well_formedness
	   of defines. */
	if (check_status) {
		/* check if it is constant. */
	    for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {
			if (!(*def_it)->get_value()->is_constant(this)) {

				stringstream sstr;
				sstr << "Error: macro value not constant "
					<< " at " << *(*def_it)->get_loc() << endl;
				annot_error(sstr.str());
				check_status = false;
				break; //stop here since check_status is false, assertion fail
			}
		}
	}

	if (check_status) {
		/* check if it is well-formed. */
	    for (def_it=defs.begin(); def_it!=defs.end(); def_it++) {
			if (!(*def_it)->get_value()->is_wellformed(this)) {
				stringstream sstr;
				sstr << "Error: macro value not well-formed "
					<< " at " << *(*def_it)->get_loc() << endl;
				annot_error(sstr.str());
				check_status = false;
				break;
			}
		}
	}

	if (!check_status) return check_status;

	bool local_status = true;

	/* then we check the commands and relations. */
	vector<const annot_statement_command *>::const_iterator com_it;
	for (com_it=coms.begin(); com_it!=coms.end(); com_it++)
		if (!(*com_it)->type_check(this)) local_status = false;

	vector<const annot_statement_relation *>::const_iterator rel_it;
	for (rel_it=rels.begin(); rel_it!=rels.end(); rel_it++)
		if (!(*rel_it)->type_check(this)) local_status = false;

	check_status = local_status;

	/* if all ok, then it should be fine to call infer_type */
	return check_status;
}

bool annot_context::has_variable(Symbol sym) const
{
	assert(type_checked && check_status);
	return var_table.find(sym)!=var_table.end();
}

bool annot_context::has_define(Symbol sym) const
{
	assert(type_checked && check_status);
	return def_table.find(sym)!=def_table.end();
}

annot_type *annot_context::get_symbol_type(Symbol ident) const 
{
	assert(type_checked && check_status);
	if (this->has_variable(ident)) {
		const annot_type *vtype = 
			(*var_table.find(ident)).second->get_var_type();
		return common_super_type(vtype, vtype);
	}
	else { //define
		assert(this->has_define(ident));
		const annot_value *vvalue = 
			(*def_table.find(ident)).second->get_value();
		return vvalue->infer_type(this);
	}
}

annot_value *annot_context::get_define_value(Symbol ident) const
{
	assert(type_checked && check_status);
	assert(this->has_define(ident));
	const annot_value *vvalue = (*def_table.find(ident)).second->get_value();
	return vvalue->infer_value(this);
}

void annot_context::add_def(const annot_statement_define *def)
{
	Symbol vname = def->get_name();

	if (def_table.find(vname)!=def_table.end()) {
		stringstream sstr;
		sstr << "Error: redefine macro " << 
			symbol_get_string(vname) << " at " << *def->get_loc() << endl;
		sstr << "       previously at " << *def_table[vname]->get_loc() << endl;
		annot_error(sstr.str());
		check_status = false;
	}
	else {
		def_table[vname] = def;
	}
}

void annot_context::add_var(const annot_statement_var *var)
{
	Symbol vname = var->get_name();

	if (var_table.find(vname)!=var_table.end()) {
		stringstream sstr;
		sstr << "Error: redefine variable " << 
			symbol_get_string(vname) << " at " << *var->get_loc() << endl;
		sstr << "       previously at " << *var_table[vname]->get_loc() << endl;
		annot_error(sstr.str());
		check_status = false;
	}
	else {
		var_table[vname] = var;
	}
}

annot_context& annot_context::augment(const annot_paragraph *para)
{
	add_paragraph(para);
	return *this;
}

annot_context& annot_context::augment(const annot_owner *owner, Symbol label)
{
	const vector<AnnotParagraph *> *annots = owner->get_parsed_annotations();
	if (annots) {
		vector<AnnotParagraph *>::const_iterator an_it;
		for (an_it=annots->begin(); an_it!=annots->end(); an_it++) {
			if ((*an_it)->label_match(label))
				add_paragraph(*an_it);
		}
	}
	return *this;
}

annot_context& annot_context::operator += (const annot_context& op2)
{
	rels.insert(rels.end(), op2.rels.begin(), op2.rels.end());
	coms.insert(coms.end(), op2.coms.begin(), op2.coms.end());
	defs.insert(defs.end(), op2.defs.begin(), op2.defs.end());
	vars.insert(vars.end(), op2.vars.begin(), op2.vars.end());

	type_checked = false;

	return *this;
}

annot_context mad_ast::operator + (const annot_context& op1,
									const annot_context& op2)
{
	annot_context a(op1);
	a += op2;
	return a;
}

static int n_annot_errors, n_annot_warns;

void mad_ast::annot_error(const string& msg)
{
	cerr << "Error: " << msg << endl;
	n_annot_errors++;
}

void mad_ast::annot_warn(const string& msg)
{
	cerr << "Warning: " << msg << endl;
	n_annot_warns++;
}

void mad_ast::annot_message(const string &msg)
{
	cerr << msg << endl;
}

