
#ifndef __ANNOTATION_HPP__
#define __ANNOTATION_HPP__

#include <map>
#include <vector>
#include <set>
#include "annot_ast.hpp"
using mad::Symbol;

namespace mad_ast {

	/* error message reporting */
	void annot_warn(const std::string& msg);
	void annot_error(const std::string& msg);
	void annot_message(const std::string& msg);

	/** Annotation data types */
	typedef enum {
		ANNOT_TYP_STR,
		ANNOT_TYP_INT,
		ANNOT_TYP_TUPLE,
		ANNOT_TYP_SET,
	} AnnotTypeType;

	/** Class for annotation data types. */
	typedef class annot_type : public mad_ast::ast_node {

		public:
			/** Constructor.
			 *  @param type The type of the annotation.
			 *  @param loc  Location in file.
			 */
			annot_type(AnnotTypeType type, const Location *loc) :
				ast_node(loc), type(type) {}

			/** Destructor. */
			virtual ~annot_type() {}

			/** Get the type of the annotation. */
			AnnotTypeType get_type() const {return type;}

			/** Check if compatible with another type object. */
			virtual bool compatible(const annot_type *) const = 0;

		private:
			AnnotTypeType type;

	} AnnotType;

	/** Class for string data type. */
	typedef class annot_type_str : public annot_type {

		public:
			/** Constructor.
			 *  @param loc Location in file.
			 */
			annot_type_str(const Location *loc) :
				annot_type(ANNOT_TYP_STR, loc) {}

			/** Destructor. */
			~annot_type_str() {}

			/** Check if compatible with another type object. */
			bool compatible(const annot_type *) const;

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

	} AnnotTypeStr;

	/** Class for integer data type. */
	typedef class annot_type_int : public annot_type {

		public:
			/** Constructor.
			 *  @param sign  If the data type is signed.
			 *  @param width Width of the data type.
			 *  @param loc   Location in file.
			 */
			annot_type_int(bool sign, unsigned width, const Location *loc) :
				annot_type(ANNOT_TYP_INT, loc), sign(sign), width(width) {}

			/** Destructor. */
			~annot_type_int() {}

			/** Check if the type is signed int. */
			bool is_signed() const {return sign;}

			/** Get the width of the data type. */
			unsigned get_width() const {return width;}

			/** Check if compatible with another type object. */
			bool compatible(const annot_type *) const;

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

		private:
			bool sign;
			bool width;

	} AnnotTypeInt;

	/** Class for tuple data type. */
	typedef class annot_type_tuple : public annot_type {

		public:
			/** Constructor.
			 *  @param ts  Type of all the annotations.
			 *  @param loc Location in file.
			 */
			annot_type_tuple(std::vector<annot_type *> *ts,
				const Location *loc) :
				annot_type(ANNOT_TYP_TUPLE, loc), child_types(ts) {}

			/** Destructor. */
			~annot_type_tuple();

			/** Check if compatible with another type object. */
			bool compatible(const annot_type *) const;

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the sub-types. */
			const std::vector<annot_type *> *get_child_types() const {
				return child_types;
			}

		private:
			std::vector<annot_type *> *child_types;

	} AnnotTypeTuple;

	/** Class for set data type. */
	typedef class annot_type_set : public annot_type {

		public:
			/** Constructor.
			 *  @param type The set type.
			 *  @param loc  Location in file.
			 */
			annot_type_set(annot_type *type, const Location *loc) :
				annot_type(ANNOT_TYP_SET, loc), child_type(type) {}

			/** Destructor. */
			~annot_type_set() {delete child_type;}

			/** Check if compatible with another type object. */
			bool compatible(const annot_type *) const;

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the sub-type. */
			const annot_type *get_child_type() const {return child_type;}

		private:
			annot_type *child_type;

	} AnnotTypeSet;

	/** Get the least common super type of two types.
	 *  @return NULL if two are not compatible.
	 */
	annot_type *common_super_type(const annot_type *op1, const annot_type *op2);

	/** Annotation value types. */
	typedef enum {
		ANNOT_VAL_STR,
		ANNOT_VAL_INT,
		ANNOT_VAL_IDENT,
		ANNOT_VAL_SET,
		ANNOT_VAL_TUPLE,
	} AnnotValueType;

	// forward declaration
	class annot_context;

	/** Class for annotation value */
	typedef class annot_value : public ast_node {

		public:

			/** Constructor.
			 *  @param t   Type of the annotation value.
			 *  @param loc Location in the source file.
			 */
			annot_value(AnnotValueType t, const Location *loc) :
					ast_node(loc), type(t) {}

			/** Base class has virtual destructor. */
			virtual ~annot_value() {}

			/** Get the type of the value. */
			AnnotValueType get_type() const {return type;}

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			virtual annot_type *infer_type(const annot_context *ctxt) const = 0;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			virtual annot_value *infer_value(const annot_context *ctxt) const=0;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			virtual bool is_constant(const annot_context *ctxt) const = 0;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			virtual bool is_wellformed(const annot_context *ctxt) const = 0;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			virtual void get_used_symbols(std::set<Symbol> &syms) const = 0;

		protected:
			AnnotValueType type;
			AnnotType *data_type;

	} AnnotValue;


	/** Annotation value identifier */
	typedef class annot_value_ident : public annot_value {

		public:

			/** Constructor.
			 *  @param ident The identifier.
			 *  @param loc   Location in source file.
			 */
			annot_value_ident(Symbol ident, const Location *loc) :
				annot_value(ANNOT_VAL_IDENT, loc), ident(ident) {}

			/** Destructor. */
			~annot_value_ident() {}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the identifier. */
			Symbol get_ident() const {return ident;}

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			annot_type *infer_type(const annot_context *ctxt) const;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			annot_value *infer_value(const annot_context *ctxt) const;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_constant(const annot_context *ctxt) const;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_wellformed(const annot_context *ctxt) const;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			void get_used_symbols(std::set<Symbol> &syms) const;

		private:
			Symbol ident;

	} AnnotValueIdent;

	/** Annotation value integer */
	typedef class annot_value_integer : public annot_value {

		public:
			/** Constructor.
			 *  @param val   The integer value in string.
			 *  @param loc   Location in source file.
			 */ 
			annot_value_integer(num_value *val, const Location *loc) :
				annot_value(ANNOT_VAL_INT, loc), val(val) {}

			/** Destructor. */
			~annot_value_integer() {}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			annot_type *infer_type(const annot_context *ctxt) const;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			annot_value *infer_value(const annot_context *ctxt) const;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_constant(const annot_context *ctxt) const;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_wellformed(const annot_context *ctxt) const;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			void get_used_symbols(std::set<Symbol> &syms) const;

		private:
			num_value *val;

	} AnnotValueInteger;

	/** Annotation value string. */
	typedef class annot_value_string : public annot_value {

		public:
			/** Constructor.
			 *  @param val   The string value.
			 *  @param loc   Location in source file.
			 */ 
			annot_value_string(const std::string& val, const Location *loc) :
				annot_value(ANNOT_VAL_STR, loc), val(val) {}

			/** Destructor. */
			~annot_value_string() {}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the string value. */
			const std::string& get_value() const {return val;}

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			annot_type *infer_type(const annot_context *ctxt) const;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			annot_value *infer_value(const annot_context *ctxt) const;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_constant(const annot_context *ctxt) const;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_wellformed(const annot_context *ctxt) const;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			void get_used_symbols(std::set<Symbol> &syms) const;

		private:
			const std::string val;

	} AnnotValueString;


	/** Annotation value set. */
	typedef class annot_value_set : public annot_value {

		public:
			/** Constructor.
			 *  @param val   The annot_value list.
			 *  @param loc   Location in source file.
			 */ 
			annot_value_set(std::vector<annot_value *> *val,
				const Location *loc) :
				annot_value(ANNOT_VAL_SET, loc), val(val) {}

			/** Destructor. */
			~annot_value_set();

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the values. */
			const std::vector<annot_value *> *get_values() const {return val;}

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			annot_type *infer_type(const annot_context *ctxt) const;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			annot_value *infer_value(const annot_context *ctxt) const;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_constant(const annot_context *ctxt) const;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_wellformed(const annot_context *ctxt) const;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			void get_used_symbols(std::set<Symbol> &syms) const;

		private:
			std::vector<annot_value *> *val;

	} AnnotValueSet;

	/** Annotation value tuple. */
	typedef class annot_value_tuple : public annot_value {

		public:
			/** Constructor.
			 *  @param val   The annot_value list.
			 *  @param loc   Location in source file.
			 */ 
			annot_value_tuple(std::vector<annot_value *> *val,
				const Location *loc) :
				annot_value(ANNOT_VAL_TUPLE, loc), val(val) {}

			/** Destructor. */
			~annot_value_tuple();

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the values. */
			const std::vector<annot_value *> *get_values() const {return val;}

			/** Infer the type under a context.
			 *  Callable only if well_formed.
			 *  @param ctxt The annotation context.
			 */
			annot_type *infer_type(const annot_context *ctxt) const;

			/** Infer the value under a context.
			 *  Callable only if well_formed and is_constant.
			 *  @param ctxt The annotation context.
			 */
			annot_value *infer_value(const annot_context *ctxt) const;

			/** Check constness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_constant(const annot_context *ctxt) const;

			/** Check well-formedness under a context.
			 *  @param ctxt The annotation context.
			 */
			bool is_wellformed(const annot_context *ctxt) const;

			/** Get all used symbols in the value. 
			 *  @param syms The set of symbols to be filled.
			 */
			void get_used_symbols(std::set<Symbol> &syms) const;

		private:

			std::vector<annot_value *> *val;

	} AnnotValueTuple;

	/** Class for annotation argument. */
	typedef class annot_argument : public ast_node {

		public:
			/** Constructor.
			 *  @param name Name of the argument.
			 *  @param val  Value of the argument.
			 *  @param loc  Location in source file.
			 */
			annot_argument(Symbol name, annot_value *val, const Location *loc) :
				ast_node(loc), name(name), value(val) {}

			/** Destructor. */
			~annot_argument() {delete value;}

			/** Get the name of the argument. */
			Symbol get_name() const {return name;}

			/** Get the value of the argument. */
			const annot_value *get_value() const {return value;}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

		private:
			Symbol name;
			annot_value *value;

	} AnnotArgument;

	/** Statement type. */
	typedef enum {
		ANNOT_STM_CMD,
		ANNOT_STM_REL,
		ANNOT_STM_VAR,
		ANNOT_STM_DEF,
	} AnnotStatementType;

	/** A statement in annotation.
	 *  A statement is either a command or a relation expression or
		a variable declaration or a define (like a macro). */
	typedef class annot_statement : public ast_node {

		public:

			/** Constructor.
			 *  @param type The type of the statement.
			 *  @param loc  Location in source code.
			 */
			annot_statement(AnnotStatementType type, const Location *loc) :
				ast_node(loc), type(type) {}

			/** Destructor. */
			virtual ~annot_statement() {}

			/** Get the type of the statement. */
			AnnotStatementType get_type() const {return type;}

		private:
			AnnotStatementType type;

	} AnnotStatement;

	/** A command statement. */
	typedef class annot_statement_command : public annot_statement {

		public:

			/** Constructor.
			 *  @param name Name of the command.
			 *  @param args Arguments of the command.
			 *  @param loc  Location of the command.
			 */
			annot_statement_command(Symbol name,
				std::vector<annot_argument *> *args, const Location *loc) :
				annot_statement(ANNOT_STM_CMD, loc), name(name), args(args) {}

			/** Destructor. */
			~annot_statement_command();

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the name of the command. */
			Symbol get_name() const {return name;}

			/** Get the value of an argument.
			 *  @param arg_name Name of the argument.
			 */
			const annot_value *get_argument_value(Symbol arg_name) const;

			/** Get the argument.
			 *  @param arg_name Name of the argument.
			 */
			const annot_argument *get_argument(Symbol arg_name) const;

		private:
			/** Type check a statement. */
			bool type_check(const annot_context *ctxt) const;

			Symbol name;
			std::vector<annot_argument *> *args;

			friend class annot_context;

	} AnnotStatementCommand;

	typedef enum {
		ANNOT_REL_EQ,
		ANNOT_REL_IN,
	} AnnotRelOp;

	/** A relationship expression. */
	typedef class annot_statement_relation : public annot_statement {

		public:
			/** Constructor.
			 *  @param left Left-hand-side of the expression.
			 *  @param right Right-hand-side of the expression.
			 *  @param op   The operator.
			 *  @param loc  Locaiton of the statement.
			 */
			annot_statement_relation(annot_value *left, annot_value *right,
				AnnotRelOp op, const Location *loc) :
				annot_statement(ANNOT_STM_REL, loc),
				left(left), right(right), op(op) {}

			/** Destructor. */
			~annot_statement_relation() {
				delete left;
				delete right;
			}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the left-hand side value. */
			const annot_value *get_left() const {return left;}

			/** Get the right-hand side value. */
			const annot_value *get_right() const {return right;}

			/** Get the operator. */
			AnnotRelOp get_operator() const {return op;}

		private:
			/** Type check a statement. */
			bool type_check(const annot_context *ctxt) const;

			annot_value *left;
			annot_value *right;
			AnnotRelOp op;

			friend class annot_context;

	} AnnotStatementRelation;

	/** A variable declaration statement. */
	typedef class annot_statement_var : public annot_statement {

		public:
			/** Constructor.
			 *  @param name Name of the variable.
			 *  @param tp   Data type of the variable.
			 *  @param loc  Location of the statement.
			 */
			annot_statement_var(Symbol name, annot_type *tp,
				const Location* loc) :
				annot_statement(ANNOT_STM_VAR, loc), name(name), vtype(tp) {}

			/** Destructor. */
			~annot_statement_var() {delete vtype;}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the name of the variable. */
			Symbol get_name() const {return name;}

			/** Get the data type of the variable. */
			const annot_type *get_var_type() const {return vtype;}

		private:
			Symbol name;
			annot_type *vtype;

	} AnnotStatementVar;

	/** A macro definition statement. */
	typedef class annot_statement_define : public annot_statement {

		public:
			/** Constructor.
			 *  @param name Name of the macro.
			 *  @param val  Value of the macro definition.
			 *  @param loc  Location of the macro.
			 */
			annot_statement_define(Symbol name, annot_value *val,
				const Location *loc) :
				annot_statement(ANNOT_STM_DEF, loc), name(name), val(val) {}

			/** Destructor. */
			~annot_statement_define() {delete val;}

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get the name of the statement. */
			Symbol get_name() const {return name;}

			/** Get the value of the statement. */
			const annot_value *get_value() const {return val;}

		private:
			Symbol name;
			annot_value *val;

	} AnnotStatementDef;

	/** A paragraph in annotation:
	 *  label: list of statements;
	 */
	typedef class annot_paragraph : public ast_node {

		public:
			/** Constructor.
			 *  @param label Label of the paragraph.
			 *  @param loc   Location of the paragraph.
			 */
			annot_paragraph(Symbol label, const Location *loc) :
				ast_node(loc), label(label) {}

			/** Destructor. */
			~annot_paragraph();

			/** Get the label of the paragraph. */
			Symbol get_label() const {return label;}

			/** See if label matches. */
			bool label_match(Symbol lab) const {
				// SymNULL means for all.
				return label==SymNULL || label==lab;
			}

			/** Append one statement.
			 *  @param stat Statement to add.
			 */
			void append_statement(annot_statement *stat);

			/** Virtual print out function. */
			std::ostream& print(std::ostream &os) const;

			/** Get all the statements in the paragraph. */
			std::vector<const annot_statement *> get_statements() const;

			/** Get all the commands in the paragraph. */
			std::vector<const annot_statement_command *> get_commands() const;

			/** Get all the defines in the paragraph. */
			std::vector<const annot_statement_define *> get_defines() const;

			/** Get all the relations in the paragraph. */
			std::vector<const annot_statement_relation *> get_relations() const;

			/** Get all the vars in the paragraph. */
			std::vector<const annot_statement_var *> get_vars() const;

			/** Get a command of given name. */
			const annot_statement_command *get_command(Symbol name) const;

		private:
			Symbol label;
			std::vector<annot_statement *> statements;

	} AnnotParagraph;

	/* annotation context, a group of annotations sharing
	   the same scope. */
	typedef class annot_context {

		public:

			/** Constructor. */
			annot_context(const mad_ast::annot_paragraph *para);

			/** Constructor. */
			annot_context(const mad_ast::annot_owner *owner, Symbol label);

			/** Constructor. */
			annot_context() : type_checked(true), check_status(true) {}

			/** Destructor. */
			~annot_context() {}

			/** Augment the annotation context. */
			annot_context& augment(const mad_ast::annot_paragraph *para);

			/** Augment the annotation context. */
			annot_context& augment(const mad_ast::annot_owner *owner,
									Symbol label);

			/** Add to another context. */
			annot_context& operator += (const annot_context& op2);

			/** Check if the context has a variable. Callable after type check.
			 *  @param var Variable name.
			 */
			bool has_variable(Symbol var) const;

			/** Check if the context has a macro. Callable after type check.
			 *  @param var The macro name.
			 */
			bool has_define(Symbol var) const;

			/** Get the type of the variable/def. Callable after type check.
			 *  @param ident The variable or define name.
			 */
			annot_type *get_symbol_type(Symbol ident) const;

			/** Get the type of the macro. Callable after type check.
			 *  @param ident The macro name.
			 */
			annot_value *get_define_value(Symbol ident) const;

			/** Type check the context.
			 *  @return True if checking ok, false otherwise.
			 */
			bool type_check();

		private:
	
			void add_paragraph(const mad_ast::annot_paragraph *para);
			void add_var(const mad_ast::annot_statement_var *var);
			void add_def(const mad_ast::annot_statement_define *def);
	
			/* defines are here. */
			std::map<Symbol, const mad_ast::annot_statement_define *> def_table;
	
			/* variables are here. */
			std::map<Symbol, const mad_ast::annot_statement_var *> var_table;
	
			/* defines. */
			std::vector<const mad_ast::annot_statement_define *> defs;
	
			/* variables */
			std::vector<const mad_ast::annot_statement_var *> vars;
	
			/* commands are here. */
			std::vector<const mad_ast::annot_statement_command *> coms;
	
			/* relationships are here. */
			std::vector<const mad_ast::annot_statement_relation *> rels;
	
			bool type_checked;
			bool check_status;

	};

	/* Addition of annotation context. */
	annot_context operator + (const annot_context& op1,
								const annot_context& op2);

}

#endif

