
#include "data_type_ast.hpp"

using namespace mad_ast;
using namespace mad;
using std::ostream;
using std::vector;
using std::pair;
using std::map;
using std::endl;
using std::string;

bool mad_ast::operator == (const DataType& type1, const DataType& type2)
{
	if (type1.get_type()!=type2.get_type())
		return false;

	switch(type1.get_type()) {

		case DT_VOID:
		case DT_STRING:
			return true;

		case DT_BIT: {
			const DataTypeBit* t1 = static_cast<const DataTypeBit *>(&type1);
			const DataTypeBit* t2 = static_cast<const DataTypeBit *>(&type2);

			if (t1->get_size()==t2->get_size() &&
				t1->get_width()==t2->get_width() &&
				t1->is_signed()==t2->is_signed())
				return true;
			else return false;
		}
		case DT_FLOAT: {
			const DataTypeFloat* t1 = 
				static_cast<const DataTypeFloat *>(&type1);
			const DataTypeFloat* t2 =
				static_cast<const DataTypeFloat *>(&type2);

			if (t1->get_size()==t2->get_size() &&
				t1->is_double()==t2->is_double())
				return true;
			else return false;
		}
		case DT_GROUP: {
			const DataTypeGroup* t1 = static_cast<const DataTypeGroup*>(&type1);
			const DataTypeGroup* t2 = static_cast<const DataTypeGroup*>(&type2);

			const vector<DataType *> *v1 = t1->get_data();
			const vector<DataType *> *v2 = t2->get_data();

			if (v1->size()!=v2->size()) return false;

			for (unsigned i=0; i<v1->size(); i++) {
				if (!(*((*v1)[i])==*((*v2)[i]))) return false;
			}
			return true;
		}
		case DT_OPER: {
			const DataTypeOper * t1 = static_cast<const DataTypeOper*>(&type1);
			const DataTypeOper * t2 = static_cast<const DataTypeOper*>(&type2);

			const vector<Symbol> *v1 = t1->get_opers();
			const vector<Symbol> *v2 = t2->get_opers();

			if (v1->size()!=v2->size()) return false;

			for (unsigned i=0; i<v1->size(); i++) {
				if (v1[i]!=v2[i]) return false;
			}

			if (t1->get_default_oper()!=t2->get_default_oper())
				return false;

			return false;
		}
	}
	return false;
}

ostream& data_type_void::print(ostream& os) const
{
	os << "void";
	return os;
}

DataType *data_type_void::clone(const Location *loc) const
{
	DataType *ret = new DataTypeVoid(loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

ostream& data_type_bit::print(ostream& os) const
{
	if (!sign) os << "u";
	if (size<1) {
		os << "int<" << width << ">";
	}
	else {
		os << "int<" << width << ">[" << size << "]";
	}
	return os;
}

DataType *data_type_bit::clone(const Location *loc) const
{
	DataType *ret = new DataTypeBit(sign, width, size, loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

ostream& data_type_float::print(ostream& os) const
{
	if (dbl) os << "double";
	else os << "float";
	
	if (size>0)
		os << "[" << size << "]";
	return os;
}

DataType *data_type_float::clone(const Location *loc) const
{
	DataType *ret = new DataTypeFloat(dbl, size, loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

ostream& data_type_string::print(ostream& os) const
{
	os << "string";
	if (size>0) os << "[" << size << "]";
	return os;
}

DataType *data_type_string::clone(const Location *loc) const
{
	DataType *ret = new DataTypeString(size, loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

ostream& data_type_group::print(ostream& os) const
{
	os << "(";
	print_object_list(os, ", ", data);
	return os << ")";
}

DataType *data_type_group::clone(const Location *loc) const
{
	vector<DataType *> *nd = new vector<DataType *>;

	vector<DataType *>::iterator it;
	for (it=data->begin(); it!=data->end(); it++)
		nd->push_back((*it)->clone(loc));

	DataType *ret = new DataTypeGroup(nd, loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

ostream& data_type_oper::print(ostream& os) const
{
	os << "{";
	print_symbol_list(os, ", ", opers);
	os << "}";

	if (def!=SymNULL) os << "(" << symbol_get_string(def) << ")";

	return os;
}

DataType *data_type_oper::clone(const Location *loc) const
{
	vector<Symbol> *o = new vector<Symbol>(*opers);
	DataType *ret = new DataTypeOper(o, def, loc);
	ret->set_readonly(this->is_readonly());
	return ret;
}

bool mad_ast::valid_bit_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();

	if (t==DT_BIT) {
		const DataTypeBit *bt = static_cast<const DataTypeBit *>(tp);
		if (bt->get_width()==0) return false;
		return true;
	}
	return false;
}

bool mad_ast::valid_scalar_value_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();

	if (valid_arith_type(tp)) return true;

	if (t==DT_STRING && tp->is_scalar()) return true;

	return false;
}

bool mad_ast::valid_tindex_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();

	if (valid_scalar_value_type(tp) || t==DT_VOID) return true;

	if (t==DT_GROUP) {

		const DataTypeGroup *dg = static_cast<const DataTypeGroup *>(tp);
		const vector<DataType *> *data = dg->get_data();

		vector<DataType *>::const_iterator it;
		for (it = data->begin(); it!=data->end(); it++) {
			if (!valid_scalar_value_type(*it)) return false;
		}
		return true;
	}

	return true;
}

bool mad_ast::valid_tindex_match(const DataType *actual, const DataType *formal)
{
	return valid_implicit_conversion(actual, formal);
}

bool mad_ast::valid_tvalue_type(const DataType *tp)
{
	/* void, bit/sbit/float/double scalar, group are valid to types */
	EnumDataType t = tp->get_type();

	if (t==DT_VOID) return true;

	if (valid_scalar_value_type(tp)) return true;

	if (t==DT_GROUP) {

		const DataTypeGroup *dg = static_cast<const DataTypeGroup *>(tp);
		const vector<DataType *> *data = dg->get_data();

		vector<DataType *>::const_iterator it;
		for (it = data->begin(); it!=data->end(); it++) {
			if (!valid_scalar_value_type(*it)) return false;
		}
		return true;
	}

	return false;
}

bool mad_ast::valid_mvar_type(const DataType *tp)
{
	return valid_scalar_value_type(tp);
}

bool mad_ast::valid_fvar_type(const DataType *tp)
{
	return valid_scalar_value_type(tp);
}

bool mad_ast::valid_var_type(const DataType *tp)
{
	return valid_scalar_value_type(tp) || (tp->get_type()==DT_OPER);
}

bool mad_ast::valid_arg_type(const DataType *tp)
{
	return valid_scalar_value_type(tp);
}

bool mad_ast::valid_def_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();

	if (valid_bit_type(tp) || t==DT_STRING || t==DT_FLOAT) {
		return true;
	}
	else if (t==DT_GROUP) { // function prototype

		const DataTypeGroup *dg = static_cast<const DataTypeGroup *>(tp);
		const vector<DataType *> *data = dg->get_data();

		vector<DataType *>::const_iterator it;
		for (it = data->begin(); it!=data->end(); it++) {
			if (!valid_scalar_value_type(*it)) return false;
		}
		return true;
	}

	return false;
}

bool mad_ast::valid_coding_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();
	if (valid_bit_type(tp) && tp->is_scalar()) return true;

	/* we still need to check if the operation has coding */
	if (t == DT_OPER) return true;

	return false;
}

bool mad_ast::valid_match_var_type(const DataType *tp)
{
	if (valid_bit_type(tp) && tp->is_scalar()) return true;
	return false;
}

bool mad_ast::valid_syntax_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();
	if (valid_scalar_value_type(tp)) return true;

	/* we still need to check if the operation has syntax */
	if (t == DT_OPER) return true;

	return false;
}

bool mad_ast::valid_arith_type(const DataType *tp)
{
	EnumDataType t = tp->get_type();
	if (valid_bit_type(tp) && tp->is_scalar()) return true;
	if (t == DT_FLOAT && tp->is_scalar()) return true;
	return false;
}

DataType *mad_ast::common_super_type(const DataType *t1, const DataType *t2)
{
	EnumDataType tp1, tp2;

	/* we don't consider vector for now */
	if (!valid_scalar_value_type(t1) || !valid_scalar_value_type(t2))
		 return NULL;

	tp1 = t1->get_type();
	tp2 = t2->get_type();

	/* bit type can be casted to float type */
	if (tp1==DT_BIT && tp2==DT_FLOAT) {
		return common_super_type(t2, t2);
	}
	else if (tp1==DT_FLOAT && tp2==DT_BIT) {
		return common_super_type(t1, t1);
	}

	/* arithmetic type can be casted to string, both are already 
	 * valid scalar type so no too much checking  */
	if (tp1==DT_STRING || tp2==DT_STRING) {
		DataType *ret = new DataTypeString(0, NULL);
		ret->set_readonly(true);
		return ret;
	}
		
	/* two types differ, no common super type */
	if (tp1!=tp2) return NULL;

	if (tp1==DT_BIT) {

		const DataTypeBit *bt1 = static_cast<const DataTypeBit *>(t1);
		const DataTypeBit *bt2 = static_cast<const DataTypeBit *>(t2);

		int w1 = bt1->get_width();
		int w2 = bt2->get_width();

		int width;
		bool sign;

		if (w1>w2) {
			width = w1;
			sign = bt1->is_signed();
		}
		else if (w2>w1) {
			width = w2;
			sign = bt2->is_signed();
		}
		else {
			width = w1;
			/* if both signed -> signed, o.w. unsigned */
			sign = bt1->is_signed() && bt2->is_signed();
		}

		DataType *ret = new DataTypeBit(sign, width, 0, NULL);
		ret->set_readonly(true);
		return ret;
	}
	else if (tp1==DT_FLOAT) {
		const DataTypeFloat *ft1 = static_cast<const DataTypeFloat *>(t1);
		const DataTypeFloat *ft2 = static_cast<const DataTypeFloat *>(t2);

		bool dbl = ft1->is_double() || ft2->is_double();

		DataType *ret = new DataTypeFloat(dbl, 0, NULL);
		ret->set_readonly(true);
		return ret;
	}

	return NULL;
}

DataType *mad_ast::bit_concat_type(const DataTypeBit *t1, const DataTypeBit *t2)
{
	/* we don't consider vector for now */
	assert(t1->is_scalar() && t2->is_scalar());

	DataType *ret = new DataTypeBit(false,
		t1->get_width()+t2->get_width(), 0, NULL);

	ret->set_readonly(true);

	return ret;
}

bool mad_ast::valid_implicit_conversion(const DataType *from,
	const DataType *to)
{
	if (!from->is_scalar() || !to->is_scalar()) return false;

	EnumDataType tp1 = from->get_type();
	EnumDataType tp2 = to->get_type();

	/* bit type can be casted to float type */
	if (tp1==DT_BIT && tp2==DT_FLOAT) 
		return true;

	if ((tp1==DT_BIT || tp1==DT_FLOAT) && tp2==DT_STRING)
		return true;

	if (tp1!=tp2) return false;

	if (tp1==DT_STRING || tp1==DT_VOID)
		return true;
	else if (tp1==DT_BIT) {
		const DataTypeBit *dfrom = static_cast<const DataTypeBit *>(from);
		const DataTypeBit *dto = static_cast<const DataTypeBit *>(to);
		return (dfrom->get_width()<=dto->get_width());
	}
	else if (tp1==DT_FLOAT) {
		const DataTypeFloat *ffrom = static_cast<const DataTypeFloat *>(from);
		const DataTypeFloat *fto = static_cast<const DataTypeFloat *>(to);
		return (!ffrom->is_double() || fto->is_double());
	}
	else if (tp1==DT_GROUP) {
		const DataTypeGroup *lfrom = static_cast<const DataTypeGroup *>(from);
		const DataTypeGroup *lto = static_cast<const DataTypeGroup *>(to);

		const vector<DataType *> *lv = lfrom->get_data();
		const vector<DataType *> *rv = lto->get_data();

		if (lv->size()!=rv->size()) return false;

		for (unsigned i=0; i<lv->size(); i++) {
			if (!valid_implicit_conversion((*lv)[i], (*rv)[i]))
				return false;
		}
		return true;
	}

	return false;
}

bool mad_ast::valid_farg_match(const DataType *actual, const DataType *formal)
{
	if (formal->is_readonly())
		return valid_implicit_conversion(actual, formal);

	if (actual->is_readonly()) return false;

	/* If pass by reference, must be identical type*/
	return *formal==*actual;
}


bool mad_ast::valid_osm_assign_match(const DataType *left,
			const DataType *right)
{
	/* Readonly bit is not an issue here since all variable's in osm
	 * section are writable. */

	/* only scalar should appear in OSM section */
	assert (left->is_scalar());
	assert (right->is_scalar());

	EnumDataType tp1 = left->get_type();
	EnumDataType tp2 = right->get_type();

	if (tp1!=tp2) return false;

	if (tp1 == DT_STRING) return true;
	else if (tp1 == DT_BIT) {

		int size1 = static_cast<const DataTypeBit *>(left)->get_width();
		int size2 = static_cast<const DataTypeBit *>(right)->get_width();

		return size1>=size2;
	}
	else if (tp1 == DT_FLOAT) {

		const DataTypeFloat *fl = static_cast<const DataTypeFloat *>(left);
		const DataTypeFloat *fr = static_cast<const DataTypeFloat *>(right);

		return (!fr->is_double() || fl->is_double());
	}
	else if (tp1 == DT_GROUP) {

		const DataTypeGroup *ltp = static_cast<const DataTypeGroup *>(left);
		const DataTypeGroup *rtp = static_cast<const DataTypeGroup *>(right);

		const vector<DataType *> *lv = ltp->get_data();
		const vector<DataType *> *rv = rtp->get_data();

		if (lv->size()!=rv->size()) return false;

		for (unsigned i=0; i<lv->size(); i++) {
			if (!valid_osm_assign_match((*lv)[i], (*rv)[i]))
				return false;
		}
		return true;
	}
	else if (tp1 == DT_VOID) return true;

	return false;
}

sym_context::sym_context() : context()
{
}

bool sym_context::has(Symbol sym) const
{
	return (context.find(sym)!=context.end());
}

void sym_context::insert(const named_typed_node *node)
{
	context[node->get_name()] = node;
}

const Location *sym_context::get_loc(Symbol sym) const
{
	map<Symbol, const named_typed_node *>::const_iterator
		it = context.find(sym);
	if (it==context.end()) return NULL;
	return (*it).second->get_loc();
}

const DataType *sym_context::get_data_type(Symbol sym) const
{
	map<Symbol, const named_typed_node *>::const_iterator
		it = context.find(sym);
	if (it==context.end()) return NULL;
	return (*it).second->get_data_type();
}

const named_typed_node *sym_context::get_object(Symbol sym) const
{
	map<Symbol, const named_typed_node *>::const_iterator
		it = context.find(sym);
	if (it==context.end()) return NULL;
	return (*it).second;
}

sym_context& sym_context::operator += (const sym_context& op2)
{
	context.insert(op2.context.begin(), op2.context.end());
	return *this;
}

SymContext mad_ast::operator + (const SymContext& context1,
		const SymContext& context2)
{
	SymContext ctxt(context1);
	ctxt += context2;
	return ctxt;
}

bool mad_ast::operator == (const NumValue& o1, const NumValue& o2)
{
	return o1.get_val_inbase(16) == o2.get_val_inbase(16); ///@@@ flt
}

#if 0
/** Print out the context. */
ostream& sym_context::print(ostream& os) const
{
	map<Symbol, pair<const DataType *, const Location *> >::const_iterator it;
	for (it = context.begin(); it!=context.end(); it++) {
		os << symbol_get_string((*it).first) << " --- " << 
			*((*it).second.second) << endl;	
	}
	return os;
}
#endif

void data_type_void::emit_C(ostream& os, bool af) const
{
	os << "void";
}

void data_type_bit::emit_C(ostream& os, bool af) const
{
	//if (this->is_readonly()) os << "const ";

	if (sign)
		os << "_INT_T";
	else
		os << "_UINT_T";

	os << "(" << std::dec << width << ")";

	if (size && af) {
		os << "[" << std::dec << size << "]";
	}
}

void data_type_float::emit_C(ostream& os, bool af) const
{
	//if (this->is_readonly()) os << "const ";

	if (dbl)
		os << "_DBL_T";
	else
		os << "_FLT_T";
	if (size && af) {
		os << "[" << std::dec << size << "]";
	}
}

void data_type_string::emit_C(ostream& os, bool af) const
{
	//if (this->is_readonly()) os << "const ";

	os << "_STR_T";
	if (size && af) {
		os << "[" << std::dec << size << "]";
	}
}

void data_type_group::emit_C(ostream& os, bool af) const
{
	
	//if (this->is_readonly()) os << "const ";

	if (data->size()==2)
		os << "_TUPLE_T(";
	else if (data->size()==3)
		os << "_TRIPLE_T(";
	else if (data->size()==4)
		os << "_QUADRUPLE_T(";
	else if (data->size()==5)
		os << "_QUINTUPLE_T(";
	else
		assert(0);

	vector<DataType *>::iterator it = data->begin();
	(*it)->emit_C(os);
	for (it++; it!=data->end(); it++) {
		os << ", ";
		(*it)->emit_C(os);
	}
	os << ")";
}

void data_type_oper::emit_C(ostream& os, bool af) const
{
	assert(0);
}

bool mad_ast::num_value_check(const DataType *tp, const num_value *val,
	const Location *loc) 
{
	bool ret = true;

	if (tp->get_type()==DT_BIT) {

		if (val->is_float()) {
			generic_error("Floating value defined as integer type", loc);
			return false;
		}

		const DataTypeBit *btype = static_cast<const DataTypeBit *>(tp);
		unsigned width = val->get_width_inbase(2);

		if (btype->get_width()<width) {
			generic_error("Constant longer than data type width", loc);
			ret = false;
		}
		
		if (!btype->is_signed() && val->is_negative()) {
			generic_warn("Negative value defined as unsigned type", loc);
		}
	}
	else if (tp->get_type()==DT_STRING) {
		generic_error("Constant type mismatch", loc);
		ret = false;
	}

	/* no need to check float type */
	return ret;
}

