
#ifndef __MAD_DATA_TYPE_AST_HPP__
#define __MAD_DATA_TYPE_AST_HPP__

#include <vector>
#include <iostream>
#include <map>
#include "mad_ast.hpp"

using mad::Symbol;

namespace mad_ast {

	/* we have the following types:
	 *    1. void, used only in manager declaration
	 *    2. bit<width>, used anywhere
	 *    3. sbit<width>, used anywhere
	 *    4. bit<witdh>[size], used in define only
	 *    5. sbit<width>[size], used in define only
	 *    8. group, used in manager declaration only, can group only
	 *       scalar bit,sbit,string
	 *    9. oper with or without binding, used in operation var only
	 */

	/** Enumerate of data types */
	typedef enum {

		DT_VOID,	/**< void data type. */
		DT_STRING,	/**< string data type. */
		DT_BIT,		/**< bit data type. */
		DT_FLOAT,	/**< float data type. */
		DT_GROUP,	/**< group data type, (a, b, ...). */
		DT_OPER		/**< oper data type, {oper...}(a,b,c...) */

	} EnumDataType;

	/* type of a variable reference */
	typedef enum {

		VT_VAR,	/**< local variable. */
		VT_DEF,	/**< global declaration. */
		VT_MAC,	/**< machine variable. */
		VT_ARG	/**< argument. */

	} EnumVarType;

	/** Base class of data type nodes */
	typedef class data_type : public ast_node {

		public:
			/** Constructor.
			 *  @param dt   Type of the node.
			 *  @param loc  Location of the node.
			 */
			data_type (const EnumDataType &dt, const Location *loc) :
				   ast_node(loc), dt(dt), readonly(false) {}

			/** Destructor */
			virtual ~data_type() {}

			/** Get the data type */
			EnumDataType get_type() const {return dt;}

			/** check if scalar type */
			virtual bool is_scalar() const {return true;}

			/** check if the type is read only */
			bool is_readonly() const {return readonly;}

			/** set the type to be readonly */
			void set_readonly(bool c) {readonly = c;}

			/** Clone one datatype with given location information.
			 *  @param loc Location for the cloned structure.
			 */
			virtual data_type *clone(const Location *loc) const = 0;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10] //@@@ what is this?
			 */  
			virtual void emit_C(std::ostream& os, bool af=true) const = 0;

		private:
			EnumDataType dt;

			/** access control, some arguments and defs are readonly */
			bool readonly;

	} DataType;

	/** comparison */
	bool operator == (const DataType& type1, const DataType& type2);


	/** Class of void data type. */
	typedef class data_type_void : public data_type {

		public:
			/** Constructor.
			 *  @param loc  Locaton of the node.
			 */
			data_type_void(const Location *loc) : data_type(DT_VOID, loc) {}

			/** Destructor. */
			~data_type_void() {}

			/** Print out the type. */
			std::ostream& print(std::ostream &) const;

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

	} DataTypeVoid;


	/** Class of data type bit, bit<32>, bit<32>[16], bit<32> option. 
	 *  Option and array cannot appear at the same time.
	 */
	typedef class data_type_bit : public data_type {

		public:
			/** Constructor.
			 *  @param s      If value of the type is signed bit.
			 *  @param width  Bit width of the type.
			 *  @param size   Array size of the type, 0 for scalar.
			 *  @param loc    Location of the type.
			 */
			data_type_bit(bool s, unsigned width, unsigned size,
				const Location *loc) :
				data_type(DT_BIT, loc), width(width),
		   		size(size), sign(s) {}

			/** Destructor. */
			~data_type_bit() {}

			/** Print out the bit type. */
			std::ostream& print(std::ostream &) const;

			/** Get the size. */
			unsigned get_size() const {return size;}

			/** Get the width. */
			unsigned get_width() const {return width;}

			/** If a scalar type : size==0. */
			bool is_scalar() const {return size==0;}

			/** If a signed type. */
			bool is_signed() const {return sign;}

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

		private:

			/* The scanner ensures that numbers are non-negative. */
			int width, size;
			bool sign;

	} DataTypeBit;

	/** Class of data type bit, bit<32>, bit<32>[16], bit<32> option. 
	 *  Option and array cannot appear at the same time.
	 */
	typedef class data_type_float: public data_type {

		public:
			/** Constructor.
			 *  @param dbl    Is double precision.
			 *  @param size   Array size of the type, 0 for scalar.
			 *  @param loc    Location of the type.
			 */
			data_type_float(bool dbl, unsigned size, const Location *loc) :
				data_type(DT_FLOAT, loc), size(size), dbl(dbl) {}

			/** Destructor. */
			~data_type_float() {}

			/** Print out the bit type. */
			std::ostream& print(std::ostream &) const;

			/** Get the size. */
			unsigned get_size() const {return size;}

			/** If a double precision number */
			bool is_double() const {return dbl;}

			/** If a scalar type : size==0. */
			bool is_scalar() const {return size==0;}

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

		private:

			/* The scanner ensures that number is non-negative. */
			int size;
			bool dbl;

	} DataTypeFloat;

	/** Class of data type string.
	 *  Allowed in DEFINE section only, referred from SYNTAX paragraphs.
	 */
	typedef class data_type_string : public data_type {

		public:
			/** Constructor.
			 *  @param size  Array size of the type, 0 for scalar.
			 *  @param loc   Location of the type.
			 */
			data_type_string(unsigned size, const Location *loc) :
				data_type(DT_STRING, loc), size(size) {}

			/** Destructor. */
			~data_type_string() {}

			/** Print out the string type. */
			std::ostream& print(std::ostream &) const;

			/** Return the size of the type.
			 *  0 -- scalar type.
			 *  other -- vector type.
			 */
			unsigned get_size() const {return size;}

			/** If a scalar type: size==0. */
			bool is_scalar() const {return size==0;}

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

		private:
			int size;

	} DataTypeString;

	
	/** Class of data type group, (a, b, ...). */
	typedef class data_type_group : public data_type {

		public:
			/** Constructor.
			 *  @param g   Array of data types.
			 *  @param loc Location of the type node.
			 */
			data_type_group(std::vector<data_type *> *g,
				const Location *loc) :
				data_type(DT_GROUP, loc), data(g) {}

			/** Destructor. */
			~data_type_group() {
				delete_object_list(data);
			}

			/** Print out the group type */
			std::ostream& print(std::ostream &) const;

			/** Get the details of the type. */
			const std::vector<data_type *> *get_data() const {return data;}

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

		private:

			std::vector<data_type *> *data;

	} DataTypeGroup;

	/** Class of data type operation. */
	typedef class data_type_oper : public data_type {

		public:
			/** Constructor.
			 *  @param opers Array of operation names.
			 *  @param def   Default operation name.
			 *  @param loc   Location of the node.
			 */
			data_type_oper(std::vector<Symbol> *opers, Symbol def,
				const Location *loc) :
				data_type(DT_OPER, loc), opers(opers), def(def) {}

			/** Destructor. */
			~data_type_oper() {
				delete opers;
			}

			/** Print out the node. */
			std::ostream& print(std::ostream &) const;

			/** Get the opers */
			const std::vector<Symbol> *get_opers() const {return opers;}

			/** Get the default operation */
			Symbol get_default_oper() const {return def;}

			/* Clone one datatype with given location information */
			data_type *clone(const Location *loc) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 *  @param af Emit array if true, etc [10]
			 */  
			void emit_C(std::ostream& os, bool af=true) const;

		private:
			std::vector<Symbol> *opers;
			Symbol def;	/* the default symbol */

	} DataTypeOper;

	/** Valid bit type, check if width==0.
	 *  @param tp The type to check.
	 */
	bool valid_bit_type(const DataType* tp);

	/** Valid type to be used as a scalar value.
	 *    scalar string/bit/sbit/float/double.
	 *  @param tp The type to check.
	 */
	bool valid_scalar_value_type(const DataType* tp);

	/** Valid token index type.
	 *  Valid token index types include:
	 *    void, scalar int/uint, or group of scalar int/uint types.
	 *  @param tp The type to check.
	 */
	bool valid_tindex_type(const DataType *tp);

	/** See if token index type matches formal specificaton.
	 *  @param actual The type to check.
	 *  @param formal The formal specification.
	 */
	bool valid_tindex_match(const DataType *actual, const DataType *formal);

	/** Valid token value type.
     *  Valid token value types include:
	 *    void, scalar int/uint/float/double/string, or group type
	 *  @param tp The type to check. 
	 */
	bool valid_tvalue_type(const DataType *tp);

	/** Valid variable type to be used in machine section.
	 *  Valid machine variable types include:
	 *    scalar string/int/uint/float/double.
	 *  @param tp The type to check.
	 */
	bool valid_mvar_type(const DataType *tp);

	/** Valid variable type to be used in function section.
	 *  Valid variable types include:
	 *	  scalar string/int/uint/float/double.
	 */
	bool valid_fvar_type(const DataType *tp);

	/** Valid variable type to be used in operation section.
	 *  Valid variable types include:
	 *	  scalar string/int/uint/float/double, oper.
	 */
	bool valid_var_type(const DataType *tp);

	/** Valid variable type to be used as argument.
	 *  Valid variable types include:
	 *	  scalar string/int/uint/float/double.
	 */
	bool valid_arg_type(const DataType *tp);

	/** Valid define variable type.
	 *  Valid variable types include:
	 *	  scalar/vector string/int/uint/float/double,
	 *	   or group of scalar string/int/uint/float/double
	 *     (as function prototype)
	 */
	bool valid_def_type(const DataType *tp);

	/** Valid variable types to be used in coding.
	 *  Valid types include scalar int/uint and operation
	 */
	bool valid_coding_type(const DataType *tp);

	/** Valid variable types to be used as matching variable in coding.
	 *  Valid types include scalar int/uint
	 */
	bool valid_match_var_type(const DataType *tp);

	/** Valid variable types to be used in syntax.
	 *  Valid types include scalar string/int/uint and operation
	 */
	bool valid_syntax_type(const DataType *tp);

	/** Valid variable types to be used in arithmetic computation.
	 *  Valid types include scalar int/uint/float/double.
	 */
	bool valid_arith_type(const DataType *tp);

	/** The lowest common super type of two types.
	 *  @return NULL if two types incompatible, o.w a readonly super type.
	 */
	DataType *common_super_type(const DataType *tp1, const DataType *tp2);

	/** Construct a datatype whose width is the sum of the two.
	 *  @return A newly constructed readonly type.
	 */
	 
	DataType *bit_concat_type(const DataTypeBit *t1, const DataTypeBit *t2);

	/** Check if operation type binding is valid.
	 *  Valid operation type binding includes:
	 *	  string  -> string,
	 *    bit<width> -> bit/sbit<width>
	 *    sbit<width> -> bit/sbit<width>
	 */
	bool valid_binding_match(const DataType *actual, const DataType *formal);

	/** Check if it is a valid implicit cast.
	 *  Implicit cast can be performed from scalar
	 *   string type, shorter bit/type to longer bit type,
	 *   bit type to string type, float type to double type.
	 *  This one does not check read_only field.
	 */
	bool valid_implicit_conversion(const DataType *from, const DataType *to);

	/** Check if function call argument type is valid.
	 *  Valid call type if actual can be casted to formal implicitly and
	 *  actual does not violate formal's read_only property.
	 */
	bool valid_farg_match(const DataType *actual, const DataType *formal);

	/** Check if assignment type is valid.
	 *  @param left  The left side of =.
	 *  @param right The right side of =.
	 */
	bool valid_osm_assign_match(const DataType *left, const DataType *right);

	/** Base class for those AST nodes with a name and a data type */
	typedef class named_typed_node : public ast_node {

		public:

			/** Constructor.
			 *  @param sym	Name of the node.
			 *  @param typ	Type of the node.
			 *  @param loc  Location of the node.
			 */
			named_typed_node(Symbol sym,
				DataType *typ, const Location *loc) :
				ast_node(loc), name(sym), type(typ) {}

			/** Destructor */
			~named_typed_node() {
				delete type;
			}

			/** Get the name of the node. */
			Symbol get_name() const {return name;}

			/** Get the data type of the node. */
			const DataType *get_data_type() const {return type;}

		protected:
			Symbol name;
			DataType *type;

	} NamedTypedNode;

	/** Symbol context class, name->named_typed_node. */
	typedef class sym_context  {

		public:

			/** Default constructor. */
			sym_context();

			/** Insert an entry into the context.
			 *  @param node The named and typed AST node.
			 */
			void insert(const named_typed_node *node);

			/** Check if a symbol exists in the context. */
			bool has(Symbol sym) const;

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The location of the symbol, NULL if symbol not found.
			 */
			const Location *get_loc(Symbol sym) const;

			/** Get the data type of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The data type of the symbol, NULL if symbol not found.
			 */
			const DataType *get_data_type(Symbol sym) const;

			/** GEt the node */
			const named_typed_node *get_object(Symbol sym) const;

			/** Print out the context. */
			//std::ostream& print(std::ostream &) const;

			typedef std::map<Symbol, const named_typed_node *>::const_iterator
					const_iterator;

			const_iterator begin() const {return context.begin();}
			const_iterator end() const {return context.end();}

			sym_context& operator += (const sym_context& op2);

		private:
			std::map<Symbol, const named_typed_node *> context;

	} SymContext;

	/* Addition of contexts. */
	SymContext operator + (const SymContext& context1,
			const SymContext& context2);

	/** Object context class, name->(data_type, location).
	 *  Same as sym_context, except that this one is template based. */
	template <class T>
	class obj_context {

		public:

			/** Default constructor. */
			obj_context() : context() {};

			/** Insert an entry into the context.
			 *  @param node The named and typed AST node.
			 */
			void insert(T *node) {
				context[node->get_name()] = node;
			}

			/** Check if a symbol exists in the context. */
			bool has(Symbol sym) const {
				return (context.find(sym)!=context.end());
			}

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The location of the symbol, NULL if symbol not found.
			 */
			const Location *get_loc(Symbol sym) const {
				typename std::map<Symbol, T *>::const_iterator
					it = context.find(sym);
				if (it==context.end()) return NULL;
			    return (*it).second->get_loc();
			}

			/** Get the data type of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The data type of the symbol, NULL if symbol not found.
			 */
			const DataType *get_data_type(Symbol sym) const {
				typename std::map<Symbol, T *>::const_iterator
					it = context.find(sym);
				if (it==context.end()) return NULL;
			    return (*it).second->get_data_type();
			}

			/** Get the object itself */
			T *get_object(Symbol sym) {
				typename std::map<Symbol, T *>::const_iterator
					it = context.find(sym);
				if (it==context.end()) return NULL;
			    return (*it).second;
			}

			/** Get the object itself */
			const T *get_object(Symbol sym) const {
				typename std::map<Symbol, T *>::const_iterator
					it = context.find(sym);
				if (it==context.end()) return NULL;
			    return (*it).second;
			}
			/** Print out the context. */
			//std::ostream& print(std::ostream &) const;

			typedef typename std::map<Symbol, T *>::const_iterator
					const_iterator;

			const_iterator begin() const {return context.begin();}
			const_iterator end() const {return context.end();}

		private:
			std::map<Symbol, T *> context;
	};


	/** Type check a numerical value. */
	bool num_value_check(const DataType *tp, const NumValue *val,
		const Location *loc);
}

#endif


