
#include "define_ast.hpp"
#include "misc.hpp"
#include "emit_c.hpp"

using namespace mad_ast;
using namespace mad;
using std::ostream;
using std::vector;
using std::string;
using std::endl;
using std::map;

ostream& num_value::print(ostream& os) const
{
	if (neg) os << "-";
	os << *value;
	return os;
}

ostream& define_value_num::print(ostream& os) const
{
	value->print(os);
	return os;
}

ostream& define_value_str::print(ostream& os) const
{
	os << *value;
	return os;
}

ostream& define_value_num_array::print(ostream& os) const
{
	os << "{";
	print_object_list(os, ", ", value);
	os << "}";
	return os;
}

ostream& define_value_str_array::print(ostream& os) const
{
	os << "{\"";
	print_string_list(os, "\", \"", value);
	os << "\"}";
	return os;
}

ostream& define_clause::print(ostream& os) const
{
	os << symbol_get_string(name);
	if (dirty) os << "*";
	os << ":";

	// see if a function declaration
	if (type->get_type()==DT_GROUP) {

		DataTypeGroup *dtype = static_cast<DataTypeGroup *>(type);
		const vector<DataType *> *types = dtype->get_data();

		os << "(";
		if (types->size()>0) {
			types->front()->print(os);
			if (!types->front()->is_readonly()) os << "*";

			vector<DataType *>::const_iterator it;
			for (it = types->begin()+1; it!=types->end(); it++) {
				os << ", ";
				(*it)->print(os);
				if (!(*it)->is_readonly()) os << "*";
			}
		}
		os << ")";
	}
	else {
		type->print(os);
	}

	if (value) {
		os << " = ";
		value->print(os);
	}

	os << "; ";

	print_object_list(os, "\n", annots);

	os << endl;

	return os;
}

ostream& define_section::print(ostream& os) const
{
	os << "DEFINE" << endl;
	print_object_list(os, "\n", clauses);
	return os;
}

void define_section::merge(define_section& two)
{
	clauses->insert(clauses->end(), two.clauses->begin(), two.clauses->end());
	two.clauses->erase(two.clauses->begin(), two.clauses->end());
}

bool define_section::check_annots()
{
	bool ret = true;
	vector<DefineClause *>::const_iterator it;	
	for (it=clauses->begin(); it!=clauses->end(); it++) {
		if (!(*it)->check_annotations())
			ret = false;
	}

	return ret;
}

bool define_value_num::type_check(const DataType *tp) const
{
	bool ret = true;

	if (!tp->is_scalar()) {
		generic_error("Expecting array, get scalar", this->get_loc());
		ret = false;
	}

	ret = num_value_check(tp, value, this->get_loc()) && ret;

	return ret;
}

bool define_value_str::type_check(const DataType *tp) const
{
	bool ret = true;

	if (!tp->is_scalar()) {
		generic_error("Expecting array, get scalar", this->get_loc());
		ret = false;
	}
	if (tp->get_type()!=DT_STRING) {
		generic_error("Constant type mismatch", this->get_loc());
		ret = false;
	}
	return ret;
}

bool define_value_num_array::type_check(const DataType *tp) const
{
	bool ret = true;

	if (tp->is_scalar()) {
		generic_error("Expecting scalar, get array", this->get_loc());
		ret = false;
	}

	unsigned size;

	if (tp->get_type()==DT_BIT) {
		size = static_cast<const DataTypeBit *>(tp)->get_size();
	}
	else if (tp->get_type()==DT_FLOAT) {
		size = static_cast<const DataTypeFloat *>(tp)->get_size();
	}
	else {
		generic_error("Constant type mismatch", this->get_loc());
		return false;
	}

	if (size != value->size()) {
		generic_error("Array size mismatch", this->get_loc());
		ret = false;
	}

	vector<NumValue *>::const_iterator cit;
	for (cit=value->begin(); cit!=value->end(); cit++) {
		ret = num_value_check(tp, *cit, this->get_loc()) && ret;
	}

	return ret;
}

bool define_value_str_array::type_check(const DataType *tp) const
{
	bool ret = true;

	if (tp->get_type()!=DT_STRING) {
		generic_error("Constant type mismatch", this->get_loc());
		ret = false;
	}
	else {
		unsigned size1, size2;

		size1 = static_cast<const DataTypeString *>(tp)->get_size();
		size2 = value->size();

		if (size1!=size2) {
			generic_error("Array size mismatch", this->get_loc());
			ret = false;
		}
	}
	return ret;
}

bool define_clause::type_check()
{
	bool ret = true;

	if (!valid_def_type(type)) {
		generic_error("Bad define variable type", type->get_loc());
		ret = false;
	}

	if (!ret || value==NULL) return ret;

	/* check value */
	return value->type_check(type);
}

bool define_section::type_check()
{
	bool ret = true;

	/* naming table */
	vector<DefineClause *>::iterator it;
	for (it=clauses->begin(); it!=clauses->end(); it++) {

		Symbol name = (*it)->get_name();
		if (def_cache.has(name)) {
			naming_conflict(name, def_cache.get_loc(name), (*it)->get_loc());
			ret = false;
		}
		else {
			def_cache.insert(*it);
		}
		ret = (*it)->type_check() && ret;
	}
	return ret;
}

void define_value_num::emit_C(ostream& os)
{
	value->emit_C(os);
}

void define_value_str::emit_C(ostream& os)
{
	os << "\"" << *value << "\"";
}

void define_value_num_array::emit_C(ostream& os)
{
	os << "{";
	vector<NumValue *>::iterator vit;
	for (vit=value->begin(); vit!=value->end(); vit++) {
		(*vit)->emit_C(os);
		os << ", ";
	}
	os << "}";
}

void define_value_str_array::emit_C(ostream& os)
{
	os << "{";
	vector<string *>::iterator vit;
	for (vit=value->begin(); vit!=value->end(); vit++) {
		os << "\"" << (**vit) << "\"" << ", ";
	}
	os << "}";
}

void define_clause::emit_C_dec(ostream& os)
{
	const DataType *the_type = this->get_data_type();
	// see if function
	if (the_type->get_type()!=DT_GROUP) {
		os << "extern ";
		the_type->emit_C(os, false);
		os << " ";
		os << DEF_NAME(this->get_name());
		if (the_type->get_type()==DT_BIT) {
			int size = static_cast<const DataTypeBit *>(the_type)->get_size();
			if (size) os << "[" << size << "]";
		}
		else if (the_type->get_type()==DT_STRING) {
			int size = static_cast<const DataTypeString *>(the_type)->get_size();
			if (size) os << "[" << size << "]";
		}
	}
	else {
		const DataTypeGroup *gtype =
			static_cast<const DataTypeGroup *>(the_type);
		const vector<DataType *> *args = gtype->get_data();

		if (dirty) os << DIRTY_E_FUN_DEC;
		else os << E_FUN_DEC;

		os << "(" << symbol_get_string(this->get_name());

		vector<DataType *>::const_iterator arg_it;
		for (arg_it=args->begin(); arg_it!=args->end(); arg_it++) {

			os << ", ";
			if ((*arg_it)->is_readonly()) os << "const ";
			(*arg_it)->emit_C(os);
			os << "&";
		}

		os << ")";
	}
	os << ";" << endl << endl;
}

void define_clause::emit_C(ostream& os)
{
	if (value==NULL) return;

	const DataType *the_type = this->get_data_type();
	the_type->emit_C(os, false);
	os << " ";
	os << DEF_NAME(this->get_name());
	if (the_type->get_type()==DT_BIT) {
		int size = static_cast<const DataTypeBit *>(the_type)->get_size();
		if (size) os << "[" << size << "]";
	}
	else if (the_type->get_type()==DT_STRING) {
		int size = static_cast<const DataTypeString *>(the_type)->get_size();
		if (size) os << "[" << size << "]";
	}
	os << " ";
	os << " = ";
	
	value->emit_C(os);

	os << ";" << endl << endl;
}

void define_section::emit_C_dec(ostream& os)
{
	os << endl << endl;
	vector<DefineClause *>::iterator dit;
	for (dit = clauses->begin(); dit!=clauses->end(); dit++) {
		(*dit)->emit_C_dec(os);
	}
}

void define_section::emit_C(ostream& os)
{
	os << endl << endl;
	vector<DefineClause *>::iterator dit;
	for (dit = clauses->begin(); dit!=clauses->end(); dit++) {
		(*dit)->emit_C(os);
	}
}

