
#ifndef __MAD_DEFINE_AST_HPP__
#define __MAD_DEFINE_AST_HPP__

#include <vector>
#include <string>
#include <iostream>
#include <map>
#include "program_ast.hpp"
#include "data_type_ast.hpp"

using mad::Symbol;

namespace mad_ast {

	/** Value for defines */
	typedef enum {
		DEF_VAL_NUM,        /**< Number literal value. */
		DEF_VAL_STR,        /**< String literal value. */
		DEF_VAL_NUM_ARRAY,  /**< Number array value. */
		DEF_VAL_STR_ARRAY   /**< String array value. */
	} EnumDefValueType;

	/** Base class for define values. */
	typedef class define_value : public ast_node {

		public:
			/** Constructor.
			 *  @param tp   Category of the value.
			 *  @param loc  Locatio of the value.
			 */
			define_value(EnumDefValueType tp,
				const Location *loc) : ast_node(loc), tp(tp) {}

			/** Destructor. */
			virtual ~define_value() {}

			/** Get the type of the value. */
			EnumDefValueType get_type() const {return tp;}

			/** Check the type of the constant value if defined. */
			virtual bool type_check(const DataType *type) const = 0;

			/** emit C declarations. */
			virtual void emit_C(std::ostream& os) = 0;

		private:
			EnumDefValueType tp;

	} DefineValue;

	/** Class for numerical define values. */
	typedef struct define_value_num : public define_value {

		public:
			/** Constructor.
			 *  @param value  The number value.
			 *  @param loc    Location of the node.
			 */
			define_value_num(NumValue *value, const Location *loc) :
				define_value(DEF_VAL_NUM, loc), value(value) {}

			/** Destructor. */
			~define_value_num() {
				delete value;
			}

			/** Print out the number. */
			std::ostream& print(std::ostream &) const;

			/** Get the constant value. */
			const NumValue *get_value() const {return value;}

			/** Check the type of the constant value if defined. */
			bool type_check(const DataType *type) const;

			/** emit C declarations. */
			void emit_C(std::ostream& os);

		private:

			NumValue *value;

	} DefineValueNum;

	/** Class for string define values. */
	typedef struct define_value_str : public define_value {

		public:
			/** Constructor.
			 *  @param value  The string value.
			 *  @param loc    Location of the node.
			 */
			define_value_str(std::string *value, const Location *loc) :
				define_value(DEF_VAL_STR, loc), value(value) {}

			/** Destructor. */
			~define_value_str() {
				delete value;
			}

			/** Print out the string. */
			std::ostream& print(std::ostream &) const;

			/** Check the type of the constant values if defined. */
			bool type_check(const DataType *type) const;

			/** emit C declarations. */
			void emit_C(std::ostream& os);

		private:
			std::string *value;

	} DefineValueStr;

	/** Class for numerical array. */
	typedef struct define_value_num_array : public define_value {

		public:
			/** Constructor.
			 *  @param value  Array of string values.
			 *  @param loc    Location of the array.
			 */
			define_value_num_array(std::vector<NumValue *> *value,
						const Location *loc) :
				define_value(DEF_VAL_NUM_ARRAY, loc), value(value) {}

			/** Destructor. */
			~define_value_num_array() {
				delete_object_list(value);
			}

			/** Print out the array. */
			std::ostream& print(std::ostream &) const;

			/** Get the size of the array. */
			int get_size() const {return value->size();}

			/** Check the type of the constant values if defined. */
			bool type_check(const DataType *type) const;

			/** emit C declarations. */
			void emit_C(std::ostream& os);

		private:
			std::vector<NumValue *> *value;

	} DefineValueNumArray;

	/** Class of string array. */
	typedef struct define_value_str_array: public define_value {

		public:
			/** Constructor.
			 *  @param value  Array of string values.
			 *  @param loc    Location of the array.
			 */
			define_value_str_array(std::vector<std::string *> *value,
						const Location *loc) :
				define_value(DEF_VAL_STR_ARRAY, loc), value(value) {}

			/** Destructor. */
			~define_value_str_array() {
				delete_object_list(value);
			}

			/** Print out the array. */
			std::ostream& print(std::ostream &) const;

			/** Check the type of the constant values if defined. */
			bool type_check(const DataType *type) const;

			/** Get the size of the array. */
			int get_size() const {return value->size();}

			/** emit C declarations. */
			void emit_C(std::ostream& os);

		private:
			std::vector<std::string *> *value;

	} DefineValueStrArray;

	/** Class for a define clause in define section. */
	typedef class define_clause : public named_typed_node,
   									public annot_owner {

		public:
			/** Constructor.
			 *  @param n   Name of the define.
			 *  @param t   Data type of the define.
			 *  @param v   Value of the define.
			 *  @param loc Location of the define clause.
			 */
			define_clause(Symbol n,
				DataType *t, DefineValue *v, const Location *loc) :
				named_typed_node(n, t, loc), value(v), dirty(true) {
				t->set_readonly(true);
			}

			/** Destructor. */
			~define_clause() {
				delete value;
			}

			/** Get the value of the define clause. */
			const DefineValue *get_value() const {return value;}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** Check the type */
			bool type_check();

			/** emit C definition. */
			void emit_C_dec(std::ostream& os);

			/** emit C declarations. */
			void emit_C(std::ostream& os);

			/** Check if dirty function. */
			bool is_dirty() const {return dirty;}

		private:
			DefineValue *value;
			bool dirty;

	} DefineClause;

	/** Class for define section. */
	typedef class define_section : public prog_section {

		public:
			/** Constructor.
			 *  @param cs  Array of clauses.
			 *  @param loc Location of the section.
			 */ 
			define_section(std::vector<DefineClause *> *cs,
				const Location *loc) :
				 prog_section(SEC_DEFINE, loc), clauses(cs), def_cache() {}

			/** Destructor. */
			~define_section() {
				delete_object_list(clauses);
			}

			/** Print out the section. */
			std::ostream& print(std::ostream &) const;

			/** Merge two define sections.
			 *  After the merge, the argument's content will be empty
			 *  @param sec The section to merge.
			 */
			void merge(define_section& sec);

			/** Get the clauses */
			const std::vector<DefineClause *> *get_clauses() const {
				return clauses;
			}

			/** type check. */
			bool type_check();

			/** emit C definition. */
			void emit_C_dec(std::ostream& os);

			/** emit C declarations. */
			void emit_C(std::ostream& os);

			/** Check annotation.
			 *  @return True if the result is good.
			 */
			bool check_annots();

		private:

			std::vector<DefineClause *> *clauses;

			/** cache for typechecking results */
			SymContext def_cache;

			friend class program;

	} DefineSection;
}

#endif
