
#include <cassert>
#include <sstream>
#include "exp_ast.hpp"
#include "operation_ast.hpp"
#include "define_ast.hpp"
#include "emit_c.hpp"

using namespace mad_ast;
using namespace mad;
using std::ostream;
using std::vector;
using std::string;
using std::stringstream;
using std::endl;

void mad_ast::emit_sym_of_var_type(Symbol sym, EnumVarType vtype,
	const string& prefix, ostream& os)
{
	Symbol osym = sym;

	switch (vtype) {
		case VT_VAR:
			os << VAR_REF(prefix, osym);
			break;
		case VT_DEF:
			os << DEF_REF(osym);
			break;
		case VT_MAC:
			os << MVR_REF(osym);
			break;
		case VT_ARG:
			os << ARG_REF(osym);
			break;
	}
}

/* emit a num expression, casting if necessary */
void mad_ast::emit_num_of_type(const NumValue *num, const DataType *type,
	ostream& os)
{
	if (type->get_type()==DT_BIT) {

		const DataTypeBit *btype = static_cast<const DataTypeBit *>(type);

		if (btype->is_signed())
			os << "_INT_CTR(";
		else
			os << "_UINT_CTR(";

		os << std::dec << btype->get_width() << ", ";

	    num->emit_C(os);
		os << ")";
	}
	else if (type->get_type()==DT_FLOAT) {

		const DataTypeFloat *ftype = static_cast<const DataTypeFloat *>(type);


		if (ftype->is_double())
			os << "_DBL_CTR(";
		else
			os << "_FLT_CTR(";

	    num->emit_C(os);
		os << ")";
	}
	else {
		assert(type->get_type()==DT_STRING);

		os << "_STR_CTR(\"";
	    num->emit_C(os);
		os << "\")";
	}
}

/** Emit an string with the code to cast if necessary */
void mad_ast::emit_str_with_cast(const string& str, const DataType *from,
	const DataType *to, ostream& os)
{
	/* see if cast necessary */
	if (*from==*(to)) {
		os << str;
		return;
	}
	
	/* needs cast, must be arith type */
	assert(valid_arith_type(from));

	if (from->get_type()==DT_BIT && to->get_type()==DT_STRING) {
		os << str << ".hex()"; //default is hex
	}
	else if (from->get_type()==DT_FLOAT && to->get_type()==DT_STRING) {
		os << str << ".sci()"; //default is sci
	}
	else if (from->get_type()==DT_BIT && to->get_type()==DT_BIT) {
		const DataTypeBit *bt1, *bt2;
		bt1 = static_cast<const DataTypeBit *>(from);
		bt2 = static_cast<const DataTypeBit *>(to);

		if (bt1->is_signed())
			os << "_I2";
		else
			os << "_U2";

		if (bt2->is_signed())
			os << "I(";
		else
			os << "U(";

		os << std::dec << bt1->get_width() << "," ;
		os << std::dec << bt2->get_width() << ",";
		os << str;
		os << ")";
	}
	else if (from->get_type()==DT_BIT && to->get_type()==DT_FLOAT) {

		const DataTypeBit *bt1;
		const DataTypeFloat *ft2;
		bt1 = static_cast<const DataTypeBit *>(from);
		ft2 = static_cast<const DataTypeFloat *>(to);

		if (bt1->is_signed())
			os << "_I2";
		else
			os << "_U2";

		if (ft2->is_double())
			os << "DBL(";
		else
			os << "FLT(";

		os << std::dec << bt1->get_width() << "," ;
		os << str;
		os << ")";
	}
	else if (from->get_type()==DT_FLOAT && to->get_type()==DT_FLOAT) {

		const DataTypeFloat *ft1;
		const DataTypeFloat *ft2;
		ft1 = static_cast<const DataTypeFloat *>(from);
		ft2 = static_cast<const DataTypeFloat *>(to);

		if (ft1->is_double())
			os << "_DBL2";
		else
			os << "_FLT2";

		if (ft2->is_double())
			os << "DBL(" << str << ")";
		else
			os << "FLT(" << str << ")";

	}
	else if (from->get_type()==DT_FLOAT && to->get_type()==DT_BIT) {

		const DataTypeFloat *ft1;
		const DataTypeBit *bt2;
		ft1 = static_cast<const DataTypeFloat *>(from);
		bt2 = static_cast<const DataTypeBit *>(to);

		if (ft1->is_double())
			os << "_DBL2";
		else
			os << "_FLT2";

		if (bt2->is_signed())
			os << "I(";
		else
			os << "U(";

		os << std::dec << bt2->get_width() << "," ;
		os << str;
		os << ")";
	}
}

/** Emit an expression, also emit casting code if the type
 *  of the expression is not the same as type */
void mad_ast::emit_exp_with_cast(Expression *exp, const DataType *type,
	const string& prefix, ostream& os)
{
	const DataType *etyp = exp->get_data_type();

	/* see if cast necessary */
	if (*type==*(etyp)) {
		exp->emit_C(os, "", prefix);
		return;
	}

	/* needs cast, must be arithmetic type */
	assert(valid_arith_type(etyp));

	if (exp->get_type()==OP_NUM) {
		NumExpression *nexp = static_cast<NumExpression *>(exp);
		emit_num_of_type(nexp->get_num(), type, os);
		return;
	}

	stringstream sstr;
	exp->emit_C(sstr, "", prefix);
	emit_str_with_cast(sstr.str(), etyp, type, os);

}

/* emit the value of an expression */
void mad_ast::emit_val_of_exp(const Expression *exp,
	const string& prefix, ostream& os)
{
	assert(exp->get_data_type()->get_type()==DT_BIT);
	if (exp->get_type()==OP_NUM) {
		const NumExpression *nexp = static_cast<const NumExpression *>(exp);
		nexp->get_num()->emit_C(os);
	}
	else {
		os << GET_VALUE "(";
		exp->emit_C(os, "", prefix);
		os << ")";
	}
}

ostream& ternary_expression::print(ostream& os) const
{
	e1->print(os);
	os << "?";
	e2->print(os);
	os << ":";
	e3->print(os);
	return os;
}

DataType *ternary_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	DataType *t1, *t2, *t3;

	if (type) return type;

	t1 = e1->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (t1==NULL) return NULL;

	if (!valid_arith_type(t1)) {
		generic_error("Bad condition data type", t1->get_loc());
	}

	t2 = e2->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	t3 = e3->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (t2==NULL || t3==NULL) return NULL;

	type = common_super_type(t2, t3);
	if (!type) {
		generic_error("Branch types incompatible", this->get_loc());
	}

	return type;
}

ostream& unary_expression::print(ostream& os) const
{
	switch (op) {
		case OP_NEG:
			os << "-";
			e->print(os);
			break;
		case OP_COM:
			os << "~";
			e->print(os);
			break;
		case OP_NOTHING:
			os << "(";
			e->print(os);
			os << ")";
			break;
		default:
			assert(0);
	}
	return os;
}

DataType *unary_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	DataType *tp = e->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (tp==NULL) return NULL;

	switch (op) {
		case OP_NEG: 
		case OP_COM:
			if (!valid_arith_type(tp)) {
				generic_error("Bad arithmetic data type", e->get_loc());
				return NULL;
			}
			else {
				this->type = common_super_type(tp, tp);
			}
			break;
		case OP_NOTHING:
			this->type = common_super_type(tp, tp);
			break;
		default:
			assert(0);
	}
	return type;
}

static char *binop[] = {
 "+", "-", "*", "/", "%", "|", "&", "^", "::", ">>", "<<",
 "<", ">", "<=", ">=", "==", "!="};

ostream& binary_expression::print(ostream& os) const
{
	assert(op<=OP_NEQ);
	e1->print(os);
	os << binop[op];
	e2->print(os);
	return os;
}

DataType *binary_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	DataType *t1, *t2;
	
	t1 = e1->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt,  otxt, ftxt);
	t2 = e2->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt,  otxt, ftxt);
	if (t1==NULL || t2==NULL) return NULL;

	if (!valid_scalar_value_type(t1) || !valid_scalar_value_type(t2))
		return NULL;

	/* if this is string computation, already scalar value type
	 * so not too much checking  */
	if (op==OP_ADD &&
		(t1->get_type()==DT_STRING || t2->get_type()==DT_STRING)) {
		type = common_super_type(t1, t2);
		return type;
	}

	switch (op) {
		case OP_ADD:
		case OP_SUB:
		case OP_MUL:
		case OP_DIV:
		case OP_MOD:
		case OP_OR:
		case OP_AND:
		case OP_XOR:
			if (!valid_arith_type(t1)) {
				generic_error("Bad arithmetic data type", e1->get_loc());
				return NULL;
			}
			if (!valid_arith_type(t2)) {
				generic_error("Bad arithmetic data type", e2->get_loc());
				return NULL;
			}
			break;
		case OP_LT:
		case OP_GT:
		case OP_LTE:
		case OP_GTE:
		case OP_EQ:
		case OP_NEQ:
			if (t1->get_type()==DT_STRING && t2->get_type()!=DT_STRING ||
				t1->get_type()!=DT_STRING && t2->get_type()==DT_STRING) {
				generic_error("Bad comparison between string with number",
					this->get_loc());
				return NULL;
			}
			break;
		case OP_CONCAT: // only bit type allowed
		case OP_RSH:
		case OP_LSH:
			if (!valid_bit_type(t1)) {
				generic_error("Bad bit operation type", e1->get_loc());
				return NULL;
			}
			if (!valid_bit_type(t2)) {
				generic_error("Bad bit operation type", e2->get_loc());
				return NULL;
			}
			break;
		default:
			assert(0);
	}

	switch(op) {

		case OP_ADD:
		case OP_SUB:
		case OP_MUL:
		case OP_DIV:
		case OP_MOD:
		case OP_OR:
		case OP_AND:
		case OP_XOR:
			type = common_super_type(t1,t2);
			break;

		case OP_LT:
		case OP_GT:
		case OP_LTE:
		case OP_GTE:
		case OP_EQ:
		case OP_NEQ: {
			if (t1->get_type()==DT_BIT && t2->get_type()==DT_BIT) {

				bool s1 = static_cast<DataTypeBit *>(t1)->is_signed();
				bool s2 = static_cast<DataTypeBit *>(t2)->is_signed();
				if (s1 != s2) {
					generic_warn("Comparing signed with unsigned",
						this->get_loc());
				}
			}

			/* creat the smallest bit type as bool */
			type = new DataTypeBit(false, 1, 0, NULL);
			type->set_readonly(true);
			break;
		}
		case OP_CONCAT: // only bit type allowed
			assert(t1->get_type()==DT_BIT);
			assert(t2->get_type()==DT_BIT);
			type = bit_concat_type(static_cast<DataTypeBit *>(t1),
					static_cast<DataTypeBit *>(t2)); 
			break;

		case OP_RSH:
		case OP_LSH:
			type = common_super_type(t1, t1);
			break;

		default:
			assert(0);
	}

	return type;
}

ostream& num_expression::print(ostream& os) const
{
	num->print(os);
	return os;
}

DataType *num_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	if (num->is_float()) {
		type = new DataTypeFloat(num->is_double(), 0, NULL);
	}
	else {
		int width;
		width = num->get_width_inbase(2);
		type = new DataTypeBit(false, width, 0, NULL);
	}
	type->set_readonly(true);
	return type;
}

ostream& str_expression::print(ostream& os) const
{
	os << "\"" << *str << "\"";
	return os;
}

DataType *str_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	type = new DataTypeString(0, NULL);
	type->set_readonly(true);
	return type;
}

ostream& var_ref_expression::print(ostream& os) const
{
	os << symbol_get_string(sym);
	return os;
}

DataType *var_ref_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	const DataType *tp = NULL;
	/* see if the variable is defined */
	if (var_ctxt && var_ctxt->has(sym)) {
		tp = var_ctxt->get_data_type(sym);
		vtype = VT_VAR;
	}
	else if (arg_ctxt && arg_ctxt->has(sym)) {
		tp = arg_ctxt->get_data_type(sym);
		vtype = VT_ARG;
	}
	else if (mvr_ctxt && mvr_ctxt->has(sym)) {
		tp = mvr_ctxt->get_data_type(sym);
		vtype = VT_MAC;
	}
	else if (def_ctxt && def_ctxt->has(sym)) {
		tp = def_ctxt->get_data_type(sym);
		vtype = VT_DEF;
	}
	else {
		undefined_ident(sym, this->get_loc());
		return NULL;
	}

	type = tp->clone(NULL);
	this->is_oper = tp->get_type()==DT_OPER;

	return type;
}

ostream& call_expression::print(ostream& os) const
{
	os << symbol_get_string(sym) << "(";
	print_object_list(os, ", ", el);
	os << ")";

	return os;
}

DataType *call_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	if (ftxt->has(sym)) {
		ftype = ftxt->get_data_type(sym)->clone(NULL);
		const FunctionSection *dc = static_cast<const FunctionSection*>
				(ftxt->get_object(sym));
		dirty = dc->is_dirty();
		ext = false;
	}
	else if (def_ctxt->has(sym)) {

		const DataType *tp = def_ctxt->get_data_type(sym);
		if (tp->get_type()!=DT_GROUP) {
			generic_error("Bad function type", this->get_loc());
			return NULL;
		}
		ftype = tp->clone(NULL);

		const DefineClause *dc = static_cast<const DefineClause*>
			(def_ctxt->get_object(sym));
		dirty = dc->is_dirty();
		ext = true;
	}
	else {
		undefined_ident(sym, this->get_loc());
		return NULL;
	}
	
	const vector<DataType *> *formals = 
			static_cast<const DataTypeGroup *>(ftype)->get_data();

	if (formals->size()!=el->size()) {
		generic_error("Argument list length mismatch", this->get_loc());
		return NULL;
	}
	else {

		for (unsigned i=0; i<formals->size(); i++) {

			DataType *tp = (*el)[i]->infer(var_ctxt, arg_ctxt,
					mvr_ctxt, def_ctxt, otxt, ftxt);

			if (!tp || !valid_farg_match(tp, (*formals)[i])) {

				generic_error("Argument type mismatch",
						(*el)[i]->get_loc());
				return NULL;
#ifdef DEBUG
if (tp) {
					tp->print(std::cerr);
					(*formals)[i]->print(std::cerr);
					std::cerr << tp->is_readonly() <<
						(*formals)[i]->is_readonly() << std::endl;
}
#endif
			}
		}
	}

	/* check to see if the expression sets string value only */
	bool str_exp = true, has_return=false;
	for (unsigned i=0; i<formals->size(); i++) {
		const DataType *dtype = (*formals)[i];
		if (!dtype->is_readonly()) {
			has_return = true;
			if (dtype->get_type()!=DT_STRING)
				str_exp = false;
		}
	}
	str_func = has_return && str_exp;


	type = new DataTypeVoid(NULL);
	type->set_readonly(true);

	return type;
}

ostream& cast_expression::print(ostream& os) const
{
	os << "(";
	ctype->print(os);
	os << ")";
	e->print(os);
	return os;
}

DataType *cast_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	DataType *tp = e->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (tp==NULL) return NULL;

	if (!valid_arith_type(tp)) {
		generic_error("Bad expression to cast", e->get_loc());
	}
	if (!valid_arith_type(ctype)) {
		generic_error("Bad type for cast", e->get_loc());
	}

	type = ctype->clone(NULL);
	type->set_readonly(true);
	return type;
}

ostream& mod_expression::print(ostream& os) const
{
	e->print(os);
	os << "." << symbol_get_string(mod);
	return os;
}

DataType *mod_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	DataType *tp = e->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (tp==NULL) return NULL;

	if (tp->get_type()==DT_OPER) {

		/* parser has converted mod to upper case */
		if (mod==Sym_syn) {

			type = new DataTypeString(0, NULL);
			type->set_readonly(true);
		}
		else if (mod==Sym_cod) {

			const DataTypeOper *otp = static_cast<const DataTypeOper *>(tp);
			const vector<Symbol> *opers = otp->get_opers();

			assert(opers->size()>0);

			const OperationSection *oper = otxt->get_object(opers->front());

			assert(oper!=NULL);

			if (!oper->has_coding()) {
				generic_error("Refering non-existing coding of operation",
						this->get_loc());
				return NULL;
			}

			type = new DataTypeBit(false, oper->get_coding_width(), 0, NULL);
			type->set_readonly(true);
		}
		else {
			generic_error("For operation variables, "
						"use modifier syn or cod to "
						"specify what you want", this->get_loc());
			return NULL;
		}
	}
	else if (mod==Sym_hex || mod==Sym_oct || mod==Sym_bin || mod==Sym_dec) {

		if (tp->get_type()!=DT_BIT) {
			generic_error("Bad type for modifier", this->get_loc());
		}
		else {
			type = new DataTypeString(0, NULL);
			type->set_readonly(true);
		}
	}
	else if (mod==Sym_sci || mod==Sym_fix) {

		if (tp->get_type()!=DT_FLOAT) {
			generic_error("Bad type for modifier", this->get_loc());
		}
		else {
			type = new DataTypeString(0, NULL);
			type->set_readonly(true);
		}
	}
	else if (mod==Sym_flt) {
		if (tp->get_type()!=DT_BIT) {
			generic_error("Bad type for modifier", this->get_loc());
		}
		else {
			const DataTypeBit *tbit = static_cast<const DataTypeBit *>(tp);
			if (tbit->get_width()==32) {
				type = new DataTypeFloat(false, 0, NULL);
				type->set_readonly(true);
			}
			else if (tbit->get_width()==64) {
				type = new DataTypeFloat(true, 0, NULL);
				type->set_readonly(true);
			}
			else {
				generic_error("Bad integer width for modifier, "
					"must be 32 or 64", this->get_loc());
			}
		}
	}
	else if (mod==Sym_bit) {
		if (tp->get_type()!=DT_FLOAT) {
			generic_error("Bad type for modifier", this->get_loc());
		}
		else {
			const DataTypeFloat *tflt = static_cast<const DataTypeFloat *>(tp);
			type = new DataTypeBit(false, tflt->is_double()?64:32, 0, NULL);
			type->set_readonly(true);
		}
	}
	else {
		generic_error("Bad modifier", this->get_loc());
	}

	return type;
}

ostream& assign_expression::print(ostream& os) const
{
	lexp->print(os);
	os << "=";
	rexp->print(os);
	return os;
}

DataType *assign_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	DataType *right = rexp->infer(var_ctxt, arg_ctxt, mvr_ctxt,
				def_ctxt, otxt, ftxt);
	DataType *left = lexp->infer(var_ctxt, arg_ctxt, mvr_ctxt,
				def_ctxt, otxt, ftxt);

	if (left==NULL || right==NULL) return NULL;

	if (left->is_readonly()) {
		generic_error("Assignment to read-only", lexp->get_loc());
			return NULL;
	}
	else {
		if (!valid_implicit_conversion(right, left)) {
			generic_error("Assignment type mismatch", this->get_loc());
#ifdef DEBUG
			left->print(std::cerr);
			right->print(std::cerr);
			std::cerr << std::endl;
#endif
			return NULL;
		}
	}

	type = new DataTypeVoid(NULL);
	return type;
}

ostream& tab_ref_expression::print(ostream& os) const
{
	base->print(os) << "[";
	ind1->print(os);
	if (ind2) ind2->print(os<<" .. ");
	os << "]";
	return os;
}

DataType *tab_ref_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{

	if (type) return type;

	DataType *tp = 
		base->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);

	if (tp==NULL) return NULL;

	DataType *t1, *t2;

	t1 = ind1->infer(var_ctxt, arg_ctxt, mvr_ctxt, def_ctxt, otxt, ftxt);
	if (t1==NULL) return NULL; 

	/* must be field extraction */
	if (tp->is_scalar()) {

		tab_ref = false;

		if (!valid_bit_type(tp)) {
			generic_error("Bad data type for field extraction",
				this->get_loc());
			return NULL;
		}

		if (!valid_bit_type(t1)) {
			generic_error("Bad data type for field indexing",
				ind1->get_loc());
			return NULL;
		}

		unsigned width = static_cast<const DataTypeBit *>(tp)->get_width();
		unsigned val1, val2;

		if (ind1->get_type()==OP_NUM) {

			const NumValue *v1 = static_cast<NumExpression *>(ind1)->get_num();

			assert(!v1->is_float());
			assert(!v1->is_negative());

			val1 = get_ulong_val(*(v1->get_value()));
			if (v1->get_width_inbase(2)>(sizeof(unsigned)*8) || val1>=width) {
				generic_error("Out-of-bound bit field access", ind1->get_loc());
				return NULL;
			}
		}
		/*  If only ind1, then it can be an expression since we know
		 *  statically the result type as 1bit.
		 *  Otherwise, if ind2 exist, then both must be number.
		 */
		else if (ind2!=NULL) {
			generic_error("Index must be constant", ind1->get_loc());
			return NULL;
		}

		/* single bit extraction */
		if (!ind2) {
			type = new DataTypeBit(false, 1, 0, NULL);
			type->set_readonly(tp->is_readonly());
		}
		else {

			t2 = ind2->infer(var_ctxt, arg_ctxt, mvr_ctxt,
						 	def_ctxt, otxt, ftxt);
			if (t2==NULL) return NULL;

			if (!valid_bit_type(t2)) {
				generic_error("Bad data type for field indexing",
					ind2->get_loc());
				return NULL;
			}

			if (ind2->get_type()==OP_NUM) {

				const NumValue *v2 =
					static_cast<NumExpression *>(ind2)->get_num();

				assert(!v2->is_float());
				assert(!v2->is_negative());

				val2 = get_ulong_val(*(v2->get_value()));
				if (v2->get_width_inbase(2)>(sizeof(unsigned)*8) ||val2>=width){
					generic_error("Out-of-bound bit field access",
						ind2->get_loc());
					return NULL;
				}
			}
			else {
				generic_error("Index must be constant", ind1->get_loc());
				return NULL;
			}

			if (val1<val2) {
				//generic_warn("Reorder index", this->get_loc());
				expression *swap = ind1;
				ind1 = ind2;
				ind2 = swap;
			}

			fwidth = 1+((val1>val2)?(val1-val2):(val2-val1));
			type = new DataTypeBit(false, fwidth, 0, NULL);
			type->set_readonly(tp->is_readonly());
		}
	}
	/* must be array access */
	else {

		tab_ref = true;

		if (!valid_bit_type(t1)) {
			generic_error("Bad arithmetic data type for array indexing",
				ind1->get_loc());
			return NULL;
		}

		if (tp->get_type()==DT_STRING) {
			type = new DataTypeString(0, NULL);
			type->set_readonly(true);
		}
		else if (tp->get_type()==DT_BIT) {

			const DataTypeBit *tbit = static_cast<const DataTypeBit *>(tp);

			type = new DataTypeBit(tbit->is_signed(),
				tbit->get_width(), 0, NULL);
			type->set_readonly(true);
		}
		else if (tp->get_type()==DT_FLOAT) {

			const DataTypeFloat *tflt = static_cast<const DataTypeFloat *>(tp);

			type = new DataTypeFloat(tflt->is_double(), 0, NULL);
			type->set_readonly(true);
		}
	}

	return type;
}

DataType *operation_expression::infer(const SymContext *var_ctxt,
	const SymContext *arg_ctxt,
	const SymContext *mvr_ctxt, const SymContext *def_ctxt,
	const OperContext *otxt, const FunContext *ftxt)
{
	if (type) return type;

	if (!var_ctxt->has(oper)) {
		undefined_ident(oper, this->get_loc());
		return NULL;
	}

	const DataType *tp = var_ctxt->get_data_type(oper);
	if (tp->get_type()!=DT_OPER) {
		generic_error("Bad operation type", this->get_loc());
		return NULL;
	}

	const DataTypeOper *otp = static_cast<const DataTypeOper *>(tp);

	// the coding matcher
	if (arg!=NULL) {

		const DataType *atp = arg->infer(var_ctxt, arg_ctxt, mvr_ctxt,
			def_ctxt, otxt, ftxt);

		if (atp==NULL)
			return NULL;

		if (!valid_coding_type(atp)) {
			generic_error("Bad argument type", this->get_loc());
			return NULL;
		}
		else {

			/* still need to check if width of coding matches */
			const DataTypeBit *btp = static_cast<const DataTypeBit *>(atp);

			unsigned width = btp->get_width();
			const vector<Symbol> *opers = otp->get_opers();

		    vector<Symbol>::const_iterator sit;
		    for (sit=opers->begin(); sit!=opers->end(); sit++) {

				const OperationSection *oper = otxt->get_object(*sit);
				if (oper && oper->get_coding_width()<width) {

					stringstream sstr;                                
					sstr << "Coding width mismatch for operation " <<
						symbol_get_string(*sit) << ", expecting " << width
						<< ", got " << oper->get_coding_width();

					generic_error(sstr.str(), this->get_loc());
					return NULL;
				}
			}
		}
	}

	type = new DataTypeVoid(NULL);
	return type;
}

ostream& decode_expression::print(ostream& os) const
{
	os << "+" << symbol_get_string(oper);
	if (arg!=NULL) {
		os << " = ";
		arg->print(os);
	}
	return os;
}

ostream& activate_expression::print(ostream& os) const
{
	os << "@" << symbol_get_string(oper);
	if (arg!=NULL) {
		os << " = ";
		arg->print(os);
	}
	return os;
}

void ternary_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	os << "_TOF(";
	type->emit_C(os);
	os << ", ";
	emit_val_of_exp(e1, prefix, os);
	os << ", ";
	emit_exp_with_cast(e2, type, prefix, os);
	os << ", ";
	emit_exp_with_cast(e3, type, prefix, os);
	os << ")";
}

void binary_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{

	switch(op) {
		case OP_ADD:
		case OP_SUB:
		case OP_MUL:
		case OP_DIV:
		case OP_MOD:
		case OP_OR:
		case OP_AND:
		case OP_XOR:
			os << "(";
			emit_exp_with_cast(e1, type, prefix, os);
			os << binop[op];
			emit_exp_with_cast(e2, type, prefix, os);
			os << ")";
			break;
		case OP_CONCAT: {
			const DataTypeBit *tbit =
				static_cast<const DataTypeBit *>(e2->get_data_type());
			os << "((";
			emit_exp_with_cast(e1, type, prefix, os);
			os << "<<" << tbit->get_width() << ") | ";
			emit_exp_with_cast(e2, type, prefix, os);
			os << ")";
			break;
		}
		case OP_RSH:
		case OP_LSH: {
			os << "(";
			emit_exp_with_cast(e1, type, prefix, os);
			os << binop[op];
			emit_val_of_exp(e2, prefix, os);
			os << ")";
			break;
		}
		case OP_LT:
		case OP_GT:
		case OP_LTE:
		case OP_GTE:
		case OP_EQ:
		case OP_NEQ: {

			DataType *tp =
				common_super_type(e1->get_data_type(), e2->get_data_type());

			os << "(";
			emit_exp_with_cast(e1, tp, prefix, os);
			os << binop[op];
			emit_exp_with_cast(e2, tp, prefix, os);
			os << ")";

			delete tp;
			break;
		}
		default:
			assert(0);
			break;
	}
}

void unary_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	os << "(";
	switch (op) {
		case OP_NEG:
			os << "-";
			e->emit_C(os, "", prefix);
			break;
		case OP_COM:
			os << "~";
			e->emit_C(os, "", prefix);
			break;
		case OP_NOTHING:
			e->emit_C(os, "", prefix);
			break;
		default:
			assert(0);
			break;
	}
	os << ")";
}

void num_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	emit_num_of_type(this->get_num(), type, os);
	return;
}

void str_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	os << "_STR_CTR(\"" << *str << "\")";
}

void var_ref_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	emit_sym_of_var_type(sym, vtype, prefix, os);
}


void call_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	os << indent;
	/* Wrap string evaluation so that simulator does not execute it.
	 * Dirty functions may have side-effects and therefore not wrapped. */
	if (str_func && !dirty)
		os << "_STRING_EVAL(";

	if (dirty && ext)
		os << DIRTY_E_FUN_REF "(" << symbol_get_string(sym);
	else if (dirty && !ext)
		os << DIRTY_FUN_REF << "(" << FUN_NAME(sym); 
	else if (!dirty && ext)
		os << E_FUN_REF "(" << symbol_get_string(sym);
	else 
		os << FUN_REF << "(" << FUN_NAME(sym); 

	if (el->size()>0) {

		int i;
		vector<expression *>::iterator it;
		const vector<DataType *> *formals = 
			static_cast<const DataTypeGroup *>(ftype)->get_data();

		for (it = el->begin(), i=0; it!=el->end(); it++, i++) {
			os << ", ";
			emit_exp_with_cast(*it, (*formals)[i], prefix, os);
		}
	}

	if (str_func && !dirty)
		os << ")";


	os << ");";
}

void cast_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	emit_exp_with_cast(e, type, prefix, os);
}

void mod_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	if (e->get_data_type()->get_type()==DT_OPER) {

		if (mod==Sym_syn) {
			e->emit_C(os, "", prefix); 
			os << "->__get_syntax()";
		}
		else {

			assert (mod==Sym_cod);

			os << "*static_cast<const ";
			
			type->emit_C(os, "");

			os << " *>(";
			e->emit_C(os, "", prefix);
			os << "->__get_coding())";
		}
	}
	else if (mod==Sym_flt) {
		os << "_U_TO_F(";
		e->emit_C(os, "", prefix); 
		os << ")";
	}
	else if (mod==Sym_bit) {
		os << "_F_TO_U(";
		e->emit_C(os, "", prefix);
		os << ")";
	}
	else {
		e->emit_C(os, "", prefix);
		os << "." << symbol_get_string(mod) << "()";
	}
}

void assign_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	os << indent;

	if (lexp->get_type()==OP_LOOKUP) {

		/* this must be deposit operation */
		TabRefExpression *le = static_cast<TabRefExpression *>(lexp);

		const DataType *ltype = le->get_data_type();

		assert(ltype->get_type()==DT_BIT);
		const DataTypeBit *btype = static_cast<const DataTypeBit *>(ltype);

		if (btype->is_signed())
			os << "_INT_DEP";
		else
			os << "_UINT_DEP";

		if (le->get_index2()!=NULL)
			os << "2(";
		else
			os << "1(";

		le->get_base()->emit_C(os, "", prefix);
		os << ", ";

		emit_val_of_exp(le->get_index1(), prefix, os);

		if (le->get_index2()!=NULL) {
			os << ", ";
			emit_val_of_exp(le->get_index2(), prefix, os);
		}

		os << ", ";
		emit_exp_with_cast(rexp, le->get_data_type(), prefix, os);
		os << ");";
	}
	else {

		/* wrap string evaluation so that simulator does not execute it*/
		if (lexp->get_data_type()->get_type()==DT_STRING)
			os << "_STRING_EVAL(";

		lexp->emit_C(os, "", prefix);
		os << " = ";
		emit_exp_with_cast(rexp, lexp->get_data_type(), prefix, os);

		if (lexp->get_data_type()->get_type()==DT_STRING)
			os << ")";

		os << ";";
	}
}

void tab_ref_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	/* must be table reference */
	if (tab_ref) {

		base->emit_C(os, "", prefix);
		os << "[";
		emit_val_of_exp(ind1, prefix, os);
		os << "]";
	}
	else {

		/* field extraction, generates uint type */
		if (ind2!=NULL) {
			os << "_UINT_XTR2(";
		}
		else {
			os << "_UINT_XTR1(";
		}

		base->emit_C(os, "", prefix);

		os << ", ";

		emit_val_of_exp(ind1, prefix, os);

		if (ind2!=NULL) {
			os << ", ";
			emit_val_of_exp(ind2, prefix, os);
		}
		os << ")";
	}
}

void decode_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	if (singleton) {
		os << indent << VAR_REF(prefix, oper)
			 << " = _GET_OPERATION("
			 << std::dec << singleton_id << ");" << endl;
	}
	else if (arg) {
		os << indent << "{" << endl;
		// need to call the decoder
		os << indent << OPER_ID_TYPE " *_temp_dtab = _DECODE("
			<< DECODER_NAME(dec_name) << ", ";
		arg->emit_C(os, "", prefix);
		os << ");" << endl;

		/* if arg is more complex than a single variable, it should
		 * be assigned to a variable first since we cannot
		 * take the address of a temporary */
		if (arg->get_type()!=OP_SYM) {
			os << indent;
			arg->get_data_type()->emit_C(os);
			os << " _temp_arg = ";
			arg->emit_C(os, "", prefix);
			os << ";" << endl;
		}

		os << indent << VAR_REF(prefix, oper)
			 << " = _GET_OPERATION(_temp_dtab[0]);" << endl;
	}
	else {

		os << indent << VAR_REF(prefix, oper) << " = _GET_OPERATION("
			<< DTABLE_REF(oper) << "[0]);" << endl;
	}

	os << indent << VAR_REF(prefix, oper) << "->__init(_machine_, "; 

	if (singleton) {
		os << "NULL, NULL);" << endl;
	}
	else if (arg) {
		os << "_temp_dtab+1";
		os << ", &";
		if (arg->get_type()!=OP_SYM)
			os << "_temp_arg";
		else
			arg->emit_C(os, "", prefix);
		os  << ");" << endl;
		os << indent << "}" << endl;
	}
	else {
		os << DTABLE_REF(oper) << "+1"
			<< ", &" <<  CODING_REF(oper) << ");" << endl;
	}
}

void activate_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	if (singleton) {
		os << indent << VAR_REF(prefix, oper)
			 << " = _GET_OPERATION("
			 << std::dec << singleton_id << ");" << endl;
	}
	else if (arg) {
		os << indent << "{" << endl;
		// need to call the decoder
		os << indent << OPER_ID_TYPE " *_temp_dtab = _DECODE("
			<< DECODER_NAME(dec_name) << ", ";
		arg->emit_C(os, "", prefix);
		os << ");" << endl;

		os << indent << VAR_REF(prefix, oper)
			 << " = _GET_OPERATION(_temp_dtab[0]);" << endl;
	}
	else {
		os << indent << VAR_REF(prefix, oper) << " = _GET_OPERATION("
			<< DTABLE_REF(oper) << "[0]);" << endl;
	}

	os << indent << "_ACT_OPERATION(" << VAR_REF(prefix, oper) << ", ";

	if (singleton) {
		os << "NULL, NULL);" << endl;
	}
	else if (arg) {
		os << "_temp_dtab+1";
		os << ", &";
		arg->emit_C(os, "", prefix);
		os  << ");" << endl;
		os << indent << "}" << endl;
	}
	else {
		os << DTABLE_REF(oper) << "+1"
			<< ", &" <<  CODING_REF(oper) << ");" << endl;
	}
}
