
#ifndef __MAD_EXP_AST_HPP__
#define __MAD_EXP_AST_HPP__

#include "mad_ast.hpp"
#include "data_type_ast.hpp"
#include "program_ast.hpp"
#include <string>
#include <vector>

namespace mad_ast {

	/** Operation types. */
	typedef enum {

			OP_ADD,
			OP_SUB,
			OP_MUL,
			OP_DIV,
			OP_MOD,

			OP_OR,
			OP_AND,
			OP_XOR,
			OP_CONCAT,		// concatenate

			OP_RSH,
			OP_LSH,

		// --------------------------

			OP_LT,
			OP_GT,
			OP_LTE,
			OP_GTE,
			OP_EQ,
			OP_NEQ,

		// --------------------------

			OP_NEG,		// negation
			OP_COM,		// complement
			OP_NOTHING,	// (exp), paranthesis

		// --------------------------

			OP_CTEST, // ? :

		// --------------------------

			OP_CALL,		// function call, should be root

		// --------------------------

			OP_CAST,	// cast
			OP_MDFY,		// modifier, from float/bit to string

		// --------------------------

			OP_ASGN,	// assign, should be root

		// --------------------------

			OP_LOOKUP,	// table lookup

		// --------------------------

			OP_NUM,
			OP_STR,
			OP_SYM,
			//OP_NONE,		// constant

		// --------------------------
			OP_DEC,		// decode operation
			OP_ACT		// activate operation

	} EnumOpType;

	/** Base class of expressions. */
	typedef class expression : public ast_node, public annot_owner {

		public:

			/** Constructor.
			 *  @param op  Operator of the operation.
			 *  @param loc Location of the operation.
			 */
			expression(const EnumOpType& op, const Location *loc) :
				ast_node(loc),  op(op), type(NULL) {}

			/** Destructor. */
			virtual ~expression() {if (type) delete type;};

			/** type infer.
			 *  @param var_ctxt Variable context.
			 *  @param arg_ctxt Argument context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define variable context.
			 *  @param otxt Operation context.
			 *  @param ftxt Function context.
			 */
			virtual DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt) = 0;

			/** get the type of the expression */
			EnumOpType get_type() const {return op;}

			/** Get the data type of the expression. */
			const DataType *get_data_type() const {return type;}

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			virtual void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix)
				   	const = 0;

		protected:

			EnumOpType op;
			DataType *type;

	} Expression;

	/** Class for conditional test operation. */
	typedef class ternary_expression : public expression {

		public:
			/** Constructor.
			 *  @param e1  Condition expression.
			 *  @param e2  True branch expression.
			 *  @param e3  False branch expression.
			 *  @param loc Location of the operation.
			 */
			ternary_expression(expression *e1, expression *e2,
				expression *e3, const Location *loc) :
				expression(OP_CTEST, loc), e1(e1), e2(e2), e3(e3) {}

			/** Destructor. */
			~ternary_expression() {
				delete e1;
				delete e2;
				delete e3;
			}

			/** Print out the expression */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

		private:
			expression *e1, *e2, *e3;

	} TernaryExpression;

	/** Class for binary operations. */
	typedef class binary_expression : public expression {

		public:
			/** Constructor.
			 *  @param op  The operator.
			 *  @param e1  The left operand.
			 *  @param e2  The right operand.
			 *  @param loc Location of the expression.
			 */
			binary_expression(const EnumOpType& op,
				expression *e1, expression *e2, const Location *loc) :
				expression(op, loc), e1(e1), e2(e2) {}

			/** Destructor. */
			~binary_expression() {
				delete e1;
				delete e2;
			}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			expression *e1, *e2;

	} BinaryExpression;

	/** Class for unary expressions. */
	typedef class unary_expression : public expression {

		public:
			/** Constructor.
			 *  @param op  The operator.
			 *  @param e   The operand.
			 *  @param loc Location of the expression.
			 */
			unary_expression(const EnumOpType& op,
				expression *e, const Location *loc) :
				expression(op, loc), e(e) {}

			/** Destructor. */
			~unary_expression() {
				delete e;
			}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			expression *e;

	} UnaryExpression;


	/** Class for numeric constant expression. */
	typedef class num_expression : public expression {

		public:
			/** Constructor.
			 *  @param s   The number value.
			 *  @param loc Location of the expression.
			 */
			num_expression(NumValue *s, const Location *loc) :
				expression(OP_NUM, loc), num(s) {}

			/** Destructor. */
			~num_expression() {delete num;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** Get the number */
			const NumValue *get_num() const {return num;}

			/** See if it is float. */
			bool is_float() const {return num->is_float();}

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			NumValue *num;

	} NumExpression;


	/** Class for std::string constant expression. */
	typedef class str_expression : public expression {

		public:
			/** Constructor.
			 *  @param s   The number value.
			 *  @param loc Location of the expression.
			 */
			str_expression(std::string *s, const Location *loc) :
				expression(OP_STR, loc), str(s) {}

			/** Destructor. */
			~str_expression() {delete str;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			std::string *str;

	} StrExpression;


	/** Class for variable reference expression. */
	typedef class var_ref_expression : public expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the variable.
			 *  @param loc  Location of the expression.
			 */
			var_ref_expression(Symbol sym, const Location *loc) :
				expression(OP_SYM, loc), sym(sym) {}

			/** Destructor. */
			~var_ref_expression() {}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


			/** Get the scope type of the variable. */
			EnumVarType get_var_type() const {return vtype;}

		private:
			Symbol sym;
			EnumVarType vtype;
			bool is_oper;

	} VarRefExpression;


	/** Class for call expression. */
	typedef class call_expression : public expression {

		public:

			/** Constructor.
			 *  @param sym Name of the function.
			 *  @param el  List of actuals.
			 *  @param loc Location of the expression.
			 */
			call_expression(Symbol sym, 
				std::vector<expression *> *el, const Location *loc) :
		   		expression(OP_CALL, loc), sym(sym), el(el), ftype(NULL) {}

			/** Destructor. */
			~call_expression() {
				delete_object_list(el);
				delete ftype;
			}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


			/** Get the name of the function. */
			Symbol get_func_name() const {return sym;}

		private:
			Symbol sym;
			std::vector<expression *> *el;

			const DataType *ftype;	/* formal type */
			bool str_func;	/* the function evaluates string only. */
			bool dirty;		/* the function is a dirty one. */
			bool ext;	/* the function is an external function. */

	} CallExpression;

	/** Unsigned cast expression. */
	typedef class cast_expression : public expression {

		public:
			/** Constructor.
			 *  @param tp  The data type to cast to.
			 *  @param e   The expression to cast.
			 *  @param loc Location of the expression.
			 */
			cast_expression(DataType *tp, expression *e,
				const Location *loc) :
				expression(OP_CAST, loc), e(e), ctype(tp) {}

			/** Destructor. */
			~cast_expression() {delete e; delete ctype;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			expression *e;
			DataType *ctype;	// the type to cast to.

	} CastExpression;

	/** Unsigned cast expression. */
	typedef class mod_expression : public expression {

		public:
			/** Constructor.
			 *  @param e   The expression to cast.
			 *  @param mod The modifier
			 *  @param loc Location of the expression.
			 */
			mod_expression(expression *e, Symbol mod,
				const Location *loc) :
				expression(OP_MDFY, loc), e(e), mod(mod) {}

			/** Destructor. */
			~mod_expression() {delete e;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

		private:
			expression *e;
			Symbol mod;

	} ModExpression;

	/** Class for assign expression. */
	typedef class assign_expression : public expression {

		public:
			/** Constructor.
			 *  @param left  Left side expression.
			 *  @param right Right side expression.
			 *  @param loc   Location of the expresison.
			 */
			assign_expression(expression *left, expression *right,
				const Location *loc) :
				expression(OP_ASGN, loc), lexp(left), rexp(right) {}

			/** Destructor. */
			~assign_expression() {delete lexp; delete rexp;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


		private:
			expression *lexp, *rexp;

	} AssignExpression;

	/** Class for table reference expression. */
	typedef class tab_ref_expression : public expression {

		public:
			/** Constructor.
			 *  @param bas  Expression of the table.
			 *  @param ind1 Index 1.
			 *  @param ind2 Index 2, may be NULL.
			 *  @param loc  Location of the expression.
			 */
			tab_ref_expression(expression *bas,
				expression *ind1, expression *ind2,
				const Location *loc) :
				expression(OP_LOOKUP, loc), base(bas), ind1(ind1), ind2(ind2) {}

			/** Destructor. */
			~tab_ref_expression() {
				delete base;
				delete ind1;
				delete ind2;
			}

			/* Get the base expression */
			const expression *get_base() const {return base;}

			/* Get index 1*/
			const expression *get_index1() const {return ind1;}

			/* Get index 2*/
			const expression *get_index2() const {return ind2;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


#if 0
			/** Get the scope type of the variable. */
			EnumVarType get_var_type() const {return vtype;}
#endif

		private:
			expression *base;
			expression *ind1;
			expression *ind2;

			unsigned fwidth;		//width of field extraction
			bool tab_ref;	//table reference(T) or field extraction(F)

//			EnumVarType vtype;

	} TabRefExpression;

	/** Class for operation decoding expression. */
	typedef class operation_expression : public expression {

		public:
			/** Constructor.
			 *  @param op   The operation.
			 *  @param sym  Name of the operation.
			 *  @param arg  Argument, the coding of the operation.
			 *  @param loc  Location of the expression.
			 */
			operation_expression(const EnumOpType& op,
				Symbol sym, Expression *arg,
				const Location *loc) :
				expression(op, loc), oper(sym), arg(arg),
		   		dec_name(mad::SymNULL), singleton(false) {}

			/** Destructor. */
			virtual ~operation_expression() {delete arg;}

			/** Get the name of the operation. */
			Symbol get_oper() const {return oper;}

			/** Get the argument. */
			const Expression *get_arg() const {return arg;}

			/** type infer */
			DataType *infer(const SymContext *var_ctxt, 
				const SymContext *arg_ctxt,
				const SymContext *mvr_ctxt, const SymContext *def_ctxt,
				const OperContext *otxt, const FunContext *ftxt);

			/** Set the name of the decoder */
			void set_decoder_name(Symbol sym) {dec_name=sym;}

			/** Set the singleton property of the operation. */
			void set_singleton(bool s, OperID id) {
				singleton = s;
				singleton_id = id;
			}

			/** Get the name of the decoder. */
			Symbol get_decoder_name() const {return dec_name;}

			/** Test if this is a singleton. */
			bool is_singleton() const {return singleton;}

			/** Get singleton id. */
			OperID get_singleton_id() const {return singleton_id;}

		protected:
			Symbol oper;
			Expression *arg;

			Symbol dec_name;
			bool singleton;
			OperID singleton_id;

	} OperationExpression;

	/** Class for operation decode expression. */
	typedef class decode_expression : public operation_expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the operation.
			 *  @param arg  Argument, the coding of the operation.
			 *  @param loc  Location of the expression.
			 */
			decode_expression(Symbol sym, Expression *arg,
				const Location *loc) :
				operation_expression(OP_DEC, sym, arg, loc) {}

			/** Destructor. */
			~decode_expression() {}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


	} DecodeExpression;

	/** Class for operation activation expression. */
	typedef class activate_expression : public operation_expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the operation.
			 *  @param arg  Argument, the coding of the operation.
			 *  @param loc  Location of the expression.
			 */
			activate_expression(Symbol sym, Expression *arg,
				const Location *loc) :
				operation_expression(OP_ACT, sym, arg, loc) {}

			/** Destructor. */
			~activate_expression() {}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** emit C code
			 *  @param os     The output stream.
			 *  @param indent The indentation for statements.
			 *  @param prefix The prefix for variable names.
			 */ 
			void emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;


	} ActivateExpression;


	/** Emit a variable name of vartype.
	 *  @param sym    The name of the variable.
	 *  @param vtype  The vartype of the variable.
	 *  @param prefix The prefix to the variable name.
	 *  @param os     The output stream.
	 */
	void emit_sym_of_var_type(Symbol sym, EnumVarType vtype,
				const std::string& prefix, std::ostream& os);

	/** Emit C++ code of a constant number of the type.
	 *  @param num The number.
	 *  @param type The type.
	 *  @param os  The output stream.
	 */
	void emit_num_of_type(const NumValue *num, const DataType *type,
				std::ostream& os);

	/** Emit a string with the C++ code to cast if necessary.
	 *  @param str The string.
	 *  @param from The cast from type.
	 *  @param to  The cast to type.
	 *  @param os  The output stream.
	 */
	void emit_str_with_cast(const std::string& str, const DataType *from,
			    const DataType *to, std::ostream& os);

	
	/** Emit an expression.
	 *  @param exp    The expression to emit.
	 *  @param type   The datatype of the variable.
	 *  @param prefix The prefix to the variable name.
	 *  @param os     The output stream.
	 */
	void emit_exp_with_cast(Expression *exp, const DataType *type,
				const std::string& prefix, std::ostream& os);

	/** Emit the value of an expression, _VAL(exp).
	 *  @param exp    The expression to emit.
	 *  @param prefix The prefix to the variable name.
	 *  @param os     The output stream.
	 */
	void emit_val_of_exp(const Expression *exp, 
				const std::string& prefix, std::ostream& os);


}

#endif
