
#include "mad_dag.hpp"
#include "machine_ast.hpp"
#include "operation_ast.hpp" //for var_clause
#include "manager_ast.hpp"	// for buf_clause
#include "emit_c.hpp"
#include <sstream>

using namespace mad;
using namespace mad_ast;
using std::ostream;
using std::vector;
using std::endl;
using std::stringstream;
using std::set;
using std::map;
using std::string;

edge_context::edge_context() : context()
{
}

void edge_context::insert(const EdgeClause *edge)
{
	assert(!this->has(edge->get_name()));
	context[edge->get_name()] = edge;
}

bool edge_context::has(Symbol sym) const
{
	return (context.find(sym)!=context.end());
}

const EdgeClause *edge_context::get_edge(Symbol sym) const
{
	map<Symbol, const EdgeClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;
	return (*it).second;
}

const Location *edge_context::get_loc(Symbol sym) const
{
	map<Symbol, const EdgeClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	const EdgeClause *a = (*it).second;
	return a->get_loc();
}

Symbol edge_context::get_from_state(Symbol sym) const
{
	map<Symbol, const EdgeClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return SymNULL;

	const EdgeClause *a = (*it).second;
	return a->get_from_state();
}

Symbol edge_context::get_to_state(Symbol sym) const
{
	map<Symbol, const EdgeClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return SymNULL;

	const EdgeClause *a = (*it).second;
	return a->get_to_state();
}


state_context::state_context() : context() 
{
}

state_context::~state_context()
{
	map<Symbol, vector<Symbol> *>::const_iterator it;
	for (it=link_context.begin(); it!=link_context.end(); it++)
		delete (*it).second;
}

void state_context::insert(const StateClause *state)
{
	assert(!this->has(state->get_name()));
	context[state->get_name()] = state;
}

void state_context::insert(Symbol state, Symbol edge)
{
	map<Symbol, vector<Symbol> *>::const_iterator it = link_context.find(state);

	if (it==link_context.end() || (*it).second==NULL) {
		vector<Symbol> *es = new vector<Symbol>(1, edge);
		link_context[state] = es;
	}
	else {
		(*it).second->push_back(edge);
	}
}

bool state_context::has(Symbol sym) const
{
	return (context.find(sym)!=context.end());
}

const StateClause *state_context::get_state(Symbol sym) const
{
	map<Symbol, const StateClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	return (*it).second;
}

const Location *state_context::get_loc(Symbol sym) const
{
	map<Symbol, const StateClause *>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	const StateClause *a = (*it).second;
	return a->get_loc();
}

const vector<Symbol> *state_context::get_edges(Symbol sym) const 
{
	map<Symbol, vector<Symbol> *>::const_iterator it = link_context.find(sym);
	if (it==link_context.end()) return NULL;

	return (*it).second;
}

ostream& buffer_clause::print(ostream& os) const
{
	os << symbol_get_string(name) << " : " << symbol_get_string(type) << "; ";
	print_object_list(os, "\n", annots);
	return os;
}

ostream& buffer_paragraph::print(ostream& os) const
{
	os << "    BUFFER  ";
	print_object_list(os, "\n            ", clauses);
	return os << endl;
}

bool buffer_paragraph::check_annots()
{
	bool ret = true;

	vector<BufferClause *>::const_iterator bit;
	for (bit=clauses->begin(); bit!=clauses->end(); bit++) {
		if (!(*bit)->check_annotations())
			ret = false;
	}
	return ret;
}

ostream& edge_clause::print(ostream& os) const
{
	os << symbol_get_string(name) << " : "
		<< symbol_get_string(from_state) << "->"
	   	<< symbol_get_string(to_state) << "; ";
	print_object_list(os, "\n", annots);
	return os;
}

ostream& edge_paragraph::print(ostream& os) const
{
	os << "    EDGE  ";
	print_object_list(os, "\n          ", edges);
	return os << endl;
}

bool edge_paragraph::check_annots()
{
	bool ret = true;

	vector<EdgeClause *>::const_iterator eit;
	for (eit=edges->begin(); eit!=edges->end(); eit++) {
		if (!(*eit)->check_annotations())
			ret = false;
	}
	return ret;
}

ostream& state_clause::print(ostream& os) const
{
	os << symbol_get_string(name);
	print_object_list(os, " ", annots);
	return os;
}

ostream& state_paragraph::print(ostream& os) const
{
	os << "    STATE  ";
	print_object_list(os, ", ", states);
	return os << ";" << endl;
}

bool state_paragraph::check_annots()
{
	bool ret = true;

	vector<StateClause *>::const_iterator sit;
	for (sit=states->begin(); sit!=states->end(); sit++) {
		if (!(*sit)->check_annotations())
			ret = false;
	}
	return ret;
}

ostream& initial_paragraph::print(ostream& os) const
{
	os << "    INITIAL  ";
	state->print(os);
	os << ";" << endl;
	return os;
}

bool initial_paragraph::check_annots()
{
	if (state) return state->check_annotations();
	return false;
}

mvariable_paragraph::~mvariable_paragraph()
{
	delete_object_list(vars);
}

ostream& mvariable_paragraph::print(ostream& os) const
{
	os << "    VAR  ";
	print_object_list(os, "\n         ", vars);
	os << endl;
	return os;
}

bool mvariable_paragraph::check_annots()
{
	bool ret = true;

	vector<VarClause *>::const_iterator vit;
	for (vit=vars->begin(); vit!=vars->end(); vit++) {
		if (!(*vit)->check_annotations())
			ret = false;
	}
	return ret;
}

ostream& machine_section::print(ostream& os) const
{
	os << "MACHINE " << symbol_get_string(name) << endl;
   
	print_object_list(os, "\n", annots);

	os << endl;

	print_object_list(os, "\n", paras);
	return os << endl;
}


void machine_section::set_paragraphs(vector<MachineParagraph *> *ps)
{
	paras = ps;
	vector<MachineParagraph *>::iterator it;
	for (it=paras->begin(); it!=paras->end(); it++) {

		switch ((*it)->get_type()) {

			case MACH_PARA_BUFFER:
				buffers.push_back(static_cast<BufferParagraph *>(*it));
				break;
			case MACH_PARA_STATE:
				states.push_back(static_cast<StateParagraph *>(*it));
				break;
			case MACH_PARA_EDGE:
				edges.push_back(static_cast<EdgeParagraph *>(*it));
				break;
			case MACH_PARA_INIT:
				inits.push_back(static_cast<InitialParagraph *>(*it));
				break;
			case MACH_PARA_VAR:
				vars.push_back(static_cast<MVarParagraph *>(*it));
				break;
		}
	}
}

machine_section::~machine_section()
{
	delete_object_list(paras);
	delete state_graph;
	delete edge_graph;
}

bool machine_section::type_check(const manager_context *m_ctxt,
	const SymContext *def_ctxt)
{
	bool ret = true;

	/* We make sure that there is one and only one initial state */
	if (inits.size()==0) {
		generic_fatal("No INITIAL state defined in MACHINE ",
			symbol_get_string(name), " at ", this->get_loc());
	}

	if (inits.size()>1) {
		stringstream sstr;
		sstr << "More than one INITIAL states in Machine "
		   << symbol_get_string(name) << endl;
		sstr << "    first at " <<  *(inits[0]->get_loc()) << endl;
		sstr << "    second at " << *(inits[1]->get_loc()); 
		fatal(sstr.str());
	}

	/* We check the naming conflict between states, edges */

	/* put initial state name in set */
	init_cache = inits[0]->get_state()->get_name();

	/* initial state is a state */
	state_cache.insert(inits[0]->get_state());
	state_set.insert(init_cache);

	/* check states names */
	for (vector<StateParagraph *>::iterator sit=states.begin();
			sit!=states.end(); sit++) {

		const vector<StateClause *> *state_clauses = (*sit)->get_states();
		vector<StateClause *>::const_iterator sym_it;

		for (sym_it = state_clauses->begin(); sym_it!=state_clauses->end();
			sym_it++) {

			Symbol state_name = (*sym_it)->get_name();
			if (state_cache.has(state_name)) {
				naming_conflict(state_name, (*sym_it)->get_loc(),
				   	state_cache.get_loc(state_name));
				ret = false;
			}
			else {
				/* cache state names */
				state_cache.insert(*sym_it);
				state_set.insert(state_name);
			}
		}
	}

	/* check edge names, in the same space as state names */
	for (vector<EdgeParagraph *>::iterator eit=edges.begin();
			eit!=edges.end(); eit++) {

		const vector<EdgeClause *> *edge_clauses = (*eit)->get_edges();
		vector<EdgeClause *>::const_iterator edge_it;

		for (edge_it = edge_clauses->begin();
				edge_it!=edge_clauses->end(); edge_it++) {

			Symbol e_name = (*edge_it)->get_name();
			if (edge_cache.has(e_name)) {
				naming_conflict(e_name, (*edge_it)->get_loc(),
						edge_cache.get_loc(e_name));
				ret = false;
			}
			else if (state_cache.has(e_name)) {
				naming_conflict(e_name, (*edge_it)->get_loc(),
						state_cache.get_loc(e_name));
				ret = false;
			}
			else {

				Symbol from = (*edge_it)->get_from_state();
				Symbol to  = (*edge_it)->get_to_state();

				/* check that the state exists */
				if (!state_cache.has(from)) {
					undefined_ident(from, (*edge_it)->get_loc());
					ret = false;
				}
				else if (!state_cache.has(to)) {
					undefined_ident(to, (*edge_it)->get_loc());
					ret = false;
				}
				else {
					state_cache.insert(from, e_name);
				}

				/* cache edges */
				edge_cache.insert(*edge_it);
				edge_set.insert(e_name);
			}
		}
	}

	/* check buffer names */
	for (vector<BufferParagraph *>::iterator bit=buffers.begin();
			bit!=buffers.end(); bit++) {

		const vector<BufferClause *> *buffer_clauses = (*bit)->get_clauses();
		vector<BufferClause *>::const_iterator buf_it;

		for (buf_it = buffer_clauses->begin();
			   	buf_it!=buffer_clauses->end(); buf_it++) {

			Symbol b_name = (*buf_it)->get_name();
			Symbol m_name = (*buf_it)->get_manager_name();

			if (!m_ctxt || !m_ctxt->has(m_name)) {
				generic_error("Undefined manager class name",
						(*buf_it)->get_loc());
				ret = false;
				continue;
			}

			if (buf_cache.has(b_name)) {
				naming_conflict(b_name, (*buf_it)->get_loc(),
					buf_cache.get_loc(b_name));
				ret = false;
			}
			else {

				(*buf_it)->dtype = m_ctxt->get_token_type(m_name)->clone(NULL);

				/* cache buffer */
				buf_cache.insert(*buf_it);

				/* check buffer types, same as to type of managers */
				if (!valid_tvalue_type((*buf_it)->dtype)) {
					generic_error("Bad buffer type", (*buf_it)->get_loc());
					ret = false;
				}
			}
		}
	}

	/* check variable names, cannot conflict with defines */
	for (vector<MVarParagraph *>::iterator vit=vars.begin();
			vit!=vars.end(); vit++) {

		const vector<VarClause *> *var_clauses = (*vit)->get_clauses();
		vector<VarClause *>::const_iterator v_it;

		for (v_it=var_clauses->begin(); v_it!=var_clauses->end(); v_it++) {

			Symbol v_name = (*v_it)->get_name();
			const DataType *vtype = (*v_it)->get_data_type();

			/* check the type of variables */
			if (!valid_mvar_type(vtype)) {
				generic_error("Bad variable type", (*v_it)->get_loc());
				ret = false;
			}

			bool conflict = false;

			/* reserved variable. */
			if (name==Sym_coding) conflict = true;

			/* reserved variable ends. */
			if (conflict) {
				generic_error("Using reserved variable name",
					(*v_it)->get_loc());
				ret = false;
			}
			else if (var_cache.has(v_name)) {
				naming_conflict(v_name, (*v_it)->get_loc(),
					var_cache.get_loc(v_name));
				ret = false;
			}
			else if (def_ctxt && def_ctxt->has(v_name)) {
				naming_conflict(v_name, (*v_it)->get_loc(),
					def_ctxt->get_loc(v_name));
				ret = false;
			}
			else {
				/* cache the variable name */
				var_cache.insert(*v_it);
			}
		}
	}

	// avoid crashing later test.
	if (!ret) return ret;

	if (state_graph) delete state_graph;
	state_graph = new mad_dag();

	/* this as a copy of initial state, used as the sink */
	state_graph->insert_vertex(SymNULL);

	set<Symbol>::iterator state_it;
	for (state_it=state_set.begin(); state_it!=state_set.end(); state_it++) {
		state_graph->insert_vertex(*state_it);
	}

	set<Symbol>::iterator edge_it;
	for (edge_it=edge_set.begin(); edge_it!=edge_set.end(); edge_it++) {
		if (edge_cache.get_to_state(*edge_it)==init_cache)
			state_graph->insert_edge(*edge_it,
				edge_cache.get_from_state(*edge_it),
				edge_cache.get_to_state(SymNULL));
		else
			state_graph->insert_edge(*edge_it,
				edge_cache.get_from_state(*edge_it),
				edge_cache.get_to_state(*edge_it));
	}

	Symbol err_state;
	if (state_graph->cycle_check(err_state)) {
		generic_error("State diagram error, cycle exists involving state ",
			symbol_get_string(err_state), this->get_loc());
		ret = false;
	}

	if (!state_graph->all_reachable(init_cache, err_state) &&
		err_state!=SymNULL) { // this sort of error will be covered next
		generic_error("State diagram error, state ",
			symbol_get_string(err_state),
			" not reachable from initial state", this->get_loc());
		ret = false;
	}

	if (!state_graph->reverse_all_reachable(SymNULL, err_state)) {
		generic_error("State diagram error, state ",
			symbol_get_string(err_state),
			" cannot return to initial state", this->get_loc());
		ret = false;
	}

	// avoid crashing later test
	if (!ret) return ret;

	if (edge_graph) delete edge_graph;
	edge_graph = new mad_dag();

	/* construct an edge graph. */
	for (edge_it=edge_set.begin(); edge_it!=edge_set.end(); edge_it++) {
		edge_graph->insert_vertex(*edge_it);
	}

	for (state_it=state_set.begin(); state_it!=state_set.end(); state_it++) {

		/** do not use initial state */
		if (*state_it==init_cache) continue;

		const mad_dag_vertex *astate = state_graph->get_vertex(*state_it);
		const vector<mad_dag_edge *> *outs = astate->get_out_edges();
		const vector<mad_dag_edge *> *ins = astate->get_in_edges();
		string state_name = symbol_get_string(*state_it);
		unsigned ind = 0;

		vector<mad_dag_edge *>::const_iterator out_it;
		for (out_it = outs->begin(); out_it!=outs->end(); out_it++) {
			vector<mad_dag_edge *>::const_iterator in_it;
			for (in_it = ins->begin(); in_it!=ins->end(); in_it++) {
				stringstream new_name;
				new_name << state_name << "$" << ind;
				edge_graph->insert_edge(symbol_find_or_insert(new_name.str()),
					(*in_it)->get_name(), (*out_it)->get_name());
				ind++;
			}
		}
	}

	return ret;
}

bool machine_section::check_annots()
{
	bool ret = check_annotations();
	
	for (vector<InitialParagraph *>::const_iterator iit=inits.begin();
		iit!=inits.end(); iit++) {
		if (!(*iit)->check_annots()) ret = false;
	}

	for (vector<StateParagraph *>::const_iterator sit=states.begin();
		sit!=states.end(); sit++) {
		if (!(*sit)->check_annots()) ret = false;
	}

	for (vector<EdgeParagraph *>::const_iterator eit=edges.begin();
		eit!=edges.end(); eit++) {
		if (!(*eit)->check_annots()) ret = false;
	}

	for (vector<BufferParagraph *>::const_iterator bit=buffers.begin();
		bit!=buffers.end(); bit++) {
		if (!(*bit)->check_annots()) ret = false;
	}

	for (vector<MVarParagraph *>::const_iterator vit=vars.begin();
		vit!=vars.end(); vit++) {
		if (!(*vit)->check_annots()) ret = false;
	}

	return false;
}

void machine_section::emit_C_dec(ostream& os) const
{
	os << endl << endl;

	os << "class " << MACH_CLASS(name)
		<< " : public _BASE_MACHINE {" << endl << endl;

	os << "  public:" << endl;

	/* emit enum for states */
	os << "    enum _STATES {" << endl;

	int i = 0;
	set<Symbol>::const_iterator sym_it;
	for (sym_it = state_set.begin(); sym_it!=state_set.end(); sym_it++,i++) {
		os << "      " << LOCAL_IDENT(*sym_it) << ",\t/*" << i << "*/" << endl;
	}

	os << "    };" << endl << endl;

	/* emit enum for edges */
	os << "    enum _EDGES {" << endl;

	i = 0;
	for (sym_it = edge_set.begin(); sym_it!=edge_set.end(); sym_it++,i++) {
		os << "      " << LOCAL_IDENT(*sym_it) << ",\t/*" << i << "*/" << endl;
	}

	os << "    };" << endl << endl;

	os << "    " << MACH_CLASS(name) << "(_SIMULATOR *);" <<  endl << endl;

	os << "    ~" << MACH_CLASS(name) << "();" <<  endl << endl;

	os << "    bool __activate();" << endl << endl;
	os << "    void __retire();" << endl << endl;
	os << "    void __flush();" << endl << endl;

	//os << "  private:" << endl;

	os << "    //variables" << endl;
	SymContext::const_iterator sc_it;
	for (sc_it = var_cache.begin(); sc_it!=var_cache.end(); sc_it++) {

		const named_typed_node *vnode = (*sc_it).second;

		os << "    ";
		vnode->get_data_type()->emit_C(os);
		os << " " << MVR_NAME(vnode->get_name()) << ";" << endl;
	}

	os << endl;
		
	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	os << "    //buffers" << endl;
	BufContext::const_iterator bf_it;
	for (bf_it = buf_cache.begin(); bf_it!=buf_cache.end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		os << "    " << MANG_CLASS(vnode->get_manager_name()) << "::TOKEN_T *"
			<< BUF_NAME(vnode->get_name()) << ";" << endl;
	}
	os << "#endif" << endl;
	

	os << "};" << endl;
}

void machine_section::emit_C(ostream& os) const
{
	os << endl << endl;

	/* emit constructor */
	os << MACH_CLASS(name) << "::" 
		<< MACH_CLASS(name) << "(_SIMULATOR *sim) : " 
		<< "_BASE_MACHINE(sim, " << std::dec
	   	<< id << ", " << edge_set.size() << ") {"
		<<  endl << endl;

	os << "  _state_ = " << LOCAL_IDENT(init_cache) << ";" <<endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	BufContext::const_iterator bf_it;
	for (bf_it = buf_cache.begin(); bf_it!=buf_cache.end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		os << "  " << BUF_NAME(vnode->get_name()) << " = NULL;";
	   	os << endl;
	}
	os << "#endif" << endl;

	os << endl;
	
	os << "}" << endl << endl;

	os << MACH_CLASS(name) << "::" 
		<< "~" << MACH_CLASS(name) << "() {" <<  endl << endl;

	os << "}" << endl << endl;

	os << "bool " << MACH_CLASS(name) << "::__activate() {" << endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	os << "  switch (_state_) {" << endl;

	set<Symbol>::const_iterator sym_it;
	for (sym_it = state_set.begin(); sym_it!=state_set.end(); sym_it++) {

		os << "    case " << LOCAL_IDENT(*sym_it) << " : " << endl;

		const vector<Symbol> *edges = state_cache.get_edges(*sym_it);
		if (edges) {
			os << "      if (false) {}" << endl;
			vector<Symbol>::const_iterator eit;
			for (eit=edges->begin(); eit!=edges->end(); eit++) {
				os << "      else if (_FIRE_EDGE(" << LOCAL_IDENT(*eit)
					<< ")) " << endl;
				os << "        _state_ = " <<
					LOCAL_IDENT(edge_cache.get_to_state(*eit)) << ";" << endl;
			}
			os << "      else" << endl
				<< "        return true;" << endl;
		}

		os << "      break;";
	   	os << endl << endl;
	}

	os << "  }" << endl;
	os << "#endif" << endl;

	os << "  return (_state_!=" << LOCAL_IDENT(init_cache) << ");"
		  << endl;

	os << "}" << endl << endl;

	os << "void " << MACH_CLASS(name) << "::__retire() {" << endl << endl;

	os << "  _state_ = " << LOCAL_IDENT(init_cache) << ";"
		<< endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	for (bf_it = buf_cache.begin(); bf_it!=buf_cache.end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		os << "  _RETIRE_BUFFER(" << BUF_NAME(vnode->get_name()) << ");"
			<< endl<< endl;
	}
	os << "#endif" << endl;

	os << "  _BASE_MACHINE::__retire();" << endl << endl;

	os << "}" << endl << endl;

	os << "void " << MACH_CLASS(name) << "::__flush() {" << endl << endl;

	os << "  _state_ = " << LOCAL_IDENT(init_cache) << ";"
		<< endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	for (bf_it = buf_cache.begin(); bf_it!=buf_cache.end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		os << "  _FLUSH_BUFFER(" << BUF_NAME(vnode->get_name()) << ");"
			<< endl<< endl;
	}
	os << "#endif" << endl;

	os << "  _BASE_MACHINE::__flush();" << endl << endl;

	os << "}" << endl << endl;

}

bool machine_section::edge_dominate(Symbol a, Symbol b) const
{
	assert(edge_graph);

	if (!edge_cache.has(a) || !edge_cache.has(b)) return false;
	return edge_graph->dominate(a,b);
}

bool machine_section::edge_reachable(Symbol a, Symbol b) const
{
	assert(edge_graph);

	if (!edge_cache.has(a) || !edge_cache.has(b)) return false;
	return edge_graph->reachable(a,b);
}

const EdgeClause *machine_section::get_edge(Symbol name) const
{
	return edge_cache.get_edge(name);
}

const StateClause *machine_section::get_state(Symbol name) const
{
	return state_cache.get_state(name);
}
