
#ifndef __MAD_MACHINE_AST_H__
#define __MAD_MACHINE_AST_H__

#include <vector>
#include <iostream>
#include <map>
#include <set>
#include "mad_dag.hpp"
#include "program_ast.hpp"
#include "data_type_ast.hpp"

/* forward declaration */
namespace mad_ast {

	/** Class for OSM edge clause in machine section */
	typedef struct edge_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param name Name of the edge.
			 *  @param from Name of the from state.
			 *  @param to   Name of the to state.
			 *  @param loc  Location of the edge clause.
			 */
			edge_clause(Symbol name, Symbol from,
				Symbol to, const Location *loc) :
				ast_node(loc), name(name), from_state(from), to_state(to) {}

			/** Destructor. */
			~edge_clause() {}

			/** Print out the edge clause. */
			std::ostream& print(std::ostream &) const;

			/** Get the name of the edge */
			Symbol get_name() const {return name;}

			/** Get the from state */
			Symbol get_from_state() const {return from_state;}

			/** Get the return state */
			Symbol get_to_state() const {return to_state;}

		private:
			Symbol name;
			Symbol from_state, to_state;

	} EdgeClause;

	/** Edge context class, symbol->(symbol, symbol, location). */
	typedef class edge_context  {

		public:

			/** Default constructor. */
			edge_context();

			/** Insert an entry into the context.
			 *  @param edge The edge clause.
			 */
			void insert(const EdgeClause *edge);

			/** Check if a symbol exists in the context. */
			bool has(Symbol sym) const;

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The edge clause, NULL if symbol not found.
			 */
			const EdgeClause *get_edge(Symbol sym) const;

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The location of the symbol, NULL if symbol not found.
			 */
			const Location *get_loc(Symbol sym) const;

			/** Get the index type of a manager.
			 *  @param sym  The symbol to lookup.
			 *  @return The index type of the symbol, NULL if symbol not found.
			 */
			Symbol get_from_state(Symbol sym) const;

			/** Get the token type of a manager.
			 *  @param sym  The symbol to lookup.
			 *  @return The token type of the symbol, NULL if symbol not found.
			 */
			Symbol get_to_state(Symbol sym) const;

		private:

			std::map<Symbol, const EdgeClause *> context;

	} EdgeContext;

	/** Class for OSM state clause in machine section */
	typedef struct state_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param name Name of the edge.
			 *  @param loc  Location of the edge clause.
			 */
			state_clause(Symbol name, const Location *loc) :
				ast_node(loc), name(name) {}

			/** Destructor. */
			~state_clause() {}

			/** Print out the edge clause. */
			std::ostream& print(std::ostream &) const;

			/** Get the name of the edge */
			Symbol get_name() const {return name;}

		private:
			Symbol name;

	} StateClause;

	/** State context class, symbol->(symbol, vector<symbol>). */
	typedef class state_context  {

		public:

			/** Default constructor. */
			state_context();

			/** Destructor. */
			~state_context();

			/** Insert an entry into the context.
			 *  @param state The state.
			 */
			void insert(const StateClause *state);

			/** Insert an entry into the context.
			 *  @param state The from state.
			 *  @param edge  The edge.
			 */
			void insert(Symbol state, Symbol edge);

			/** Check if a symbol exists in the context. */
			bool has(Symbol sym) const;

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The state clause, NULL if symbol not found.
			 */
			const StateClause *get_state(Symbol sym) const;

			/** Get the location of a state symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The location of the symbol, NULL if symbol not found.
			 */
			const Location *get_loc(Symbol sym) const;

			/** Get the edges connecting to a state */
			const std::vector<Symbol> *get_edges(Symbol sym) const;

		private:

			std::map<Symbol, const StateClause *> context;
			std::map<Symbol, std::vector<Symbol> *> link_context;

	} StateContext;

	typedef enum {
		MACH_PARA_BUFFER,
		MACH_PARA_STATE,
		MACH_PARA_EDGE,
		MACH_PARA_INIT,
		MACH_PARA_VAR
	} EnumMachineParaType;

	/** Base class for paragraphs in MACHINE section */
	typedef class machine_paragraph : public ast_node {

		public:

			/** Constructor.
			 *  @param tp   Category of the paragraph.
			 *  @param loc  Location of the paragraph.
			 */
			machine_paragraph(EnumMachineParaType tp,
				const Location *loc) : ast_node(loc), tp(tp) {}

			/** Destructor. */
			virtual ~machine_paragraph() {}

			/** get the type of the paragraph. */
			EnumMachineParaType get_type() const {return tp;}

		private:
			EnumMachineParaType tp;

	} MachineParagraph;
	
	class manager_context;

	typedef class buffer_paragraph BufferParagraph;
	typedef class state_paragraph StateParagraph;
	typedef class edge_paragraph EdgeParagraph;
	typedef class initial_paragraph InitialParagraph;
	typedef class mvariable_paragraph MVarParagraph;

	/* create a class instead of simply typedef to minimize dependency
	 * between header files. */
	/** class for buffer environment */
	typedef class buf_context : public obj_context<class buffer_clause> {
	} BufContext;

	/** Class for machine section */
	typedef class machine_section : public prog_section, public annot_owner {

		public:
			/** Constructor.
			 */
			machine_section() : prog_section(SEC_MACHINE, NULL),
				name(SymNULL), paras(NULL),
				state_graph(NULL), edge_graph(NULL) {}

			/** Destructor. */
			~machine_section();

			/** Set the name of the machine.
			 *  @param n Name of the machine.
			 */
			void set_name(Symbol n) {name=n;}

			/** Set the paragraphs. 
			 *  @param ps The paragraphs.
			 */
			void set_paragraphs(std::vector<MachineParagraph *> *ps);

			/** Get the name of the machine. */
			Symbol get_name() const {return name;}

			/** Get the name of the initial state.
			 *  Callable only after type_check.
			 */
			Symbol get_initial_state() const {return init_cache;}

			/** Get all state names.
			 *  Callable only after type_check.
			 */
			const std::set<Symbol> *get_states() const {return &state_set;}

			/** Get all edge names.
			 *  Callable only after type_check.
			 */
			const std::set<Symbol> *get_edges() const {return &edge_set;}

			/** Get one edge clause. Callable only after type_check.
			 *  @param name Name of the edge.
			 */
			const EdgeClause *get_edge(Symbol name) const;

			/** Get one state clause. Callable only after type_check.
			 *  @param name Name of the state .
			 */
			const StateClause *get_state(Symbol name) const;

			/** Print out the section */
			std::ostream& print(std::ostream &) const;

			/** Checking the declaration.
			 *  @param m_ctxt The managers context.
			 *  @param d_ctxt The defines context.
			 */
			bool type_check(const manager_context *m_ctxt,
						   	const SymContext *d_ctxt);

			/** Emit C++ code */
			void emit_C(std::ostream &os) const;

			/** Emit C++ class declaration. */
			void emit_C_dec(std::ostream &os) const;

			/** Get the id of the machine. */
			MachID get_id() const {return id;}

			/** Set the id of the machine. */
			void set_id(MachID nid) {id=nid;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

			/** Check if edge a dominates edge b.
			 *  @param a Edge 1.
			 *  @param b Edge 2.
			 *  @return True if a dominates b.
			 */
			bool edge_dominate(Symbol a, Symbol b) const;

			/** Check if edge b reachable from a.
			 *  @param a Edge 1.
			 *  @param b Edge 2.
			 *  @return True if a can reach b.
			 */
			bool edge_reachable(Symbol a, Symbol b) const;

			/** Get variables. */
			const SymContext *get_variables() const {return &var_cache;}

			/** Get the buffers. */
			const BufContext *get_buffers() const {return &buf_cache;}

			/** Get the state graph. */
			const mad::mad_dag *get_state_graph() const {return state_graph;}

			/** Get the edge graph. */
			const mad::mad_dag *get_edge_graph() const {return edge_graph;}

		private:

			Symbol name;
			std::vector<MachineParagraph *> *paras;

			std::vector<InitialParagraph *> inits;
			std::vector<BufferParagraph *> buffers;
			std::vector<StateParagraph *> states;
			std::vector<EdgeParagraph *> edges;
			std::vector<MVarParagraph *> vars;

			/* cache for type checking result */
			Symbol init_cache;
			EdgeContext edge_cache;
			StateContext state_cache;

			BufContext buf_cache;
			SymContext var_cache;

			/* For the convenience of C code emitting */
			std::set<Symbol> state_set;
			std::set<Symbol> edge_set;

			MachID id;

			mad::mad_dag *state_graph;
			mad::mad_dag *edge_graph;

			//friend class program;
			friend class operation_section;

	} MachineSection;


	/** Class for buffer clause in buffer paragraph of machine section. */
	typedef struct buffer_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param name  Name of the buffer.
			 *  @param type  Manager class name of the buffer.
			 *  @param loc   Location of the buffer.
			 */
			buffer_clause(Symbol name, Symbol type,
				const Location *loc) :
				ast_node(loc), name(name), type(type), dtype(NULL) {}

			/** Destructor. */
			~buffer_clause() {delete dtype;}

			/** Print out the buffer clause. */
			std::ostream& print(std::ostream &) const;

			/** Get the name of the buffer. */
			Symbol get_name() const {return name;}

			/** Get the name of the manager class. */
			Symbol get_manager_name() const {return type;}

			/** Get the data type of the buffer.
			 *  Can be called only after typechecking.
			 */
			const DataType *get_data_type() const {return dtype;}

		private:
			Symbol name;
			Symbol type;
			const DataType *dtype;	// token type

			friend class machine_section;

	} BufferClause;

	/** Class for a buffer paragraph in machine section. */
	typedef class buffer_paragraph : public machine_paragraph {

		public:

			/** Constructor.
			 *  @param cl  Array of buffer clauses.
			 *  @param loc Location of the paragraph.
			 */
			buffer_paragraph(std::vector<buffer_clause *> *cl,
				const Location *loc) :
				machine_paragraph(MACH_PARA_BUFFER, loc), clauses(cl) {}

			/** Destructor */
			~buffer_paragraph() {
				delete_object_list(clauses);
			}

			/** Print out the paragraph */
			std::ostream& print(std::ostream &) const;

			/** Get the clauses */
			const std::vector<BufferClause *> *get_clauses() const {
				return clauses;
			}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

			std::vector<BufferClause *> *clauses;

	} BufferParagraph;


	/** Class for OSM edge paragraph declaration in machine section*/
	typedef class edge_paragraph : public machine_paragraph {

		public:
			/** Constructor.
			 *  @param el  Array of edge clauses.
			 *  @param loc Location of the paragraph.
			 *  @sa See also edge_clause.
			 */
			edge_paragraph(std::vector<EdgeClause *> *el, const Location *loc) :
				machine_paragraph(MACH_PARA_EDGE, loc), edges(el) {}

			/** Destructor. */
			~edge_paragraph() {
				delete_object_list(edges);
			}

			/** Print out the paragraph */
			std::ostream& print(std::ostream &) const;

			/** get the edge clauses */
			const std::vector<EdgeClause *> *get_edges() const {return edges;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

			std::vector<EdgeClause *> *edges;

	} EdgeParagraph;


	/** Class for OSM state paragraph declaration in machine section. */
	typedef class state_paragraph : public machine_paragraph {

		public:
			/** Constructor.
			 *  @param s   Array of state names.
			 *  @param loc Location of the paragraph.
			 */
			state_paragraph(std::vector<StateClause *> *s,
				const Location *loc) :
				machine_paragraph(MACH_PARA_STATE, loc), states(s) {}
	
			/** Destructor. */
			~state_paragraph() {
				delete_object_list(states);
			}

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** get the names of the states */
			const std::vector<StateClause *> *get_states() const {return states;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

			std::vector<StateClause *> *states;

	} StateParagraph;


	/** Class for initial state paragraph in machine section. */
	typedef class initial_paragraph : public machine_paragraph {

		public:
			/** Constructor.
			 *  @param s   The initial state.
			 *  @param loc Location of the declaration.    
			 */
			initial_paragraph(StateClause *s, const Location *loc) :
				machine_paragraph(MACH_PARA_INIT, loc), state(s) {}

			/** Destructor. */
			~initial_paragraph() {}

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** Get the name of the state */
			StateClause *get_state() const {return state;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

			StateClause *state;

	} InitialParagraph;


	typedef class variable_clause VarClause;

	/** Class of variable paragraph declaration. */           
	typedef class mvariable_paragraph : public machine_paragraph {

		public:
			/** Constructor.
			 *  @param vl   Array of var clauses
			 *  @param loc  Location of the paragraph 
			 */
			mvariable_paragraph(std::vector<VarClause *> *vl,
				const Location *loc) :
				machine_paragraph(MACH_PARA_VAR, loc), vars(vl) {};

			/** Destructor. */
			~mvariable_paragraph();

			/** Print out the paragraph. */                   
			std::ostream& print(std::ostream &) const;
			
			/** get the variable clauses */
			const std::vector<VarClause *> *get_clauses() const {return vars;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

	private:
			std::vector<VarClause *> *vars;                        

	} MVarParagraph;             
}

#endif
