
#include "manager_ast.hpp"
#include "emit_c.hpp"

using namespace mad;
using namespace mad_ast;

using std::ostream;
using std::vector;
using std::endl;
using std::pair;
using std::map;

manager_context::manager_context() : context()
{
}

void manager_context::insert(Symbol sym, Symbol cn,
	const DataType *i, const DataType *t, const Location *l)
{
	ManagerData a = {cn,i,t,l};
	context[sym] = a;
}

bool manager_context::has(Symbol sym) const
{
	return (context.find(sym)!=context.end());
}

const Location *manager_context::get_loc(Symbol sym) const
{
	map<Symbol, ManagerData>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	ManagerData a = (*it).second;
	return a.loc;
}

const Symbol manager_context::get_class_name(Symbol sym) const
{
	map<Symbol, ManagerData>::const_iterator it = context.find(sym);
	if (it==context.end()) return SymNULL;

	ManagerData a = (*it).second;
	return a.cname;
}

const DataType *manager_context::get_index_type(Symbol sym) const
{
	map<Symbol, ManagerData>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	ManagerData a = (*it).second;
	return a.ind_type;
}

const DataType *manager_context::get_token_type(Symbol sym) const
{
	map<Symbol, ManagerData>::const_iterator it = context.find(sym);
	if (it==context.end()) return NULL;

	ManagerData a = (*it).second;
	return a.tok_type;
}

ostream& manager_class_clause::print(ostream& os) const
{
	os << symbol_get_string(cname) << " : ";
	from_type->print(os);
	os << " -> ";
	to_type->print(os);
	os << "; ";

	print_object_list(os, "\n", annots);

	return os;
}

ostream& manager_inst_clause::print(ostream& os) const
{
	os << symbol_get_string(iname) << " : " << symbol_get_string(cname);
	os << "; ";

	print_object_list(os, "\n", annots);

	return os;
}

ostream& manager_section::print(ostream& os) const
{
	os << "MANAGER" << endl;
	assert (clss->size()>0);
	assert (insts->size()>0);

	os << endl;
	os << "    CLASS" << endl << "         ";
	print_object_list(os, "\n         ",  clss);
	os << endl << endl;

	os << "    INSTANCE" << endl << "         ";
	print_object_list(os, "\n         ",  insts);
	os << endl << endl;
	return os;
}

void manager_section::emit_C_dec(ostream& os) const
{
	vector<ManagerInstClause *>::const_iterator iit;
	for (iit=insts->begin(); iit!=insts->end(); iit++) {

		Symbol iname = (*iit)->get_name();
		Symbol cname = (*iit)->get_class_name();

		os << "MANG_DEF(" << symbol_get_string(cname) << ", "
			<< symbol_get_string(iname) << ")" << endl;
	}
	os << "#undef MANG_DEF" << endl;
}

void manager_section::merge(manager_section& two)
{
	clss->insert(clss->end(), two.clss->begin(), two.clss->end());
	two.clss->erase(two.clss->begin(), two.clss->end());

	insts->insert(insts->end(), two.insts->begin(), two.insts->end());
	two.insts->erase(two.insts->begin(), two.insts->end());
}

bool manager_section::check_annots()
{
	bool ret = true;

	vector<ManagerClassClause *>::const_iterator cit;
	for (cit=clss->begin(); cit!=clss->end(); cit++) {
		if (!(*cit)->check_annotations())
			ret = false;
	}

	vector<ManagerInstClause *>::const_iterator iit;
	for (iit=insts->begin(); iit!=insts->end(); iit++) {
		if (!(*iit)->check_annotations())
			ret = false;
	}

	return ret;
}

bool manager_section::type_check()
{
	bool ret = true;

	/* We check the naming conflict between managers and manager classes*/

	vector<ManagerClassClause *>::iterator it;
	for (it=clss->begin(); it!=clss->end(); it++) {

		Symbol cname = (*it)->get_name();
		const DataType *from = (*it)->get_from_type();
		const DataType *to = (*it)->get_to_type();

		if (!valid_tindex_type(from)) {
			generic_error("Bad token index type ", from->get_loc());
			ret = false;
		}

		if (!valid_tvalue_type(to)) {
			generic_error("Bad token value type ", to->get_loc());
			ret = false;
		}

		if (class_cache.has(cname)) {
			naming_conflict(cname,
				class_cache.get_loc(cname), (*it)->get_loc());
			ret = false;
		}
		else {
			class_cache.insert(cname, cname, from, to, (*it)->get_loc());
		}
	}

	vector<ManagerInstClause *>::iterator iit;
	for (iit=insts->begin(); iit!=insts->end(); iit++) {

		Symbol iname = (*iit)->get_name();
		Symbol cname = (*iit)->get_class_name();
		const Location *iloc = (*iit)->get_loc();

		/* manager instance should not have the same name as manager class */
		if (class_cache.has(iname)) {
			naming_conflict(iname, class_cache.get_loc(cname), iloc);
			ret = false;
		}
		/* cannot conflict with class name since in emitted C++ code, they
		 * are in the same name space */
		else if (man_cache.has(iname)) {
			naming_conflict(iname, man_cache.get_loc(iname), iloc);
			ret = false;
		}
		else if (!class_cache.has(cname)) {
			undefined_ident(cname, iloc);
			ret = false;
		}
		else {
			const DataType *from = class_cache.get_index_type(cname);
			const DataType *to = class_cache.get_token_type(cname);
			man_cache.insert(iname, cname, from, to, iloc);
		}
	}
	return ret;
}

