
#ifndef __MAD_MANAGER_AST_H__
#define __MAD_MANAGER_AST_H__

#include <vector>
#include <iostream>
#include <map>
#include "program_ast.hpp"
#include "data_type_ast.hpp"

namespace mad_ast {


	/** Manager context class, symbol->(data_type, data_type, location). 
	 *  The class can be sued to store manager classes or manager instances.
	 *  When used for manager classes, the get_class_name() method will
	 *  return the same thing as its argument. */
	typedef class manager_context  {

		public:
			/** Default constructor. */
			manager_context();

			/** Insert an entry into the context.
			 *  @param sym  The symbol.
			 *  @param cn   Class name of the manager.
			 *  @param i    Index type of the manager.
			 *  @param t    Token type of the symbol.
			 *  @param l    Location of the symbol.
			 */
			void insert(Symbol sym, Symbol cn, const DataType *i,
						const DataType *t, const Location *l); 

			/** Check if a symbol exists in the context. */
			bool has(Symbol sym) const;

			/** Get the location of a symbol.
			 *  @param sym  The symbol to lookup.
			 *  @return The location of the symbol, NULL if symbol not found.
			 */
			const Location *get_loc(Symbol sym) const;

			/** Get the class name of a manager.
			 *  @param sym  The symbol to lookup.
			 *  @return The name of the class.
			 */
			const Symbol get_class_name(Symbol sym) const;

			/** Get the index type of a manager.
			 *  @param sym  The symbol to lookup.
			 *  @return The index type of the symbol, NULL if symbol not found.
			 */
			const DataType *get_index_type(Symbol sym) const;

			/** Get the token type of a manager.
			 *  @param sym  The symbol to lookup.
			 *  @return The token type of the symbol, NULL if symbol not found.
			 */
			const DataType *get_token_type(Symbol sym) const;

		private:

			/** Manager information. */
			typedef struct manager_data {
				Symbol cname;
				const DataType *ind_type;
				const DataType *tok_type;
				const Location* loc;
			} ManagerData;

			std::map<Symbol, ManagerData> context;

	} ManagerContext;


	/** Class for manager class declaration. */
	typedef class manager_class_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param cname  Name of the manager class.
			 *  @param t1     Id type of the manager.
			 *  @param t2     Token type of the manager.
			 *  @param loc    Location of the declaration.
			 */
			manager_class_clause(Symbol cname,
				DataType *t1, DataType *t2, const Location *loc) :
				ast_node(loc), cname(cname), from_type(t1), to_type(t2) {}

			/** Destructor */
			~manager_class_clause() {
				delete from_type;
				delete to_type;
			}

			/** Print out the manager class. */
			std::ostream& print(std::ostream &) const;

			/** get the name of the manager class */
			Symbol get_name() {return cname;}

			/** Get the from type of the manager class */
			DataType *get_from_type() {return from_type;}

			/** Get the to type of the manager class */
			DataType *get_to_type() {return to_type;}

		private:
			Symbol cname;	//class name
			DataType *from_type, *to_type;
		
	} ManagerClassClause;


	/** Class for manager instance declaration clauses. */
	typedef class manager_inst_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param iname  Name of the manager class.
			 *  @param cname  Name of the manager class.
			 *  @param loc    Location of the declaration.
			 */
			manager_inst_clause(Symbol iname, Symbol cname,
				const Location *loc) :
				ast_node(loc), iname(iname), cname(cname) {}

			/** Destructor */
			~manager_inst_clause() {}

			/** Print out the manager instant clause. */
			std::ostream& print(std::ostream &) const;

			/** get the name of the manager class */
			Symbol get_name() {return iname;}

			/** Get the from type of the manager class */
			Symbol get_class_name() {return cname;}

		private:
			Symbol iname;	//instance name
			Symbol cname;	//class name

	} ManagerInstClause;

#if 0
	/** Class of manager declaration. */
	typedef struct manager_clause : public ast_node  {

		public:
			/** Constructor.
			 *  @param cname  Name of the manager class.
			 *  @param t1     Id type of the manager.
			 *  @param t2     Token type of the manager.
			 *  @param ms     Instances of the manager class.
			 *  @param loc    Location of the declaration.
			 */
			manager_clause(Symbol cname, DataType *t1, DataType *t2,
				std::vector<Symbol> *ms, const Location *loc) :
				ast_node(loc), cname(cname), from_type(t1), to_type(t2),
				ms(ms) {}

			/** Destructor */
			~manager_clause() {
				delete from_type;
				delete to_type;
				delete ms;
			}

			/** Print out the manager */
			std::ostream& print(std::ostream &) const;

			/** get the name of the manager class */
			Symbol get_name() {return cname;}

			/** Get the from type of the manager class */
			DataType *get_from_type() {return from_type;}

			/** Get the to type of the manager class */
			DataType *get_to_type() {return to_type;}

			/** Get the name of the managers */
			const std::vector<Symbol> *get_managers() {return ms;}

		private:
			Symbol cname;	//class name
			DataType *from_type, *to_type;
			std::vector<Symbol> *ms;
		
	} ManagerClause;
#endif

	/** Class of the manager section declaration */
	typedef class manager_section : public prog_section {

		public:
			/** Constructor.
			 *  @param cs   Array of manager class declaration clauses.
			 *  @param is   Array of manager instance declaration clauses.
			 *  @param loc  Location of the section.
			 *  @sa See also manager_clause.
			 */  
			manager_section(std::vector<ManagerClassClause *> *cs,
				std::vector<ManagerInstClause *> *is, 
				const Location *loc) :
				prog_section(SEC_MANAGER, loc), clss(cs), insts(is) {}

			/** Destructor */
			~manager_section() {
				delete_object_list(clss);
				delete_object_list(insts);
			}

			/** Print out the section */
			std::ostream& print(std::ostream &) const;

			/** Checking the declaration */
			bool type_check();

			/** Merge two define sections.
			 *  After the merge, the argument's content will be empty
			 *  @param sec The section to merge.
			 */
			void merge(manager_section& sec);

			/** emit C declarations. */
			void emit_C_dec(std::ostream &) const;

			/** Check annotation.
			 *  @return True if the result is good.
			 */
			bool check_annots();

			/** Return all the classes. */
			const std::vector<ManagerClassClause *> *get_classes() const {
				return clss;
			}

			/** Return all the instances. */
			const std::vector<ManagerInstClause *>  *get_instances() const {
				return insts;
			}

		private:

			std::vector<ManagerClassClause *> *clss;
			std::vector<ManagerInstClause *>  *insts;

			// manager class -> (index type, token type)
			ManagerContext class_cache;

			// manager -> (index type, token type)
			ManagerContext man_cache;

			friend class program;

	} ManagerSection;
}

#endif
