
#include <sstream>
#include "misc.hpp"
#include <gmp.h>

using std::ostream;
using std::string;
using std::vector;
using std::endl;
using std::cerr;
using std::stringstream;
using namespace mad;
using namespace mad_ast;
using mad::symbol_get_string;

static int n_errors, n_warns;
static bool show_warn;

#define SYM_DEC(_a) Symbol mad_ast::Sym_ ## _a;
#include "symbol_list.hpp"

void mad_ast::mad_ast_init(bool quiet)
{
	n_errors = 0;
	n_warns = 0;
	show_warn = !quiet;

#define SYM_DEC(_a) Sym_ ## _a = symbol_find_or_insert(#_a);
#include "symbol_list.hpp"
}

ostream& mad_ast::operator << (ostream& os, const struct yyltype_t& loc) {
	if (loc.first_line != loc.last_line)
		os << loc.first_line << "." << loc.first_column << "-"
			<< loc.last_line << "." << loc.last_column - 1;
	else if (loc.first_column < loc.last_column - 1)
		os << loc.first_line << "." << loc.first_column << "-"
			<< loc.last_column - 1;
	else
		os << loc.first_line << "." << loc.first_column;
	return os;
}

ostream& mad_ast::operator << (ostream& os, const Location &loc)
{
	os << symbol_get_string(loc.file_name) << ":" << *loc.loc;
	return os;
}

void mad_ast::print_symbol_list (ostream &os, const string &sep,
	const vector<Symbol> *ss)
{
	if (ss==NULL || ss->size()==0) return;

	vector<Symbol>::const_iterator it = ss->begin();
	os << symbol_get_string(*it);

	for (it++; it!=ss->end(); it++)
		os << sep << symbol_get_string(*it);
}

void mad_ast::print_string_list (ostream &os, const string &sep,
	const vector<string *> *ss)
{
	if (ss==NULL || ss->size()==0) return;

	vector<string *>::const_iterator it = ss->begin();
	os << *(*it); 

	for (it++; it!=ss->end(); it++)
		os << sep << *(*it); 
}

unsigned long mad_ast::get_ulong_val(const string& num)
{
	const char *s = num.c_str();
	unsigned long val;

	if (s[0]=='0' && (s[1]=='b' || s[1]=='B'))
		val = strtoul(s+2, NULL, 2);
	else
		val = strtoul(s, NULL, 0);
	return val;
}

void mad_ast::fatal(const string& msg)
{
	cerr << "Fatal: " << msg << endl;
	exit(1);
}

void mad_ast::error(const string& msg)
{
	cerr << "Error: " << msg << endl;
	n_errors++;
}

void mad_ast::warn(const string& msg)
{
	if (show_warn) cerr << "Warning: " << msg << endl;
	n_warns++;
}

void mad_ast::message(const string &msg)
{
	cerr << msg << endl;
}

void mad_ast::naming_conflict(Symbol name,
		const Location *loc1, const Location *loc2)
{
	stringstream sstr;
	sstr << "Identifier naming conflict \"" << symbol_get_string(name) 
			<< "\",\n    at " << *loc1 << " and " << *loc2;
	error(sstr.str());
}

void mad_ast::undefined_ident(Symbol name, const Location *loc)
{
	stringstream sstr;
	sstr << "Undefined identifier \"" << symbol_get_string(name) 
			<< "\",\n    at " << *loc;
	error(sstr.str());
}

void mad_ast::generic_error(const string& msg, const Location *loc)
{
	stringstream sstr;
	sstr << msg << endl << "    at " << *loc;
	error(sstr.str());
}

void mad_ast::generic_error(const string& msg1, const string& msg2,
	const Location *loc)
{
	stringstream sstr;
	sstr << msg1 << msg2 << endl << "    at " << *loc;
	error(sstr.str());
}

void mad_ast::generic_error(const string& msg1, const string& msg2,
	const string& msg3, const Location *loc)
{
	stringstream sstr;
	sstr << msg1 << msg2 << msg3 << endl << "    at " << *loc;
	error(sstr.str());
}

void mad_ast::generic_error(const string& msg1, const string& msg2,
	const string& msg3, const string& msg4, const Location *loc)
{
	stringstream sstr;
	sstr << msg1 << msg2 << msg3 << msg4 << endl << "    at " << *loc;
	error(sstr.str());
}

void mad_ast::generic_warn(const string& msg, const Location *loc)
{
	stringstream sstr;
	sstr << msg << endl << "    at " << *loc;
	warn(sstr.str());
}

void mad_ast::generic_fatal(const string& msg, const Location *loc)
{
	stringstream sstr;
	sstr << msg << endl << "    at " << *loc;
	fatal(sstr.str());
}

void mad_ast::generic_fatal(const string& msg1, const string& msg2,
	const Location *loc)
{
	stringstream sstr;
	sstr << msg1 << msg2 << endl << "    at " << *loc;
	fatal(sstr.str());
}

void mad_ast::generic_fatal(const string& msg1, const string& msg2,
	const string& msg3, const Location *loc)
{
	stringstream sstr;
	sstr << msg1 << msg2 << msg3 << endl << "    at " << *loc;
	fatal(sstr.str());
}

bool mad_ast::summarize_error_warn()
{
	if (n_errors>1)
		cerr << n_errors << " errors. ";
	else if (n_errors)
		cerr << n_errors << " error. ";

	if (n_warns>1)
		cerr << n_warns << " warnings. ";
	else if (n_warns)
		cerr << n_warns << " warning. ";

	cerr << endl;
	return n_errors>0;
}

#if 0
bool binary_pattern::overlap(const struct binary_pattern &x) const
{
	assert(this->get_length()==x.get_length());
	const char *str1 = this->data.c_str();
	const char *str2 = x.data.c_str();

	for (unsigned int i = 0; i<this->get_length(); i++) {
		if (str1[i]=='0' && str2[i]=='1') return false;
		if (str1[i]=='1' && str2[i]=='0') return false;
	}
	return true;
}

bool binary_pattern::contain(const struct binary_pattern &x) const
{
	assert(this->get_length()==x.get_length());
	const char *str1 = this->data.c_str();
	const char *str2 = x.data.c_str();

	for (unsigned int i = 0; i<this->get_length(); i++) {
		if (str2[i]=='-' && str1[i]!='-') return false;
		if (str1[i]!='-' && str1[i]!=str2[i]) return false;
	}
	return true;
}

vector<BinPattern> binary_pattern::diff(const struct binary_pattern& x) const
{
	assert(this->contain(x));

	vector<BinPattern> ret;
	const char *str1 = this->data.c_str();
	const char *str2 = x.data.c_str();

	for (unsigned int i = 0; i<this->get_length(); i++) {

		if (str1[i]=='-' && str2[i]=='0') {
			string s(str1);
			s[i] = '1';
			ret.push_back(s);
		}
		else if (str1[i]=='-' && str2[i]=='1') {
			string s(str1);
			s[i] = '0';
			ret.push_back(s);
		}
	}
	return ret;
}

string binary_pattern::get_mask() const
{
	string ret = data;
	for (unsigned int i=0; i<data.length(); i++) {
		if (ret[i]=='-') ret[i] = '0';
		else ret[i] = '1';
	}

	return ret;
}

string binary_pattern::get_signature() const
{
	string ret = data;
	for (unsigned int i=0; i<data.length(); i++) {
		if (ret[i]=='-') ret[i] = '0';
	}
	return ret;
}

#define hex_char(_v) ((_v<10)?(_v+'0'):(_v+'A'-10))

string binary_pattern::get_hex_mask() const
{
	string ret;
	for (int i=data.length()-1; i>=0; i-=4) {
		unsigned val = 0;
		for (int j=0; j<4 && i-j>=0; j++) {
			if (data[i-j]!='-') val |= 1<<j;
		}
		ret.insert(ret.begin(), hex_char(val));	
	}
	return ret;
}

string binary_pattern::get_hex_signature() const
{
	string ret;
	for (int i=data.length()-1; i>=0; i-=4) {
		unsigned val = 0;
		for (int j=0; j<4 && i-j>=0; j++) {
			if (data[i-j]=='1') val |= 1<<j;
		}
		ret.insert(ret.begin(), hex_char(val));	
	}
	return ret;
}

ostream& binary_pattern::print(ostream& os) const
{
	return os << data;
}

const BinPattern mad_ast::operator + (const BinPattern& x, const BinPattern& y)
{
	BinPattern pat((x.data)+(y.data));
	return pat;
}

bool mad_ast::operator == (const BinPattern& x, const BinPattern& y)
{
	assert(x.get_length()==y.get_length());
	return (x.data)==(y.data);
}

bool mad_ast::operator != (const BinPattern& x, const BinPattern& y)
{
	assert(x.get_length()==y.get_length());
	return (x.data)!=(y.data);
}

ostream& mad_ast::operator << (ostream& os, const BinPattern& x)
{
	return os << (x.data);
}
#endif

void mad_ast::lower_case(char *str)
{
	while(*str) {
		*str = tolower(*str);
		str++;
	}
}

string mad_ast::wrapped_symbol(const string& wrap, Symbol sym)
{
	return wrap + "(" + symbol_get_string(sym) + ")";
}

string mad_ast::wrapped_string(const string& wrap, const string& val)
{
	return wrap + "(" + val + ")";
}

void num_value::emit_C(ostream& os) const
{
	if (flt) {
		if (neg) os << "-";
		os << *value;
	}
	else if (get_width_inbase(2)>64) {
		os << "\"";
		if (neg) os << "-";
		else os << *value;
		os << "\"";
	} else {
		string val = get_val_inbase(16);
		if (neg)
			os << "-0x" << val.c_str()+1;
		else
			os << "0x" << val.c_str();
	}
}

unsigned num_value::get_width_inbase(int base) const
{
	if (flt) return this->is_double()?64:32;

	const char *s = value->c_str();
	mpz_t m;
	unsigned len;
	
	if (s[0]=='0') {
		if (s[1]=='b' || s[1]=='B') {
			mpz_init_set_str(m, s+2, 2);
		}
		else if (s[1]=='x' || s[1]=='X') {
			mpz_init_set_str(m, s+2, 16);
		}
		else { // octal
			mpz_init_set_str(m, s+1, 8);
		}
	}
	else  //decimal
		mpz_init_set_str(m, s, 10);
 
	if (neg) mpz_neg(m, m);

	len = mpz_sizeinbase(m, base); // returns at least 1
	mpz_clear(m);
	return len;
}

string num_value::get_val_inbase(int base) const
{
	mpz_t m;
	const char *s = value->c_str();

	// should not be called here
	assert (!flt); /* {
		double x = strtod(s);
		if (neg) x = -x;

		char b[1024];
		sprintf(b, "%e", x);
		return b;
	}*/
	
	if (s[0]=='0') {
		if (s[1]=='b' || s[1]=='B') {
			mpz_init_set_str(m, s+2, 2);
		}
		else if (s[1]=='x' || s[1]=='X') {
			mpz_init_set_str(m, s+2, 16);
		}
		else { // octal
			mpz_init_set_str(m, s+1, 8);
		}
	}
	else  //decimal
		mpz_init_set_str(m, s, 10);
 
	if (neg) mpz_neg(m, m);

	unsigned len;
	len = mpz_sizeinbase(m, base); // returns at least 1
	char *b = new char[len+2];
	mpz_get_str(b, base, m);
	mpz_clear(m);

	string ret(b);
	delete [] b;

	return ret;
}

bool num_value::is_double() const
{
	if (!flt || *(value->rbegin())=='f' || *(value->rbegin())=='F')
		return false;
	return true;
}

