
#ifndef __MAD_AST_MISC_HPP__
#define __MAD_AST_MISC_HPP__

#include <iostream>
#include <string>
#include <cassert>
#include "symbol.hpp"
#include "config.h"

using mad::Symbol;

namespace mad_ast {

#define SYM_DEC(_a) extern Symbol Sym_ ## _a;
#include "symbol_list.hpp"

	/** Structure to keep track of location in bison parsing */
	struct yyltype_t {

		int first_line;		/**< line # of first character */
		int first_column;	/**< column # of first character */
		int last_line;		/**< line # of last character */
		int last_column;	/**< column # of last charater */

		/** Reset the positions to head of file. */
		void reset() {
			first_column = first_line = 1;
			last_column = last_line = 1;
		}

		/** Advance column # of the last character. */
		void advance_column(int num) {
			last_column += num;
		}

		/** Advance line # of the last character. */
		void advance_line(int num) {
			last_column = 1;
			last_line += num;
		}

		/** Step to next node */
		void step() {
			first_column = last_column;
			first_line = last_line;
		}

		/** Union the locations of a and b as the current location. */
		void propogate(struct yyltype_t& a, struct yyltype_t& b) {
			first_column = a.first_column;
			first_line = a.first_line;
			last_column = b.last_column;
			last_line = b.last_line;
		}
	};

	/** Output the location */
	std::ostream& operator << (std::ostream& os, const struct yyltype_t& loc);

	/** Structure to keep track of ast_node locations. */
	typedef struct src_location {

		/** Constructor.
		 *  @param name Source file name.
		 *  @param loc  Location in the file.
		 */
		src_location(Symbol name, const struct yyltype_t& loc) :
			file_name(name), loc(new yyltype_t(loc)) {}

		/** Destructor */
		~src_location() {delete loc;}

		/** Copy constructor. */
		src_location(const src_location& l) {
			file_name = l.file_name;
			loc = new yyltype_t(*(l.loc));
		}

		Symbol file_name;
		const struct yyltype_t *loc;

	} Location;

	/** Print out the locatin structure */
	std::ostream& operator << (std::ostream&, const Location &);

	/** Class to store a numerical value, including float. 
	 *  Value is stored in string, so no loss of resolution due to finite width.
	 */
	typedef class num_value {

		public:
			/** Constructor.
			 *  @param val The pointer to the value.
			 *  @param neg The value is preceded by a negative sign.
			 *  @param flt The value is float.
			 */
			num_value(std::string *val, bool neg, bool flt) :
				value(val), neg(neg), flt(flt) {}

			/** Destructor. */
			~num_value() {delete value;}

			/** Print out the node. */
			std::ostream& print(std::ostream& os) const;

			/** Get the value. */
			const std::string *get_value() const {return value;}

			/** See if having a negative sign. */
			bool is_negative() const {return neg;}

			/** See if this is a floating point number, including double. */
			bool is_float() const {return flt;}

			/** See if this is a double precision floating point number. */
			bool is_double() const;

			/** Get the width of the constant under base.
			 *  @param base The base used to evaluate the width.
			 */
			unsigned get_width_inbase(int base) const;

			/** Convert the constnat to string according to the given base.
			 @  @param base The base for conversion.
			 */
			std::string get_val_inbase(int base) const;

			/** Emit constant in C syntax. */
			void emit_C(std::ostream& os) const;

		private:
			std::string *value;
			bool neg, flt;

	} NumValue;

	bool operator == (const NumValue& o1, const NumValue& o2);

	/** Print an array of symbols.
	 *  @param os  The output stream.
	 *  @param sep The separator to print between symbols.
	 *  @param ss  Array of symbols to print.
	 */
	void print_symbol_list (std::ostream &os, const std::string& sep,
			const std::vector<Symbol> *ss);

	/** Print an array of strings.
	 *  @param os  The output stream.
	 *  @param sep The separator to print between strings.
	 *  @param ss  Array of strings to print.
	 */
	void print_string_list (std::ostream &os, const std::string& sep,
			const std::vector<std::string *> *ss);

	/** Turn an array of pointers to an array of const pointers.
	 *  @param  s The array of pointers.
	 *  @return The array of const pointers.
	 */
	template<class T>
	inline std::vector<const T *> const_vector_transform
		(const std::vector<T *>& s) {

		std::vector<const T *> ret;
		typename std::vector<T *>::const_iterator it;
		for (it = s.begin(); it != s.end(); it++)
			ret.push_back(*it);
		return ret;	
	}


	/** Print an array of objects.
	 *  @param os  The output stream.
	 *  @param sep The separator to print between objects.
	 *  @param s   The object array to print.
	 */
	template<class T>
	inline void print_object_list(std::ostream & os, const std::string& sep,
		const std::vector<T *> *s)
	{
		if (s==NULL || s->size()==0) return;

		typename std::vector<T *>::const_iterator it = s->begin();
		(*it)->print(os);

		for (it++; it!=s->end(); it++) {
			os << sep;
			(*it)->print(os);
		}
	}

	/** Delete an array of objects.
	 *  @param s  The object pointer arry to delete.
	 */
	template<class T>
	inline void delete_object_list(std::vector<T *> *s) {
		if (s==NULL) return;

		typename std::vector<T *>::iterator it;
		for (it=s->begin(); it!=s->end(); it++)
			delete *it;

		delete s;
	}

	/* Initialize a few global variables.
	 * @param quiet True if not reporting warnings.
	 */
	void mad_ast_init(bool quiet);

	/** Get the value of the unsigned number represented by the string.
	 *  @param num The numeric string.
	 */
	unsigned long get_ulong_val(const std::string& num);

	/** Report a fatal error and exit.
	 *  @param msg The error message to report.
	 */
	void fatal(const std::string& msg);

	/** Report an error message but do not quit.
	 *  @param msg The error message to report.
	 */
	void error(const std::string& msg);

	/** Report an warning message.
	 *  @param msg The warning message to report.
	 */
	void warn(const std::string& msg);

	/** Report an normal message.
	 *  @param msg The message to report.
	 */
	void message(const std::string& msg);

	/** Report naming conflict error.
	 *	@param name The conflict name.
	 *  @param loc1 Location 1 where the name is defined.
	 *  @param loc2 Location 2 where the name is defined.
	 */
	void naming_conflict(Symbol name,
		const Location *loc1, const Location *loc2);

	/** Report undefined identifier error.
	 *  @param name The conflict name.
	 *  @param loc  The location of the reference.
	 */
	void undefined_ident(Symbol name, const Location *loc);

	/** Report a generic error.
	 *  @param msg The message to report.
	 *  @param loc The location of the error.
	 */
	void generic_error(const std::string& msg, const Location *loc);

	/** Report a generic error.
	 *  @param msg1 The message to report.
	 *  @param msg2 The message to report, concatenated with msg1.
	 *  @param loc The location of the error.
	 */
	void generic_error(const std::string& msg1, const std::string& msg2,
		const Location *loc);

	/** Report a generic error.
	 *  @param msg1 The message to report.
	 *  @param msg2 The message to report, concatenated with msg1.
	 *  @param msg3 The message to report, concatenated with msg1 & msg2.
	 *  @param loc The location of the error.
	 */
	void generic_error(const std::string& msg1, const std::string& msg2,
		const std::string& msg3, const Location *loc);

	/** Report a generic error.
	 *  @param msg1 The message to report.
	 *  @param msg2 The message to report, concatenated with msg1.
	 *  @param msg3 The message to report, concatenated with msg1 & msg2.
	 *  @param msg4 The message to report, concatenated with msg1 - msg3.
	 *  @param loc The location of the error.
	 */
	void generic_error(const std::string& msg1, const std::string& msg2,
		const std::string& msg3, const std::string& msg4, const Location *loc);

	/** Report a generic warning.
	 *  @param msg The message to report.
	 *  @param loc The location of the error.
	 */
	void generic_warn(const std::string& msg, const Location *loc);

	/** Report a generic fatal error and exit.
	 *  @param msg1 The error message to report.
	 *  @param loc The location of the error.
	 */
	void generic_fatal(const std::string& msg1, const Location *loc);

	/** Report a generic fatal error and exit.
	 *  @param msg1 The error message to report.
	 *  @param msg2 The error message to report, concatenated with msg1.
	 *  @param loc The location of the error.
	 */
	void generic_fatal(const std::string& msg1, const std::string& msg2,
		const Location *loc);

	/** Report a generic fatal error and exit.
	 *  @param msg1 The error message to report.
	 *  @param msg2 The error message to report, concatenated with msg1.
	 *  @param msg3 The error message to report, concatenated with msg1/msg2.
	 *  @param loc The location of the error.
	 */
	void generic_fatal(const std::string& msg1, const std::string& msg2,
		const std::string& msg3, const Location *loc);

	/** Summarize error and warning messages.
	 *  @return True if there are errors. */
	bool summarize_error_warn();

	/** Converting C-string to lower case. */
	void lower_case(char *str);

	/** String of macro wrapped symbol, etc wrap(name).
	 *  @param wrap The wrapper.
	 *  @param sym  The name to be wrapped.
	 */
	std::string wrapped_symbol(const std::string& wrap, Symbol sym);

	/** String of macro wrapped string, etc wrap(name).
	 *  @param wrap The wrapper.
	 *  @param val  The string to be wrapped.
	 */
	std::string wrapped_string(const std::string& wrap, const std::string& val);

}

#endif
