
#include <set>
#include "oper_graph.hpp"
#include "exp_ast.hpp"
#include "operation_ast.hpp"
#include "data_type_ast.hpp"

using std::list;
using std::set;
using std::map;
using std::vector;
using std::ostream;
using std::cerr;
using std::endl;
using namespace mad;
using namespace mad_ast;


oper_graph_and_node::~oper_graph_and_node()
{
	vector<oper_graph_or_node *>::iterator oit;
	for (oit=children.begin(); oit!=children.end(); oit++) {
		delete *oit;
	}

	vector<oper_tree *>::iterator tit;
	for (tit=expansions.begin(); tit!=expansions.end(); tit++) {
		delete *tit;
	}
}

Symbol oper_graph_and_node::get_name() const {
	return oper->get_name();
}

vector<const oper_tree *> oper_graph_and_node::get_expansions() const
{
	return const_vector_transform<oper_tree>(expansions);
}

void oper_graph_and_node::connect_decode
	(map<Symbol, oper_graph_and_node *> *and_nodes)
{
	map<Symbol, DecodeEntity *>::const_iterator it;
	for (it=oper->dec_ents.begin(); it!=oper->dec_ents.end(); it++) {

		const DecodeEntity *dent = (*it).second;
		if (!dent->is_decoded()) continue;

		const variable_clause *var = dent->get_variable();
		const DataType *dtype = var->get_data_type();
		assert(dtype->get_type()==DT_OPER);
		const DataTypeOper *dt = static_cast<const DataTypeOper *>(dtype);

		/* decoded oper-type variable, creat an or_node */
		oper_graph_or_node *onode = new oper_graph_or_node(this,
			var->get_name(), dt, and_nodes);

		children.push_back(onode);
	}
}

void oper_graph_and_node::connect_coding
	(map<Symbol, oper_graph_and_node *> *and_nodes)
{
	map<Symbol, DecodeEntity *>::const_iterator it;
	for (it=oper->dec_ents.begin(); it!=oper->dec_ents.end(); it++) {

		const DecodeEntity *dent = (*it).second;
		if (dent->is_single()) continue;

		const variable_clause *var = dent->get_variable();
		const DataType *dtype = var->get_data_type();
		assert(dtype->get_type()==DT_OPER);
		const DataTypeOper *dt = static_cast<const DataTypeOper *>(dtype);

		/* decoded oper-type variable, creat an or_node */
		oper_graph_or_node *onode = new oper_graph_or_node(this,
			var->get_name(), dt, and_nodes);

		children.push_back(onode);
	}
}

void oper_graph_and_node::expand()
{
	assert(!cyclic_flag);

	if (expanded) return;

	expanded = true;

	if (is_leaf()) {
		oper_tree *atree = new oper_tree(this);
		expansions.push_back(atree);
		return;
	}

	cyclic_flag = true;
	vector<oper_graph_or_node *>::iterator cit;
	for (cit=children.begin(); cit!=children.end(); cit++) {
		(*cit)->expand();
	}
	cyclic_flag = false;

	int n_exp = 1;
	vector<int> n_sizes;
	for (cit=children.begin(); cit!=children.end(); cit++) {
		n_exp *= (*cit)->expansions.size();
		n_sizes.push_back((*cit)->expansions.size());
	}

	/* there are n_exp expansions, we compute them. */
	int n_elem = children.size();							
	for (int i=0; i<n_exp; i++) {							 
 
		int cnt = i;										  
		oper_tree *atree = new oper_tree(this);
		expansions.push_back(atree);

		for (int j=0; j<n_elem; j++) {
			atree->insert_child(children[j]->expansions[cnt%n_sizes[j]],
				children[j]->vname);
			cnt = cnt/n_sizes[j];
		}
	}
}

bool oper_graph_and_node::type_check()
{
	std::vector<oper_tree *>::iterator tit;
	for (tit=expansions.begin(); tit!=expansions.end(); tit++) {
		if (!(*tit)->type_check()) return false;
	}
	return true;
}

oper_graph_and_node *oper_graph_and_node::check_cycle()
{
	if (cyclic_flag) return this;

	cyclic_flag = true;
	cyclic_reached = true;
	vector<oper_graph_or_node *>::iterator ait;
	for (ait=children.begin(); ait!=children.end(); ait++) {
		oper_graph_and_node *anode = (*ait)->check_cycle();
		if (anode!=NULL) return anode;
	}
	cyclic_flag = false;
	return NULL;
}

void oper_graph_and_node::topological_sort(list<operation_section *> &sorted)
{
    if (cyclic_reached) return;
    cyclic_reached = true;

    vector<oper_graph_or_node *>::iterator eit;
    for (eit=children.begin(); eit!=children.end(); eit++) {
    	vector<oper_graph_and_node *>::iterator ait;
		for (ait=(*eit)->children.begin(); ait!=(*eit)->children.end(); ait++)
			(*ait)->topological_sort(sorted);
	}

	sorted.push_front(this->oper);
}

void oper_graph_and_node::print(ostream &os)
{
	std::vector<oper_tree *>::iterator tit;
	for (tit=expansions.begin(); tit!=expansions.end(); tit++) {
		(*tit)->print(os);
		os << endl;
	}
}

oper_graph_or_node::oper_graph_or_node(oper_graph_and_node *parent,
	Symbol name, const data_type_oper *type,
	map<Symbol, oper_graph_and_node *> *and_nodes) :
	oper(parent), vname(name)
{
	oper = parent;
	const vector<Symbol> *opers = type->get_opers();

	vector<Symbol>::const_iterator sit;
	for (sit=opers->begin(); sit!=opers->end(); sit++) {
		assert(and_nodes->find(*sit)!=and_nodes->end());
		children.push_back((*and_nodes)[*sit]);
		(*and_nodes)[*sit]->parents.push_back(this);
	}

	Symbol def_oper = type->get_default_oper();
	if (def_oper!=SymNULL) {
		assert(and_nodes->find(def_oper)!=and_nodes->end());
		children.push_back((*and_nodes)[def_oper]);
		(*and_nodes)[def_oper]->parents.push_back(this);
	}
}

oper_graph_or_node::~oper_graph_or_node()
{
}

void oper_graph_or_node::expand()
{
	vector<oper_graph_and_node *>::iterator ait;
	for (ait=children.begin(); ait!=children.end(); ait++) {
		(*ait)->expand();
		expansions.insert(expansions.begin(),
			(*ait)->expansions.begin(), (*ait)->expansions.end());
	}
}

oper_graph_and_node *oper_graph_or_node::check_cycle()
{
	vector<oper_graph_and_node *>::iterator ait;
	for (ait=children.begin(); ait!=children.end(); ait++) {
		oper_graph_and_node *anode = (*ait)->check_cycle();
		if (anode!=NULL) return anode;
	}
	return NULL;
}

void oper_tree::print(ostream& os)
{
	os << symbol_get_string(node->get_name());
	vector<oper_tree *>::iterator tit;
	for (tit=children.begin(); tit!=children.end(); tit++) {
		os << " ";
		(*tit)->print(os);
	}
}

bool oper_tree::check_causal(Symbol init_edge)
{
	if (!node->oper->check_causal(init_edge))
		return false;

	unsigned index = 0;
	vector<oper_tree *>::iterator tit;
	for (tit=children.begin(); tit!=children.end(); tit++, index++) {

		const DecodeEntity *dent = 
			node->oper->get_decode_entity(vnames[index]);

		assert(dent);

		const vector<DecodeInfo *>& dinfos = dent->get_infos();
		vector<DecodeInfo *>::const_iterator iit;
		for (iit=dinfos.begin(); iit!=dinfos.end(); iit++) {

			Symbol edge = (*iit)->edge;
			if (edge==SymNULL) edge = init_edge;

			if (!(*tit)->check_causal(edge)) {
				generic_error("Non-causal decoding operation ", 
					symbol_get_string((*tit)->node->oper->get_name()),
					(*iit)->exp->get_loc());
				return false;
			}
		}
	}
	return true;
}

bool oper_tree::type_check()
{
	// need to check that operations timing is correct
	if (!check_causal(SymNULL)) return false;

	// next we check if an operation appears multiple times in the tree
	set<Symbol> node_names;

	list<oper_tree *> work_list;
	work_list.push_back(this);
	while (!work_list.empty()) {

		// get the first from work_list
		oper_tree *anode = work_list.front();
		work_list.pop_front();

		// this should be new to the list
		if (node_names.find(anode->get_name())!=node_names.end()) {
			generic_error("Multiple apprearances of OPERATION ",
				symbol_get_string(anode->get_name()),
				" found in expanded operation tree with root ",
				symbol_get_string(this->get_name()),
				anode->node->oper->get_loc());

			return false;
		}

		node_names.insert(anode->node->get_name());

		vector<oper_tree *>::iterator it;
		for (it=anode->children.begin(); it!=anode->children.end(); it++)
			work_list.push_back(*it);
	}
	
	return true;
}

void oper_tree::insert_child(oper_tree *tree, Symbol vname)
{
	children.push_back(tree);
	vnames.push_back(vname);
}

vector<const operation_section *> oper_tree::get_child_opers() const
{
	vector<const operation_section *> opers;
	opers.push_back(node->get_oper());

	vector<oper_tree *>::const_iterator oper_it;
	for (oper_it=children.begin(); oper_it!=children.end(); oper_it++) {
		vector<const operation_section *> c_opers =
				(*oper_it)->get_child_opers();
		opers.insert(opers.end(), c_opers.begin(), c_opers.end());
	}
	return opers;
}

oper_graph::oper_graph(oper_context *opers, bool decode) : decode(decode)
{
	/* first, create an and-node for each operation section. */
	OperContext::const_iterator oit;
	for (oit=opers->begin(); oit!=opers->end(); oit++) {
		oper_graph_and_node *anode;
		anode = new oper_graph_and_node((*oit).second);
		and_nodes[(*oit).first] = anode;
	}

	/* then, connect the and-node to each other. */
	map<Symbol, oper_graph_and_node *>::iterator nit;
	for (nit=and_nodes.begin(); nit!=and_nodes.end(); nit++) {
		if (decode)
			(*nit).second->connect_decode(&and_nodes);
		else
			(*nit).second->connect_coding(&and_nodes);
	}

	/* Then collect root nodes. */
	for (nit=and_nodes.begin(); nit!=and_nodes.end(); nit++) {
		if ((*nit).second->is_root()) roots.push_back((*nit).second);
	}
}

oper_graph::~oper_graph()
{
	map<Symbol, oper_graph_and_node *>::iterator nit;
	for (nit=and_nodes.begin(); nit!=and_nodes.end(); nit++) {
		delete (*nit).second;
	}
}

void oper_graph::expand()
{
	/* start to expand from the roots */
	vector<oper_graph_and_node *>::iterator ait;
	for (ait=roots.begin(); ait!=roots.end(); ait++) {
		(*ait)->expand();
	//	(*ait)->print(std::cerr);
	}
}

bool oper_graph::type_check()
{
	bool ret = true;

	if (and_nodes.size() >0 && roots.size()==0) {
		error("Cycle detected in operataion graph, no root found");
		return false;
	}

	/* need to check that this is a DAG, no cycle should exiest.  */
	map<Symbol, oper_graph_and_node *>::iterator ait;
	for (ait=and_nodes.begin(); ait!=and_nodes.end(); ait++) {
		(*ait).second->cyclic_flag = false;
		(*ait).second->cyclic_reached = false;
	}

	/* Two cycle situations can occur:
	 * 1. A node is marked twice in the marking process.
	 * 2. A node is never marked -- the cycle can be reached from no root.
	 */
	vector<oper_graph_and_node *>::iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {
		oper_graph_and_node *anode = (*rit)->check_cycle();
		if (anode!=NULL) {
			generic_error("Cycle detected in operataion graph, involving ",
				symbol_get_string(anode->oper->get_name()),
				anode->oper->get_loc());
			return false;
		}
	}

	for (ait=and_nodes.begin(); ait!=and_nodes.end(); ait++) {
		if ((*ait).second->cyclic_reached==false) {
			generic_error("Cycle detected in operataion graph, operation ",
				symbol_get_string((*ait).second->oper->get_name()),
				" not reachable from any root",
				(*ait).second->oper->get_loc());
			return false;
		}
		/* clear the flags for future use. */
		(*ait).second->cyclic_flag = false;
		(*ait).second->cyclic_reached = false;
	}

	if (decode) {

		/* expand the trees */
		expand();

		/* start to type-check from the roots */
		for (rit=roots.begin(); rit!=roots.end(); rit++) {
			ret = (*rit)->type_check() && ret;
		}
	}

	return ret;
}

void oper_graph::topological_sort(list<operation_section *>& sorted)
{
	map<Symbol, oper_graph_and_node *>::iterator ait;
	for (ait=and_nodes.begin(); ait!=and_nodes.end(); ait++) {
		(*ait).second->cyclic_reached = false;
	}

	vector<oper_graph_and_node *>::iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {
		(*rit)->topological_sort(sorted);
	}
}

vector<const operation_section *> oper_graph::get_root_opers() const
{
	vector<const operation_section *> result;
	vector<oper_graph_and_node *>::const_iterator it;
	for (it=roots.begin(); it!=roots.end(); it++)
		result.push_back((*it)->oper);
	return result;
}

vector<const oper_graph_and_node *> oper_graph::get_root_nodes() const
{
	return const_vector_transform<oper_graph_and_node>(roots);
}

vector<operation_section *> oper_graph::get_reachables
	(const operation_section *root) const
{

	vector<operation_section *> result;
	set<Symbol> node_names;

	assert(and_nodes.find(root->get_name())!=and_nodes.end());
	
	list<oper_graph_and_node *> work_list;
	work_list.push_back((*and_nodes.find(root->get_name())).second);
	while (!work_list.empty()) {

		// get the first from work_list
		oper_graph_and_node *anode = work_list.front();
		work_list.pop_front();

		// see if this has already been handled
		if (node_names.find(anode->get_name())!=node_names.end()) continue;

		node_names.insert(anode->get_name());
		vector<oper_graph_or_node *>::const_iterator it;
		for (it=anode->children.begin(); it!=anode->children.end(); it++)
			work_list.insert(work_list.end(),
				(*it)->children.begin(), (*it)->children.end());
	}

	set<Symbol>::const_iterator sit;
	for (sit=node_names.begin(); sit!=node_names.end(); sit++) {
		assert(and_nodes.find(*sit)!=and_nodes.end());
		result.push_back((*and_nodes.find(*sit)).second->oper);
	}
	return result;
}
