
#ifndef __MAD_OPER_GRAPH_HPP__
#define __MAD_OPER_GRAPH_HPP__


#include <vector>
#include <map>
#include <list>
#include "symbol.hpp"

using mad::Symbol;

namespace mad_ast {

	class operation_section;
	class oper_context;
	class data_type_oper;

	class oper_tree;
	class oper_graph_or_node;

	/** An and node in the graph. It may have multiple or-node parents. */
	typedef class oper_graph_and_node {

		public:
			/** Constructor.
			 *  @param oper The operation section for this and_node.
			 */
			oper_graph_and_node(operation_section *oper) :
				oper(oper), cyclic_flag(false), cyclic_reached(false),
				expanded(false) {}

			/** Destructor. */
			~oper_graph_and_node();

			/** Test if an and node is a leaf. */
			bool is_leaf() const { return children.size()==0; }

			/** Test if an and node is a root. */
			bool is_root() const { return parents.size()==0; }

			/** To expand. */
			void expand();

			/** Type check the operation tree. */
			bool type_check();

			/** Print the expansions */
			void print(std::ostream& os);

			/** Get the name of the node. */
			Symbol get_name() const;

			/** Get the expansions. */
			std::vector<const oper_tree *> get_expansions() const;

			/** Get the operation. */
			const operation_section *get_oper() const {return oper;}

		private:

			/** Connect to the its children accoding to decoding relationships.
			 */
			void connect_decode
				(std::map<Symbol, oper_graph_and_node *> *and_nodes);

			/** Connect to the its children accoding to coding reference
			 *  relationships.
			 */
			void connect_coding
				(std::map<Symbol, oper_graph_and_node *> *and_nodes);

			/** Check if cycle exists in the graph.
			 *  @return NULL if no cycle, a node in the cycle if exists.
			 */
			oper_graph_and_node *check_cycle();

			/** Topological sort the graph. */
			void topological_sort(std::list<operation_section *>& result);

			/* the operation section corresponding to this node. */
			operation_section *oper;

			/* flags used to detect cyclic reference in and-or graph */
			bool cyclic_flag;
			bool cyclic_reached;

			/* expand called already */
			bool expanded;

			/* the children nodes. */
			std::vector<oper_graph_or_node *> children;

			/* the parent nodes. */
			std::vector<oper_graph_or_node *> parents;

			/* the expanded result, contain children only */
			std::vector<oper_tree *> expansions;

		friend class oper_graph_or_node;
		friend class oper_graph;
		friend class oper_tree;

	} OGAndNode;


	/** An or node in the graph. It has only one and-node parent. */
	typedef class oper_graph_or_node {

		public:
			/** Constructor for an or node.
			 *  @param parent The parent and-node.
			 *  @param vname  Name of the oper-type variable.
			 *  @param type   Type of the variable.
			 *  @param and_nodes All operations.
			 */
			oper_graph_or_node(oper_graph_and_node *parent,
				Symbol vname, const data_type_oper *type,
				std::map<Symbol, oper_graph_and_node *> *and_nodes);

			/* Destructor. */
			~oper_graph_or_node();

			/** To expand. */
			void expand();

		private:

			/** Check if cycle exists in the graph.
			 *  @return NULL if no cycle, a node in the cycle if exists.
			 */
			oper_graph_and_node *check_cycle();

			/* the operation section corresponding to this node. */
			oper_graph_and_node *oper;

			/* the name of the variable. */
			Symbol vname;

			/* the children nodes. */
			std::vector<oper_graph_and_node *> children;

			/* the expanded result, contain children only */
			std::vector<oper_tree *> expansions;

		friend class oper_graph_and_node;
		friend class oper_graph;

	} OGOrNode;

	/** An oper tree is an expansion of an oper graph. */
	typedef class oper_tree {

		public:
			/** Constructor. */
			oper_tree(oper_graph_and_node *node) : node(node) {}

			/** Destructor. Children will be freed by others. */
			~oper_tree() {}

			/** Type check the operation tree.
			 *  Check two things -- causailty, multiple copy of same node.
			 */
			bool type_check();

			/** Recursively print the tree, depth first. */
			void print(std::ostream& os);

			/** Check causal relationship.
			 *  @param init_edge The initialization edge in the state DAG.
			 *  @return True if ok.
			 */
			bool check_causal(Symbol init_edge);

			/** Get the name of the tree node. */
			Symbol get_name() const {return node->get_name();}

			/** Get all children operations. */
			std::vector<const operation_section *> get_child_opers() const;

		private:

			/** Add one child.
			 *  @param tree  The child tree node.
			 *  @param vname The variable name corresponding to the child.
			 */
			void insert_child(oper_tree *tree, Symbol vname);

			oper_graph_and_node *node;
			std::vector<oper_tree *> children;
			std::vector<Symbol> vnames;	// the variable names for children

		friend class oper_graph_and_node;

	} OGTree;


	/** An oper graph is an acyclic and-or graph of operation sections. 
	 *  Each operation corresponds to an and-node in the graph. Each
	 *  oper-typed variable corresponds to an or-node in the graph.
	 */
	typedef class oper_graph {


		public:
			/** Constructor.
			 *  @param opers  All operation sections.
			 *  @param decode True if decoding tree, otherwise coding tree.
			 */
			oper_graph(oper_context *opers, bool decode);

			/** Destructor. */
			~oper_graph();

			/** Type check the operation tree. */
			bool type_check();

			/** Get the roots in the graph. 
			 *  Callable after type_check.
			 *  @param result The vector reference holding the roots.
			 */
			std::vector<const operation_section *> get_root_opers() const;

			/** Get the root and_nodes in the graph. 
			 *  Callable after type_check.
			 *  @return The vector reference holding the roots.
			 */
			std::vector<const oper_graph_and_node *> get_root_nodes() const;


		private:

			/** Get the nodes that is reachable from a node
			 *  @param root   The from node.
			 *  @return The vector holding the result.
			 */
			std::vector<operation_section *> get_reachables
					(const operation_section *root) const;

			/** Topological sort the graph. */
			void topological_sort(std::list<operation_section *>& result);

			bool decode;

			/** Expand the graph. */
			void expand();

			std::map<Symbol, oper_graph_and_node *> and_nodes;

			std::vector<oper_graph_and_node *> roots;

			friend class program;

	} OperGraph;


};


#endif	
