
#include <string>
#include <cassert>
#include <sstream>
#include <fstream>
#include <set>
#include "operation_ast.hpp"
#include "machine_ast.hpp"
#include "manager_ast.hpp"
#include "data_type_ast.hpp"
#include "exp_ast.hpp"
#include "osm_exp_ast.hpp"
#include "define_ast.hpp"
#include "emit_c.hpp"

using namespace mad_ast;
using std::ostream;
using std::vector;
using std::endl;
using std::cerr;
using std::pair;
using std::map;
using std::set;
using std::string;
using std::stringstream;
using std::ofstream;
using std::list;
using namespace mad;

extern map<OperID, Symbol> oper_name_table;

bool decode_pattern::contain(const decode_pattern& pat) const
{
	/* if ids are the same then no conflict */
	if (ids==pat.get_ids()) return false;

	/** containment */
	return bin.contain(pat.bin);
}

bool decode_pattern::conflict(const decode_pattern& pat) const
{
	/* if ids are the same then no conflict */
	if (ids==pat.get_ids()) return false;

	/** Equal is conflict */
	if (bin==pat.bin) return true;

	/** containment is ok */
	if (bin.contain(pat.bin) || pat.bin.contain(bin)) return false;

	/** overlap? bad */
	return bin.overlap(pat.bin);
}

void decode_pattern::append_ids(const vector<OperID>& nids)
{
	ids.insert(ids.end(), nids.begin(), nids.end());
}

void decode_pattern::prepend_id(const OperID& nid)
{
	ids.insert(ids.begin(), nid);
}

void decode_pattern::pad_ids_with_zeros(unsigned length)
{
	assert(ids.size()<=length);
	while (ids.size()<length)
		ids.push_back(0);
}

void decode_pattern::print(ostream& os) const
{
	bin.print(os);
}

void decode_pattern::print_names(ostream& os) const
{
	if (ids.empty()) return;
	os << symbol_get_string(oper_name_table[ids[0]]);
	for (unsigned i=1; i<ids.size(); i++)
		if (oper_name_table[ids[i]]!=SymNULL)
			os << " " << symbol_get_string(oper_name_table[ids[i]]);
}

const DecodePattern mad_ast::operator + (const DecodePattern& x,
										const DecodePattern& y)
{
	DecodePattern ret(x.get_pattern()+y.get_pattern());
	ret.append_ids(x.get_ids());
	ret.append_ids(y.get_ids());
	return ret;
}

bool mad_ast::DecPatPtrCmp(const DecodePattern *x, const DecodePattern *y)
{
	const vector<OperID>& xids = x->get_ids();
	const vector<OperID>& yids = y->get_ids();

	unsigned minsize = xids.size()<yids.size()?xids.size():yids.size();
	for (unsigned ii=0; ii<minsize; ii++) {
		if (xids[ii]>yids[ii]) return false;
		if (xids[ii]<yids[ii]) return true;
	}
	/* if x is shorter, x is smaller */
	return xids.size()<=yids.size();
}

bool mad_ast::check_decode_patterns(vector<DecodePattern *> *patterns,
	const Location *loc) 
{
	/* we check if there is any conflict */
	vector<DecodePattern *>::const_iterator it, jt;
	for (it = patterns->begin(); it!=patterns->end(); it++) {
		for (jt = patterns->begin(); jt!=it; jt++) {

			if ((*it)->conflict(**jt)) {

				stringstream sstr;
				sstr << "Patterns overlap, between ";
				sstr << endl;
				(*it)->print(sstr);
				sstr << " : ";
				(*it)->print_names(sstr);
				sstr << endl;
				(*jt)->print(sstr);
				sstr << " : ";
				(*jt)->print_names(sstr);

				generic_error(sstr.str(), loc);
				return false;
			}
			else if ((*it)->contain(**jt) || (*jt)->contain(**it)) {

				stringstream sstr;
				sstr << "Pattern contains other, between ";
				sstr << endl;
				(*it)->print(sstr);
				sstr << " : ";
				(*it)->print_names(sstr);
				sstr << endl;
				(*jt)->print(sstr);
				sstr << " : ";
				(*jt)->print_names(sstr);

				generic_warn(sstr.str(), loc);
			}
		}
	}

	for (it = patterns->begin(); it!=patterns->end();) {

		(*it)->set_hflag(false);
		for (jt = it, jt++; jt!=patterns->end(); jt++) {
	 		if ((*it)->get_ids()==(*jt)->get_ids())
				(*jt)->set_hflag(true);
			else break;
		}
		it = jt;
	}

	return true;
}

/* these are not inlined since exp type is not defined in header file. */
bool decode_info::has_arg() const {
	return exp->get_arg()!=NULL;
}

bool decode_info::is_activate() const {
	return exp->get_type()==OP_ACT;
}

decode_entity::decode_entity(const operation_section *oper,
	const variable_clause *v, const OperContext *oper_context, bool& ret) :
	oper(oper), var(v), cc(NULL), has_cc(false),
	need_dec(false), is_singleton(false),
	is_heterogeneous(false), machine_id(0), default_name(SymNULL),
	default_oper(NULL)
{

	const DataType *dtype = v->get_data_type();
	assert(dtype->get_type()==DT_OPER);

	const DataTypeOper *dt = static_cast<const DataTypeOper *>(dtype);

	const vector<Symbol> *opers = dt->get_opers();
	vector<Symbol>::const_iterator oper_it;
	for (oper_it=opers->begin(); oper_it!=opers->end(); oper_it++) {

		/* see if the operations exists */
		if (!oper_context->has(*oper_it)) {
			undefined_ident(*oper_it, dt->get_loc());
			ret = false;
		}
		else {

			/* check to see if the variable is a 
			 * singleton, and if the operations are of
			 * more than one machine type. */ 
			const OperationSection *op = oper_context->get_object(*oper_it);

			/* machine id not set yet (first time) */
			if (machine_id==0)
				machine_id = op->get_mach_id();
			else if (machine_id!=op->get_mach_id())
				is_heterogeneous = true;

			/* Singleton test */
			if (opers->size()==1 && !op->has_coding()) {
				is_singleton = true;
				singleton_id = op->get_id();
			}
		}
	}

	/* Check if the default operation exists*/
	default_name = dt->get_default_oper();
	if (default_name!=SymNULL) {

		if	(!oper_context->has(default_name)) {
			undefined_ident(default_name, dt->get_loc());
			ret = false;
		}
		else {

			/* check to see if the variable is a 
			 * singleton, and if the operations are of
			 * more than one machine type. */ 
			default_oper = oper_context->get_object(default_name);
			default_id = default_oper->get_id();

			/* singleton does not need decode, so default ignored. */
			if (is_singleton) 
				generic_warn("Default operation ignored for singleton",
					dt->get_loc());
			else if (machine_id!=default_oper->get_mach_id())
				is_heterogeneous = true;
		}
	}

	/* Assign a name for the decoder entity */
	decoder_name = symbol_find_or_insert
		(symbol_get_string(oper->get_name()) + "_" +
		symbol_get_string(v->get_name()));

}

decode_entity::~decode_entity() 
{
	if (has_cc) delete cc;
	vector<DecodeInfo *>::iterator it;
	for (it=infos.begin(); it!=infos.end(); it++) {
		delete *it;
	}
}

bool decode_entity::type_check()
{
	if (default_oper && !default_oper->is_decode_leaf()) {
		generic_error("Non-leaf operation in decoding tree used as"
			" default operation", symbol_get_string(default_name),
			var->get_loc());
		return false;
	}

	if (infos.size()==0) {
		generic_error("Unused operation variable ",
			symbol_get_string(var->get_name()), var->get_loc());
		return false;
	}

	bool has_arg = infos[0]->has_arg();
	bool is_activate = infos[0]->is_activate();

	vector<DecodeInfo *>::iterator it;
	for (it=infos.begin(); it!=infos.end(); it++) {
		/* if used consistently */
		if ((*it)->has_arg()!=has_arg || (*it)->is_activate()!=is_activate) {
			generic_error("Variable ", symbol_get_string(var->get_name()),
				" used inconsistently", var->get_loc());
			return false;
		}

		/* If decode, then must use the same machine.
		 * On the other hand, activate is fine with
		 * heterogeneous machine type. */
		if ((*it)->exp->get_type()==OP_DEC) {
			if (is_heterogeneous || machine_id!=oper->get_mach_id()) {
				generic_error("Decoding operations with different machines",
					var->get_loc());
				return false;
			}
		}

		/* has a coding clause in coding paragraph, then should not have
		 * argument anymore. */
		if (cc!=NULL && !has_cc  && (*it)->has_arg()) {
			generic_error("Invalid decoding of variable ",
				symbol_get_string(var->get_name()),
				": no argument needed for variable appearing in coding",
				(*it)->exp->get_loc());
			return false;
		}
		const MachineSection *mach = oper->get_machine();

		/* If decoded more than once, we need to check if they
		 * occurred on mutual exclusive paths. */
		vector<DecodeInfo *>::iterator jt;
		for (jt=infos.begin(); jt!=infos.end(); jt++) {

			if (jt==it) continue;

			if ((*it)->edge==SymNULL || (*jt)->edge==SymNULL ||
				mach->edge_reachable((*it)->edge, (*jt)->edge) ||
				mach->edge_reachable((*jt)->edge, (*it)->edge)) {
				generic_error("Variable ", symbol_get_string(var->get_name()),
					" decoded more than once at non-exclusive locations",
					var->get_loc());
				return false;
			}
		}
	}

	/* in the decoding tree, we cannot have one OPER at two nodes,
  	   this should be checked on the expanded operation trees. */

	need_dec = has_arg;
	decoded = !is_activate;

	/* if singleton and with arg, then
	 * errors will be reported by coding width
	 * checking. */

	return true;
}

const vector<DecodePattern *> *decode_entity::get_patterns() const {
	if (cc) return cc->get_patterns();
	return NULL;
}

void decode_entity::emit_pat_file(const char *prefix) const
{
	ofstream ofs;
	string dfname = prefix;
	dfname += DECODER_PAT_FILE(decoder_name);
	ofs.open(dfname.c_str());
	message(dfname+" ...");

	ofs << std::hex;

	list<BinPattern> all_pats, default_pats;
	const vector<DecodePattern *> *patterns = cc->get_patterns();
	vector<DecodePattern *>::const_iterator pit;
	for (pit=patterns->begin(); pit!=patterns->end(); pit++) {

		ofs << "IDEF(";

		const vector<OperID>& ids = (*pit)->get_ids();
		ofs << DECODER_NAME(decoder_name) << "_";
		for (unsigned i=0; i<ids.size(); i++) {               
			if (ids[i]) 
				ofs << symbol_get_string(oper_name_table[ids[i]]) << "_";
		}
		BinPattern patt = (*pit)->get_pattern();
		all_pats.push_back(patt);

		ofs << ", 0x" << patt.get_hex_mask()
			<< ", 0x" << patt.get_hex_signature() << ", 1)\n"; 
	}

	inverse(all_pats, default_pats);
	list<BinPattern>::const_iterator bin_it;
	for (bin_it=default_pats.begin(); bin_it!=default_pats.end(); bin_it++) {

		ofs << "IDEF(";
		if (default_name!=SymNULL) {
			ofs << DECODER_NAME(decoder_name) << "_"
				<< symbol_get_string(default_name);
		}
		else {
			ofs << "_func_default";
		}
		ofs << ", 0x" << (*bin_it).get_hex_mask()
			<< ", 0x" << (*bin_it).get_hex_signature() << ", 0.001)\n"; 
	}

	ofs.close();

}

void decode_entity::emit_header_file(const char *prefix) const
{
	ofstream ofs;
	string hfname = prefix;
	hfname += DECODER_HDR_FILE(decoder_name);
	ofs.open(hfname.c_str());
	message(hfname+" ...");

	ofs << "#undef _MAIN_DECODE_ENTRY" << endl;
	ofs << "#undef _FUNC_DEFAULT" << endl << endl;

	ofs << "typedef _UINT_T(" << std::dec << cc->get_width()
	   	<< ")::BASE_T _INST_T;" << endl;
	ofs << "#define _MAIN_DECODE_ENTRY  " OPER_ID_TYPE " *"
	 	<< DECODER_NAME(decoder_name)
	   	<< "(_INST_T inst)" << endl;
	ofs << "extern _MAIN_DECODE_ENTRY;" << endl;

	if (default_name!=SymNULL) {
		ofs << "#define _FUNC_DEFAULT "
			<< DECODER_NAME(decoder_name) << "_"
			<< symbol_get_string(default_name) << endl;
	}

	ofs.close();

}

void decode_entity::emit_opt_header_file(const char *prefix) const
{
	ofstream ofs;
	string hfname = prefix;
	hfname += DECODER_HDR_FILE(decoder_name);
	ofs.open(hfname.c_str());
	message(hfname+" ...");


	ofs << "#undef _MAIN_DECODE_ENTRY" << endl;
	ofs << "#undef _FUNC_DEFAULT" << endl << endl;

	ofs << "typedef _UINT_T(" << std::dec << cc->get_width()
	   	<< ")::BASE_T _INST_T;" << endl;
	ofs << "#define _MAIN_DECODE_ENTRY  unsigned " 
	 	<< DECODER_NAME(decoder_name)
	   	<< "(_INST_T inst)" << endl;
	ofs << "extern _MAIN_DECODE_ENTRY;" << endl;

	if (default_name!=SymNULL) {
		ofs << "#define _FUNC_DEFAULT "
			<< DECODER_NAME(decoder_name) << "_"
			<< symbol_get_string(default_name) << endl;
	}

	ofs.close();

}

void decode_entity::emit_tab_file(const char *prefix) const
{
	ofstream ofs;
	string tfname = prefix;
	tfname += DECODER_TAB_FILE(decoder_name);
	ofs.open(tfname.c_str());
	message(tfname+" ...");

	ofs << std::dec;

	ofs << OPER_ID_TYPE << " ";
	ofs << DECODER_TABLE(decoder_name) << "[] = {" << endl;

	const vector<DecodePattern *> *patterns = cc->get_patterns();
	vector<DecodePattern *>::const_iterator pit;
	for (pit=patterns->begin(); pit!=patterns->end(); pit++) {

		/* skip duplicate ones. */
		if ((*pit)->get_hflag()) continue;

		vector<OperID> ids = (*pit)->get_ids();
		for (unsigned i=0; i<ids.size(); i++) {               
			ofs << ids[i] << ",";
		}
		ofs << endl;
	}

	if (default_name!=SymNULL) {
		ofs << default_id << "," << endl;
	}
	ofs << "};" << endl << endl;

	int offset = 0;

	ofs << "#ifdef _DUMP_DECODE" << endl;
	ofs << "#include <iostream>" << endl;
	ofs << "#endif" << endl;
	for (pit=patterns->begin(); pit!=patterns->end(); pit++) {

		/* skip duplicate ones. */
		if ((*pit)->get_hflag()) continue;

		ofs << OPER_ID_TYPE << " *";

		vector<OperID> ids = (*pit)->get_ids();
		ofs << DECODER_NAME(decoder_name) << "_";
		for (unsigned i=0; i<ids.size(); i++) {               
			if (ids[i]) 
				ofs << symbol_get_string(oper_name_table[ids[i]]) << "_";
		}
		ofs << "(_INST_T inst) {" << endl;

		ofs << "#ifdef _DUMP_DECODE" << endl;
		ofs << "std::cerr << \"";
		for (unsigned i=0; i<ids.size(); i++) {               
			if (ids[i]) 
				ofs << symbol_get_string(oper_name_table[ids[i]]) << "_";
		}
		ofs << "\" << std::endl;" << endl;
		ofs << "#endif" << endl;

		ofs << "  return "
			<<	DECODER_TABLE(decoder_name)
		   	<< " + " << offset << ";" << endl;
		ofs << "}" << endl << endl;

		offset += ids.size();
	}

	if (default_name!=SymNULL) {
		ofs << OPER_ID_TYPE << " *"
			<< DECODER_NAME(decoder_name) << "_"
			<< symbol_get_string(default_name);
		ofs << "(_INST_T inst) {" << endl;

		ofs << "#ifdef _DUMP_DECODE" << endl;
		ofs << "std::cerr << \""
			<< symbol_get_string(default_name)
		   	<< "\" << std::endl;" << endl;
		ofs << "#endif" << endl;

		ofs << "  return "
			<<	DECODER_TABLE(decoder_name)
		   	<< " + " << offset << ";" << endl;
		ofs << "}" << endl << endl;
	}
	ofs.close();
}

void decode_entity::emit_opt_tab_file(const char *prefix) const
{
	ofstream ofs;

	string tfname = prefix;
	tfname += DECODER_TAB_FILE(get_decoder_name());
	ofs.open(tfname.c_str());
	message(tfname+" ...");
	ofs << std::dec;

	unsigned index=0;
	ofs << "#ifdef _DUMP_DECODE" << endl;
	ofs << "#include <iostream>" << endl;
	ofs << "#endif" << endl;

	const vector<DecodePattern *> *patterns = cc->get_patterns();
	vector<DecodePattern *>::const_iterator pit;
	for (pit=patterns->begin(); pit!=patterns->end(); pit++) {

		/* skip duplicate ones. */
		if ((*pit)->get_hflag()) continue;

		ofs << "unsigned ";

		vector<OperID> ids = (*pit)->get_ids();
		ofs << DECODER_NAME(get_decoder_name()) << "_";
		for (unsigned i=0; i<ids.size(); i++) {               
			if (ids[i]) 
				ofs << symbol_get_string(oper_name_table[ids[i]]) << "_";
		}
		ofs << "(_INST_T inst) {" << endl;

		ofs << "#ifdef _DUMP_DECODE" << endl;
		ofs << "std::cerr << \"";
		for (unsigned i=0; i<ids.size(); i++) {               
			if (ids[i]) 
				ofs << symbol_get_string(oper_name_table[ids[i]]) << "_";
		}
		ofs << "\" << std::endl;" << endl;
		ofs << "#endif" << endl;

		ofs << "  return " << index << ";" << endl;
		ofs << "}" << endl << endl;
		index++;
	}

	if (has_default()) {
		ofs << "unsigned "
			<< DECODER_NAME(get_decoder_name()) << "_"
			<< symbol_get_string(default_name);
		ofs << "(_INST_T inst) {" << endl;

		ofs << "#ifdef _DUMP_DECODE" << endl;
		ofs << "std::cerr << \""
			<< symbol_get_string(default_name)
		   	<< "\" << std::endl;" << endl;
		ofs << "#endif" << endl;


		ofs << "  return " << index <<";" << endl;
		ofs << "}" << endl << endl;
	}
	ofs.close();
}

void decode_entity::emit_main_file(const char *prefix) const
{
	ofstream ofs;
	string cfname = prefix;
	cfname += DECODER_MAIN_FILE(decoder_name);
	ofs.open(cfname.c_str());
	message(cfname+" ...");

	ofs << "#include \"" << prefix << OSM_LIB_INCLUDE << "\"" << endl;
	ofs << "#include \""
		<< prefix << DECODER_HDR_FILE(decoder_name) << "\"" << endl;
	ofs << "#include \"" << DECODER_COMMON_HDR << "\"" << endl;
	ofs << "#include \""
		<< prefix << DECODER_TAB_FILE(decoder_name) << "\"" << endl;
	ofs << "#include \""
		<< prefix << DECODER_DEC_FILE(decoder_name) << "\"" << endl;

	ofs.close();
}

ostream& arg_clause::print(ostream& os) const
{
	os << symbol_get_string(name) << ":";
	type->print(os);
	if (!type->is_readonly()) os << "*";
	print_object_list(os, " ", annots);
	return (os);
}

ostream& variable_clause::print(ostream& os) const
{
	os << symbol_get_string(name) << " : ";
	type->print(os);
	os << "; ";
	print_object_list(os, "\n", annots);
	return (os);
}

ostream& variable_paragraph::print(ostream& os) const
{
	os << "    VAR  ";
	print_object_list(os, "\n         ", vars);
	os << endl;
	return os;
}

ostream& eval_paragraph::print(ostream& os) const
{
	os << "    EVAL" << endl << "        ";

	vector<Expression *>::const_iterator eit;
	for (eit = clauses->begin(); eit != clauses->end(); eit++) {
		(*eit)->print(os);
		os << "; ";
		print_object_list(os, "\n         ", (*eit)->get_annotations());
		os << endl;
	}

	os << endl;
	return os;
}

eval_paragraph::~eval_paragraph() {
	delete_object_list(clauses); 
} 

bool eval_paragraph::type_check(
	const sym_context *var_context, const sym_context *arg_context,
	const sym_context *mvr_context, const sym_context *def_context,
	const OperContext *otxt, const FunContext *ftxt)
{
	bool ret = true;

	vector<Expression *>::iterator it;
	for (it=clauses->begin(); it!=clauses->end(); it++)
		ret = ((*it)->infer(var_context, arg_context, mvr_context, def_context,
				otxt, ftxt)!=NULL) && ret;

	return ret;
}

ostream& trans_clause::print(ostream& os) const
{
	os << "      " << symbol_get_string(edge) << ":" << endl;

	if (osm_exps) {
		os << "        {";
		print_object_list(os, ", ", osm_exps);
		os << "}";
	}

	if (exps) {
		os << endl << "        ";	
		print_object_list(os, ";\n        ", exps);
	}

	os << ";";
	return os;
}

void trans_clause::append_exp(expression *exp)
{
	if (exps==NULL) {
		exps = new vector<expression *>;
	}
	exps->push_back(exp);
}

bool trans_clause::type_check(const manager_context *man_context,
	const edge_context *edge_context, const buf_context *buf_context,
	const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context,
	const OperContext *oper_context, const FunContext *fun_context)
{
	bool ret = true;

	if (!edge_context->has(edge)) {
		undefined_ident(edge, this->get_loc());
		ret = false;
	}

	/* For osm can access all variables. */
	if (osm_exps) {
		vector<OSMExpression *>::iterator it;
		for (it=osm_exps->begin(); it!=osm_exps->end(); it++)
			if ((*it)->infer(man_context,
				buf_context, var_context, mvr_context, def_context)==NULL)
			ret = false;
	}

	if (exps) {
		vector<Expression *>::iterator it;
		for (it=exps->begin(); it!=exps->end(); it++)
			if ((*it)->infer(var_context, NULL, mvr_context,
						def_context, oper_context, fun_context)==NULL)
			ret = false;
	}

	/* We sort the osm expressions if everything is successful*/
	if (ret && osm_exps)
		sort(osm_exps->begin(), osm_exps->end(), OSMExpPtrCmp);

	return ret;
}

ostream& trans_paragraph::print(ostream& os) const
{
	os << "    TRANS" << endl;
	print_object_list(os, "\n\n",  clauses);
	return os;
}

trans_paragraph::~trans_paragraph() {
	delete_object_list(clauses);
}

bool trans_paragraph::type_check(const manager_context *man_context,
	const edge_context *edge_context, const buf_context *buf_context,
	const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context,
   	const OperContext *oper_context, const FunContext *fun_context)
{
	bool ret = true;

	set<Symbol> edge_names;

	vector<TransClause *>::iterator it;
	for (it=clauses->begin(); it!=clauses->end(); it++) {
		ret = (*it)->type_check(man_context, edge_context,
				buf_context, var_context, mvr_context,
				def_context, oper_context, fun_context) && ret;

		Symbol edge_name = (*it)->edge;
		if (edge_names.find(edge_name)!=edge_names.end()) {
			generic_error("Edge ", symbol_get_string(edge_name),
				" referred more than once", (*it)->get_loc());
			ret = false;
		}
		else edge_names.insert(edge_name);
	}

	return ret;
}


operation_section::~operation_section() {

	delete_object_list(paras);

	map<Symbol, DecodeEntity *>::iterator dit;
	for (dit=dec_ents.begin(); dit!=dec_ents.end(); dit++) {
		delete (*dit).second;
	}
	delete vcoding;
}

MachID operation_section::get_mach_id() const {
	return machine->get_id();
}

ostream& operation_section::print(ostream& os) const
{
	os << "OPERATION  " << symbol_get_string(name) << endl;
	print_object_list(os, "\n", annots);
    os << endl;
	print_object_list(os, "\n", paras);
	os << endl;
	return os;
}

void function_section::set_args(vector<ArgClause *> *as)
{
	args = as;
	vector<ArgClause *>::const_iterator it;
	vector<DataType *> *dtypes = new vector<DataType *>;

	for (it=args->begin(); it!=args->end(); it++)
		dtypes->push_back((*it)->get_data_type()->clone(NULL));

	type = new DataTypeGroup(dtypes, NULL);
}

function_section::~function_section() {
	delete_object_list(args);
	delete_object_list(paras);
	delete type;
}

ostream& function_section::print(ostream& os) const
{
	os << "FUNCTION  " << symbol_get_string(name);
	if (args) {
		os << "(";
		print_object_list(os, ", ", args);
		os << ")";
	}
	os << endl;
	print_object_list(os, "\n", annots);
	os << endl;
	print_object_list(os, "\n", paras);
	os << endl;
	return os;
}

coding_clause::~coding_clause() {
	delete_object_list(pats);
}

ostream& coding_clause_pattern::print(ostream& os) const
{
	(*pats)[0]->print(os);
	print_object_list(os, "\n", annots);
	return os;
}

ostream& coding_clause_var_ref::print(ostream& os) const
{
	os << symbol_get_string(var_ref);
	print_object_list(os, "\n", annots);
	return os;
}

bool coding_clause_var_ref::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;

    const DataType *tp = NULL;
    /* see if the variable is defined */
    if (var_ctxt && var_ctxt->has(var_ref)) {
        tp = var_ctxt->get_data_type(var_ref);
        vtype = VT_VAR;
    }
    else if (mvr_ctxt && mvr_ctxt->has(var_ref)) {
        tp = mvr_ctxt->get_data_type(var_ref);
        vtype = VT_MAC;
    }
    else if (def_ctxt && def_ctxt->has(var_ref)) {
        tp = def_ctxt->get_data_type(var_ref);
        vtype = VT_DEF;
    }
    else {
        undefined_ident(var_ref, this->get_loc());
        return false;
    }

	dtype = tp->clone(NULL);
	if (!valid_coding_type(tp)) {
		generic_error("Bad variable type for coding", this->get_loc());
		ret = false;
		return ret;
	}

	/* see if operation */
	is_oper = tp->get_type()==DT_OPER;

	/* VT_DEF allowed */
	if (tp->is_readonly()) {

		assert(vtype==VT_DEF);
		const DefineClause *dclause = static_cast<const DefineClause *>
				(def_ctxt->get_object(var_ref));

		if (dclause->get_value()==NULL) {
			generic_error("Readonly variable with unknown value"
						  " used for coding", this->get_loc());
			ret = false;
		}
		else {
			assert(dclause->get_value()->get_type()==DEF_VAL_NUM);
			const DefineValueNum *dvalue = static_cast<const DefineValueNum *>
					(dclause->get_value());

			const NumValue *val = dvalue->get_value();

			/* should have been checked by valid_coding_type */
			assert(!val->is_float());
			assert(tp->get_type()==DT_BIT);

			const DataTypeBit *btype =
				static_cast<const DataTypeBit *>(tp);

			width = btype->get_width();

			string str = val->get_val_inbase(2);

 			// this should fit, checked by define
			assert(str.size()<=(unsigned)width);

			/* pack zeros */
			if (str.size()<width) {
				str.insert(str.begin(), width-str.size(), '0');
			}
			pats->push_back(new DecodePattern(str));
		}
	}
	/* if other non-oper variables, we set the pattern to '-'s */
	else if (!is_oper) {

		assert(tp->get_type()==DT_BIT);
		const DataTypeBit *btype =
				static_cast<const DataTypeBit *>(tp);

		width = btype->get_width();

		string str(width, '-');

		pats->push_back(new DecodePattern(str));
	}

	return ret;
}

bool coding_clause_var_ref::check_patterns
   (const OperContext *op_ctxt,
	const sym_context *var_ctxt,
	const manager_context *man_ctxt,
	const sym_context *def_ctxt,
	const FunContext *fun_ctxt)
{
	/* a non-operation variable */
	if (!is_oper) return true;

	if (!var_ctxt || !var_ctxt->has(var_ref)) {
		undefined_ident(var_ref, this->get_loc());
		return false;
	}

	/* must be operation typed variable by now */
	const DataType *vtp = var_ctxt->get_data_type(var_ref);
	assert(vtp->get_type()==DT_OPER);

	const DataTypeOper *otp = static_cast<const DataTypeOper *>(vtp);
	const vector<Symbol> *opers = otp->get_opers();

	bool width_set = false;
	unsigned max_coding_depth = 0;
	vector<Symbol>::const_iterator sit;
	for (sit=opers->begin(); sit!=opers->end(); sit++) {

		/* in case no such operation,
		 * error should have been reported by type checker */
		if (!op_ctxt->has(*sit)) return false;

		const OperationSection *oper = op_ctxt->get_object(*sit);
		const vector<DecodePattern *> *bp = oper->get_patterns();

		if (bp==NULL) return false;

		if (!width_set) {
			width = oper->get_coding_width();
			width_set = true;
		}
		else if (width != oper->get_coding_width()) {
			generic_error("Operation coding lengths mismatch, near ",
				symbol_get_string(*sit), vtp->get_loc());
			return false;
		}

		if ((unsigned)(oper->get_coding_depth())>max_coding_depth) {
			max_coding_depth = oper->get_coding_depth();
		}
	}

	Symbol def_oper = otp->get_default_oper();
	if (def_oper!=SymNULL) {
		/* in case no such operation,
		 * error should have been reported by type checker */ //@@@ ignore warn
		if (!op_ctxt->has(def_oper)) return false;

		const OperationSection *oper = op_ctxt->get_object(def_oper);
		if (oper->has_coding()) {

			//if default has no coding, fine
			//if it has, we need to check width

			const vector<DecodePattern *> *bp = oper->get_patterns();
			if (bp==NULL) return false;

			/* if the default oper has coding */
			if (oper->get_coding_width()!=width) {
				generic_error("Operation coding lengths mismatch, near ",
					symbol_get_string(def_oper), vtp->get_loc());
				return false;
			}

			/* if the default oper has depth, wrong */
			if (oper->get_coding_depth()>0) {
				generic_error("Default operation \'",
					symbol_get_string(def_oper),
					"\' contain oper-typed variable in CODING", vtp->get_loc());
				return false;
			}
		}
	}

	/* now generate the patterns. */
	for (sit=opers->begin(); sit!=opers->end(); sit++) {

		const OperationSection *oper = op_ctxt->get_object(*sit);
		const vector<DecodePattern *> *bp = oper->get_patterns();

		vector<DecodePattern *>::const_iterator it;
		for (it=bp->begin(); it!=bp->end(); it++) {
			DecodePattern *npat = new DecodePattern(**it);
			npat->pad_ids_with_zeros(max_coding_depth);
			npat->prepend_id(oper->get_id());
			pats->push_back(npat);
		}
	}

	/* need one more to decode this itself */
	oper_depth = max_coding_depth+1;

	sort(pats->begin(), pats->end(), DecPatPtrCmp);
	return check_decode_patterns(pats, this->get_loc());
}

ostream& coding_clause_or_patterns::print(ostream& os) const
{
	os << "(";
	print_object_list(os, " | ", pats);
	os << ")";
	print_object_list(os, "\n", annots);
	return os;
}

bool coding_clause_or_patterns::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;

	assert(pats->size()>0);

	vector<DecodePattern *>::const_iterator p_it = pats->begin();
	width = (*p_it)->get_length();

	for (p_it++; p_it!=pats->end(); p_it++) {
		if (width!=(*p_it)->get_length()) {
			generic_error("Pattern lengths mismatch", this->get_loc());
			ret = false;
		}
	}

	return ret;
}

coding_paragraph::~coding_paragraph() {
	delete_object_list(elements);
	delete_object_list(patterns);
}

ostream& coding_paragraph::print(ostream & os) const
{
	os << "    CODING  ";
	print_object_list(os, " ", elements);

	os << "; ";
	os << endl;
	return os;
}

bool coding_paragraph::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;
	vector<CodingClause *>::iterator c_it;

	for (c_it=elements->begin(); c_it!=elements->end(); c_it++) {
		ret = (*c_it)->type_check(var_ctxt, mvr_ctxt, def_ctxt) && ret;
	}

	return ret;
}

bool coding_paragraph::check_patterns
   (const OperContext *op_ctxt,
	const sym_context *var_ctxt,
	const manager_context *man_ctxt,
	const sym_context *def_ctxt,
	const FunContext *fun_ctxt)
{
	/* if pattern is already checked */
	if (pat_checked) return check_result;

	pat_checked = true;
	check_result = false;

	vector<CodingClause *>::iterator c_it;
	map<unsigned, CodingClauseVarRef *> svars;

	/* Assume that there is no naming conflict between all contexts */
	for (c_it=elements->begin(); c_it!=elements->end(); c_it++) {

		/* check variable reference */
		if ((*c_it)->get_type()==EC_VAR_REF) {

			CodingClauseVarRef *cvar = static_cast<CodingClauseVarRef *>(*c_it);
			if (!cvar->check_patterns(op_ctxt, var_ctxt,
				man_ctxt, def_ctxt, fun_ctxt))
				return false;

			if (cvar->is_operation()) {
				assert(svars.find(cvar->get_index())==svars.end());
				svars[cvar->get_index()] = cvar;
			}
		}
	}

	coding_depth = 0;
	map<unsigned, CodingClauseVarRef *>::iterator cvar_it;
	for(cvar_it = svars.begin(); cvar_it!=svars.end(); cvar_it++) {
		(*cvar_it).second->set_table_offset(coding_depth);
		coding_depth += (*cvar_it).second->get_depth();
	}

	coding_width = 0;
	int n_pat = 1;
	vector<int> n_sizes;

	for (c_it=elements->begin(); c_it!=elements->end(); c_it++) {

		int pcnt = (*c_it)->get_patterns()->size();

		if (pcnt==0) {return false;}

		coding_width += (*c_it)->get_width();

		n_sizes.push_back(pcnt);
		n_pat *= pcnt;
	}

	unsigned offset = 0;
	vector<CodingClause *>::reverse_iterator r_it;
	for (r_it=elements->rbegin(); r_it!=elements->rend(); r_it++) {
		(*r_it)->set_offset(offset);
		offset += (*r_it)->get_width();
	}

	/* now we know that there are n_pat total patterns, we compute all these */
	int n_elem = elements->size();
	for (int i=0; i<n_pat; i++) {

		int cnt = i;
		DecodePattern bp("");

		map<unsigned, vector<OperID> > idses;

		for (int j=0; j<n_elem; j++) {

			CodingClause *cls = (*elements)[j];
			const vector<DecodePattern *> *dpats = cls->get_patterns();
			const DecodePattern *the_pat = (*dpats)[cnt%n_sizes[j]];
			bp = bp + *the_pat;

			if (cls->get_type()==EC_VAR_REF) {

				CodingClauseVarRef *cvar = 
					static_cast<CodingClauseVarRef *>(cls);

				if (cvar->is_operation()) {
					assert(idses.find(cvar->get_index())==idses.end());
					idses[cvar->get_index()] = the_pat->get_ids();
				}
			}
			cnt = cnt/n_sizes[j];
		}

		vector<OperID> ids;
		map<unsigned, vector<OperID> >::iterator vcit;
		for (vcit = idses.begin(); vcit!=idses.end(); vcit++) {
			ids.insert(ids.end(), (*vcit).second.begin(), (*vcit).second.end());
		}

		patterns->push_back(new DecodePattern(bp.get_pattern(), ids));
	}

	sort(patterns->begin(), patterns->end(), DecPatPtrCmp);
	check_result = check_decode_patterns(patterns, this->get_loc());

	return check_result;
}

ostream& syntax_table_lookup::print(ostream& os) const
{
	os << symbol_get_string(tab) << "[" << symbol_get_string(ind) << "]";
	return os;
}

bool syntax_table_lookup::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;

	Symbol tab_name = tab;
	Symbol ind_name = ind;

	/* check if table defined */
	if (!def_ctxt->has(tab_name)) {
		undefined_ident(tab_name, this->get_loc());
		ret = false;
	}

    const DataType *tp = NULL;
    /* see if the variable is defined */
    if (var_ctxt && var_ctxt->has(ind_name)) {
        tp = var_ctxt->get_data_type(ind_name);
        vtype = VT_VAR;
    }
    else if (mvr_ctxt && mvr_ctxt->has(ind_name)) {
        tp = mvr_ctxt->get_data_type(ind_name);
        vtype = VT_MAC;
    }
    else if (def_ctxt && def_ctxt->has(ind_name)) {
        tp = def_ctxt->get_data_type(ind_name);
        vtype = VT_DEF;
    }
    else {
        undefined_ident(ind_name, this->get_loc());
        ret = false;
    }

	/* Proceed only if good so far */
	if (!ret) return ret;

	t_type = def_ctxt->get_data_type(tab_name)->clone(NULL);
	d_type = tp->clone(NULL);

	/* index must be bit scalar */
	if (d_type->get_type()!=DT_BIT || !d_type->is_scalar()) {
		generic_error("Bad index type for syntax", this->get_loc());
		ret = false;
	}
	else {
		unsigned size = static_cast<const DataTypeString *>(t_type)->get_size();
		unsigned width = static_cast<const DataTypeBit *>(d_type)->get_width();

		if (width>31 || ((unsigned)1<<width)>size) {
			generic_warn("Potential out-of-bound array access",
				this->get_loc());
		}
	}

	/* must be string array */
	if (t_type->is_scalar() || 
		(t_type->get_type()!=DT_BIT && t_type->get_type()!=DT_FLOAT &&
		 t_type->get_type()!=DT_STRING)) {
		generic_error("Bad table type for syntax", this->get_loc());
		ret = false;
	 }

	/* also need to check modifier */
	if (mod!=SymNULL) {

		if (t_type->get_type()==DT_BIT) {
			if (mod!=Sym_hex && mod!=Sym_oct && mod!=Sym_bin && mod!=Sym_dec) {
				generic_error("Bad modifier", this->get_loc());
				ret = false;
			}
		}
		else if (t_type->get_type()==DT_FLOAT) {
			if (mod!=Sym_sci && mod!=Sym_fix) {
				generic_error("Bad modifier", this->get_loc());
				ret = false;
			}
		}
		else {
			generic_error("Bad modifier", this->get_loc());
			ret = false;
		}
	}

	return ret;
}

void syntax_table_lookup::emit_C(ostream& os) const
{
	emit_sym_of_var_type(tab, VT_DEF, "", os);
	os << "[ " GET_VALUE "(";
	emit_sym_of_var_type(ind, vtype, "", os);
	os << ")]";

	if (mod!=SymNULL) os << "." << symbol_get_string(mod) << "()";
	else if (t_type->get_type()==DT_BIT) os << ".hex()";
	else if (t_type->get_type()==DT_FLOAT) os << ".sci()";
}

ostream& syntax_clause_literal::print(ostream& os) const
{
	if (follow && !first) os << "^ ";
	os << "\"" << *literal << "\"";
	print_object_list(os, "\n", annots);
	return os;
}

void syntax_clause_literal::emit_C(ostream& os) const
{
	os << "\"" << *literal << "\"";
	print_object_list(os, "\n", annots);
}

ostream& syntax_clause_var_ref::print(ostream& os) const
{
	if (follow && !first) os << "^ ";
	os << symbol_get_string(var_ref);
	if (mod!=SymNULL) os << "." << symbol_get_string(mod);
	print_object_list(os, "\n", annots);
	return os;
}

bool syntax_clause_var_ref::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;

    const DataType *tp = NULL;
    /* see if the variable is defined */
    if (var_ctxt && var_ctxt->has(var_ref)) {
        tp = var_ctxt->get_data_type(var_ref);
        vtype = VT_VAR;
    }
    else if (mvr_ctxt && mvr_ctxt->has(var_ref)) {
        tp = mvr_ctxt->get_data_type(var_ref);
        vtype = VT_MAC;
    }
    else if (def_ctxt && def_ctxt->has(var_ref)) {
        tp = def_ctxt->get_data_type(var_ref);
        vtype = VT_DEF;
    }
    else {
        undefined_ident(var_ref, this->get_loc());
		return false;
	}

	/* if an oper type, may still need to check if it has syntax section */
	dtype = tp->clone(NULL);
	if (!valid_syntax_type(dtype)) {
		generic_error("Bad variable type for syntax", this->get_loc());
		ret = false;
	}

	/* also need to check modifier */
	if (mod!=SymNULL) {

		if (dtype->get_type()==DT_BIT) {
			if (mod!=Sym_hex && mod!=Sym_oct && mod!=Sym_bin && mod!=Sym_dec) {
				generic_error("Bad modifier", this->get_loc());
				ret = false;
			}
		}
		else if (dtype->get_type()==DT_FLOAT) {
			if (mod!=Sym_sci && mod!=Sym_fix) {
				generic_error("Bad modifier", this->get_loc());
				ret = false;
			}
		}
		else if (dtype->get_type()==DT_OPER) {
			if (mod!=Sym_syn) { /* useless but acceptable */
				generic_error("Bad modifier", this->get_loc());
				ret = false;
			}
		}
		else {
			generic_error("Bad modifier", this->get_loc());
			ret = false;
		}
	}
	return ret;
}

void syntax_clause_var_ref::emit_C(ostream& os) const
{
	emit_sym_of_var_type(var_ref, vtype, "", os);
	if (dtype->get_type()==DT_BIT) {
		if (mod!=SymNULL) os << "." << symbol_get_string(mod) << "()";
		else os << ".hex()";	// default is hex
	}
	else if (dtype->get_type()==DT_FLOAT) {
		if (mod!=SymNULL) os << "." << symbol_get_string(mod) << "()";
		else os << ".sci()";	// default is sci
	}
	else if (dtype->get_type()==DT_OPER) {
		os << "->__get_syntax()";
	}
}

ostream& syntax_clause_tab_ref::print(ostream& os) const
{
	if (follow && !first) os << "^ ";
	tab_ref->print(os);
	print_object_list(os, "\n", annots);
	return os;
}

bool syntax_clause_tab_ref::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	return tab_ref->type_check(var_ctxt, mvr_ctxt, def_ctxt);
}

void syntax_clause_tab_ref::emit_C(ostream& os) const
{
	tab_ref->emit_C(os);
}

ostream& syntax_clause_or_literals::print(ostream& os) const
{
	if (follow && !first) os << "^ ";
	os << "(\"";
	print_string_list(os, "\" | \"", literals);
	os << "\")";
	print_object_list(os, "\n", annots);
	return os;
}

void syntax_clause_or_literals::emit_C(ostream& os) const
{
	os << "\"" <<  *(literals->begin()) << "\"";
}

ostream& syntax_clause_or_tab_refs::print(ostream& os) const
{
	if (follow && !first) os << "^ ";
	os << "(";
	print_object_list(os, " | ", trs);
	os << ")";
	print_object_list(os, "\n", annots);
	return os;
}

syntax_clause_or_literals::~syntax_clause_or_literals() {
	delete_object_list(literals);
}

syntax_clause_or_tab_refs::~syntax_clause_or_tab_refs() {
	delete_object_list(trs);
} 

bool syntax_clause_or_tab_refs::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;

	vector<syntax_table_lookup *>::iterator it;
	for (it=trs->begin(); it!=trs->end(); it++) {
		ret = (*it)->type_check(var_ctxt, mvr_ctxt, def_ctxt) && ret;
	}

	return ret;
}

void syntax_clause_or_tab_refs::emit_C(ostream& os) const
{
	(*(trs->begin()))->emit_C(os);
}

ostream& syntax_paragraph::print(ostream& os) const
{
	os << "    SYNTAX  ";
	print_object_list(os, " ", elements);
	os << "; ";
	os << endl;
	return os;
}

syntax_paragraph::~syntax_paragraph() {
	delete_object_list(elements);
} 

bool syntax_paragraph::type_check(
	const sym_context *var_ctxt, const sym_context *mvr_ctxt,
   	const sym_context *def_ctxt)
{
	bool ret = true;
	vector<SyntaxClause *>::iterator c_it;

	/* Assume that there is no naming conflict between all contexts */
	for (c_it=elements->begin(); c_it!=elements->end(); c_it++) {
		ret = (*c_it)->type_check(var_ctxt, mvr_ctxt, def_ctxt) && ret;
	}

	return ret;
}

void syntax_paragraph::emit_C(ostream& os) const
{
	assert(elements->size()>0);

	os << "#ifdef _GEN_DISASSEMBLER" << endl;

	vector<syntax_clause *>::const_iterator it = elements->begin();
	os << "  _STR_T syntax = ";
	(*it)->emit_C(os);
	os << ";" << endl;

	for (it++; it!=elements->end(); it++) {
		if (!(*it)->get_follow()) os << "  syntax += \" \";" << endl;
		os << "  syntax += ";
		(*it)->emit_C(os);
		os << ";" << endl;
	}

	os << "  return syntax;" << endl;
	os << "#else" << endl;
	os << "  return \"\";" << endl;
	os << "#endif" << endl;
}

bool operation_section::classify_paragraphs()
{
	bool ret = true;

	if (classified) return classify_result;
	classified = true;

	vector<OperationParagraph *>::iterator it;
	for (it=paras->begin(); it!=paras->end(); it++) {
		switch((*it)->get_type()) {
			case OPER_PARA_VAR:
				if (var) {
					generic_error("More than one VAR's in OPERATION",
						this->get_loc());
					ret = false;
				}
				else var = static_cast<variable_paragraph *>(*it);
				break;
			case OPER_PARA_TRANS:
				if (trans) {
					generic_error("More than one TRANS's in OPERATION",
						this->get_loc());
					ret = false;
				}
				else trans = static_cast<TransParagraph *>(*it);
				break;
			case OPER_PARA_SYNTAX:
				if (syntx) {
					generic_error("More than one SYNTAX's in OPERATION",
						this->get_loc());
					ret = false;
				}
				else syntx = static_cast<SyntaxParagraph *>(*it);
				break;
			case OPER_PARA_CODING:
				if (codng) {
					generic_error("More than one CODING's in OPERATION",
						this->get_loc());
					ret = false;
				}
				else
					codng = static_cast<CodingParagraph *>(*it);
				break;
			case OPER_PARA_EVAL:
				if (eval) {
					generic_error("More than one EVAL's in OPERATION",
						this->get_loc());
					ret = false;
				}
				else eval = static_cast<EvalParagraph *>(*it);
				break;
		}
	}

	classify_result = ret;
	return ret;
}

bool function_section::classify_paragraphs()
{
	bool ret = true;

	if (classified) return classify_result;
	classified = true;

	vector<OperationParagraph *>::iterator it;
	for (it=paras->begin(); it!=paras->end(); it++) {
		switch((*it)->get_type()) {
			case OPER_PARA_VAR:
				if (var) {
					generic_error("More than one VAR's in FUNCTION",
						this->get_loc());
					ret = false;
				}
				else var = static_cast<variable_paragraph *>(*it);
				break;
			case OPER_PARA_EVAL:
				if (eval) {
					generic_error("More than one ASSIGN's in FUNCTION",
						this->get_loc());
					ret = false;
				}
				else eval = static_cast<EvalParagraph *>(*it);
				break;
			default:
				generic_error("Illegal content in FUNCTION", this->get_loc());
				ret = false;
				break;
		}
	}

	if (!eval) {
		generic_error("Missing EVAL in FUNCTION", this->get_loc());
		ret = false;
	}

	classify_result = ret;
	return ret;
}

variable_paragraph::~variable_paragraph() {
	delete_object_list(vars);
}

bool variable_paragraph::type_check(bool is_oper, sym_context *context)
{
	bool ret = true;
	map<Symbol, const Location *> var_names;

	vector<VarClause *>::iterator it;
	for (it=vars->begin(); it!=vars->end(); it++) {

		Symbol name = (*it)->get_name();

		bool conflict = false;

		/*reserved variable */
		if (name==Sym_coding)
			conflict = true;

		/*reserved variable end */

		if (conflict) {
			generic_error("Using reserved variable name", (*it)->get_loc());
			ret = false;
		}
		/* check if two variables have the same name */
		else if (var_names.find(name)!=var_names.end()) {
			naming_conflict(name, (*it)->get_loc(), var_names[name]);
			ret = false;
		}
		/* check if variable name clashes with existing identifier */
		else if (context->has(name)) {
			naming_conflict(name, (*it)->get_loc(),
				context->get_loc(name));
			ret = false;
		}
		else {
			var_names[name] = (*it)->get_loc();

			const DataType *dtype = (*it)->get_data_type();

			// check if a name appear multiple times in the oper list
			if (dtype->get_type()==DT_OPER) {
				const DataTypeOper *otype =
					static_cast<const DataTypeOper *>(dtype);

				const vector<Symbol> *opers = otype->get_opers();
				set<Symbol> oper_set;
				vector<Symbol>::const_iterator sit;
				for (sit=opers->begin(); sit!=opers->end(); sit++) {
					if (oper_set.find(*sit)!=oper_set.end()) {
						generic_error("Multiple references of ",
							symbol_get_string(*sit),
							" in list", dtype->get_loc());
						ret = false;
					}
					else oper_set.insert(*sit);
				}
			}

			if ((is_oper && !valid_var_type(dtype)) ||
				(!is_oper && !valid_fvar_type(dtype))) {
				generic_error("Bad variable type", (*it)->get_loc());
				ret = false;
			}
			else {
				var_cache.insert(*it);		
			}
		}
	}
	return ret;
}

int operation_section::dec_entity_cmp(const DecodeEntity *op1,
									const DecodeEntity *op2)
{
	// now we try to determine the order of the two nodes
	const vector<decode_info *>& ias = op1->get_infos();
	const vector<decode_info *>& ibs = op2->get_infos();

	bool a_dom_b = false;
	bool b_dom_a = false;
	bool no_comp = false;

	vector<decode_info *>::const_iterator ias_it;
	vector<decode_info *>::const_iterator ibs_it;
	for (ias_it=ias.begin(); ias_it!=ias.end(); ias_it++) {

		for (ibs_it=ibs.begin(); ibs_it!=ibs.end(); ibs_it++) {

			Symbol aedge = (*ias_it)->edge;
			Symbol bedge = (*ibs_it)->edge;

			if (aedge==SymNULL && bedge!=SymNULL)
				a_dom_b = true;
			else if (bedge==SymNULL && aedge!=SymNULL)
				b_dom_a = true;
			else if (aedge==bedge) {
				if ((*ias_it)->index<(*ibs_it)->index)
					a_dom_b = true;
				else
					b_dom_a = true;
			}
			else if (machine->edge_dominate(aedge, bedge))
				a_dom_b = true;
			else if (machine->edge_dominate(bedge, aedge))
				b_dom_a = true;
			else
				no_comp = true;
		}
	}

	if ((a_dom_b && b_dom_a) || no_comp) return 0;
	if (a_dom_b) return -1; // a is earlier than b
	return 1;
}


bool operation_section::pre_type_check(const manager_context *man_context,
	const sym_context *def_context, const OperContext *oper_context,
	const FunContext *func_context)
{
	if (type_checked) return check_result;

	bool ret = true;

	/* Note arguments, mvariables, defines, variables should not have
	 * naming clashes.
	 * We create a lump sum context
	 */
	sym_context ctxt = machine->var_cache;

	sym_context::const_iterator sym_it;
	for (sym_it=ctxt.begin(); sym_it!=ctxt.end(); sym_it++) {
		var_types[(*sym_it).second->get_name()] = VT_MAC;
	}

	if (def_context) {
   		ctxt += *def_context;
		for (sym_it=def_context->begin();
			sym_it!=def_context->end(); sym_it++) {
			var_types[(*sym_it).second->get_name()] = VT_DEF;
		}
	}

	if (var) {

		if (!var->type_check(true, &ctxt)) ret = false;

		ctxt += var->var_cache;

		/* now check Oper type variable */
		vector<VarClause *>::iterator it;
		for (it=var->vars->begin(); it!=var->vars->end(); it++) {

			var_types[(*it)->get_name()] = VT_VAR;

			const DataType *dtype = (*it)->get_data_type();
			if (dtype->get_type()==DT_OPER) {

				/* allocate decoding entry and then cache it*/
				DecodeEntity *ndec = new 
					DecodeEntity(this, *it, oper_context, ret);

				/* each oper typed variable has a decoding entity */
				dec_ents[(*it)->get_name()] = ndec;
			}
		}
	}

	decode_leaf = true;

	// check if the coding paragraph has any oper-typed variable
	// if so, then the operation cannot be used as default operation
	if (codng && var) {

		const vector<CodingClause *> *clss = codng->get_clauses();
		vector<CodingClause *>::const_iterator cit;
		for (cit = clss->begin(); cit!=clss->end(); cit++) {
			if ((*cit)->get_type()==EC_VAR_REF) {
				const CodingClauseVarRef *cvar = 
					static_cast<const CodingClauseVarRef *>(*cit);
				Symbol vname = cvar->get_name();
				if (var->var_cache.has(vname) &&
					var->var_cache.get_data_type(vname)->get_type()==DT_OPER) {
					decode_leaf = false;
				}
			}
		}
	}

	/* set the need_decoder flag dec_ents */
	if (trans && trans->clauses) {

		vector<TransClause *>::iterator tit;
		for (tit=trans->clauses->begin(); tit!=trans->clauses->end(); tit++) {

			vector<expression *> *exps = (*tit)->exps;
			if (exps) {

				unsigned index = 0;
				vector<expression *>::iterator eit;
				for (eit=exps->begin(); eit!=exps->end(); eit++, index++) {

					if ((*eit)->get_type()==OP_DEC ||
						(*eit)->get_type()==OP_ACT) {

						OperationExpression *oexp =
							static_cast<OperationExpression *>(*eit);

						Symbol op_name = oexp->get_oper();

						if (dec_ents.find(op_name)!=dec_ents.end()) {

							DecodeEntity *dent = dec_ents[op_name];

							oexp->set_singleton(dent->is_singleton,
								dent->singleton_id);

							DecodeInfo *info = 
								new DecodeInfo((*tit)->get_edge(), index, oexp);

							dent->infos.push_back(info);

							/* If argument is not null, then need decoder */
							if (oexp->get_arg()!=NULL) {
								oexp->set_decoder_name(dent->decoder_name);
							}
						}
						/* else the error will be reported by trans typecheck*/
					}
				}
			}
		}
	}

	if (eval && eval->clauses) {

		unsigned index = 0;
		vector<expression *> *exps = eval->clauses;
		vector<expression *>::iterator eit;
		for (eit=exps->begin(); eit!=exps->end(); eit++, index++) {

			if ((*eit)->get_type()==OP_DEC ||
				(*eit)->get_type()==OP_ACT) {

				OperationExpression *oexp =
					static_cast<OperationExpression *>(*eit);

				Symbol op_name = oexp->get_oper();
				if (dec_ents.find(op_name)!=dec_ents.end()) {

					DecodeEntity *dent = dec_ents[op_name];
					oexp->set_singleton(dent->is_singleton, dent->singleton_id);

					DecodeInfo *info = new DecodeInfo(SymNULL, index, oexp);
					dent->infos.push_back(info);

					/* If argument is not null, then need decoder */
					if (oexp->get_arg()!=NULL) {
						oexp->set_decoder_name(dent->decoder_name);
					}
				}
				/* else the error will be reported by eval typecheck*/
			}
		}
	}

	/* now we bubble sort the dec_ents */
	map<Symbol, DecodeEntity *>::iterator sym_it1, sym_it2;
	for (sym_it1 = dec_ents.begin(); sym_it1!=dec_ents.end(); sym_it1++){
		(*sym_it1).second->index = 0;
	}

	for (sym_it1 = dec_ents.begin(); sym_it1!=dec_ents.end(); sym_it1++){

		DecodeEntity *d1 = (*sym_it1).second;
		for (sym_it2=sym_it1, sym_it2++; sym_it2!=dec_ents.end(); sym_it2++){

			DecodeEntity *d2 = (*sym_it2).second;

			/* compare these two; */ 
			int cmp = dec_entity_cmp(d1, d2);

			/* cmp==0 means tie, bad for optimization ambiguous. */ //@@@
			/* >0 means 1 should be later than 2, swap index if necessary*/
			if (cmp<=0) d2->index++;
			else if (cmp>0) d1->index++;
		}
	}
	
	/* need decoder flags done */
	return ret;
}

bool operation_section::type_check(const manager_context *man_context,
	const sym_context *def_context, OperContext *oper_context,
	const FunContext *func_context)
{
	if (type_checked) return check_result;

	bool ret = true;

	type_checked = true;
	check_result = false;

	/* Note arguments, mvariables, defines, variables should not have
	 * naming clashes.
	 * We create a lump sum context
	 */
	sym_context ctxt = machine->var_cache;

	if (def_context) ctxt += *def_context;

	if (var) ctxt += var->var_cache; 

	/* check Coding type */
	if (codng)
		ret = codng->type_check(var?&(var->var_cache):NULL,
					&(machine->var_cache), def_context) && ret;


	/* set the coding clause field for dec_ents explicitly existing in
	 * coding paragraph. */
	if (codng) {

		vector<CodingClause *> *ccls = codng->elements;
		vector<CodingClause *>::iterator it;
		for (it=ccls->begin(); it!=ccls->end(); it++) {

			if ((*it)->get_type()==EC_VAR_REF) {

				CodingClauseVarRef *vcls =
					static_cast<CodingClauseVarRef *>(*it);

				/* this is set by coding type_check */
				if (!vcls->is_operation()) continue;

				Symbol op_name = vcls->get_name();
				assert (dec_ents.find(op_name)!=dec_ents.end());
			//	if (dec_ents.find(op_name)!=dec_ents.end()) {
					dec_ents[op_name]->cc = vcls;
					vcls->set_index(dec_ents[op_name]->index);
			//	}
				/* else the error will be reported by trans typecheck*/
			}
		}
	}

	/* check pattern if everything is ok */
	if (ret && codng) {
		ret = codng->check_patterns(oper_context, &ctxt,
			man_context, def_context, func_context) && ret;
	}

	map<Symbol, DecodeEntity *>::const_iterator dit;
	for (dit=dec_ents.begin(); dit!=dec_ents.end(); dit++)
		if (!(*dit).second->type_check()) ret = false;

	/* check the patterns of other decoding entities */
	map<Symbol, DecodeEntity *>::iterator de_it;
	for (de_it=dec_ents.begin(); de_it!=dec_ents.end(); de_it++) {
		DecodeEntity *de = (*de_it).second;

		/* if not in coding paragraph, but explicitly specified with
		   decoding command and argument */
		if (de->cc==NULL && de->need_dec) {

			/* construct a fake coding clause for this entry */
			de->has_cc = true;
			de->cc = new CodingClauseVarRef((*de_it).first, 
						new Location(*(de->var->get_loc())));

			ret = de->cc->type_check(var?&(var->var_cache):NULL,
				&(machine->var_cache), def_context) && ret;

			ret = de->cc->check_patterns(oper_context, &ctxt,
				man_context, def_context, func_context) && ret;

		}
		/* if neither in coding, nor with decoding command */
		else if (!de->cc && !de->is_singleton) {
			generic_error("Don't know how to decode ",
				symbol_get_string(de->var->get_name()),
				de->var->get_loc());
			ret = false;
		}
	}

	/* Eval and trans are the last since they may need coding knowledge
	 * such as the width of an operation's coding
	 */

	sym_context var_cash;
	if (var) var_cash = var->var_cache;

	if (codng) { 
		vcoding = new
		   	VarClause(Sym_coding,
				new DataTypeBit(false, codng->get_width(), 0, NULL),
				new Location(*(codng->get_loc())));
		var_cash.insert(vcoding);
	}

	/* this one needs coding variable possibly, so do it later. */
	if (syntx)
		ret = syntx->type_check(&var_cash,
					&(machine->var_cache), def_context) && ret;

	/* stop here if there are errors */
	if (!ret) return ret;

	/* check TRANS type */
	if (trans)
		ret = trans->type_check(man_context,
				&(machine->edge_cache), &(machine->buf_cache), &var_cash,
				&(machine->var_cache), def_context,
			   	oper_context, func_context) && ret;

	/* check EVAL type */
	if (eval)
		ret = eval->type_check(&var_cash, NULL,
				&(machine->var_cache), def_context,
				oper_context, func_context) && ret;

	check_result = ret;
	return ret;
}


const vector<DecodePattern *> *operation_section::get_patterns() const
{
	if (!codng) return NULL;
	else return codng->get_patterns();
}

bool function_section::type_check(const sym_context *def_context,
	const FunContext *fnc_context)
{
	bool ret = true;

	/* should have been classified by now. */
	assert(classified);
	ret = classify_result;

	/* build up the argument context */
	vector<ArgClause *>::iterator arg_it;
	for (arg_it=args->begin(); arg_it!=args->end(); arg_it++) {

		Symbol name = (*arg_it)->get_name();

		/* see if conflict with define */
		if (def_context && def_context->has(name)) {
			naming_conflict(name, (*arg_it)->get_loc(),
				def_context->get_loc(name));
			ret = false;
		}
		/* see if conflict with other arguments */
		else if (arg_context.has(name)) {
			naming_conflict(name, (*arg_it)->get_loc(),
				arg_context.get_loc(name));
			ret = false;
		}
		else {
			arg_context.insert(*arg_it);
		}
	}

	sym_context ctxt = arg_context;
	if (def_context) ctxt += *def_context;

	if (var) {
		var->type_check(true, &ctxt);
		ctxt += var->var_cache;
	}

	/* check EVAL type */
	if (eval) {
		ret = eval->type_check(var?&(var->var_cache):NULL, &arg_context,
			 NULL, def_context, NULL, fnc_context) && ret;
	}

	return ret;
}

bool function_section::propogate_dirty(const sym_context *def_context,
	const FunContext *fnc_context)
{
	if (!eval || dirty) return false;

	/* see if the function calls a dirty external function. */
	vector<Expression *>::const_iterator it;
	for (it=eval->clauses->begin(); it!=eval->clauses->end(); it++) {
		if ((*it)->get_type()==OP_CALL) {
			const CallExpression *ce =
				static_cast<const CallExpression *>(*it);

			Symbol cname = ce->get_func_name();
			if (def_context && def_context->has(cname)) {
				const DefineClause *dc = static_cast<const DefineClause *>
					(def_context->get_object(cname));
				if (dc->is_dirty())
					dirty = true;
			}
			else if (fnc_context->has(cname)) {
				const FunctionSection *dc = static_cast<const FunctionSection*>
					(fnc_context->get_object(cname));
				if (dc->is_dirty())
					dirty = true;
			}
			/* else the error will be reported by type_check*/
		}
	}

	return dirty;
}

void arg_clause::emit_C(ostream& os) const
{
	if (type->is_readonly()) os << "const ";
	type->emit_C(os);
	os << "& " << ARG_NAME(name);
}

void variable_clause::emit_C(ostream& os) const
{
	type->emit_C(os);
	os << " " << VAR_NAME(name);
}

void variable_paragraph::emit_C(ostream& os) const
{
	vector<VarClause *>::iterator it;

	for (it=vars->begin(); it!=vars->end(); it++) {
		os << "  ";
	   	(*it)->emit_C(os);
	    os << ";" << endl;
	}
}

void eval_paragraph::emit_C(ostream& os) const
{
	vector<Expression *>::iterator it;

	for (it=clauses->begin(); it!=clauses->end(); it++) {
	   	(*it)->emit_C(os, "  ", "");
	    os << endl;
	}
}

void operation_section::emit_C_dec(ostream& os) const
{
	os << endl << endl;
	os << "class " << OPER_CLASS(name)
		<< " : public _BASE_OPERATION {" << endl << endl;

	os << "  public:" << endl << endl;

	os << "    " << OPER_CLASS(name) << "();" << endl << endl;
	os << "    void _OPER_INIT(_BASE_MACHINE *mach, " 
		<< OPER_ID_TYPE	<< " *dtab, void *cdng);" << endl << endl;
	os << "    bool __test_osm(unsigned int id);" << endl << endl;
	os << "    void __run_osm_pre(unsigned int id);" << endl << endl;
	os << "    void __run_osm_post(unsigned int id);" << endl << endl;
	os << "    void __run_exp(unsigned int id);" << endl << endl;
	os << "    bool __run_all(unsigned int id);" << endl << endl;

	if (codng)
		os << "    const void *__get_coding() {return &"
				<< VAR_REF("", Sym_coding) << ";}"
			   	<< endl << endl;

	if (syntx)
		os << "    std::string __get_syntax();" << endl << endl;

	os << "  private:" << endl << endl;

	os << "    " << MACH_CLASS(machname) << " *_machine_;" << endl << endl;

	// if have coding section, then should have a coding variable
	if (codng) {
		DataTypeBit *tmp =
			new DataTypeBit(false, codng->get_width(), 0, NULL);

		os << "    ";
		tmp->emit_C(os);
		os << " " << VAR_NAME(Sym_coding) << ";" << endl << endl;

		delete tmp;
	}

	if (var) {

		/* first emit normal variables */
		vector<VarClause *>::iterator it;
		for (it=var->vars->begin(); it!=var->vars->end(); it++) {
			const DataType *dtype = (*it)->get_data_type();
			if (dtype->get_type()!=DT_OPER) {

				os << "    ";
				dtype->emit_C(os);
				os << " " << VAR_NAME((*it)->get_name()) << ";" << endl << endl;
			}
		}
		
		/* second emit oper variable*/
		for (it=var->vars->begin(); it!=var->vars->end(); it++) {
			const DataType *dtype = (*it)->get_data_type();
			if (dtype->get_type()==DT_OPER) {

				os << "    _BASE_OPERATION *"
					<< VAR_NAME((*it)->get_name()) << ";" << endl;

				os << endl;
			}
		}
	}

	if (codng) {

		const vector<CodingClause *> *ccls = codng->elements;
		vector<CodingClause *>::const_iterator it;

		for (it=ccls->begin(); it!=ccls->end(); it++) {

			int width = (*it)->get_width();

			if ((*it)->get_type()==EC_VAR_REF) {

				const CodingClauseVarRef *vcls =
					static_cast<const CodingClauseVarRef *>(*it);

				if (!vcls->is_operation()) continue;

				Symbol vname = vcls->get_name();

				DataTypeBit *tmp =
					new DataTypeBit(false, width, 0, NULL);

				os << "    ";
				tmp->emit_C(os);
				os << " " << CODING_NAME(vname) << ";" << endl; 

				delete tmp;

				/* An operation requiring further decoding */
				if (vcls->get_depth()) {
					os << "    " << OPER_ID_TYPE
					  << " *" << DTABLE_NAME(vname) << ";" << endl;
				}
			}
			os << endl;
		}
	}
	os << "};" << endl << endl;
}

void operation_section::emit_C_init(ostream& os) const
{

	os << "  _mach->__register_oper(this);" << endl;
	os << "  _machine_ = static_cast<" << MACH_CLASS(machname)
		 	<< " *>(_mach);" << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	// then set osm edges
	if (trans && trans->clauses) {

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {

			int reg_val=0;

			if ((*trans_it)->osm_exps) reg_val = 2;
			if ((*trans_it)->exps || (*trans_it)->osm_exps==NULL) reg_val+=1;

			os << "  _REGISTER_OPER_ON_EDGE("
				<< MACH_CLASS(machine->get_name()) << ","
				<< LOCAL_IDENT((*trans_it)->edge) << ","
				<< reg_val << ");" << endl;

		}
	}
	os << "#endif" << endl;

	// next set the coding clauses
	if (codng) {

		DataTypeBit *tmp =
			new DataTypeBit(false, codng->get_width(), 0, NULL);

		os << "  " << VAR_REF("", Sym_coding) << " = *(static_cast<";
		tmp->emit_C(os);
		os << " *>(_cdng));" << endl << endl;

		delete tmp;

		int current = codng->get_width()-1;
		//int cur_dtab_ptr = 0;

		const vector<CodingClause *> *ccls = codng->elements;
		vector<CodingClause *>::const_iterator it;

		for (it=ccls->begin(); it!=ccls->end(); it++) {

			int width = (*it)->get_width();

			if ((*it)->get_type()==EC_VAR_REF) {

				const CodingClauseVarRef *vcls =
					static_cast<const CodingClauseVarRef *>(*it);

				const DataType *dtype = vcls->get_data_type();

				if (dtype->is_readonly()) {
					assert(vcls->get_var_type()==VT_DEF);
					continue;
				}

				Symbol vname = vcls->get_name();

				os << "  ";
				if (vcls->is_operation()) 
					os << CODING_NAME(vname);
				else
					emit_sym_of_var_type(vname, vcls->get_var_type(), "", os); 

				/** add __coding suffix for operation. */

				os << " = ";

				bool is_sign = false;

				if (dtype->get_type()==DT_BIT) {
					const DataTypeBit *bdtype =
						static_cast<const DataTypeBit *>(dtype);

					is_sign = bdtype->is_signed();
				}

				if (is_sign)
					os << "_INT_";
				else
					os << "_UINT_";

				if (width==1) {
					os << "XTR1(" << VAR_REF("", Sym_coding)
					   	<< ", " << current << ");" << endl;
				}
				else {
					os << "XTR2(" << VAR_REF("", Sym_coding)
						<< ", " << current << ", "
						<< current - width + 1 << ");" << endl;
				}

				/* An operation requiring further decoding,
				 * cannot be singleton since singleton does not have coding. */
				if (vcls->is_operation()) {

					if (vcls->get_depth()) {
						os << "  " << DTABLE_NAME(vname);
						os << " = _dtab + " << vcls->get_table_offset() 
							<< ";" << endl;
						//assert(cur_dtab_ptr==vcls->get_table_offset());
					}
					//cur_dtab_ptr += vcls->get_depth();
				}
			}

			current -= width;
			assert(current>=-1);
		}
	//	os << cur_dtab_ptr << " " << this->get_coding_depth() << endl;
		//assert(cur_dtab_ptr==this->get_coding_depth());
	}

	// then run eval 
	if (eval) eval->emit_C(os);

	os << endl;

}

void operation_section::emit_C_osm_test_element(ostream& os,
	TransClause *tp) const
{
	vector<osm_expression *>::const_iterator eit;
	for (eit=tp->osm_exps->begin(); eit!=tp->osm_exps->end(); eit++) {
		os << "      if (!";
		(*eit)->emit_C_test(os, "");
		os << ") return false;" << endl;
	}
}

void operation_section::emit_C_osm_test(ostream& os) const
{
	if (trans && trans->clauses) {
		os << "  switch(id) {" << endl << endl;

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {
			
			if ((*trans_it)->osm_exps) {

				os << "    case " << MACH_CLASS(machine->get_name()) << "::"
					<< LOCAL_IDENT((*trans_it)->edge) << ":" << endl << endl;

				emit_C_osm_test_element(os, *trans_it);
				os << "      break;" << endl << endl;
			}
		}
		os << "    default: break;" << endl;
		os << "  }" << endl;
	}
}

void operation_section::emit_C_osm_pre_element(ostream& os,
	TransClause *tp) const
{
	vector<osm_expression *>::const_iterator eit;
	for (eit=tp->osm_exps->begin(); eit!=tp->osm_exps->end(); eit++) {
		if ((*eit)->is_pre_eval())
			if ((*eit)->emit_C(os, "      ", ""))
				os << endl << endl;
	}
}

void operation_section::emit_C_osm_pre(ostream& os) const
{

	if (trans && trans->clauses) {

		os << "  switch(id) {" << endl << endl;

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {
			
			if ((*trans_it)->osm_exps) {

				os << "    case " << MACH_CLASS(machine->get_name()) << "::"
					<< LOCAL_IDENT((*trans_it)->edge) << ":" << endl << endl;

				/* The emit class 1 and class 2 */
				emit_C_osm_pre_element(os, *trans_it);
				os << "      break;" << endl << endl;
			}
		}
		os << "    default: break;" << endl;
		os << "  }" << endl;
	}
}

void operation_section::emit_C_osm_post_element(ostream& os,
	TransClause *tp) const
{
	vector<osm_expression *>::const_iterator eit;
	for (eit=tp->osm_exps->begin(); eit!=tp->osm_exps->end(); eit++) {
		if (!(*eit)->is_pre_eval())
			if ((*eit)->emit_C(os, "      ", ""))
				os << endl << endl;
	}
}

void operation_section::emit_C_osm_post(ostream& os) const
{
	if (trans && trans->clauses) {

		os << "  switch(id) {" << endl << endl;

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {
			
			if ((*trans_it)->osm_exps) {

				os << "    case " << MACH_CLASS(machine->get_name()) << "::"
					<< LOCAL_IDENT((*trans_it)->edge) << ":" << endl << endl;

				/* The emit class 3 and class 4 */
				emit_C_osm_post_element(os, *trans_it);
				os << "      break;" << endl << endl;
			}
		}

		os << "    default: break;" << endl;
		os << "  }" << endl;
	}
}

void operation_section::emit_C_exp_element(ostream& os,
	TransClause *tp) const
{
	vector<expression *>::const_iterator eit;
	for (eit=tp->exps->begin(); eit!=tp->exps->end(); eit++) {
		(*eit)->emit_C(os, "      ", "");
		os << endl << endl;
	}
}

void operation_section::emit_C_exp(ostream& os) const
{
	if (trans && trans->clauses) {

		os << "  switch(id) {" << endl;

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {
			
			if ((*trans_it)->exps) {
				os << "    case " << MACH_CLASS(machine->get_name()) << "::"
					<< LOCAL_IDENT((*trans_it)->edge) << ":" << endl << endl;

				emit_C_exp_element(os, *trans_it);
				os << "      break;" << endl << endl;
			}
		}

		os << "    default: break;" << endl;

		os << "  }" << endl;
	}
}

void operation_section::emit_C(ostream& os) const
{
	os << endl << endl;


	os << OPER_CLASS(name) << "::" << OPER_CLASS(name)
		<< "() : _BASE_OPERATION (" << std::dec << id << ", "
	   	<< machine->get_id() << ") {" << endl << endl;

	os << "}" << endl << endl;

	if (syntx) {
		os << "_STR_T " << OPER_CLASS(name) << "::__get_syntax() {" << endl;
		syntx->emit_C(os);
		os << "}" << endl;
	}

	// initialization function
	os << "void " << OPER_CLASS(name) << "::"
			"_OPER_INIT(_BASE_MACHINE *_mach, " 
		<< OPER_ID_TYPE << " *_dtab, void *_cdng) {" << endl << endl;

	emit_C_init(os);

	os << "}" << endl << endl;

	/* emitting the osm testing code */
	os << "bool " << OPER_CLASS(name) << "::" << "__test_osm(unsigned int id) {"
		<< endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	emit_C_osm_test(os);
	os << "#endif" << endl;

	os << "  return true;" << endl;
	os << "}" << endl << endl;

	// emit the osm function
	os << "void " << OPER_CLASS(name) << "::"
		<< "__run_osm_pre(unsigned int id) {" << endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	emit_C_osm_pre(os);
	os << "#endif" << endl;

	os << "}" << endl << endl;

	// emit the osm function
	os << "void " << OPER_CLASS(name) << "::"
		<< "__run_osm_post(unsigned int id) {" << endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	emit_C_osm_post(os);
	os << "#endif" << endl;

	os << "}" << endl << endl;

	// emit the exp evalation function
	os << "void " << OPER_CLASS(name) << "::" << "__run_exp(unsigned int id) {"
		<< endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	emit_C_exp(os);
	os << "#endif" << endl;

	os << "}" << endl << endl;

	/* emitting the run_all method */
	os << "bool " << OPER_CLASS(name) << "::" << "__run_all(unsigned int id) {"
		<< endl << endl;

	os << "#ifndef _GEN_DISASSEMBLER" << endl;
	if (trans && trans->clauses) {

		os << "  switch(id) {" << endl << endl;

		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
				trans_it!=trans->clauses->end(); trans_it++) {
			
			if ((*trans_it)->osm_exps || (*trans_it)->exps) {

				os << "    case " << MACH_CLASS(machine->get_name()) << "::"
					<< LOCAL_IDENT((*trans_it)->edge) << ":" << endl << endl;

				if ((*trans_it)->osm_exps) {
					emit_C_osm_test_element(os, *trans_it);
					emit_C_osm_pre_element(os, *trans_it);
				}

				if ((*trans_it)->exps)
					emit_C_exp_element(os, *trans_it);

				if ((*trans_it)->osm_exps)
					emit_C_osm_post_element(os, *trans_it);

				os << "      break;" << endl << endl;
			}
		}

		os << "    default: break;" << endl;
		os << "  }" << endl;
	}
	os << "#endif" << endl;

	os << "  return true;" << endl;
	os << "}" << endl << endl;

}

void operation_section::emit_decode_patterns(const char *prefix)
{

	map<Symbol, DecodeEntity *>::iterator de_it;
	for (de_it=dec_ents.begin(); de_it!=dec_ents.end(); de_it++) {

		DecodeEntity *de = (*de_it).second;

		if (de->need_dec && !de->is_singleton) {

			de->emit_pat_file(prefix);
			de->emit_header_file(prefix);
			de->emit_tab_file(prefix);
			de->emit_main_file(prefix);
		}
	}
}

void function_section::emit_C(ostream& os) const
{
	os << endl << endl;

	if (dirty)
		os << DIRTY_FUN_DEC;
	else 
		os << FUN_DEC;

	os << "(" << FUN_NAME(name); 

	assert(args && args->size()>0);

	vector<ArgClause *>::iterator it = args->begin();
	for (it = args->begin(); it!=args->end(); it++) {
		os << ", ";
		(*it)->emit_C(os);
	}
	os << ")" << endl;

	os << "{" << endl;

	os << endl;

	if (var) {
		var->emit_C(os);
		os << endl;
	}

	eval->emit_C(os);

	os << "}" << endl << endl;

}

bool operation_section::check_annots()
{
	bool ret = true;

	if (check_annotations()) ret = false;
	if (var && !var->check_annots()) ret = false;
	if (trans && !trans->check_annots()) ret = false;
	if (syntx && !syntx->check_annots()) ret = false;
	if (codng && !codng->check_annots()) ret = false;
	if (eval  && !eval->check_annots()) ret = false;

	return ret;
}

bool function_section::check_annots()
{
	bool ret = true;

	if (check_annotations()) ret = false;
	if (var && !var->check_annots()) ret = false;
	if (eval  && !eval->check_annots()) ret = false;

	vector<ArgClause *>::const_iterator ait;
	for (ait=args->begin(); ait!=args->end(); ait++) {
		if (!(*ait)->check_annotations())
			ret = false;
	}
	return ret;
}

bool variable_paragraph::check_annots()
{
	bool ret = true;

	vector<VarClause *>::const_iterator vit;
	for (vit=vars->begin(); vit!=vars->end(); vit++) {
		if (!(*vit)->check_annotations())
			ret = false;
	}
	return ret;
}

bool syntax_clause::check_annots()
{
	return check_annotations();
}

bool coding_clause::check_annots()
{
	return check_annotations();
}

bool syntax_paragraph::check_annots()
{
	bool ret = true;

	vector<SyntaxClause *>::const_iterator vit;
	for (vit=elements->begin(); vit!=elements->end(); vit++) {
		if (!(*vit)->check_annotations())
			ret = false;
	}
	return ret;
}

bool coding_paragraph::check_annots()
{
	bool ret = true;

	vector<CodingClause *>::const_iterator vit;
	for (vit=elements->begin(); vit!=elements->end(); vit++) {
		if (!(*vit)->check_annotations())
			ret = false;
	}
	return ret;
}

bool trans_paragraph::check_annots()
{
	bool ret = true;

	vector<TransClause *>::const_iterator tit;
	for (tit=clauses->begin(); tit!=clauses->end(); tit++) {
		if (!(*tit)->check_annots())
			ret = false;
	}
	return ret;
}

bool eval_paragraph::check_annots()
{
	bool ret = true;

	vector<Expression *>::const_iterator eit;
	for (eit=clauses->begin(); eit!=clauses->end(); eit++) {
		if (!(*eit)->check_annotations())
			ret = false;
	}
	return ret;
}

trans_clause::~trans_clause() {
	delete_object_list(osm_exps);
	delete_object_list(exps);
}

bool trans_clause::check_annots()
{
	bool ret = check_annotations();

	if (exps) {
		vector<Expression *>::const_iterator eit;
		for (eit=exps->begin(); eit!=exps->end(); eit++) {
			if (!(*eit)->check_annotations())
				ret = false;
		}
	}

	if (osm_exps) {
		vector<OSMExpression *>::const_iterator oit;
		for (oit=osm_exps->begin(); oit!=osm_exps->end(); oit++) {
			if (!(*oit)->check_annotations())
				ret = false;
		}
	}
	return ret;
}

bool operation_section::check_causal(Symbol init_edge)
{
	/* if initialization occurs very first */
	if (init_edge==SymNULL) return true;

	if (trans && trans->clauses) {
		vector<TransClause *>::const_iterator trans_it;
		for (trans_it=trans->clauses->begin();
			trans_it!=trans->clauses->end(); trans_it++) {

			Symbol edge = (*trans_it)->get_edge();
			if (init_edge==edge || !machine->edge_reachable(init_edge, edge)) {
				return false;
			}
		}
	}
	return true;
}

const DecodeEntity *operation_section::get_decode_entity(Symbol vname) const
{
	//std::cerr << symbol_get_string(this->get_name()) << ':'  << symbol_get_string(vname) << std::endl;
	if (dec_ents.find(vname)==dec_ents.end()) return NULL;
	return (*dec_ents.find(vname)).second;
}

