
#ifndef __MAD_OPERATION_AST_HPP__
#define __MAD_OPERATION_AST_HPP__

#include "program_ast.hpp"
#include "bin_pattern.hpp"
#include <map>

using mad::Symbol;
using mad::BinPattern;

/** Name space for abstract syntax tree stuffs.
 *  The ast nodes are classified into three levels:
 *	1. section
 *	   Includes operation, function, diagram, manager, etc. These are
 *	   first level keywords in a mad file.
 *
 *	2. paragraph
 *	   Includes var, trans, eval, buffer, etc. Theses are second level
 *	   keywords in a mad file.
 *
 *	3. clause
 *	   Clauses form the contents of paragraphs.
 */
namespace mad_ast {

	/** Enumerate of operation paragraph types */
	typedef enum {
		OPER_PARA_VAR,		/**< VAR paragraph */
		OPER_PARA_TRANS,	/**< TRANS paragraph */
		OPER_PARA_SYNTAX,	/**< SYNTAX paragraph */
		OPER_PARA_CODING,	/**< CODING paragraph */
		OPER_PARA_EVAL		/**< EVAL paragraph */
	} EnumOperPara;

	/* forward declarations */
	class sym_context;
	class edge_context;
	class buf_context;
	class data_type;
	class machine_section;
	class manager_context;
	class expression;
	class operation_expression;
	class osm_expression;

	class operation_section;

	/** Class for operation pattern. */
	typedef class decode_pattern {

		public:

			/** Constructor. */
			decode_pattern(char *s) :
				bin(s), ids(), hflag(false) {}

			/** Constructor. */
			decode_pattern(const BinPattern& d) :
				bin(d), ids(), hflag(false) {}

			/** Constructor. */
			decode_pattern(const BinPattern& d, const std::vector<OperID>& ids):
				bin(d), ids(ids), hflag(false) {}

			/** Destructor. */
			~decode_pattern() {}

			/** Get the length of the pattern */
			unsigned get_length() const {return bin.get_length();}

			/** Get the pattern */
			const BinPattern& get_pattern() const {return bin;}

			/** Get the IDS */
			const std::vector<OperID>& get_ids() const {return ids;}

			/** Conflict, check if A contains B.
			 *  If ids is the same, return false*/
			bool contain(const decode_pattern& pat) const;

			/** Conflict, check if A is in conflict with B.
			 *  If the patterns overlap but ids mismatch, contain is fine */
			bool conflict(const decode_pattern& pat) const;

			/** Augment the ids. */
			void append_ids(const std::vector<OperID>& nids);

			/** Augment the ids. */
			void prepend_id(const OperID& nid);

			/** Padding ids to length.
			 *  @param len The length of the padded ids.
			 */
			void pad_ids_with_zeros(unsigned len);

			/** Print out the pattern. */
			void print(std::ostream& os) const;

			/** Print out the pattern. */
			void print_names(std::ostream& os) const;

			/** Set the hflag as given value. */
			void set_hflag(bool nf) {hflag = nf;}

			/** Get the hflag value. */
			bool get_hflag() const {return hflag;}

		private:
			BinPattern bin;
			std::vector<OperID> ids;
			bool hflag; //hide flag, used to hide duplicate patterns

	} DecodePattern;

	/** Concatenate two patterns. */
	const DecodePattern operator + (const DecodePattern& x,
									const DecodePattern& y);

	/** To sort the patterns so that identical ones are grouped together,
	 *  making it easy to set the hflags */
	bool DecPatPtrCmp(const DecodePattern *x, const DecodePattern *y);

	/** Check the consistency of a list of sorted patterns, 
	 *  set hflag for duplicate ones.
	 *  @param pats The array of patterns.
	 *  @param loc  Location requesting the check, of error reporting only.
	 */
	bool check_decode_patterns(std::vector<DecodePattern *> *pats,
								const Location *loc);

	/* foward declarations */
	class coding_clause_var_ref; 
	class variable_clause;

	/** Decoding information, for each oper decoding/activating expression. */
	typedef struct decode_info {

		/** Constructor. 
		 *  @param edge  The decoding/actvating edge.
		 *  @param ind   The order index of the decoding expression on edge.
		 *  @param exp   The expression itself.
		 */
		decode_info(Symbol edge, unsigned ind,
			const operation_expression *exp) :
			edge(edge), index(1,ind), exp(exp) {}

		/** Test if the decoding/activating has argument. */
		bool has_arg() const;

		/** Test if this is an activate or decode. */
		bool is_activate() const;

		Symbol edge;                 // the edge when this takes place
		std::vector<unsigned> index; // the order of the action on the edge
		const operation_expression *exp; // the expression

	} DecodeInfo;


	/** A decoder entity is associated with each oper variable. */
	typedef class decode_entity {

		public:

			/** Constructor.
			 *  @param oper The operation owning the entity.
			 *  @param var  The oper-typed variable for the entity.
			 *  @param oper_context The operation context.
			 *  @param result The construction result, as a return value.
			 */
			decode_entity(const operation_section *oper,
				const variable_clause *var,
				const OperContext *oper_context,
				bool& result);

			/** Destructor. */
			~decode_entity();

			/** Type check the decode entity and all its references. */
			bool type_check();

			/** Get the decoding patterns.
			 *  Useful only after type_check is called.
			 */
			const std::vector<DecodePattern *> *get_patterns() const;

			/** Need decoder.
			 *  Those variables referred by a explicit decoding
			 *  expression (with argument) needs decoder.
			 */
			bool need_decoder() const {return need_dec;}

			/** Is decoded, not activated.
			 *  Callable after type check.
			 */
			bool is_decoded() const {return decoded;}

			/** Is activated, not decoded.
			 *  Callable after type check.
			 */
			bool is_activated() const {return !decoded;}

			/** Is singlegon, no need to decode.
			 *  Callable after type check.
			 */
			bool is_single() const {return is_singleton;}

			/** Get the variable clause corresponding to this. */
			const variable_clause *get_variable() const {return var;}

			/** Get the decode infos. */
			const std::vector<DecodeInfo *>& get_infos() const {return infos;}

			/** Test to see if this has default operation. */
			bool has_default() const {return default_name!=SymNULL;}

			/** Get the default operation's name. */
			Symbol get_default_name() const {return default_name;}

			/** Get the default operation's id. */
			Symbol get_default_id() const {return default_id;}

			/** Get the coding clauses of the variable.
			 *  Callable only after type check.
			 */
			const coding_clause_var_ref *get_coding_clause() const {return cc;}

			/** Get the name of the decoder. */
			Symbol get_decoder_name() const {return decoder_name;}

			/** Emit pattern definition file. */
			void emit_pat_file(const char *prefix) const;

			/** Emit decoder header file for optimized osm. */
			void emit_header_file(const char *prefix) const;

			/** Emit decoder header file. */
			void emit_opt_header_file(const char *prefix) const;

			/** Emit decoding table and wrapper file.
			 */
			void emit_tab_file(const char *prefix) const;

			/** Emit decoding table and wrapper file for optimzed osm.
			 */
			void emit_opt_tab_file(const char *prefix) const;

			/** Emit the main decoder cpp file */
			void emit_main_file(const char *prefix) const;

		private:

			const operation_section *oper;	// operation owning this
			const variable_clause *var;		// the oper variable.

			/* if the variable appears in coding paragraph, then cc points
			   to that coding clause. Otherwise a new one will be allocated
			   and has_cc will be set. */
			coding_clause_var_ref *cc;	// the coding clause

			bool has_cc;				// owner of the coding clause
			bool need_dec;				// needs decoder 
			bool is_singleton;			// only one operation without coding
			bool is_heterogeneous;		// heterogeneous machine
			bool decoded;

			MachID machine_id;			// useful only if heterogeneous==false
			OperID singleton_id;		// the singleton operation
										// useful if is_singleton==true

			Symbol default_name;		// the default operation 
			OperID default_id;			// the default operation's id
															  
			Symbol decoder_name;		// name of the decoder

			unsigned index;				//for sorting use

			std::vector<DecodeInfo *> infos;
			const operation_section *default_oper;

		friend class operation_section;					   
															  
	} DecodeEntity;

	/** Base class for paragraphs in OPERATION section. */
	typedef class operation_paragraph : public ast_node {

		public:
			/** Constructor.
			 *  @param t    Category of the paragraph.
			 *  @param loc  Location of the node in source code
			 */
			operation_paragraph(EnumOperPara t, const Location *loc) :
				ast_node(loc), tp(t) {}

			/** Destructor. */
			virtual ~operation_paragraph() {}

			/** Get the type of the paragraph */
			EnumOperPara get_type() const {return tp;}

		private:
			EnumOperPara tp;

	} OperationParagraph;


	/** Enumerate of clause types */
	typedef enum {
		EC_PATTERN,		/**< pattern encoding type */
		EC_VAR_REF,		/**< variable reference encoding type */
		EC_OR_PATTERN,	/**< ORed pattern encoding types */
	} EnumCodingClause;

	/** Base class of coding clause declaration.
	 *
	 *  An coding clause can be either a
	 *  1. pattern string {0,1,-}^n
	 *  2. variable reference w/o field extraction
	 *  3. table lookup
	 *  4. ORed results of 1
	 *  5. ORed results of 3
	 *
	 *  Limitation of current implementation: 
	 *  1.  3 and 5 not implemeneted yet,
	 *  2.  Only allows for OR of basic elements
	 */
	typedef class coding_clause : public ast_node, public annot_owner {

		public:

			/** Constructor.
			 *  @param tp   Category of the clause.
			 *  @param pats Array of decoding patterns.
			 *  @param loc  Location of the clause.
			 */
			coding_clause(EnumCodingClause tp, 
		   		std::vector<DecodePattern *> *pats,
				const Location *loc) :
				ast_node(loc), tp(tp), width(0), pats(pats), offset(0) {}

			/** Destructor. */
			virtual ~coding_clause();

			/** Get the type of the clause. */
			EnumCodingClause get_type() const {return tp;}

			/** Get the width of the clause */
			unsigned get_width() const {return width;}

			/** Get the offset of the clause in overall coding. */
			unsigned get_offset() const {return offset;}

			/** Set the offset in overall coding. 
			 *  @param ofs The offset.
			 */
			void set_offset(unsigned ofs) {offset = ofs;}

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			virtual bool type_check(
					const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt) = 0;

			/** get patterns */
			const std::vector<DecodePattern *> *get_patterns() const
				{return pats;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			EnumCodingClause tp;

		protected:
			unsigned width;	// set by derived classes
			std::vector<DecodePattern *> *pats;
			unsigned offset;

	} CodingClause;

	/* The scanner ensures that the pattern contains only 0, 1, - */

	/** Class for pattern clause declaration, 100--01 etc. */
	typedef class coding_clause_pattern : public coding_clause {

		public:

			/** Constructor.
			 *  @param s   The pattern string
			 *  @param loc Location of the clause.
			 */
			coding_clause_pattern(DecodePattern *s, const Location *loc) :
				coding_clause(EC_PATTERN,
					new std::vector<DecodePattern *>(1,s), loc) {
				width = s->get_length();
			}

			/** Destructor. */
			~coding_clause_pattern() {}

			/** Print out the pattern. */
			std::ostream& print(std::ostream &) const;

			/** type check, do nothing for now.
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(
					const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt) {return true;}

		private:

	} CodingClausePattern;

	/** Class for variable/operation reference coding clause, ident */
	typedef class coding_clause_var_ref : public coding_clause {

		public:

			/** Constructor.
			 *  @param s   The name of the variable.
			 *  @param loc The location of the clause.
			 */
			coding_clause_var_ref(Symbol s, const Location *loc) :
				coding_clause(EC_VAR_REF,
					new std::vector<DecodePattern *>, loc),
				var_ref(s), dtype(NULL),
		   		is_oper(false) {}

			/** Destructor. */
			~coding_clause_var_ref() {delete dtype;}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** type check.
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(
					const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** check the width of the patterns and compute the patterns.
			 *  Must be called after type_check.
			 *  @param oper_context Operation context.
			 *  @param var_context  Variable context.
			 *  @param man_context  Manager context.
			 *  @param def_context  Define context.
			 *  @param func_context Function context.
			 *  @return True if successful.
			 */
			bool check_patterns(
				const OperContext *oper_context,
				const sym_context *var_context,
				const manager_context *man_context,
				const sym_context *def_context,
				const FunContext *func_context);

			/** Get the name of the variable. */
			Symbol get_name() const {return var_ref;}

			/** Get the data type of the variable. */
			const data_type *get_data_type() const {return dtype;}

			/** Get the variable type.*/
			EnumVarType get_var_type() const {return vtype;}

			/** If it is an operation reference. */
			bool is_operation() const {return is_oper;}

			/** return the depth of the coding, useful if is_oper is true.*/
			unsigned get_depth() const {return is_oper?oper_depth:0;}

			/** Get the offset of the clause in decoding table. */
			unsigned get_table_offset() const {return tab_offset;}

			/** Set the offset of the clause in decoding table. */
			void set_table_offset(unsigned i) {tab_offset = i;}

			/** Get the order index. */
			unsigned get_index() const {return ind;}

			/** Set the index of the clause, for sorting purpose. */
			void set_index(unsigned i) {ind = i;}

		private:

			Symbol var_ref;
			EnumVarType vtype;	// type of the variable, set by type_check
			data_type *dtype;	// set by type_check

			bool is_oper;		// set if the symbol refers to an oper
			unsigned oper_depth;	// depth of decoding table usage.
			unsigned tab_offset;	// offset in parent's decoding table

			unsigned ind;	//index for sorting
			
	} CodingClauseVarRef;


	/** Class for OR pattern coding clause declaration, (100-- | 00011). */
	typedef class coding_clause_or_patterns : public coding_clause {

		public:
			/** Constructor.
			 *  @param s   Array of patterns.
			 *  @param loc Location of the clause.
			 */
			coding_clause_or_patterns(std::vector<DecodePattern *> *s,
				const Location *loc) :
				coding_clause(EC_OR_PATTERN, s, loc) {}

			/** Destructor. */
			~coding_clause_or_patterns() {}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** type check.
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(
					const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

		private:

	} CodingClauseOrPatterns;


	/** Class of coding paragraph declaration. */
	typedef class coding_paragraph : public operation_paragraph {

		public:

			/** Constructor.
			 *  @param e   Array of coding clauses.
			 *  @param loc Location of the paragraph.
			 */
			coding_paragraph(std::vector<CodingClause *> *e,
				const Location *loc) :
				operation_paragraph(OPER_PARA_CODING, loc),
		   		elements(e), 
			   	patterns(new std::vector<DecodePattern *>),
			   	pat_checked(false) {}

			/** Destructor. */
			~coding_paragraph();

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** type check.
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(
					const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** check the patterns.
			 *  @param oper_context Operation context.
			 *  @param var_context  Variable context.
			 *  @param def_context  Global constant variable context.
			 *  @param man_context  Manager context.
			 *  @param def_context  Define context.
			 *  @param func_context Function context.
			 *  @return True if successful.
			 */
			bool check_patterns(
				const OperContext *oper_context,
				const sym_context *var_context,
				const manager_context *man_context,
				const sym_context *def_context,
				const FunContext *func_context);

			/** Get the pattterns */
			const std::vector<DecodePattern *> *get_patterns() const 
				{return patterns;}

			/** Get the pattern width.
			 *  Should be called only after check_patterns called. */
			unsigned get_width() const {return coding_width;}

			/** Get coding depth of the coding.
			 *  This is actually the number of decoding table entries
			 *  that this operation will consume.
			 */
			int get_depth() const {return coding_depth;}

			/** Get the coding clauses */
			const std::vector<CodingClause *> *get_clauses() const {
				return elements;
			}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			std::vector<CodingClause *> *elements;
			std::vector<DecodePattern *> *patterns;

			bool pat_checked;   // check_patterns called
			bool check_result;  // result of check_patterns

			unsigned coding_width;
			int coding_depth;

		friend class operation_section;

	} CodingParagraph;


	/** Class for table lookup declaration in SYNTAX, id[id] */
	typedef class syntax_table_lookup : public ast_node {

		public:
			/** Constructor.
			 *  @param s1  Name of the table
			 *  @param s2  Name of the index variable.
			 *  @param mod Modifier if any.
			 *  @param loc Location of the declaration.
			 */
			syntax_table_lookup(Symbol s1, Symbol s2,
				Symbol mod, const Location *loc) :
				ast_node(loc), tab(s1), ind(s2), mod(mod),
				t_type(NULL), d_type(NULL) {}

			/** Destructor. */
			~syntax_table_lookup() {delete t_type; delete d_type;}

			/** Print out the table reference */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** Get the name of the table */
			Symbol get_table_name() const {return tab;}

			/** Get the name of the index variable. */
			Symbol get_index_name() const {return ind;}

			/** Get the type of the table */
			const data_type *get_table_data_type() const {return t_type;}

			/** Get the type of the table */
			const data_type *get_index_data_type() const {return d_type;}

			/** Get the var type of the variable. */
			EnumVarType get_index_var_type() const {return vtype;}

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:

			Symbol tab, ind, mod;
			data_type *t_type;	// this should be def_var.
			data_type *d_type;

			EnumVarType vtype;

	} SyntaxTableLookup;

	typedef enum {
		ES_STRING,		/**< String syntax */
		ES_VAR_REF,		/**< Variable reference syntax */
		ES_TAB_REF,		/**< Table reference syntax */
		ES_OR_TAB_REF,	/**< ORed variable reference syntax */
		ES_OR_STRING,	/**< ORed table reference syntax */
	} EnumSyntaxClause;

	/** Base class of syntax clause declaration.
	 *  A syntax clause can be either a
	 *  1. string literal
	 *  2. variable reference
	 *  3. table lookup
	 *  4. ORed results of 1
	 *  5. ORed results of 3
	 *
	 *  Limitation:
	 *  Only allows for OR of basic elements
	 */
	typedef class syntax_clause : public ast_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param tp  Syntax-category of the syntax clause.
			 *  @param loc Location of the syntax clause.
			 */
			syntax_clause(EnumSyntaxClause tp, const Location *loc) :
				ast_node(loc), tp(tp), follow(false), first(false) {}

			/** Destructor. */
			virtual ~syntax_clause() {};

			/** Get the type of the clause. */
			EnumSyntaxClause get_type() const {return tp;}

			/** Set the first attribute . */
			void set_first(bool f) {first = f;}

			/** Set the follow attribute to f. */
			void set_follow(bool f) {follow = f;}

			/** Get the follow attribute. */
			bool get_follow() const {return follow;}

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			virtual bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt) = 0;

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			virtual void emit_C(std::ostream& os) const = 0;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			EnumSyntaxClause tp;

		protected:
			bool follow; // the element follows the preceding immediately
			bool first;  // the first element

	} SyntaxClause;


	/** Class of string literal syntax clause declaration. */
	typedef class syntax_clause_literal : public syntax_clause {

		public:

			/** Constructor.
			 *  @param s   The string literal.
			 *  @param loc Location of the string.
			 */
			syntax_clause_literal(std::string *s, const Location *loc) :
				syntax_clause(ES_STRING, loc), literal(s) {}

			/** Destructor. */
			~syntax_clause_literal() {delete literal;}

			/** Print out the string. */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt) {return true;}

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:
			std::string *literal;

	} SyntaxClauseLiteral;


	/** Class of variable/operation reference syntax clause declaration. */
	typedef class syntax_clause_var_ref : public syntax_clause {

		public:
			/** Constructor.
			 *  @param s   Name of the variable.
			 *  @param m   Modifier for output, eg dec, hex.
			 *  @param loc Location of the variale.
			 */
			syntax_clause_var_ref(Symbol s, Symbol m,
				const Location *loc) :
				syntax_clause(ES_VAR_REF, loc), var_ref(s),
		   		mod(m), dtype(NULL) {}

			/** Destructor. */
			~syntax_clause_var_ref() {delete dtype;}

			/** Print out the clause */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** Get the name of the variable. */
			Symbol get_name() const {return var_ref;}

			/** Get the modifier. */
			Symbol get_mod() const {return mod;}

			/** Get the data type of the variable. */
			const data_type *get_data_type() const {return dtype;}

			/** Get the variable type.*/
			EnumVarType get_var_type() const {return vtype;}

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:
			Symbol var_ref;
			Symbol mod;

			data_type *dtype;
			EnumVarType vtype;

	} SyntaxClauseVarRef;


	/** Class of table reference syntax clause declaration. */
	typedef class syntax_clause_tab_ref : public syntax_clause {

		public:
			/** Constructor.
			 *  @param s   The table lookup object.
			 *  @param loc Location of the table lookup object.
			 *  @sa See also syntax_table_lookup.
			 */
			syntax_clause_tab_ref(syntax_table_lookup *s, const Location *loc):
				syntax_clause(ES_TAB_REF, loc), tab_ref(s) {}

			/** Destructor. */
			~syntax_clause_tab_ref() {delete tab_ref;}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:
			syntax_table_lookup *tab_ref;
		
	} SyntaxClauseTabRef;


	/** Class of ORed string literal syntax clause declaration. */
	typedef class syntax_clause_or_literals : public syntax_clause {

		public:
			/** Constructor.
			 *  @param s   Array of string literals.
			 *  @param loc Location of the literals.
			 */
			syntax_clause_or_literals(std::vector<std::string *> *s,
				const Location *loc):
				syntax_clause(ES_OR_STRING, loc), literals(s) {}

			/** Destructor. */
			~syntax_clause_or_literals();

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt) {return true;}

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:
			std::vector<std::string *> *literals;

	} SyntaxClauseOrLiterals;


	/** Class of ORed table reference syntax clause declaration. */
	typedef class syntax_clause_or_tab_refs : public syntax_clause {

		public:
			/** Constructor.
			 *  @param s   Array of syntax_table_lookup objects.
			 *  @param loc Location of the objects.
			 *  @sa See also syntax_table_lookup.
			 */
			syntax_clause_or_tab_refs(std::vector<syntax_table_lookup *> *s,
				const Location *loc) :
				syntax_clause(ES_OR_TAB_REF, loc), trs(s) {}

			/** Destructor. */
			~syntax_clause_or_tab_refs();

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;


		private:
			std::vector<syntax_table_lookup *> *trs;

	} SyntaxClauseOrTabRefs;


	/** Class of syntax paragraph declaration. */
	typedef class syntax_paragraph : public operation_paragraph {

		public:
			/** Constructor.
			 *  @param e   Array of syntax clauses.
			 *  @param loc Location of the paragraph.
			 *  @sa See also syntax_clause, operation_paragraph.
			 */
			syntax_paragraph(std::vector<syntax_clause *> *e,
				const Location *loc) :
				operation_paragraph(OPER_PARA_SYNTAX, loc),  elements(e) {}

			/** Destructor */
			~syntax_paragraph();

			/* Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** type check. 
			 *  @param var_ctxt Local variable context.
			 *  @param mvr_ctxt Machine variable context.
			 *  @param def_ctxt Define context.
			 */
			bool type_check(const sym_context *var_ctxt,
					const sym_context *mvr_ctxt,
				   	const sym_context *def_ctxt);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			std::vector<syntax_clause *> *elements;

	} SyntaxParagraph;


	/** Class of OPERATION or FUNCTION argument declaration. */
	typedef class arg_clause : public named_typed_node, public annot_owner {

		public:
			/** Constructor.
			 *  @param name  Name of the variable.
			 *  @param type  Data type of the variable.
			 *  @param r	 Read only flag, true only for arguments w/o *.
			 *  @param loc   Location of the variable.
			 */
			arg_clause(Symbol name, data_type *type,
				const bool &r, const Location *loc) :
				named_typed_node(name, type, loc) {
				type->set_readonly(r);
			}

			/** Destructor. */
			~arg_clause() {}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

	} ArgClause;

	/** Class of a variable clause node in VAR paragraph. */
	typedef class variable_clause : public named_typed_node,
   									public annot_owner {
	
		public:
			/** Constructor.
			 *  @param name  Name of the variable.
			 *  @param type  Data type of the variable.
			 *  @param loc   Location of the variable.
			 */
			variable_clause(Symbol name, data_type *type,
				const Location *loc) :
				named_typed_node(name, type, loc) {}

			/** Destructor. */
			~variable_clause() {}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

		private:

			friend class operation_section;

	} VarClause;

	/** Class of variable paragraph declaration. */
	typedef class variable_paragraph : public operation_paragraph {

		public:
			/** Constructor.
			 *  @param vl   Array of var clauses
			 *  @param loc  Location of the paragraph
			 */
			variable_paragraph(std::vector<VarClause *> *vl,
				const Location *loc) :
				operation_paragraph(OPER_PARA_VAR, loc), vars(vl) {};

			/** Destructor. */
			~variable_paragraph();

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** Type check the paragraph
			 *  @param is_oper If in the context of OPERATION or FUNCTION.
			 *  @param context Context of identifiers.
			 */
			bool type_check(bool is_oper, sym_context *context);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

			/** Get the clauses. 
			 */
			const std::vector<VarClause *> *get_clauses() const {return vars;}

		private:
			std::vector<VarClause *> *vars;

			/* variable -> type, loc */
			sym_context var_cache;

			friend class operation_section;
			friend class function_section;
			friend class oper_graph_and_node;
			friend class oper_graph_or_node;
			friend class oper_tree;

	} VarParagraph;

	/** Class of trans clause declaration. */
	typedef class trans_clause: public ast_node, public annot_owner {

		public: 
			/** Constructor.
			 *  @param sym  Name of the edge
			 *  @param olst Array of OSM expressions
			 *  @param elst Array of expressions
			 *  @param loc  Location of the clause
			 */
			trans_clause(Symbol sym,
				std::vector<osm_expression *> *olst,
				std::vector<expression *> *elst,
				const Location *loc) :
				ast_node(loc), edge(sym), osm_exps(olst), exps(elst) {}

			/** Destructor */
			~trans_clause();

			/** Get the name of the edge. */
			Symbol get_edge() const {return edge;}

			/** append an expression, a parsing hack */
			void append_exp(expression *exp);

			/** Get the osm expressions. */
			const std::vector<osm_expression *> *get_osm_exps() const {
				return osm_exps;
			}

			/** Get the osm expressions. */
			const std::vector<expression *> *get_exps() const {
				return exps;
			}

			/** Print out the clause. */
			std::ostream& print(std::ostream &) const;

			/** Type check the OSM clause.
			 *  @param  man_context  Manager definitions.
			 *  @param  edge_context Edge definitions.
			 *  @param  buf_context  Buffere definitions.
			 *  @param  var_context  Local variable definitions
			 *  @param  mvar_context Machine variable definitions.
			 *  @param  def_context  Define definitions
			 *  @param  oper_context  Operation definitions.
			 *  @param  fun_context  Function definitions.
			 */
			bool type_check(const manager_context *man_context,
					const edge_context *edge_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvar_context,
					const sym_context *def_context,
					const OperContext *oper_context,
					const FunContext *fun_context);

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			Symbol edge;
			std::vector<osm_expression *> *osm_exps;
			std::vector<expression *> *exps;

			friend class trans_paragraph;
			friend class operation_section;

	} TransClause;

	/** Class for paragraph TRANS. */
	typedef class trans_paragraph : public operation_paragraph {

		public:
			/** Constructor.
			 *  @param cl   Array of trans clauses
			 *  @param loc  Location of the paragraph
			 */
			trans_paragraph(std::vector<TransClause *> *cl,
				const Location *loc) :
				operation_paragraph(OPER_PARA_TRANS, loc), clauses(cl) {}

			/** Destructor. */
			~trans_paragraph();

			/** Get the transition clauses. */
			const std::vector<TransClause *> *get_clauses() const {
				return clauses;
			}

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** Type check the Trans paragraph.
			 *  @param  man_context  Manager definitions.
			 *  @param  edge_context Edge definitions.
			 *  @param  buf_context  Buffere definitions.
			 *  @param  var_context  Local variable definitions
			 *  @param  mvar_context Machine variable definitions.
			 *  @param  def_context  Define definitions
			 *  @param  oper_context Operation definitions.
			 *  @param  fun_context  Function definitions.
			 */
			bool type_check(const manager_context *man_context,
					const edge_context *edge_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvar_context,
					const sym_context *def_context,
					const OperContext *oper_context,
					const FunContext *fun_context);

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:

			std::vector<TransClause *> *clauses;

			friend class operation_section;

	} TransParagraph;


	/** Class of eval paragraph declaration. */
	typedef class eval_paragraph : public operation_paragraph {

		public:
			/** Constructor.
			 *  @param a   Array of expressions
			 *  @param loc Location of the paragraph
			 */
			eval_paragraph(std::vector<expression *> *a,
				const Location *loc) :
				operation_paragraph(OPER_PARA_EVAL, loc), clauses(a) {}

			/** Destructor */
			~eval_paragraph();

			/** Print out the paragraph. */
			std::ostream& print(std::ostream &) const;

			/** Type check.
			 *  @param var_context  Local variable definitions
			 *  @param arg_context  Argument definitions, for functions only.
			 *  @param mvr_context  Machine variable definitions.
			 *  @param var_context  Argument definitions
			 *  @param def_context  Define definitions
			 *  @param otxt Operation context.
			 *  @param ftxt Function context.
			 */
			bool type_check(const sym_context *var_context,
				const sym_context *arg_context,
				const sym_context *mvr_context,
				const sym_context *def_context,
				const OperContext *otxt, const FunContext *ftxt);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

			/** Get the clauses.
			 *  @return True if result ok.
			 */
			const std::vector<expression *> *get_clauses() const {
				return clauses;
			}

		private:

			std::vector<expression *> *clauses;

			friend class operation_section;
			friend class function_section;

	} EvalParagraph;

	/** Class of operation section declaration */
	typedef class operation_section : public prog_section,
   										public annot_owner {

		public:
			/** Constructor.
			 *  @param m   Name of the machine
			 */
			operation_section(Symbol m) :
				prog_section(SEC_OPERATION, NULL), machname(m), paras(NULL),
				classified(false), var(NULL), trans(NULL), syntx(NULL),
				codng(NULL), eval(NULL), type_checked(false), vcoding(NULL) {}

			/** Destructor. */
			~operation_section();

			/** Set the name of the operation. */
			void set_name(Symbol n) {name = n;}

			/** Set the paragraphs of a machine section. */
			void set_paragraphs(std::vector<OperationParagraph *> *ps) {
				paras = ps;
			}

			/** Print out the section. */
			std::ostream& print(std::ostream &) const;

			/** Get the name of the operation. */
			Symbol get_name() const {return name;}

			/** Get the name of the machine */
			Symbol get_mach_name() const {return machname;}

			/** Classify the paragraphs.
			 *  Paragraphs are not called in constructor since I want
			 *  to separate the syntax checking phase and the type
			 *  checking phase. The parser should not report such errors.
			 */
			bool classify_paragraphs();

			/** Set the machine node.
			 *  @param m  The machine associated with the operation.
			 */  
			void set_machine(machine_section *m) {machine = m;}

			/** Get the machine node.
			 *  @return the machine node associated with this operation.
			 */
			const machine_section *get_machine() const {return machine;}

			/** Pre-type check the operation.
			 *  This will set up the decoding entities for oper-graph
			 *  building.
			 *  @param man_context  Manager context.
			 *  @param def_context  Define context.
			 *  @param oper_context Operation context.
			 *  @param func_context Function context.
			 */
			bool pre_type_check(
				const manager_context *man_context,
				const sym_context *def_context,
				const OperContext *oper_context,
				const FunContext *func_context);

			/** Type check the operation.
			 *  @param man_context  Manager context.
			 *  @param def_context  Define context.
			 *  @param oper_context Operation context.
			 *  @param func_context Function context.
			 */
			bool type_check(
				const manager_context *man_context,
				const sym_context *def_context,
				OperContext *oper_context,
				const FunContext *func_context);

			/** Get the width of the coding.
			 *  Should be called only after check patterns. */
			unsigned get_coding_width() const {
				return (codng)?codng->get_width():0;
			}

			/** Get coding depth of the coding.
			 *  This is actually the number of decoding table entries
			 *  that this operation will consume.
			 */
			int get_coding_depth() const {
				return (codng)?codng->get_depth():0;
			}

			/** Get the patterns of an operation, callable after tye_check. */
			const std::vector<DecodePattern *> *get_patterns() const;

			/** emit C++ declaration code 
			 *  @param os The output stream.
			 */  
			void emit_C_dec(std::ostream& os) const;

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** emit C++ declaration code 
			 */  
			void emit_decode_patterns(const char *prefix);

			/** Get the id of the operation. */
			OperID get_id() const {return id;}

			/** Set the id of the operation. */
			void set_id(OperID nid) {id = nid;}

			/** Get the id of the machine. */
			MachID get_mach_id() const;

			/** See if the operation has coding field,
			 *  Callable after type_checking.
			 */
			bool has_coding() const {return codng!=NULL;}

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

			/** Check if causal relationship.
			 *  @param init_edge The edge when this oper is initialized.
			 *  @return True if correct -- all trans reachable from init_edge.
			 */
			bool check_causal(Symbol init_edge);

			/** Get the decode entity of a oper-typed variable.
			 *  @param vname The name of the variable.
			 *  @return NULL if the variable is unknown.
			 */
			const DecodeEntity *get_decode_entity(Symbol vname) const;

			/** Get all the decode entities of an operation section. */
			const std::map<Symbol, DecodeEntity *>&
				get_decode_entities() const{return dec_ents;}

			/** Get the coding paragraph, Callable after type check. */
			const CodingParagraph *get_coding() const {return codng;}

			/** Get the trans paragraph. Callable after type check. */
			const TransParagraph *get_trans() const {return trans;}

			/** Get the eval paragraph. Callable after type check. */
			const EvalParagraph *get_eval() const {return eval;}

			/** Get the variable paragraph. Callable after type check. */
			const VarParagraph *get_var() const {return var;}

			/** Test if any operation is the leaf of a decoding tree.
			 *  Callable only after pre_type_check.
			 */
			bool is_decode_leaf() const {return decode_leaf;}

		private:

			/* emit the body of the virtual methods. */
			void emit_C_init(std::ostream& os) const;

			void emit_C_osm_test_element(std::ostream&, TransClause *) const;
			void emit_C_osm_test(std::ostream& os) const;

			void emit_C_osm_pre_element(std::ostream&, TransClause *) const;
			void emit_C_osm_pre(std::ostream& os) const;

			void emit_C_exp_element(std::ostream&, TransClause *) const;
			void emit_C_exp(std::ostream& os) const;

			void emit_C_osm_post_element(std::ostream&, TransClause *) const;
			void emit_C_osm_post(std::ostream& os) const;

			/** Compare the decoding order of two entities
			 *  @return 1 if a later than b, -1 reverse, 0 tie
			 */
			int dec_entity_cmp(const DecodeEntity *a, const DecodeEntity *b);

			Symbol name, machname;
			std::vector<OperationParagraph *> *paras;

			const machine_section *machine;

			/** Classify the paragraphs */
			bool classified, classify_result;

			VarParagraph *var;
			TransParagraph *trans;
			SyntaxParagraph *syntx;
			CodingParagraph *codng;
			EvalParagraph *eval;

			/* flag indicating whether type_check is called */
			bool type_checked;
			bool check_result;

			/* cache for var_types */
			std::map<Symbol, EnumVarType> var_types;

			OperID id;

			std::map<Symbol, DecodeEntity *> dec_ents;

			VarClause *vcoding;

			/* flag indicating if this is a leaf in decoding tree,
			 * useful to test if an operation can be used as default oper. */
			bool decode_leaf;

		friend class oper_graph_and_node;
		friend class oper_graph_or_node;
		friend class oper_tree;

	} OperationSection; 


	/** Class of function declaration. 
	 *  Function is the same as operation except that it has no
	 *  osm, syntax and coding paragraphs, and that is must have arguments.
	 *  Function can only be referenced in the EVAL paragraph of a
	 *  function or an operation.
	 */
	typedef class function_section : public prog_section, public annot_owner {

		public:
			/** Constructor.
			 */
			function_section() : prog_section(SEC_FUNCTION, NULL),
				args(NULL), paras(NULL), classified(false), 
				var(NULL), eval(NULL), dirty(false) {}

			/** Destructor. */
			~function_section();

			/** Set the name of the operation. */
			void set_name(Symbol n) {name = n;}

			/** Set the paragraphs of a machine section. */
			void set_paragraphs(std::vector<OperationParagraph *> *ps) {
				paras = ps;
			}

			/** Set the arguments of a function section. */
			void set_args(std::vector<ArgClause *> *a);

			/** Get the name of the operation. */
			Symbol get_name() const {return name;}

			/** Classify the paragraphs.
			 *  Paragraphs are not called in constructor since I want
			 *  to separate the syntax checking phase and the type
			 *  checking phase. The parser should not report such errors.
			 */
			bool classify_paragraphs();

			/** Get the type of the operation. */
			const data_type *get_data_type() const {return type;}

			/** type check a function
			 *  @param def_context  Define context.
			 *  @param fnc_context Function context.
			 */
			bool type_check(const sym_context *def_context,
				const FunContext *fnc_context);

			/** emit C++ code 
			 *  @param os The output stream.
			 */  
			void emit_C(std::ostream& os) const;

			/** Check if the function is dirty. */
			bool is_dirty() const {return dirty;}

			/** Propogate the dirty flag among internal functions.
			 *  @param def_context  Define context.
			 *  @param fnc_context Function context.
			 *  @return True if dirty bit is newly set.
			 */
			bool propogate_dirty(const sym_context *def_context,
				const FunContext *fnc_context);

			/** Print out the section */
			std::ostream& print(std::ostream &) const;

			/** Check annotations.
			 *  @return True if result ok.
			 */
			bool check_annots();

		private:
			Symbol name;
			std::vector<ArgClause *> *args;
			std::vector<OperationParagraph *> *paras;

			/** The signature of the operation as a group type */
			data_type *type;
			sym_context arg_context;

			/** Classify the paragraphs */
			bool classified, classify_result;

			VarParagraph *var;
			EvalParagraph *eval;

			/** Is a dirty function.
			 *  A function is dirty if it has side effect, which
			 *  may occur if this function calls other dirty function. */
			bool dirty;

	} FunctionSection;
}

#endif
