
#include <iostream>
#include <sstream>
#include <fstream>
#include <list>
#include "symbol.hpp"
#include "opti.hpp"
#include "emit_c.hpp"
#include "exp_ast.hpp"
#include "osm_exp_ast.hpp"
#include "gcolor.hpp"

using std::set;
using std::map;
using std::pair;
using std::vector;
using std::stringstream;
using std::list;
using std::ostream;
using std::ofstream;
using std::endl;
using std::cerr;
using std::string;
using namespace mad_ast;
using namespace mad;

opti_expression::~opti_expression()
{
	delete_object_list(children);
}

void opti_expression::append_child(opti_expression *exp)
{
	if (children==NULL) children = new vector<opti_expression *>;
	children->push_back(exp);
}

opti_edge::opti_edge(Symbol base, Symbol name,
	opti_state *src, opti_state *dest) :
	base_name(base), name(name), src(src), dest(dest),
	exps(NULL), osm_exps(NULL), label(SymNULL), emit(false)
{
	src->insert_out_edge(this);
	dest->insert_in_edge(this);
}

opti_edge::~opti_edge()
{
	delete_object_list(exps);
	delete_object_list(osm_exps);
}

void opti_edge::append_exps(const vector<expression *> *base_exps,
	const operation_section *oper)
{
 	//we create this vector, empty exps still mean unconditional good edge
	if (!exps) exps = new vector<OptiExpression *>;
	if (base_exps) {
		vector<expression *>::const_iterator it;
		for (it = base_exps->begin(); it!=base_exps->end(); it++) {
			exps->push_back(new OptiExpression(*it, oper));
		}
	}
}

void opti_edge::append_exps(const vector<OptiExpression *> *oexps)
{
	if (oexps) {
		if (!exps) exps = new vector<OptiExpression *>;
		vector<OptiExpression *>::const_iterator oit;
		for (oit=oexps->begin(); oit!=oexps->end(); oit++) {
			exps->push_back(new OptiExpression((*oit)->base, (*oit)->oper));
		}
	}
}

void opti_edge::append_osm_exps(const vector<osm_expression *> *base_osm_exps,
	const operation_section *oper)
{
 	//we create this vector, empty exps still means unconditional good edge
	if (!osm_exps) osm_exps = new vector<OptiOSMExpression *>;
	if (base_osm_exps) {
		vector<osm_expression *>::const_iterator oit;
		for (oit=base_osm_exps->begin(); oit!=base_osm_exps->end(); oit++) {
			osm_exps->push_back(new OptiOSMExpression(*oit, oper));
		}
	}
}

void opti_edge::append_osm_exps(const vector<OptiOSMExpression *> *oexps)
{
	// here we only copy whatever the argument offers
	if (oexps) {
		if (!osm_exps) osm_exps = new vector<OptiOSMExpression *>;
		vector<OptiOSMExpression *>::const_iterator oit;
		for (oit=oexps->begin(); oit!=oexps->end(); oit++) {
			osm_exps->push_back(new OptiOSMExpression((*oit)->base,
									(*oit)->oper));
		}
	}
}

/* two edges are equal if their base is the same, and
 * they have the same expressions on their edges */
bool opti_edge::equal(opti_edge *o2)
{
	/* see if the bases are the same */
	if (base_name!=o2->base_name) return false;

	if (osm_exps==NULL && o2->osm_exps!=NULL) return false;
	if (osm_exps!=NULL && o2->osm_exps==NULL) return false;
	if (exps==NULL && o2->exps!=NULL) return false;
	if (exps!=NULL && o2->exps==NULL) return false;

	if (osm_exps && o2->osm_exps) {

		/* compare all the osm expressions. */
		if (osm_exps->size()!=o2->osm_exps->size()) return false;
		for (unsigned ii=0; ii<osm_exps->size(); ii++) {

			/* if they are not from the same place in the source file,
			   compare content */
			if ((*osm_exps)[ii]->base!=(*(o2->osm_exps))[ii]->base) {

				int ret =
					(*osm_exps)[ii]->base->equal((*(o2->osm_exps))[ii]->base);

				if (ret==0) return false;

				/* conditionally equal, local variables has the same name.
				   then compare operation */
				if (ret==-1 &&
					(*osm_exps)[ii]->oper!=(*(o2->osm_exps))[ii]->oper)
					return false;
			}
		}
	}

	if (exps && o2->exps) {

		/* compare the expressions. */ //@@@
		if (exps->size()!=o2->exps->size()) return false;
		for (unsigned ii=0; ii<exps->size(); ii++) {
			if ((*exps)[ii]->base != (*(o2->exps))[ii]->base) return false;
		}
	}

	return true;
}

void opti_state::remove_in_edge(OptiEdge *edge)
{
	vector<OptiEdge *>::iterator eit;
	for (eit=in_edges.begin(); eit!=in_edges.end(); eit++)
		if ((*eit)==edge) {
			in_edges.erase(eit);
			break;
		}
}

void opti_state::remove_out_edge(OptiEdge *edge)
{
	vector<OptiEdge *>::iterator eit;
	for (eit=out_edges.begin(); eit!=out_edges.end(); eit++)
		if ((*eit)==edge) {
			out_edges.erase(eit);
			break;
		}
}

/* two states are compatible if their output edges are equal and
 * they point to the same set of successors.
 */
bool opti_state::compatible(opti_state *o2)
{
	if (base_name!=o2->base_name) return false;

	/* ==0 means a head when spinning off, don't merge it */ //@@@
	if (in_edges.size()==0 || o2->in_edges.size()==0) return false;

	/* there have to be the same number of edges. */
	if (out_edges.size()!=o2->out_edges.size()) return false;

	for (unsigned ii=0; ii<out_edges.size(); ii++) {
		/* label caches the equality relationship among edges. */
		if (out_edges[ii]->label!=o2->out_edges[ii]->label)
			return false;
		/* get the destination of the edge. */
		if (out_edges[ii]->get_dest()!=o2->out_edges[ii]->get_dest())
			return false;
	}

	return true;
}

opti_machine::opti_machine(const operation_section *root,
	const OperContext *opers) :
	base_mach(root->get_machine()),
	root_oper(root), oper_ctxt(opers), name_counter(0), optimized(false)
{
	/* get the initial state. */
	initial = base_mach->get_initial_state();

	name = symbol_find_or_insert(symbol_get_string(root->get_name())
			+ '_' + symbol_get_string(base_mach->get_name()));

	const set<Symbol> *state_names = base_mach->get_states();
	const set<Symbol> *edge_names = base_mach->get_edges();

	/* first we create all the states. */
	set<Symbol>::const_iterator sym_it;
	for (sym_it=state_names->begin(); sym_it!=state_names->end(); sym_it++) {
		OptiState *nstate = new opti_state(*sym_it, *sym_it);
		states[*sym_it] = nstate;
	}

	/* then we create all the edges. */
	for (sym_it=edge_names->begin(); sym_it!=edge_names->end(); sym_it++) {

		OptiState *src = states[base_mach->get_edge(*sym_it)->get_from_state()];
		OptiState *dest = states[base_mach->get_edge(*sym_it)->get_to_state()];

		OptiEdge *nedge = new opti_edge(*sym_it, *sym_it, src, dest);
		edges[*sym_it] = nedge;
	}

	/* we also introduce a special state and edge for the EVAL expressions. */
	states[SymNULL] = new opti_state(SymNULL, SymNULL);
	edges[SymNULL] = new opti_edge(SymNULL, SymNULL,
							 states[SymNULL], states[initial]);

	/* then we insert the expressions of the root into the edges */
	if (root->get_trans()) {

		const vector<TransClause *> *cls = root->get_trans()->get_clauses();
		vector<TransClause *>::const_iterator it;
		for (it=cls->begin(); it!=cls->end(); it++) {

			Symbol edge_name = (*it)->get_edge();
			edges[edge_name]->append_exps((*it)->get_exps(), root);
			edges[edge_name]->append_osm_exps((*it)->get_osm_exps(), root);

		}
	}

	/* inser root's EVAL into the special edge. */
	if (root->get_eval()) {
		edges[SymNULL]->append_exps(root->get_eval()->get_clauses(), root);
	}
}

opti_machine::~opti_machine()
{
	map<Symbol, OptiEdge *>::const_iterator eit;
	for (eit=edges.begin(); eit!=edges.end(); eit++)
		delete (*eit).second;

	map<Symbol, OptiState *>::const_iterator sit;
	for (sit=states.begin(); sit!=states.end(); sit++)
		delete (*sit).second;

	vector<DataType *>::const_iterator cvt_it;
	for (cvt_it=coalesced_var_types.begin();
		cvt_it!=coalesced_var_types.end(); cvt_it++)
		delete *cvt_it;
}

OptiState *opti_machine::spin_off(Symbol state, const vector<OperID>& ids)
{
	/* mapping from base name to newly created states*/
	map<Symbol, OptiState *> new_states;
	map<Symbol, OptiEdge *> new_edges;

	//std::cerr << "current states : " << states.size() << std::endl;
	//std::cerr << "current edges  : " << edges.size() << std::endl;

	/* we use the work_list to traverse from the state to initial */
	list<OptiState *> work_list;
	work_list.push_back(states[state]);

	/* first we duplicate the states since the creation of new edges
	 * needs state object.
	 */
	while (!work_list.empty()) {

		OptiState *cur_state = work_list.front();
		work_list.pop_front();

		/* now we duplicate the state */
		Symbol new_sname = new_state_name(cur_state->base_name);
		OptiState *new_state = new OptiState(cur_state->base_name, new_sname);

		/* insert the new state into the table */
		assert(states.find(new_sname)==states.end());
		states[new_sname] = new_state;

		/* insert this to new state table. */
		new_states[cur_state->base_name] = new_state;

		/* then we push its successors onto the stack. */
		vector<OptiEdge *>& out_edges = cur_state->out_edges;
		vector<OptiEdge *>::const_iterator eit;
		for (eit=out_edges.begin(); eit!=out_edges.end(); eit++) {

			OptiState *dest_state = (*eit)->dest;
			/* we push the destination state into work_list if not done yet */
			if (new_states.find(dest_state->base_name)==new_states.end()) {

				/* initial state is the end */
				if (dest_state->name!=initial)
					/* push this to the queue */
					work_list.push_back(dest_state);
			}
		}
	}

	/* the set that has been processed. */
	set<Symbol> done_set;

	/* then we duplicate the edges. */
	work_list.push_back(states[state]);
	while (!work_list.empty()) {

		OptiState *cur_state = work_list.front();
		work_list.pop_front();

		/* insert this to done_set */
		done_set.insert(cur_state->base_name);

		vector<OptiEdge *>& out_edges = cur_state->out_edges;
		vector<OptiEdge *>::const_iterator eit;
		for (eit=out_edges.begin(); eit!=out_edges.end(); eit++) {

			OptiState *dest_state = (*eit)->dest;

			/* we need to get the newly generated states. */
			assert(new_states.find(cur_state->base_name)!=new_states.end());
			OptiState *new_cur = new_states[cur_state->base_name];

			OptiState *new_dest;
			if (dest_state->name==initial) {
				new_dest = states[initial];
			}
			else {
				assert(new_states.find(dest_state->base_name)!=
										new_states.end());
				new_dest = new_states[dest_state->base_name];
			}

			/* then we generate an edge */
			Symbol new_ename = new_edge_name((*eit)->base_name);
			OptiEdge *new_edge = new OptiEdge((*eit)->base_name,
				new_ename, new_cur, new_dest);

			/* insert the new edge into the table */
			assert(edges.find(new_ename)==edges.end());
			edges[new_ename] = new_edge;
			new_edges[(*eit)->base_name] = new_edge;

			/* we duplicate the expressions. */
			new_edge->append_osm_exps((*eit)->osm_exps);
			new_edge->append_exps((*eit)->exps);

			if (done_set.find(dest_state->base_name)==done_set.end()) {
				/* initial state is the end */ 
				if (dest_state->name!=initial)
					/* push this to the queue */
					work_list.push_back(dest_state);
			}
		}
	}

	/* copy the expressions and osm transactions to the edge. */
	extern map<OperID, Symbol> oper_name_table;
	vector<OperID>::const_iterator id_it;
	for (id_it=ids.begin(); id_it!=ids.end() && (*id_it)>0; id_it++) {

		Symbol tname = oper_name_table[*id_it];
		const OperationSection *toper = oper_ctxt->get_object(tname);

		if (toper->get_trans()) {

			const vector<TransClause *>
					*cls = toper->get_trans()->get_clauses();

			vector<TransClause *>::const_iterator it;
			for (it=cls->begin(); it!=cls->end(); it++) {

				Symbol edge_name = (*it)->get_edge();

				/* this should be ensured by causality checking. */
				assert(new_edges.find(edge_name)!=new_edges.end());
				new_edges[edge_name]->append_exps((*it)->get_exps(), toper);
				new_edges[edge_name]->append_osm_exps
						((*it)->get_osm_exps(), toper);
			}
		}
	}

	/* return the newly created header state. */
	return new_states[state];
}

Symbol opti_machine::new_edge_name(Symbol base)
{
	stringstream sstr;
	sstr << '$' << symbol_get_string(base) << '$' << name_counter++;
	return symbol_find_or_insert(sstr.str());
}

Symbol opti_machine::new_state_name(Symbol base)
{
	return new_edge_name(base);
}

void opti_machine::label_edges()
{
	map<Symbol, vector<OptiEdge *> > edge_buckets;

	/* first we cleanup the labels. */
	map<Symbol, OptiEdge *>::iterator iit;
	for (iit=edges.begin(); iit!=edges.end(); iit++) {
		(*iit).second->label = SymNULL;
		edge_buckets[(*iit).second->base_name].push_back((*iit).second);
	}

	map<Symbol, vector<OptiEdge *> >::iterator eit;
	for (eit=edge_buckets.begin(); eit!=edge_buckets.end(); eit++) {

		vector<OptiEdge *>::iterator it;
		for (it = (*eit).second.begin(); it!=(*eit).second.end(); it++) {

			/* the first one of a kind */
			if ((*it)->label==SymNULL) (*it)->label = (*it)->name;

			/* compare with all the rest */
			vector<OptiEdge *>::iterator jit;
			for (jit=it,jit++; jit!=(*eit).second.end(); jit++) {
				if ((*it)->equal(*jit)) {
					(*jit)->label = (*it)->label;
				}
			}
		}
	}
}

void opti_machine::consolidate()
{
	label_edges();

	message("Consolidating ...");
	list<OptiState *> work_list;
	set<OptiState *> work_set;	//duplicate of work_list, for fast inquiry

	OptiState *init = states[initial];

	/* we start from init, going backwards */
	vector<OptiEdge *>::iterator eit;
	for (eit=init->in_edges.begin(); eit!=init->in_edges.end(); eit++) {
		if (work_set.find((*eit)->get_source())==work_set.end()) {
			work_list.push_back((*eit)->get_source());
			work_set.insert((*eit)->get_source());
		}
	}

	while (!work_list.empty()) {

		list<OptiState *>::iterator wit;
		for (wit=work_list.begin(); wit!=work_list.end();) {

			list<OptiState *>::iterator vit;
			for (vit=wit, vit++; vit!=work_list.end();) {

				/* see if mergeable */
				if ((*wit)->compatible(*vit)) {

					/* insert all the sources of the two into the list. */
					// except for the initial state and the special state
					for (eit=(*wit)->in_edges.begin();
						eit!=(*wit)->in_edges.end(); eit++) {

						OptiState *src = (*eit)->get_source();
						if (work_set.find(src)==work_set.end() &&
							src->base_name!=initial &&
							src->base_name!=SymNULL) {
							work_list.push_back(src);
							work_set.insert(src);
						}
					}

					for (eit=(*vit)->in_edges.begin();
						eit!=(*vit)->in_edges.end(); eit++) {

						OptiState *src = (*eit)->get_source();
						if (work_set.find(src)==work_set.end() &&
							src->base_name!=initial &&
							src->base_name!=SymNULL) {
							work_list.push_back(src);
							work_set.insert(src);
						}
					}

					/* merge this into the other. */
					merge(*wit, *vit);

					work_set.erase(*vit);
					vit = work_list.erase(vit); //go to the next
				}
				else vit++;
			}

			work_set.erase(*wit);
			wit = work_list.erase(wit); //go to the next
		}
	}
}

bool opti_machine::remove_useless()
{
	/* we cache the ones to delete since we cannot delete on site.
	 * That may screw the iterator of map */
	vector<Symbol> useless_edges;

	message ("Remove useless ...");

	/* remove empty edges: edges with no actions. */
	map<Symbol, OptiEdge *>::iterator iit;
	for (iit=edges.begin(); iit!=edges.end(); iit++) {

		// special edge should be preserved in any case.
		if ((*iit).first==SymNULL) continue;

		if ((*iit).second->exps==NULL && (*iit).second->osm_exps==NULL) {

			/* remove it from the out_edge list of the source. */
			OptiState *src = (*iit).second->get_source();
			src->remove_out_edge((*iit).second);

			OptiState *dest = (*iit).second->get_dest();
			dest->remove_in_edge((*iit).second);

			useless_edges.push_back((*iit).first);
		}
	}

	/* now we delete these useless edges. */
	vector<Symbol>::iterator uit;
	for (uit=useless_edges.begin(); uit!=useless_edges.end(); uit++) {
		delete edges[*uit];
		edges.erase(*uit);
	}

	/* samely we collect the useless states first, then delete them. */
	vector<Symbol> useless_states;

	/* remove useless states then */
	map<Symbol, OptiState *>::iterator sit;
	for (sit=states.begin(); sit!=states.end(); sit++) {

		// special state should be preserved in any case.
		if ((*sit).first==SymNULL) continue;

		// split state should be preserved.
		if ((*sit).second->get_base_name()==split_state_name) continue;

		// state unreachable, remove it
		if ((*sit).second->in_edges.size()==0)
			useless_states.push_back((*sit).first);	
	}

	for (uit=useless_states.begin(); uit!=useless_states.end(); uit++) {
		delete states[*uit];
		states.erase(*uit);
	}

	/* check dead-end state */
	for (sit=states.begin(); sit!=states.end(); sit++) {

		// dead end state, report error
		// split state is not dead end state.
		if ((*sit).second->out_edges.size()==0 &&
			(*sit).second->get_base_name()!=split_state_name) {
			string msg = "Dead-end state found during optimization: ";	
			msg += symbol_get_string((*sit).second->get_name());
			error(msg);
			return false;
		}
	}

	return true;
}

void opti_machine::merge(OptiState *s1, OptiState *s2)
{
	/* change the dest of all the incoming edges. */
	//assert(s2->in_edges.size()==s1->in_edges.size());
	for (unsigned ii=0; ii<s2->in_edges.size(); ii++) {

		/* if there is not so many edes in s1 */
		OptiEdge *edge = s2->in_edges[ii];

		bool removed = false;
		for (unsigned jj=0; jj<s1->in_edges.size(); jj++) {

			/* from the same source state, remove it */
			if (edge->src == s1->in_edges[ii]->src &&
				edge->equal(s1->in_edges[jj])) {
				edge->src->remove_out_edge(s2->in_edges[ii]);
				removed = true;
				assert(0); //this shouldn't happen ???
				delete edge;
				break;
			}
		}

		if (!removed) {
			edge->dest = s1;
			s1->insert_in_edge(edge);
		}
	}

	/* simply delete the out edges. */
	assert(s2->out_edges.size()==s1->out_edges.size());
	for (unsigned ii=0; ii<s2->out_edges.size(); ii++) {

		// should also remove it from the dest state
		s2->out_edges[ii]->dest->remove_in_edge(s2->out_edges[ii]);
		edges.erase(s2->out_edges[ii]->name);
		delete s2->out_edges[ii];
	}

	/* delete the state*/
	states.erase(s2->name);
	delete s2;
}

void opti_machine::print_dot(ostream& os)
{
	os << "digraph finite_state_machine {" << endl;
	os << "    rankdir=LR;" <<endl;
	os << "    orientation=land;" << endl;
	os << "    size=\"8,5\";" << endl;

	
	os << "    node [shape = doublecircle];";
	map<Symbol, OptiState *>::const_iterator sit;
	for (sit=states.begin(); sit!=states.end(); sit++) {
		if ((*sit).second->base_name==initial) {
			os << " " << symbol_get_string((*sit).first);
		}
	}
	os << ";" << endl;

	os << "    node [shape = circle];" << endl;
	map<Symbol, OptiEdge *>::const_iterator eit;
	for (eit=edges.begin(); eit!=edges.end(); eit++) {
			os << "    "
				<< symbol_get_string((*eit).second->get_source()->name)
				<< " -> "
				<< symbol_get_string((*eit).second->get_dest()->name)
				<< " [ label = \""
				<< symbol_get_string((*eit).first)
				<< "\" ];" << endl;
	}
	os << "}" << endl;
}

bool opti_machine::optimize()
{

	/* As a safety measure, we scan the graph first to see if it
	 * can be optimized, the current limitation is that it has
	 * less than one decoder with argument expression and it has no
	 * activate with argument expression. 
	 * Also all decode without argument should occur in EVAL.
	 */
	bool pass = true;
	unsigned n_dec = 0;
	OperContext::const_iterator oit;
	for (oit=oper_ctxt->begin(); oit!=oper_ctxt->end(); oit++) {

		const OperationSection *oper = (*oit).second;
		const map<Symbol, DecodeEntity *>& dents = oper->get_decode_entities();

		map<Symbol, DecodeEntity *>::const_iterator dit;
		for (dit=dents.begin(); dit!=dents.end(); dit++) {

			const DecodeEntity *ent = (*dit).second;

			/* see if this is decoded. */
			if (ent->is_decoded()) {

				/* need decoder. */
				if (ent->need_decoder()) {
					n_dec++;
					if (oper!=root_oper) pass = false; // this has to be root
				}
				else if (oper==root_oper) 
					pass = false;// root has only one that needs decoder

				/* let's check the decoding location */
				const vector<DecodeInfo *>& infos = ent->get_infos();
				if (infos.size()>1) pass = false;

				assert(infos.size()>0);

				/* if don't need decoder. then must be in EVAL section */
				if (!ent->need_decoder()) {
					if (infos.front()->edge!=SymNULL) pass = false;
				}
			}
			else {
				/* don't know how to handle @xx = yy; yet. */
				if (ent->need_decoder()) pass = false;
			}
		}
	}


	if (n_dec>1 || !pass) {
		warn("Cannot perform optimization due to current implementation "
				"constraint");
		return false;
	}

	message("Optimizing ...");

	const map<Symbol, DecodeEntity *>& dents = root_oper->get_decode_entities();
	map<Symbol, DecodeEntity *>::const_iterator dit;
	for (dit=dents.begin(); dit!=dents.end(); dit++) {

		const DecodeEntity *ent = (*dit).second;

		/* this is the one we are looking for. */
		if (ent->need_decoder()) {

			const vector<DecodeInfo *>& infos = ent->get_infos();
			assert(infos.size()==1);

			Symbol dec_edge = infos.front()->edge;
			OptiEdge *oedge = edges[dec_edge]; // this edge belongs to root

			split_edge = oedge;
			OptiState *split_state = oedge->get_dest();
			split_state_name = split_state->get_base_name();
			const vector<DecodePattern *> *patterns = ent->get_patterns();

			vector<OptiExpression *>::iterator exp_it;
			for (exp_it=oedge->exps->begin();
				exp_it!=oedge->exps->end(); exp_it++) {

				const expression *base_exp = (*exp_it)->get_base();
				if (base_exp->get_type()==OP_DEC) {

					decode_exp = *exp_it;
					const OperationExpression *oexp =
						static_cast<const OperationExpression *>(base_exp);
					assert(oexp->get_arg()!=NULL);
					break;
				}
			}

			assert(exp_it!=oedge->exps->end());
			OptiExpression *opt_exp = *exp_it;

			vector<DecodePattern *>::const_iterator pit;
			for (pit=patterns->begin(); pit!=patterns->end(); pit++) {
				// if a duplicate one, do nothing
				if (!(*pit)->get_hflag()) {
					OptiState *nstate = 
						spin_off(split_state->get_name(), (*pit)->get_ids());
					split_states.push_back(nstate);

					/* we get the EVALs of the operations */
					OptiExpression *exp = get_evals((*pit)->get_ids());
					opt_exp->append_child(exp);
				}
			}
			if (ent->has_default()) {
				vector<OperID> ids(1, ent->get_default_id());
				OptiState *nstate = spin_off(split_state->get_name(), ids);
				split_states.push_back(nstate);

				OptiExpression *exp = get_evals(ids);
				opt_exp->append_child(exp);
			}

			break;
		}
	}

	coalesce_variables();

	if (!remove_useless()) return false;
	//dump_stats(std::cerr);
	consolidate();
	//dump_stats(std::cerr);

	/* we emit a function for each unique label. */
	set<Symbol> labels;
	map<Symbol, OptiEdge *>::iterator iit;
	for (iit=edges.begin(); iit!=edges.end(); iit++) {

		Symbol edge_label = (*iit).second->label;
		if (labels.find(edge_label)==labels.end()) {
			labels.insert(edge_label);
			(*iit).second->emit = true;
		}
		else {
			(*iit).second->emit = false;
		}
	}

	optimized = true;

	return true;
}

void opti_machine::coalesce_variables()
{
	typedef pair<OperID, OperID> OPair;
	set<OPair> conflict_table;

	const map<Symbol, DecodeEntity *>& dents = root_oper->get_decode_entities();
	map<Symbol, DecodeEntity *>::const_iterator dit;
	for (dit=dents.begin(); dit!=dents.end(); dit++) {

		const DecodeEntity *ent = (*dit).second;

		/* this is the one we are looking for. */
		if (ent->need_decoder()) {

			const vector<DecodePattern *> *patterns = ent->get_patterns();

			vector<DecodePattern *>::const_iterator pit;
			for (pit=patterns->begin(); pit!=patterns->end(); pit++) {

				const vector<OperID> &ids = (*pit)->get_ids();
				vector<OperID>::const_iterator ait, bit;
				for (ait=ids.begin(); ait!=ids.end(); ait++) {

					if (*ait==0) continue;

					// the root oper is conflicting with everything
					conflict_table.insert(OPair(root_oper->get_id(), *ait));
					conflict_table.insert(OPair(*ait, root_oper->get_id()));

					for (bit=ait, bit++; bit!=ids.end(); bit++) {

						if (*bit==0) continue;

						conflict_table.insert(OPair(*ait, *bit));
						conflict_table.insert(OPair(*bit, *ait));

					}
				}
			}

			if (ent->has_default()) {
				OperID def_id = ent->get_default_id();
				conflict_table.insert(OPair(root_oper->get_id(), def_id));
				conflict_table.insert(OPair(def_id, root_oper->get_id()));
			}

			break;
		}
	}

	int var_counter = 1, conflict_counter=0;
	map<VarClause *, int> var_inds;
	map<int, VarClause *> var_rinds;
	map<VarClause *, OperID> var_set;
	map<VarClause *, set<VarClause *> > var_conflict;
	vector<VarClause *> vcodings;

	OperContext::const_iterator oit;
	for (oit=oper_ctxt->begin(); oit!=oper_ctxt->end(); oit++) {

		const OperationSection *toper = (*oit).second;
		const VarParagraph *var = toper->get_var();
		if (var) {
			const vector<VarClause *> *vcls = var->get_clauses();
			vector<VarClause *>::const_iterator vit;
			for (vit=vcls->begin(); vit!=vcls->end(); vit++) {
				if ((*vit)->get_data_type()->get_type()!=DT_OPER) {
					var_set[*vit] = toper->get_id();
					var_rinds[var_counter] = *vit;
					var_inds[*vit] = var_counter++;
				}
			}
		}
		/* we create a variable clause for operation's coding */
		if (toper->has_coding()) {
			VarClause *vcdng = new VarClause(Sym_coding, 
				new data_type_bit(false, toper->get_coding_width(), 0, NULL),
				NULL);
			vcodings.push_back(vcdng);
			var_set[vcdng] = toper->get_id();
			var_rinds[var_counter] = vcdng;
			var_inds[vcdng] = var_counter++;
		}
	}

	map<VarClause *, OperID>::iterator iiit, jjjt;
	for (iiit=var_set.begin(); iiit!=var_set.end(); iiit++) {

		VarClause *vi = (*iiit).first;
		set<VarClause *> vset;

		for (jjjt=iiit, jjjt++; jjjt!=var_set.end(); jjjt++) {

			OperID iid = (*iiit).second;
			OperID jid = (*jjjt).second;
			VarClause *vj = (*jjjt).first;

			/* the same operation */
			if (iid==jid) {
				vset.insert(vj);
				conflict_counter++;
			}
			/* conflict operation */
			else if (conflict_table.find(OPair(jid,iid))!=
					conflict_table.end()) {

				vset.insert(vj);
				conflict_counter++;
			}
			/* different data type */
			else if (!(*(vi->get_data_type())==*(vj->get_data_type()))) {
				vset.insert(vj);
				conflict_counter++;
			}
		}
		var_conflict[vi] = vset; 
	}


	adjacencytype graph[GRAPHSIZE];
	memset(graph, 0, sizeof(graph));
	vertextype order = var_counter-1;
	popmembertype color_result;

	ofstream ofs;
#if 0
    ofs.open("v.con");

	ofs << var_counter-1 << " " << conflict_counter << endl;
#endif
	for (iiit=var_set.begin(); iiit!=var_set.end(); iiit++) {

		VarClause *vi = (*iiit).first;
		set<VarClause *>& cset = var_conflict[vi];

		set<VarClause *>::iterator cit;
		for (cit=cset.begin(); cit!=cset.end(); cit++) {
#if 0
			ofs << var_inds[vi] << " " << var_inds[*cit] << endl;
#endif
			setedge(var_inds[vi]-1, var_inds[*cit]-1);
			setedge(var_inds[*cit]-1, var_inds[vi]-1);
		}
	}

#if 0
    ofs.close();
#endif

	greedy_color(graph, order, &color_result);

	coalesced_var_types.clear();
	for (int ii=1; ii<=color_result.clrdata.numcolors; ii++) {
		for (int jj=0; jj<order; jj++) {
			if (color_result.vc[jj].color==ii) {
				const DataType *dt = var_rinds[color_result.vc[jj].vertex+1]->
					get_data_type();
				coalesced_var_types.push_back(dt->clone(NULL));
				break;
			}
		}
	}

	extern map<OperID, Symbol> oper_name_table;
	coalesced_var_table.clear();
	for (int kk=1; kk<var_counter; kk++) {

		OperID op_id = var_set[var_rinds[kk]];
		Symbol tname = oper_name_table[op_id];
		coalesced_var_table
			[pair<Symbol, Symbol>(tname, var_rinds[kk]->get_name())] = 
				color_result.vc[kk-1].color;
	}

#if 0
    ofs.open("vc.list");
	/* now first we emit the temp variables. */
	for (int ii=1; ii<=color_result.clrdata.numcolors; ii++) {

		for (int jj=0; jj<order; jj++) {

			if (color_result.vc[jj].color==ii) {
				ofs << "CVAR_DEF(C_" << ii << ",";
				var_rinds[color_result.vc[jj].vertex+1]->
					get_data_type()->emit_C(ofs, false);
				ofs << ")" << endl;
				break;
			}
		}
	}	
    ofs.close();


	ofs.open("vc.map");
	for (int kk=1; kk<var_counter; kk++) {

		OperID op_id = var_set[var_rinds[kk]];
		Symbol tname = oper_name_table[op_id];
		ofs << "#define "
			<< VAR_REF(symbol_get_string(tname)+'_', var_rinds[kk]->get_name())
			<< " ";
		ofs << "C_" << color_result.vc[kk-1].color;
		ofs << endl;
	}
    ofs.close();
#endif

#if 0
    ofs.open("v.list");
	for (iiit=var_set.begin(); iiit!=var_set.end(); iiit++) {

		VarClause *vi = (*iiit).first;
		OperID op_id = var_set[vi];
		Symbol tname = oper_name_table[op_id];
		ofs << var_inds[vi] << " ";
		vi->get_data_type()->emit_C(ofs, false);
		ofs << " " <<  VAR_REF(symbol_get_string(tname)+'_', vi->get_name())
			<< endl;
	}
#endif

	vector<VarClause *>::iterator vc_it;
	for (vc_it=vcodings.begin(); vc_it!=vcodings.end(); vc_it++) {
		delete (*vc_it);
	}
	
    ofs.close();

}

void opti_machine::dump_stats(ostream& os)
{
	map<Symbol, vector<OptiEdge *> > edge_buckets;
	map<Symbol, OptiEdge *>::iterator iit;
	unsigned n_labels=0;

	for (iit=edges.begin(); iit!=edges.end(); iit++) {
		edge_buckets[(*iit).second->base_name].push_back((*iit).second);
		if ((*iit).second->emit) n_labels++;
	}

	map<Symbol, vector<OptiEdge *> >::iterator bit;
	for (bit=edge_buckets.begin(); bit!=edge_buckets.end(); bit++) {

		unsigned n_uniques=0;
		vector<OptiEdge *>::const_iterator cit;
		for (cit=(*bit).second.begin(); cit!=(*bit).second.end(); cit++)
			if ((*cit)->emit) n_uniques++;

		os << symbol_get_string((*bit).second.front()->base_name)
			<< " : " << (*bit).second.size()
			<< " (unique " << n_uniques << ")" <<  endl;
	}

	map<Symbol, vector<OptiState *> > state_bins;
	map<Symbol, OptiState *>::iterator sit;

	for (sit=states.begin(); sit!=states.end(); sit++) {
		state_bins[(*sit).second->base_name].push_back((*sit).second);
	}
	map<Symbol, vector<OptiState *> >::iterator tit;
	for (tit=state_bins.begin(); tit!=state_bins.end(); tit++) {
		os << symbol_get_string((*tit).second.front()->base_name)
			<< " : " << (*tit).second.size() << endl;
	}
	std::cerr << "Total states : " << states.size() << std::endl;
	std::cerr << "Total edges  : " << edges.size()
			 << "(unique " << n_labels << ")" << std::endl;

	unsigned var_count = 0;
	OperContext::const_iterator oit;
	for (oit=oper_ctxt->begin(); oit!=oper_ctxt->end(); oit++) {

		const OperationSection *oper = (*oit).second;
		const VarParagraph *var = oper->get_var();
		if (var) {

			const vector<VarClause *> *vcls = var->get_clauses();
			vector<VarClause *>::const_iterator vit;
			for (vit=vcls->begin(); vit!=vcls->end(); vit++) {
				if ((*vit)->get_data_type()->get_type()!=DT_OPER)
					var_count++;
			}
			
		}
	}
	std::cerr << "Total variables : " << var_count << std::endl;
}

OptiExpression *opti_machine::get_evals(const vector<OperID>& ids)
{
	extern map<OperID, Symbol> oper_name_table;

	assert(ids.size()>0);
	assert(ids.front()>0);

	/* get the name of the current operation name. */
	Symbol tname = oper_name_table[ids[0]];

	/* get the current operations */
	const OperationSection *toper = oper_ctxt->get_object(tname);

	/* get the evaluation paragraph of the operation. */
	const EvalParagraph *teval = toper->get_eval(); 
	OptiExpression *ret_exp = new OptiExpression(NULL, toper);

	if (teval) {

		const vector<Expression *> *clauses = teval->get_clauses();
		vector<Expression *>::const_iterator cls_it;
		for (cls_it=clauses->begin(); cls_it!=clauses->end(); cls_it++) {

			if ((*cls_it)->get_type()==OP_DEC) {

				const OperationExpression *oexp =
					static_cast<const OperationExpression *>(*cls_it);
				assert(oexp->get_arg()==NULL);

				Symbol var_name = oexp->get_oper();
				const DecodeEntity *dent = toper->get_decode_entity(var_name);
				const CodingClauseVarRef *vcls = dent->get_coding_clause();

				assert(vcls!=NULL);
				assert(vcls->is_operation());

				vector<OperID> sub_ids(ids.begin()+1+vcls->get_table_offset(),
										ids.begin()+1+vcls->get_table_offset()+
											vcls->get_depth());

#if 0
				cerr << symbol_get_string(toper->get_name()) << " "
					<< sub_ids.size() << " " << ids.size()
					<< " " << vcls->get_table_offset() << " " << vcls->get_depth()  << endl;
#endif

				OptiExpression *cexp = get_evals(sub_ids);	
				ret_exp->append_child(cexp);
			}
			else {
				ret_exp->append_child(new OptiExpression(*cls_it, toper));
			}
		}
	}
	return ret_exp;
}


void opti_expression::emit_C(const string& indent, ostream& os) const
{
	/* base==NULL means this is the child of a decode */
	if (base==NULL) {

		/* the coding of the operation is assigned by parent */
		const string oper_name_str = symbol_get_string(oper->get_name());

		const CodingParagraph *codng = oper->get_coding();

		if (codng) {

			const vector<CodingClause *> *ccls = codng->get_clauses();

			vector<CodingClause *>::const_iterator c_it;
			for (c_it=ccls->begin(); c_it!=ccls->end(); c_it++) {

				/* check variable reference */
				if ((*c_it)->get_type()==EC_VAR_REF) {

					CodingClauseVarRef *vcls =
						static_cast<CodingClauseVarRef *>(*c_it);

					unsigned width = (*c_it)->get_width();
					unsigned offset = (*c_it)->get_offset();
					const DataType *dtype = vcls->get_data_type();
					if (dtype->is_readonly()) {
						assert(vcls->get_var_type()==VT_DEF);
						continue;
					}

					Symbol vname = vcls->get_name();

					os << indent;
					if (vcls->is_operation()) {

						int count = vcls->get_index();
						/* get the name of the operation*/
						vector<OptiExpression *>::const_iterator cit;
						for (cit=children->begin();
							cit!=children->end(); cit++) {
							if ((*cit)->base==NULL) {
								if (count--==0) break;
							}
						}

						assert(cit!=children->end());

						Symbol child_oper = (*cit)->oper->get_name();	
					
						os << VAR_REF(symbol_get_string(child_oper)+'_',
								Sym_coding);
					}
					else
						emit_sym_of_var_type(vname, vcls->get_var_type(), 
							oper_name_str+'_', os);

					/** add __coding suffix for operation. */
					os << " = ";

					bool is_sign = false;

					if (dtype->get_type()==DT_BIT) {
						const DataTypeBit *bdtype =
							static_cast<const DataTypeBit *>(dtype);

						is_sign = bdtype->is_signed();
					}

					if (is_sign)
						os << "_INT_";
					else
						os << "_UINT_";

					if (width==1) {
						os << "XTR1(" << VAR_REF(oper_name_str+'_', Sym_coding)
						   	<< ", " << offset << ");" << endl;
					}
					else {
						os << "XTR2(" << VAR_REF(oper_name_str+'_', Sym_coding)
							<< ", " << offset + width -1 << ", "
							<< offset << ");" << endl;
					}
				}
			}
		}

		/* then we worry about the EVALs */
		if (children) {
			vector<OptiExpression *>::iterator c_it;
			for (c_it=children->begin(); c_it!=children->end(); c_it++) {
				(*c_it)->emit_C(indent, os);
				os << endl;
			}
		}

		return;
	}
	else if (base->get_type()==OP_ACT) {
		os << indent << "_ACT_MACHINE("
			<< oper->get_machine()->get_id() << ");" << endl;
	}
	else if (base->get_type()!=OP_DEC) {
		base->emit_C(os, indent, symbol_get_string(oper->get_name())+'_');
		os << endl;
		return;
	}
	
	const DecodeExpression *dec_exp =
		static_cast<const DecodeExpression*>(base);

#if 0
	const DecodeEntity *dent = oper->get_decode_entity(dec_exp->get_oper());
	const vector<DecodePattern *> *dec_patterns = dent->get_patterns();
#endif

	/* if decoding with argument, then emit the decoder. */
	if (dec_exp->get_arg()) {

		os << indent << "{" << endl;

		os << indent << "_temp_coding = ";
		dec_exp->get_arg()->emit_C(os, "",
			symbol_get_string(oper->get_name())+'_');

		os << ";" << endl;

		// need to call the decoder
		os << indent << "unsigned _temp_result = _DECODE("
			<< DECODER_NAME(dec_exp->get_decoder_name())
			<< ", _temp_coding);" << endl;

		os << indent << "(this->*__decs[_temp_result])();" << endl;
		os << indent << "}" << endl;
	}
	else if (children) { 
		vector<OptiExpression *>::iterator c_it;
		for (c_it=children->begin(); c_it!=children->end(); c_it++) {
			(*c_it)->emit_C(indent, os);
		}
	}
}

void opti_edge::emit_C(ostream& os) const
{

	vector<OptiOSMExpression *>::iterator oeit;

	if (osm_exps) {

		/* emit test expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			os << "  if (!";
			(*oeit)->base->emit_C_test(os, 
				symbol_get_string((*oeit)->oper->get_name())+'_');
			os << ") return false;" << endl;
		}

		/* emit pre expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			if ((*oeit)->base->is_pre_eval())
				if ((*oeit)->base->emit_C(os, "  ",
					symbol_get_string((*oeit)->oper->get_name())+'_'))
				os << endl << endl;
		}
	}

	if (exps) {

		/* emit expressions. */
		vector<OptiExpression *>::iterator eit;
		for (eit=exps->begin(); eit!=exps->end(); eit++) {
			(*eit)->emit_C("  ", os);
			os << endl;
		}
	}

	if (osm_exps) {
		/* emit post expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			if (!(*oeit)->base->is_pre_eval())
				if ((*oeit)->base->emit_C(os, "  ",
					symbol_get_string((*oeit)->oper->get_name())+'_'))
					os << endl << endl;
		}
	}
}

void opti_machine::emit_C_dec(ostream& ofs)
{

	/* emit C declarations. */
	ofs << endl << endl;

	ofs << "class " << MACH_CLASS(name)
		<< " : public _BASE_MACHINE {" << endl << endl;

	ofs << "  public:" << endl;

	/* emit enum for states */
	ofs << "    enum _STATES {" << endl;

	int i = 0;
	map<Symbol, OptiState *>::iterator sit;
	for (sit=states.begin(); sit!=states.end(); sit++) {
		if ((*sit).first==SymNULL) continue;
		ofs << "      " << LOCAL_IDENT((*sit).second->name) 
			<< ",\t/*" << i++ << "*/" << endl;
	}

	ofs << "    };" << endl << endl;

	i = 0;
	map<Symbol, OptiEdge *>::iterator eit;
	for (eit = edges.begin(); eit!=edges.end(); eit++,i++) {
		if ((*eit).first!=SymNULL && (*eit).second->emit)
			ofs << "    bool " << LOCAL_IDENT((*eit).first) << "();" << endl;
	}

	ofs << endl << endl;

	ofs << "    " << MACH_CLASS(name)
					<< "(_SIMULATOR *);" <<  endl << endl;

	ofs << "    ~" << MACH_CLASS(name)
					<< "();" <<  endl << endl;

	ofs << "    void __init();" << endl << endl;
	//ofs << "    bool __activate();" << endl << endl;
	ofs << "    void __retire();" << endl << endl;

	//ofs << "  private:" << endl;

	ofs << "    //variables" << endl;
	unsigned cv_ind = 1;
	vector<DataType *>::const_iterator cvt_it;
	for (cvt_it=coalesced_var_types.begin();
		cvt_it!=coalesced_var_types.end(); cvt_it++) {
		ofs << "    ";
		(*cvt_it)->emit_C(ofs, false);
		ofs << " C_" << cv_ind++ << ";" << endl;
	}

#if 0
	OperContext::const_iterator oit;
	for (oit=oper_ctxt->begin(); oit!=oper_ctxt->end(); oit++) {

		const OperationSection *oper = (*oit).second;
		const VarParagraph *var = oper->get_var();
		if (var) {

			const vector<VarClause *> *vcls = var->get_clauses();
			vector<VarClause *>::const_iterator vit;
			for (vit=vcls->begin(); vit!=vcls->end(); vit++) {
				if ((*vit)->get_data_type()->get_type()!=DT_OPER) {

					ofs << "    ";
					(*vit)->get_data_type()->emit_C(ofs);
					ofs << " "
						<< VAR_REF(symbol_get_string(oper->get_name())+'_',
							(*vit)->get_name())
						<< ";" << endl;
				}
			}
		}
		if (oper->has_coding()) {
			ofs << "    ";
			DataTypeBit b(false, oper->get_coding_width(), 0, NULL);
			b.emit_C(ofs);
			ofs << " "
				<< VAR_REF(symbol_get_string(oper->get_name())+'_', Sym_coding)
				<< ";" << endl;
		}
		ofs << endl;
	}
#endif

	const DecodeExpression *dec_exp =
		static_cast<const DecodeExpression*>(decode_exp->get_base());

	ofs << endl << "    "; 
	dec_exp->get_arg()->get_data_type()->emit_C(ofs);
	ofs << " _temp_coding;" << endl << endl;

	const SymContext *mvars = base_mach->get_variables();
	SymContext::const_iterator sc_it;
	for (sc_it = mvars->begin(); sc_it!=mvars->end(); sc_it++) {

		const named_typed_node *vnode = (*sc_it).second;

		ofs << "    ";
		vnode->get_data_type()->emit_C(ofs);
		ofs << " " << MVR_NAME(vnode->get_name()) << ";" << endl;
	}

	ofs << endl;
		
	ofs << "#ifndef _GEN_DISASSEMBLER" << endl;
	ofs << "    //buffers" << endl;

	const BufContext *buffers = base_mach->get_buffers();;
	BufContext::const_iterator bf_it;
	for (bf_it = buffers->begin(); bf_it!=buffers->end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		ofs << "    " << MANG_CLASS(vnode->get_manager_name()) << "::TOKEN_T *"
			<< BUF_NAME(vnode->get_name()) << ";" << endl;
	}
	ofs << "#endif" << endl;
	
#if 0
	for (sit=states.begin(); sit!=states.end(); sit++) {
		if ((*sit).first==SymNULL) continue;
		ofs << "    bool __act_" << LOCAL_IDENT((*sit).first) << "();" << endl;
	}
#endif


	const DecodeEntity *dent =
		decode_exp->oper->get_decode_entity(dec_exp->get_oper());
	const vector<DecodePattern *> *dec_patterns = dent->get_patterns();

	unsigned pcounter=0;
	vector<DecodePattern *>::const_iterator pit;
	for (pit=dec_patterns->begin(); pit!=dec_patterns->end(); pit++) {

		// if a duplicate one, do nothing
		if (!(*pit)->get_hflag()) {
			ofs << "    void __dec_" << pcounter << "();" << endl;
			pcounter++;
		}
	}
	if (dent->has_default()) {
		ofs << "    void __dec_" << pcounter << "();" << endl;
	}

	ofs << "};" << endl;


	ofs << "\n#endif\n";

}

void opti_machine::emit_C(ostream& ofs)
{
	/* have already optimized. */
	assert(optimized);

	map<pair<Symbol, Symbol>, int>::const_iterator cvt_it;
	for (cvt_it=coalesced_var_table.begin();
		cvt_it!=coalesced_var_table.end(); cvt_it++) {
		ofs << "#define "
			<< VAR_REF(symbol_get_string((*cvt_it).first.first)+'_',
					(*cvt_it).first.second)
			<< " C_" << (*cvt_it).second << endl;
	}

	/* Constructor. */
	ofs << MACH_CLASS(name) << "::" << MACH_CLASS(name)
		<< "(_SIMULATOR *sim):_BASE_MACHINE(sim, "
		<< std::dec << base_mach->get_id() << ") {" << endl << endl;
		
	ofs << "  _state_ = " << LOCAL_IDENT(initial) << ";" <<endl;

	ofs << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	const BufContext *buffers = base_mach->get_buffers();
	BufContext::const_iterator bf_it;
	for (bf_it = buffers->begin(); bf_it!=buffers->end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		ofs << "  " << BUF_NAME(vnode->get_name()) << " = NULL;";
	   	ofs << endl;
	}
	ofs << "#endif" << endl;

	ofs << endl;
	
	ofs << "}" << endl << endl;

	/* Destructor. */
	ofs << MACH_CLASS(name) << "::~"
		<< MACH_CLASS(name) << "() {}" << endl << endl;
		

	unsigned edge_count=0;
	map<Symbol, OptiState *>::iterator sit;
	for (sit=states.begin(); sit!=states.end(); sit++) {
		vector<OptiEdge *> *out_edges = &((*sit).second->out_edges);
		edge_count += out_edges->size();
	}
	if (edge_count!=edges.size()) {
		cerr << edge_count << endl;
		cerr << edges.size() << endl;
		error("count mismatch");
		assert(0);
	}

	const expression *base_exp = decode_exp->get_base();
	assert(base_exp->get_type()==OP_DEC);
	const DecodeExpression *dec_exp =
		static_cast<const DecodeExpression*>(base_exp);

	const DecodeEntity *dent =
		decode_exp->oper->get_decode_entity(dec_exp->get_oper());
	const vector<DecodePattern *> *dec_patterns = dent->get_patterns();

	unsigned pcounter=0;
	vector<DecodePattern *>::const_iterator pit;
	for (pit=dec_patterns->begin(); pit!=dec_patterns->end(); pit++) {

		// if a duplicate one, do nothing
		if (!(*pit)->get_hflag()) {

			ofs << "void " << MACH_CLASS(name) << "::__dec_"
				<< pcounter << "() {" << endl;

			const OperationSection *coper = 
				(*decode_exp->children)[pcounter]->oper;
			ofs << "  "
				<< VAR_REF(symbol_get_string(coper->get_name())+'_', Sym_coding)
				<< " = _temp_coding;" << endl;

			(*decode_exp->children)[pcounter]->emit_C("  ", ofs);
			ofs << "  _state_ = "
				<< LOCAL_IDENT(split_states[pcounter]->get_name())
				<< ";" <<endl;
			ofs << "}" << endl;
			pcounter++;	
		}
	}
	if (dent->has_default()) {
		ofs << "void " << MACH_CLASS(name) << "::__dec_"
			<< pcounter << "() {" << endl;
		const OperationSection *coper = 
			(*decode_exp->children)[pcounter]->oper;

		if (coper->has_coding())
			ofs << "  "
				<< VAR_REF(symbol_get_string(coper->get_name())+'_', Sym_coding)
				<< " = _temp_coding;" << endl;

		(*decode_exp->children)[pcounter]->emit_C("  ", ofs);
		ofs << "  _state_ = "
			<< LOCAL_IDENT(split_states[pcounter]->get_name())
			<< ";" <<endl;
		ofs << "}" << endl;
	}

	ofs << "typedef void (" << MACH_CLASS(name)
		<< "::*__DECODE_FTYPE)();" << endl;
	ofs << "static const __DECODE_FTYPE __decs[] = {" << endl;

	pcounter = 0;
	for (pit=dec_patterns->begin(); pit!=dec_patterns->end(); pit++) {

		// if a duplicate one, do nothing
		if (!(*pit)->get_hflag()) {

			ofs << "&" << MACH_CLASS(name) << "::__dec_"
				<< pcounter++ << "," << endl;
		}
	}
	if (dent->has_default()) {
		ofs << "&" << MACH_CLASS(name) << "::__dec_"
			<< pcounter++ << "," << endl;
	}
	ofs << "};" << endl << endl;

	map<Symbol, OptiEdge *>::iterator iit;
	for (iit=edges.begin(); iit!=edges.end(); iit++) {

		if ((*iit).first!=SymNULL && (*iit).second->emit) {
			ofs << "bool " << MACH_CLASS(name) << "::"
				<< LOCAL_IDENT((*iit).second->name) << "() {" << endl;
			(*iit).second->emit_C(ofs);
			ofs << "  return true;" << endl;
			ofs << "}" << endl << endl;
		}
	}

	/* we emit an activate function for all states. */
	for (sit=states.begin(); sit!=states.end(); sit++) {

		if ((*sit).first==SymNULL) continue;
		ofs << "static bool __act_"
			<< LOCAL_IDENT((*sit).first) << "(_opt_machine_ *_mach) {" << endl;

		ofs << "  " << MACH_CLASS(name) << " *_tmach = "
			<< "static_cast<" << MACH_CLASS(name)
			<< " *>(_mach);" << endl;

		vector<OptiEdge *> *out_edges = &((*sit).second->out_edges);
		if (out_edges->size()) {
	
			ofs << "	  if (false) {}" << endl;
			vector<OptiEdge *>::iterator eit;
			for (eit=out_edges->begin(); eit!=out_edges->end(); eit++) {

				ofs << "	  else if (_FIRE_EDGE("
					<< LOCAL_IDENT((*eit)->label)	 
					<< ")) " << endl;
				if (*eit==split_edge)
					ofs << "		{}" << endl;
				else
					ofs << "		_mach->__set_state("
						<<  MACH_CLASS(name) << "::"
						<< LOCAL_IDENT((*eit)->get_dest()->get_name())
						<< ");" << endl;
			}
			ofs << "	  else" << endl
				<< "		return true;" << endl;

		}
		ofs << "    return (_mach->__get_state()!="
			<< MACH_CLASS(name) << "::"
			<< LOCAL_IDENT(initial) << ");" << endl;
		ofs << "}" << endl << endl;
	}

	ofs << "typedef bool (*__ACT_FTYPE)(_opt_machine_ *);" << endl;
	ofs << "static const __ACT_FTYPE __nacts[] = {" << endl;

	for (sit=states.begin(); sit!=states.end(); sit++) {
		if ((*sit).first==SymNULL) continue;
		ofs << "&__act_"
			<< LOCAL_IDENT((*sit).first) << "," << endl;
	}

	ofs << "};" << endl << endl;

	ofs << "void " << MACH_CLASS(name)
		<< "::__init() {" << endl;
	edges[SymNULL]->emit_C(ofs);
	ofs <<  "  this->_acts_ = __nacts;" << endl;
	ofs << "}" << endl << endl;

#if 0
	ofs << "bool " << MACH_CLASS(base_mach->get_name()) << "::__activate() {"
			<< endl;
	ofs << "  return (this->*__acts[_state_])();" << endl;
	ofs << "}" << endl << endl;
#endif

	ofs << "void " << MACH_CLASS(name)
		<< "::__retire() {" << endl << endl;

	ofs << "  _state_ = " << LOCAL_IDENT(initial) << ";"
		<< endl << endl;

	ofs << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	for (bf_it = buffers->begin(); bf_it!=buffers->end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		ofs << "  _RETIRE_BUFFER(" << BUF_NAME(vnode->get_name()) << ");"
			<< endl << endl;
	}
	ofs << "#endif" << endl;

	ofs << "}" << endl << endl;

}

void opti_machine::emit_decode_patterns(const char *prefix)
{
	const map<Symbol, DecodeEntity *>& dents = root_oper->get_decode_entities();
	map<Symbol, DecodeEntity *>::const_iterator dit;
	for (dit=dents.begin(); dit!=dents.end(); dit++) {

		const DecodeEntity *ent = (*dit).second;
		if (ent->need_decoder()) {

			ent->emit_pat_file(prefix);
			ent->emit_opt_header_file(prefix);
			ent->emit_opt_tab_file(prefix);
			ent->emit_main_file(prefix);
		}
	}
}

void opti_edge::emit_C_ISS(ostream& os,
	const OptiExpression *dec_exp) const
{
	vector<OptiOSMExpression *>::iterator oeit;
	if (osm_exps) {

		os << "  else if (";
		/* emit test expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			(*oeit)->base->emit_C_test(os, 
				symbol_get_string((*oeit)->oper->get_name())+'_');
			os << " && ";
		}
		os << " true) {" << endl;

		/* emit pre expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			if ((*oeit)->base->is_pre_eval())
				if ((*oeit)->base->emit_C(os, "  ",
					symbol_get_string((*oeit)->oper->get_name())+'_'))
				os << endl << endl;
		}
	}
	else
		os << "  else {" << endl;

	if (exps) {

		/* emit expressions. */
		vector<OptiExpression *>::iterator eit;
		for (eit=exps->begin(); eit!=exps->end(); eit++) {

			(*eit)->emit_C("  ", os);
			os << endl;

			if (*eit==dec_exp) {
				os << "  return;" << endl;
				return;
			}
		}
	}

	if (osm_exps) {
		/* emit post expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			if (!(*oeit)->base->is_pre_eval())
				if ((*oeit)->base->emit_C(os, "  ",
					symbol_get_string((*oeit)->oper->get_name())+'_'))
					os << endl << endl;
		}
	}
}

void opti_edge::emit_C_ISS_rest(ostream& os,
	const OptiExpression *dec_exp) const
{
	assert(exps);

	/* emit expressions. */
	vector<OptiExpression *>::iterator eit;
	for (eit=exps->begin(); eit!=exps->end(); eit++)
		if (*eit==dec_exp) break;

	for (eit++; eit!=exps->end(); eit++) {
		(*eit)->emit_C("  ", os);
		os << endl;
	}

	if (osm_exps) {
		vector<OptiOSMExpression *>::iterator oeit;
		/* emit post expressions. */
		for (oeit=osm_exps->begin(); oeit!=osm_exps->end(); oeit++) {
			if (!(*oeit)->base->is_pre_eval())
				if ((*oeit)->base->emit_C(os, "  ",
					symbol_get_string((*oeit)->oper->get_name())+'_'))
					os << endl << endl;
		}
	}
}

void opti_machine::emit_C_dec_ISS(ostream& ofs) const
{

	/* emit C declarations. */
	ofs << endl << endl;

	ofs << "class " << MACH_CLASS(name)
		<< " : public _BASE_MACHINE {" << endl << endl;

	ofs << "  public:" << endl;

	ofs << "    " << MACH_CLASS(name)
					<< "(_SIMULATOR *);" <<  endl << endl;

	ofs << "    ~" << MACH_CLASS(name)
					<< "();" <<  endl << endl;

	ofs << "    void __activate();" << endl << endl;

	ofs << "    //variables" << endl;

	unsigned cv_ind = 1;
	vector<DataType *>::const_iterator cvt_it;
	for (cvt_it=coalesced_var_types.begin();
		cvt_it!=coalesced_var_types.end(); cvt_it++) {
		ofs << "    ";
		(*cvt_it)->emit_C(ofs, false);
		ofs << " C_" << cv_ind++ << ";" << endl;
	}

#if 0
	OperContext::const_iterator oit;
	for (oit=oper_ctxt->begin(); oit!=oper_ctxt->end(); oit++) {

		const OperationSection *oper = (*oit).second;
		const VarParagraph *var = oper->get_var();
		if (var) {

			const vector<VarClause *> *vcls = var->get_clauses();
			vector<VarClause *>::const_iterator vit;
			for (vit=vcls->begin(); vit!=vcls->end(); vit++) {
				if ((*vit)->get_data_type()->get_type()!=DT_OPER) {

					ofs << "    ";
					(*vit)->get_data_type()->emit_C(ofs);
					ofs << " "
						<< VAR_REF(symbol_get_string(oper->get_name())+'_',
							(*vit)->get_name())
						<< ";" << endl;
				}
			}
		}
		if (oper->has_coding()) {
			ofs << "    ";
			DataTypeBit b(false, oper->get_coding_width(), 0, NULL);
			b.emit_C(ofs);
			ofs << " "
				<< VAR_REF(symbol_get_string(oper->get_name())+'_', Sym_coding)
				<< ";" << endl;
		}
		ofs << endl;
	}
#endif

	const DecodeExpression *dec_exp =
		static_cast<const DecodeExpression*>(decode_exp->get_base());

	ofs << endl << "    "; 
	dec_exp->get_arg()->get_data_type()->emit_C(ofs);
	ofs << " _temp_coding;" << endl << endl;

	const SymContext *mvars = base_mach->get_variables();
	SymContext::const_iterator sc_it;
	for (sc_it = mvars->begin(); sc_it!=mvars->end(); sc_it++) {

		const named_typed_node *vnode = (*sc_it).second;

		ofs << "    ";
		vnode->get_data_type()->emit_C(ofs);
		ofs << " " << MVR_NAME(vnode->get_name()) << ";" << endl;
	}

	ofs << endl;
		
	ofs << "    //buffers" << endl;

	const BufContext *buffers = base_mach->get_buffers();;
	BufContext::const_iterator bf_it;
	for (bf_it = buffers->begin(); bf_it!=buffers->end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		ofs << "    " << MANG_CLASS(vnode->get_manager_name()) << "::TOKEN_T *"
			<< BUF_NAME(vnode->get_name()) << ";" << endl;
	}
	
	const DecodeEntity *dent =
		decode_exp->oper->get_decode_entity(dec_exp->get_oper());
	const vector<DecodePattern *> *dec_patterns = dent->get_patterns();

	unsigned pcounter=0;
	vector<DecodePattern *>::const_iterator pit;
	for (pit=dec_patterns->begin(); pit!=dec_patterns->end(); pit++) {

		// if a duplicate one, do nothing
		if (!(*pit)->get_hflag()) {
			ofs << "    void __dec_" << pcounter << "();" << endl;
			pcounter++;
		}
	}
	if (dent->has_default()) {
		ofs << "    void __dec_" << pcounter << "();" << endl;
	}

	ofs << "};" << endl;


	ofs << "\n#endif\n";
}

/* to emit ISS code in C++ */
void opti_machine::emit_C_ISS(ostream& ofs) const
{
	assert(optimized);

	map<pair<Symbol, Symbol>, int>::const_iterator cvt_it;
	for (cvt_it=coalesced_var_table.begin();
		cvt_it!=coalesced_var_table.end(); cvt_it++) {
		ofs << "#define "
			<< VAR_REF(symbol_get_string((*cvt_it).first.first)+'_',
					(*cvt_it).first.second)
			<< " C_" << (*cvt_it).second << endl;
	}

	/* Constructor. */
	ofs << MACH_CLASS(name) << "::" << MACH_CLASS(name)
		<< "(_SIMULATOR *sim):_BASE_MACHINE(sim, "
		<< std::dec << base_mach->get_id() << ") {" << endl << endl;

#if 0
	ofs << "#ifndef _GEN_DISASSEMBLER" << endl;
	/* reset buffers */
	const BufContext *buffers = base_mach->get_buffers();
	BufContext::const_iterator bf_it;
	for (bf_it = buffers->begin(); bf_it!=buffers->end(); bf_it++) {

		const BufferClause *vnode = (*bf_it).second;

		ofs << "  " << BUF_NAME(vnode->get_name()) << " = NULL;";
	   	ofs << endl;
	}
	ofs << "#endif" << endl;
#endif

	ofs << endl;
	
	ofs << "}" << endl << endl;

	/* Destructor. */
	ofs << MACH_CLASS(name) << "::~"
		<< MACH_CLASS(name) << "() {}" << endl << endl;

	/* for every split state, we emit a function */	
	unsigned pcounter = 0;
	vector<OptiState *>::const_iterator split_it;

	ofs << "typedef void (" << MACH_CLASS(name)
		<< "::*__DECODE_FTYPE)();" << endl;
	ofs << "static const __DECODE_FTYPE __decs[] = {" << endl;

	for (split_it=split_states.begin(); split_it!=split_states.end();
		split_it++) {

		ofs << "&" << MACH_CLASS(name) << "::__dec_"
			<< pcounter++ << "," << endl;
	}
	ofs << "};" << endl << endl;

	/* We first emit a main function that is common to all instructions. */
	ofs << "void " << MACH_CLASS(name) << "::__activate() {" << endl;

	assert(edges.find(SymNULL)!=edges.end());
	(*edges.find(SymNULL)).second->emit_C(ofs);

	assert(states.find(initial)!=states.end());
	set<OptiState *> temp_state_set;
	emit_C_ISS_succ_state(ofs, (*states.find(initial)).second, temp_state_set);

	ofs << "}" << endl << endl;

	pcounter = 0;
	for (split_it=split_states.begin(); split_it!=split_states.end();
		split_it++) {

		ofs << "void " << MACH_CLASS(name) << "::" << "__dec_" << pcounter
			<< "() {" << endl;

		split_edge->emit_C_ISS_rest(ofs, decode_exp);

		const OperationSection *coper = 
			(*decode_exp->children)[pcounter]->oper;

		if (coper->has_coding())
			ofs << "  "
				<< VAR_REF(symbol_get_string(coper->get_name())+'_', Sym_coding)
				<< " = _temp_coding;" << endl;

		(*decode_exp->children)[pcounter]->emit_C("  ", ofs);

		set<OptiState *> temp_state_set;
		emit_C_ISS_succ_state(ofs, *split_it, temp_state_set);
		ofs << "}" << endl << endl;
		pcounter++;
	}
}

void opti_machine::emit_C_ISS_succ_state(ostream& ofs,
	OptiState *state, set<OptiState *>& done_set) const
{
	assert(done_set.find(state)==done_set.end());

	done_set.insert(state);

	/* first we duplicate the states since the creation of new edges
	 * needs state object.
	 */
	OptiState *cur_state = state;

	/* emit a label */
	ofs << "\n__" << LOCAL_IDENT(cur_state->get_name()) << ":" << endl;

	ofs << "  if (false) {}" << endl;
	/* then we emit all expressions on the outgoing edges. */
	vector<OptiEdge *>& out_edges = cur_state->out_edges;
	vector<OptiEdge *>::const_iterator eit;
	for (eit=out_edges.begin(); eit!=out_edges.end(); eit++) {

		(*eit)->emit_C_ISS(ofs, decode_exp);

		OptiState *dest_state = (*eit)->dest;
		if (dest_state->get_name()==initial) {
			ofs << "  return;" << endl;
		}
		/* we push the destination state into work_list if not done yet */
		else if (done_set.find(dest_state)==done_set.end()) {

			/* push this to the queue */
			emit_C_ISS_succ_state(ofs, dest_state, done_set);
		}
		else {
			ofs << "  goto __" << LOCAL_IDENT(dest_state->get_name())
				<< ";" << endl;
		}
		ofs << "  }" << endl;

	//	ofs << "  else {assert(0);}" << endl;
	}
}

