
#ifndef __MAD_OPTI_HPP__
#define __MAD_OPTI_HPP__

#include "machine_ast.hpp"
#include "operation_ast.hpp"
#include <iostream>
#include <vector>
#include <set>
#include <map>

namespace mad_ast {

	/** An intermediate level osm_expression in the optimized machine.
	 *  Is is basically a wrapper. 
	 */
	typedef class opti_osm_expression {

		public:
			/** Constructor.
			 *  @param osm_exp An OSM expression.
			 *  @param op      The operation section owning osm_exp.
			 */
			opti_osm_expression(const osm_expression *osm_exp,
				const operation_section *op) :  base(osm_exp), oper(op) {};

			/** Destructor. */
			~opti_osm_expression() {}

		private:

			// The osm expression itself;
			const osm_expression *base;

			// The operation owning this osm_expression
			const operation_section *oper;

			friend class opti_edge;

	} OptiOSMExpression;


	/** An intermediate level expression in the optimized machine. 
	 *  This one normally works as a wrapper. In case when the
	 *  expression is a decode, the children will contains a list
	 *  of expressions corresponding to the EVALs along the decoding tree.
	 */
	typedef class opti_expression {

		public:
			/** Constructor.
			 *  @param exp The MAD expression.
			 *  @param op  The operation owning the exp.
			 */
			opti_expression(const expression *exp,
				const operation_section *op) :
				base(exp), oper(op), children(NULL) {}

			/** Destructor. */
			~opti_expression();

			/** Append a child expression.
			 *  @param e The child.
			 */
			void append_child(opti_expression *e);

			/** Get the base expression. */
			const expression *get_base() const {return base;}

			/** Emit the expression. */
			void emit_C(const std::string&indent, std::ostream& os) const;

		private:

			// the original mad expression
			const expression *base;	

			// The operation owning this expression
			const operation_section *oper;

			// child expressions.
			std::vector<opti_expression *> *children;

			friend class opti_edge;
			friend class opti_machine;

	} OptiExpression;

	/* forward declaration */
	class opti_state;

	/** An intermediate level edge in the optimized state diagram. */
	typedef class opti_edge {

		public:
			/** Constructor.
			 *  @param base The base name, name in original diagram.
			 *  @param name The new name, the full name of the edge.
			 *  @param src  The source state.
			 *  @param dest The destinatin state.
			 */
			opti_edge(Symbol base, Symbol name,
				opti_state *src, opti_state *dest);

			/** Destructor. */
			~opti_edge();

			/** Initialize the expressions field.
			 *  @param exps Insert expressions into the exps.
			 *  @param oper The oper owning the expressions.
			 */
			void append_exps(const std::vector<expression *> *exps,
				const operation_section *oper);

			/** Initialize the osm_expressions field.
			 *  @param oexps Insert wrapped expressions into the exps.
			 */
			void append_exps(const std::vector<OptiExpression *> *oexps);

			/** Initialize the osm_expressions field.
			 *  @param oexps Insert osm_expressions into the exps.
			 */
			void append_osm_exps(const std::vector<OptiOSMExpression *> *oexps);

			/** Initialize the osm_expressions field.
			 *  @param osm_exps Insert osm_expressions into the exps.
			 *  @param oper The oper owning the expressions.
			 */
			void append_osm_exps(const std::vector<osm_expression *> *osm_exps,
				const operation_section *oper);

			/** check equality relation of two edges. 
			 *  @param edge The edge to compare against.
			 */
			bool equal(opti_edge *edge);

			/** Get the soruce state. */
			opti_state *get_source() {return src;}

			/** Get the destination state. */
			opti_state *get_dest() {return dest;}

			/** Get the name of the edge .*/
			Symbol get_name() const {return name;}

			/** Get the base_name of the edge. */
			Symbol get_base_name() const {return base_name;}

			/** Emit C function. */
			void emit_C(std::ostream& os) const;

			/** Emit C function for ISS. */
			void emit_C_ISS(std::ostream& os,
				const OptiExpression *dec_exp) const;

			/** Emit C function for ISS. */
			void emit_C_ISS_rest(std::ostream& os,
				const OptiExpression *dec_exp) const;

		private:
			Symbol base_name;	// name of the base edge
			Symbol name;		// name
			opti_state *src;	// source state
			opti_state *dest;	// destination state

			std::vector<OptiExpression *> *exps;
			std::vector<OptiOSMExpression *> *osm_exps;

			Symbol label;		// label for C output
			bool emit;			// unique edge, emit function for this

			friend class opti_machine;
			friend class opti_state;

	} OptiEdge;

	/** An intermediate level state in the optimized state diagram. */
	typedef class opti_state {

		public:

			/** Constructor. 
			 *  @param base Base state name.
			 *  @param name The full name.
			 */
			opti_state(Symbol base, Symbol name) :
				base_name(base), name(name) {}

			/** Destructor. */
			~opti_state() {}

			/** One more outgoing edge.
			 *  @param e The outgoing edge.
			 */
			void insert_out_edge(OptiEdge *e) {
				out_edges.push_back(e);
			}

			/** One more incoming edge.
			 *  @param e The incoming edge.
			 */
			void insert_in_edge(OptiEdge *e) {
				in_edges.push_back(e);
			}

			/** Remove an outgoing edge.
			 *  @param e The outgoing edge.
			 */
			void remove_out_edge(OptiEdge *e);

			/** Remove an incoming edge.
			 *  @param e The outgoing edge.
			 */
			void remove_in_edge(OptiEdge *e);

			/** Compare two states.
			 *  @param op2 Compare the compatibility of two states.
			 */
			bool compatible(opti_state *op2);

			/** Get the name of the edge .*/
			Symbol get_name() const {return name;}

			/** Get the base_name of the edge. */
			Symbol get_base_name() const {return base_name;}

		private:

			Symbol base_name;	// name of the base state
			Symbol name;		// name
			std::vector<OptiEdge *> out_edges;	// outgoing edges
			std::vector<OptiEdge *> in_edges;	// incoming edges

			friend class opti_machine;

	} OptiState;

	/** An intermediate level state machine for optimization. */
	typedef class opti_machine {

		public:

			/** Constructor, construct an IR state machine.
			 *  @param root  The root level operation.
			 *  @param opers All the operations.
			 */
			opti_machine(const operation_section *root,
						 const OperContext *opers);

			/** Destructor. */
			~opti_machine();

			/** Get the name of the machine. */
			Symbol get_name() const {return name;}

			/** Optimize.
			 *  @return True if optimization performed.
			 */
			bool optimize();

			/** Spin off one path from the given state.
			 *  @param state The point where spin off occues.
			 *  @param ids   The operation id list to attach to the path.
			 *  @return The new beginning state.
			 */
			OptiState *spin_off(Symbol state, const std::vector<OperID>& ids);

			/** Consolidate the states.
			 */
			void consolidate();

			/** Print the state diagram as a dot file.
			 *  @param os The output stream.
			 */
			void print_dot(std::ostream& os);

			/** Dump the stats. */
			void dump_stats(std::ostream& os);

			/** Emit C code for the optimized machine. */
			void emit_C(std::ostream& os);

			/** Emit C declaration code for the optimized machine. */
			void emit_C_dec(std::ostream& os);

			/** Emit decode patterns. */
			void emit_decode_patterns(const char *prefix);

			/** Get the root operation. */
			const operation_section *get_root_oper() const {return root_oper;}

			/** Some procedures to emit ISS */
			void emit_C_dec_ISS(std::ostream& os) const;
			void emit_C_ISS(std::ostream& os) const;
			void emit_C_ISS_succ_state(std::ostream& os,
				OptiState *st, std::set<OptiState *>& done_set) const;

		private:

			/** label the edges, equal edges will get the same label. */
			void label_edges();

			/** merge s2 into s1 */
			void merge(OptiState *s1, OptiState *s2);

			void coalesce_variables();

			/* remove useless states and edges.
			 * @return True if everything ok, false means error.
			 */
			bool remove_useless();

			/** generate a new unique name for an edge. */
			Symbol new_edge_name(Symbol base);

			/** generate a new unique name for a state. */
			Symbol new_state_name(Symbol base);

			OptiExpression *get_evals(const std::vector<OperID>& ids);

			const machine_section   *base_mach;	// the base machine section
			const operation_section *root_oper;	// the root operation section
			const OperContext *oper_ctxt;

			std::map<Symbol, OptiEdge *>  edges;	// all the edges.
			std::map<Symbol, OptiState *> states;	// all the states.
			Symbol initial;			//initial state

			unsigned name_counter;	//for duplication

			Symbol name;			//name of the machine

			bool optimized;

			Symbol split_state_name;
			std::vector<OptiState *> split_states;
			OptiEdge *split_edge;
			OptiExpression *decode_exp;

			std::vector<DataType *> coalesced_var_types;
			std::map<std::pair<Symbol, Symbol>, int> coalesced_var_table;

	} OptiMachine;
};

#endif
