
#include "osm_exp_ast.hpp"
#include "exp_ast.hpp"
#include "emit_c.hpp"
#include "manager_ast.hpp"
#include "machine_ast.hpp"
#include <cassert>
#include <sstream>

using namespace mad_ast;
using namespace mad;
using std::ostream;
using std::string;
using std::stringstream;
using std::vector;
using std::endl;

static char *alphanum[] = {"first", "second", "third", "fourth", "fifth"};
static char *comp_str[] = {"<", ">", "<=", ">=", "==", "!="};
static char *multples[] = {"TUPLE", "TRIPLE", "QUADRUPLE", "QUINTUPLE"};

bool mad_ast::OSMExpPtrCmp(const OSMExpression *e1, const OSMExpression *e2)
{
	if (e1->get_type()!=e2->get_type())
		return e1->get_type()<e2->get_type();

	if (e1->get_type()==OSM_ASSIGN) {

		const OSMAssignExpression *assgn1 =
			static_cast<const OSMAssignExpression *>(e1);
		const OSMAssignExpression *assgn2 =
			static_cast<const OSMAssignExpression *>(e2);


		/* allocation */
		if (assgn1->get_lhs()->get_type()==OSM_BUF_REF)
			return true;
		else if (assgn2->get_lhs()->get_type()==OSM_BUF_REF)
			return false;

		/* read */
		if (assgn1->get_lhs()->get_type()==OSM_VAR_REF ||
			assgn1->get_lhs()->get_type()==OSM_GROUP)
			return true;
		else if (assgn2->get_lhs()->get_type()==OSM_VAR_REF ||
			assgn2->get_lhs()->get_type()==OSM_GROUP)
			return false;
			
	}

	// others doesn't matter
	return true;
}

ostream& osm_compare_expression::print(ostream& os) const
{
	lhs->print(os);
	os << comp_str[cop];
	rhs->print(os);
	print_object_list(os, " ", annots);

	return os;
}

OSMExpType *osm_compare_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	if (type) return type;

	OSMExpType *ot1 = lhs->infer(man_context, 
		buf_context, var_context, mvr_context, def_context);

	OSMExpType *ot2 = rhs->infer(man_context,
		buf_context, var_context, mvr_context, def_context);

	/* error */
	if (ot1==NULL || ot2==NULL)
		return NULL;

	/* should be guaranteed by parser */
	assert(ot1->exp_type==OET_VAL);
	assert(ot2->exp_type==OET_VAL);

	if (!valid_arith_type(ot1->data_type) ||
		!valid_arith_type(ot2->data_type))
		return NULL;

	sup_type = common_super_type(ot1->data_type, ot2->data_type);
	if (sup_type==NULL) return NULL;

	pre_eval = true;
	/* root level void type */
	type = new OSMExpType(OET_VOID, NULL);
	return type;
}

ostream& osm_var_ref_expression::print(ostream& os) const
{
	os << symbol_get_string(sym);
	return os;
}

OSMExpType *osm_var_ref_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

    const DataType *tp = NULL;
    /* see if the variable is defined */
    if (var_context && var_context->has(sym)) {
        tp = var_context->get_data_type(sym);
        vtype = VT_VAR;
	}
	else if (mvr_context && mvr_context->has(sym)) {
		tp = mvr_context->get_data_type(sym);
		vtype = VT_MAC;
	}
	else if (def_context && def_context->has(sym)) {
		tp = def_context->get_data_type(sym);
		vtype = VT_DEF;                                       
	}
	else if (buf_context && buf_context->has(sym)) {
		generic_error("Using token buffer as normal variable", this->get_loc());
		return NULL;
	}
	else {
		undefined_ident(sym, this->get_loc());
		return NULL;
	}

	type = new OSMExpType(OET_VAL, tp->clone(NULL));
	return type;
}

ostream& osm_buf_ref_expression::print(ostream& os) const
{
	os << symbol_get_string(sym);
	return os;
}

OSMExpType *osm_buf_ref_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	if (buf_context && buf_context->has(sym)) {
		const BufferClause *buf = buf_context->get_object(sym);
		type = new OSMExpType(
			buf_context->get_data_type(sym)->clone(NULL),
			buf->get_manager_name());
	}
	else if ((var_context && var_context->has(sym)) ||
			(mvr_context && mvr_context->has(sym))) {
		generic_error("Using normal variable as token buffer", this->get_loc());
	}
	else {
		undefined_ident(sym, this->get_loc());
	}

	return type;
}

ostream& osm_num_expression::print(ostream& os) const
{
	num->print(os);
	return os;
}

OSMExpType *osm_num_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	if (num->is_float()) { 
		type = new OSMExpType(OET_VAL, 
			new DataTypeFloat(num->is_double(), 0, NULL));
	}
	else {
		type = new OSMExpType(OET_VAL,
			new DataTypeBit(false, num->get_width_inbase(2), 0, NULL));
	}

	type->data_type->set_readonly(true);
	return type;
}

ostream& osm_str_expression::print(ostream& os) const
{
	os << "\"" << str << "\"";
	return os;
}

OSMExpType *osm_str_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	type = new OSMExpType(OET_VAL, new DataTypeString(0, NULL));

	type->data_type->set_readonly(true);
	return type;
}

ostream& osm_release_expression::print(ostream& os) const
{
	os << "!";
	buf->print(os);
	print_object_list(os, " ", annots);
	return os;
}

OSMExpType *osm_release_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	type = new OSMExpType(OET_VOID, NULL);
	buf->infer(man_context, buf_context, var_context,
				mvr_context, def_context);

	return type;
}

ostream& osm_discard_expression::print(ostream& os) const
{
	os << "!!";
	buf->print(os);
	print_object_list(os, " ", annots);
	return os;
}

OSMExpType *osm_discard_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	type = new OSMExpType(OET_VOID, NULL);
	buf->infer(man_context, buf_context, var_context,
				mvr_context, def_context);

	return type;
}

ostream& osm_man_ref_expression::print(ostream& os) const
{
	os << symbol_get_string(sym) << "[";
	if (ind) ind->print(os);
	os << "]";
	print_object_list(os, " ", annots);
	return os;
}

OSMExpType *osm_man_ref_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	if (!man_context || !man_context->has(sym)) {
		undefined_ident(sym, this->get_loc());
	}
	else {

		itype = man_context->get_index_type(sym)->clone(NULL);

		if (ind) {
			OSMExpType *ot = ind->infer(man_context,
				buf_context, var_context,
			   	mvr_context, def_context);
			if (ot && !valid_tindex_match(ot->data_type, itype)) {
				generic_error("Manager index type mismatch", ind->get_loc());
			}
		}
		else { //index type must be void
			if (itype->get_type()!=DT_VOID)
				generic_error("Required index missing", this->get_loc());
		}

		type = new OSMExpType(
			man_context->get_token_type(sym)->clone(NULL),
			man_context->get_class_name(sym));
	}

	return type;
}

ostream& osm_deref_expression::print(ostream& os) const
{
	os << "*";
	e->print(os);
	print_object_list(os, " ", annots);
	return os;
}

OSMExpType *osm_deref_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	OSMExpType *ot =
		e->infer(man_context, buf_context,
			var_context, mvr_context, def_context);

	if (!ot || ot->exp_type!=OET_TOKEN) {
		generic_error("Dereferencing bad type", this->get_loc());
	}
	else {
		type = new OSMExpType(OET_VAL, ot->data_type->clone(NULL));
	}

	return type;
}

ostream& osm_assign_expression::print(ostream& os) const
{
	e1->print(os);
	os << " = ";
	e2->print(os);
	print_object_list(os, " ", annots);
	return os;
}

OSMExpType *osm_assign_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	OSMExpType *ot1 = e1->infer(man_context,
		buf_context, var_context, mvr_context, def_context);

	OSMExpType *ot2 = e2->infer(man_context,
		buf_context, var_context, mvr_context, def_context);

	/* error */
	if (ot1==NULL || ot2==NULL)
		return NULL;

	if (e2->get_type()==OSM_DEREF && e2->get_data_type()->get_type()==DT_VOID) {
		generic_error("Bad right-side value type", this->get_loc());
		return NULL;
	}

	if (ot1->data_type->is_readonly()) {
		generic_error("Assignment to read only type", this->get_loc());
	}
	else if (ot1->exp_type!=ot2->exp_type) {
		generic_error("Assignment type mismatch", this->get_loc());
	}
	else if (ot1->exp_type==OET_TOKEN) {
		/* token type matching requires that the type be exactly the same */
		if (ot1->mname!=ot2->mname) {
			generic_error("Assignment type mismatch", this->get_loc());
		}
	}
	else if (ot1->exp_type==OET_VAL) {
		/* val type matching is looser on width*/
		if (!valid_osm_assign_match(ot1->data_type, ot2->data_type)) {
			generic_error("Assignment type mismatch", this->get_loc());
		}
	}

	if (e1->get_type()==OSM_BUF_REF || //allocate
		e1->get_type()==OSM_VAR_REF || //read
	  	e1->get_type()==OSM_GROUP) //read
		pre_eval = true;

	type = new OSMExpType(OET_VOID, NULL);
	return type;
}

osm_group_expression::~osm_group_expression()
{
	delete_object_list(elements);
}

ostream& osm_group_expression::print(ostream& os) const
{
	os << "(";
	print_object_list(os, ", ", elements);
	os << ")";
	return os;
}

OSMExpType *osm_group_expression::infer(const manager_context *man_context,
	const buf_context *buf_context, const sym_context *var_context,
	const sym_context *mvr_context, const sym_context *def_context)
{
	// in case this function is called before
	if (type) return type;

	bool readonly = false;

	vector<DataType *> *nt = new vector<DataType *>;
	vector<OSMExpression *>::iterator it;

	for (it=elements->begin(); it!=elements->end(); it++) {

		OSMExpType *et = (*it)->infer(man_context,
			buf_context, var_context, mvr_context, def_context);

		if (et) {
			nt->push_back(et->data_type->clone(NULL));
			if (et->data_type->is_readonly())
				readonly = true;
		}
	}

	type = new OSMExpType(OET_VAL, new DataTypeGroup(nt, NULL));
	type->data_type->set_readonly(readonly);

	return type;
}

void mad_ast::emit_osm_exp_with_cast(const OSMExpression *exp,
	const DataType *type, const string& prefix, ostream& os)
{
	const DataType *from_type = exp->get_data_type();

	if (exp->get_type()==OSM_VAR_REF) {

		const OSMVarRefExpression *vexp =
				static_cast<const OSMVarRefExpression *>(exp);

		assert(valid_arith_type(from_type));

		stringstream sstr;
		vexp->emit_C(sstr, "", prefix);
		emit_str_with_cast(sstr.str(), exp->get_data_type(), type, os);
	}
	else if (exp->get_type()==OSM_NUM) {

		const OSMNumExpression *nexp =
			 	static_cast<const OSMNumExpression *>(exp);

		assert(valid_arith_type(from_type));
		assert(valid_arith_type(type));

		emit_num_of_type(nexp->get_num(), type, os);
	}
	else if (exp->get_type()==OSM_STR) {

		const OSMStrExpression *nexp =
			 	static_cast<const OSMStrExpression *>(exp);

		os << "_STR_CTR(\"" << *(nexp->get_string()) << "\")";
	}
	else if (exp->get_type()==OSM_GROUP) {

		const OSMGroupExpression *gexp =
			 	static_cast<const OSMGroupExpression *>(exp);

		assert(from_type->get_type()==DT_GROUP);
		assert(type->get_type()==DT_GROUP);

		const DataTypeGroup *ftype =
			  	static_cast<const DataTypeGroup *>(from_type);
		const DataTypeGroup *ttype =
			  	static_cast<const DataTypeGroup *>(type);

		const vector<DataType *> *ftypes = ftype->get_data();
		const vector<DataType *> *ttypes = ttype->get_data();
		const vector<OSMExpression *> *children = gexp->get_children();

		assert(ftypes->size()==ttypes->size());
		assert(ttypes->size()>=2);

		assert(ttypes->size()<=sizeof(alphanum)/sizeof(char *));

		os << "_" << multples[ttypes->size()-2] << "_CTR(";

		unsigned int i;
		(*ttypes)[0]->emit_C(os);
		for (i=1; i<ttypes->size(); i++) {
			os << ", ";
			(*ttypes)[i]->emit_C(os);
		}

		for (i=0; i<ttypes->size(); i++) {
			os << ", ";
			emit_osm_exp_with_cast((*children)[i], (*ttypes)[i], prefix, os);
		}

		os << ")";
	}
	else
		assert(0);

}

bool osm_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	assert(0);
	return false;
}

void osm_expression::emit_C_test(ostream& os, const string& prefix) const
{
	assert(0);
}

bool osm_compare_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	return false;
}

void osm_compare_expression::emit_C_test(ostream& os, const string& prefix) const
{
	os << "_COMPARE(";
	emit_osm_exp_with_cast(lhs, sup_type, prefix, os);
	os << ", " << comp_str[cop] << ", ";
	emit_osm_exp_with_cast(rhs, sup_type, prefix, os);
	os << ")";
}

bool osm_var_ref_expression::emit_C(ostream& os,
	const string& indent, const string& prefix) const
{
	emit_sym_of_var_type(sym, vtype, prefix, os);
	return true;
}

bool osm_buf_ref_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	os << BUF_REF(sym);
	return true;
}

bool osm_release_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	os << indent;

	os << "_RELEASE(";

	os << BUF_NAME(buf->get_name());
	os << ");";
	return true;
}

void osm_release_expression::emit_C_test(ostream& os, const string& prefix) const
{
	os << "_TEST_RELEASE(";

	os << BUF_NAME(buf->get_name());
	os << ")";
	//->isTokenGoodToReclaim(machine)";
}

bool osm_discard_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	os << indent;

	os << "_DISCARD(";

	os << BUF_NAME(buf->get_name());
	os << ");";
	return true;
}

void osm_discard_expression::emit_C_test(ostream& os, const string& prefix) const
{
	//no test for discard
	os << "true";
}


/* this is called only for root level man_ref, which means temp allocate */
bool osm_man_ref_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	os << indent;

	if (ind) {
		os << "_ALLOCATE_TEMP_ID(";
		os << MANG_REF(sym);	// manager name
		os << ", ";
		emit_osm_exp_with_cast(ind, itype, prefix, os);	//var value
		os << ");";
	}
	else {

		os << "_ALLOCATE_TEMP(";
		os << MANG_REF(sym);	// manager name
		os << ");";
	}

	return true;
}

void osm_man_ref_expression::emit_C_test(ostream& os,
	const string& prefix) const
{

	if (ind) {
		os << "_TEST_ALLOCATE_ID(";
		os << MANG_REF(sym);	// manager name
		os << ", ";
		emit_osm_exp_with_cast(ind, itype, prefix, os);	//var value
		os << ")";
	}
	else {

		os << "_TEST_ALLOCATE(";
		os << MANG_REF(sym);	// manager name
		os << ")";
	}
}

bool osm_deref_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	// do nothing here
	assert(e->get_type()==OSM_MAN_REF);
	const OSMManRefExpression *child =
		static_cast<const OSMManRefExpression *>(e);

	const OSMExpression *ind = child->get_index();

	os << indent;
	if (ind) {
		os << "_INQUIRE_ID(";
		os << MANG_REF(child->get_name());	// manager name
		os << ", ";
		emit_osm_exp_with_cast(ind, child->get_index_type(), prefix, os);
		os << ");";
	}
	else {

		os << "_INQUIRE(";
		os << MANG_REF(child->get_name());	// manager name
		os << ");";
	}
	return true;
}

void osm_deref_expression::emit_C_test(ostream& os, const string& prefix) const
{

	assert(e->get_type()==OSM_MAN_REF);
	const OSMManRefExpression *child =
		static_cast<const OSMManRefExpression *>(e);

	const OSMExpression *ind = child->get_index();

	if (ind) {
		os << "_TEST_INQUIRE_ID(";
		os << MANG_REF(child->get_name());	// manager name
		os << ", ";
		emit_osm_exp_with_cast(ind, child->get_index_type(), prefix, os);
		os << ")";
	}
	else {

		os << "_TEST_INQUIRE(";
		os << MANG_REF(child->get_name());	// manager name
		os << ")";
	}
}


/** emit a dereference expression as a left hand side expression */
/** vstr is its right hand side stuff. */
void emit_lhs_deref(const OSMDeRefExpression *lexp,
	const string& vstr, const string& prefix, ostream& os)
{
	const osm_expression *child = lexp->get_child();

	if (child->get_type()==OSM_BUF_REF) {

		const OSMBufRefExpression *bref =
				static_cast<const OSMBufRefExpression *>(child);

		os << "_WRITE(";
		os << BUF_NAME(bref->get_name());
	}
	else {	// must be man ref

		assert(child->get_type()==OSM_MAN_REF);
		const OSMManRefExpression *mref =
			static_cast<const OSMManRefExpression *>(child);

		/* the index may be a group, variable, or number */
		const OSMExpression *index = mref->get_index();
		if (index) {

			os << "_WRITE_TOKEN_ID(";
			os << MANG_REF(mref->get_name()) << ", ";
			emit_osm_exp_with_cast(index, mref->get_index_type(), prefix, os);
		}
		else {
			os << "_WRITE_TOKEN(";
			os << MANG_REF(mref->get_name());
		}
	}
	os << ", " <<  vstr << ")";
}

/** emit a dereference expression as a right hand side expression */
void emit_rhs_deref(const OSMDeRefExpression *rexp,
	const string& prefix, ostream& os)
{
	const osm_expression *child = rexp->get_child();

	if (child->get_type()==OSM_BUF_REF) {

		const OSMBufRefExpression *bref =
			static_cast<const OSMBufRefExpression *>(child);

		os << "_READ(" << LOCAL_TEMP << ", "
			<< BUF_NAME(bref->get_name()) << ")";
	}
	else {	// must be man ref

		assert(child->get_type()==OSM_MAN_REF);

		const OSMManRefExpression *mref =
			static_cast<const OSMManRefExpression *>(child);

		/* the index may be a group, variable, or number */
		const OSMExpression *index = mref->get_index();
		if (index) {

			os << "_READ_TOKEN_ID(" << LOCAL_TEMP << ",";
			os << MANG_REF(mref->get_name()) << ", ";
			emit_osm_exp_with_cast(index, mref->get_index_type(), prefix, os);
			os << ")";
		}
		else {
			os << "_READ_TOKEN(" << LOCAL_TEMP << ",";
			os << MANG_REF(mref->get_name()) << ")";
		}
	}
}

bool emit_rhs_deref_with_cast(const OSMDeRefExpression *rexp,
		const DataType *type, const string& indent,
	   	const string& prefix, ostream& os)
{

	os << indent;
	rexp->get_data_type()->emit_C(os);
	os << " " LOCAL_TEMP ";" << endl;

	os << indent;
	emit_rhs_deref(rexp, prefix, os);
	os << ";" << endl;

	/* no cast needed */
	if (*type == *(rexp->get_data_type())) return false;

	/* string type will not reach here */
	if (type->get_type()==DT_GROUP) {

		os << indent;
		type->emit_C(os);
		os << " " LOCAL_RESULT " = ";

		assert(rexp->get_data_type()->get_type()==DT_GROUP);

		const DataTypeGroup *ftype =
			  	static_cast<const DataTypeGroup *>(rexp->get_data_type());
		const DataTypeGroup *ttype =
			  	static_cast<const DataTypeGroup *>(type);

		const vector<DataType *> *ftypes = ftype->get_data();
		const vector<DataType *> *ttypes = ttype->get_data();

		assert(ftypes->size()==ttypes->size());
		assert(ttypes->size()>=2);
		assert(ttypes->size()<=sizeof(alphanum)/sizeof(char *));

		os << "_" << multples[ttypes->size()-2] << "_CTR(";

		unsigned int i;
		(*ttypes)[0]->emit_C(os);
		for (i=1; i<ttypes->size(); i++) {
			os << ", ";
			(*ttypes)[i]->emit_C(os);
		}

		for (i=0; i<ttypes->size(); i++) {
			os << ", ";
			emit_str_with_cast(string(LOCAL_TEMP)+"."+alphanum[i], 
					(*ftypes)[i], (*ttypes)[i], os);
		}

		os << ")" << ";" << endl;
	}
	else if (type->get_type()==DT_BIT || type->get_type()==DT_FLOAT) {
		os << indent;
		type->emit_C(os);
		os << " " LOCAL_RESULT " = ";
		emit_str_with_cast(LOCAL_TEMP, rexp->get_data_type(), type, os);
		os << ";" << endl;
	}
	else
		assert(0);
	
	return true;
}

bool osm_assign_expression::emit_C(ostream&os,
	const string& indent, const string& prefix) const
{
	// if the right side is man_ref => allocation
	if (e2->get_type()==OSM_MAN_REF) {

		assert(e1->get_type()==OSM_BUF_REF);

		OSMManRefExpression *rexp = static_cast<OSMManRefExpression *>(e2);
		OSMBufRefExpression *lexp = static_cast<OSMBufRefExpression *>(e1);

		os << indent;
		os << BUF_REF(lexp->get_name()) << " = ";

		const OSMExpression *ind = rexp->get_index();
		Symbol sym = rexp->get_name();

		if (ind) {
			os << "_ALLOCATE_ID(";
			os << MANG_REF(sym);	// manager name
			os << ", ";
			emit_osm_exp_with_cast(ind, rexp->get_index_type(), prefix, os);	
			os << ");";
		}
		else {

			os << "_ALLOCATE(";
			os << MANG_REF(sym);	// manager name
			os << ");";
		}

		return true;
	}

	// if the left side is var_ref => read 
	if (e1->get_type()==OSM_VAR_REF) {

		assert(e2->get_type()==OSM_DEREF);

		OSMVarRefExpression *lexp = static_cast<OSMVarRefExpression *>(e1);
		OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);

		// start a new c block to localize temporary variables
		os << indent << "{" << endl;

		/* since this is simple variable, we can cast easily */
		bool cst =emit_rhs_deref_with_cast(rexp, lexp->get_data_type(),
				indent, prefix, os);

		os << indent;
		lexp->emit_C(os, "", prefix);

		if (cst)
			os << " = " LOCAL_RESULT ";" << endl;
		else
			os << " = " LOCAL_TEMP ";" << endl;

		// end a c block
		os << indent << "}" << endl;

		return true;
	}

	// if the left side is deref
	if (e1->get_type()==OSM_DEREF) {

		OSMDeRefExpression *lexp = static_cast<OSMDeRefExpression *>(e1);
		const OSMExpression *lchild = lexp->get_child();

		bool use_brace = false;
		stringstream rstr;


		/* the right hand side can be var, num/str, deref, group*/
		if (e2->get_type()==OSM_VAR_REF) {
			emit_osm_exp_with_cast(e2, lchild->get_data_type(), prefix, rstr);
		}
		else if (e2->get_type()==OSM_NUM) {
			emit_osm_exp_with_cast(e2, lchild->get_data_type(), prefix, rstr);
		}
		else if (e2->get_type()==OSM_STR) {
			emit_osm_exp_with_cast(e2, lchild->get_data_type(), prefix, rstr);
		}
		else if (e2->get_type()==OSM_DEREF) {

			// start a new c block to localize temporary variables
			os << indent << "{" << endl;
			OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);
			bool cst = emit_rhs_deref_with_cast(rexp, e1->get_data_type(),
				  	indent, prefix, os);

			if (cst) rstr << LOCAL_RESULT;
			else rstr << LOCAL_TEMP;
			use_brace = true;
		}
		else if (e2->get_type()==OSM_GROUP) {

			// start a new c block to localize temporary variables
			os << indent << "{" << endl;
			emit_osm_exp_with_cast(e2, lchild->get_data_type(), prefix, rstr);
			use_brace = true;
		}
		else assert(0);

		os << indent;
		emit_lhs_deref(lexp, rstr.str(), prefix, os);

		os << ";" << endl;

		// end a c block
		if (use_brace) os << indent << "}" << endl;

		return true;
	}

	// if the left side is group, right must be deref
	if (e1->get_type()==OSM_GROUP) {

		OSMGroupExpression *lexp = static_cast<OSMGroupExpression *>(e1);
		OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);

		const DataType *ltype = lexp->get_data_type();
		const DataType *rtype = rexp->get_data_type();

		// start a new c block to localize temporary variables
		os << indent << "{" << endl;

		os << indent;
		rexp->get_data_type()->emit_C(os);
		os << " " LOCAL_TEMP ";" << endl;

		os << indent;
		emit_rhs_deref(rexp, prefix, os);
		os << ";" << endl;

		assert(ltype->get_type()==DT_GROUP);
		assert(rtype->get_type()==DT_GROUP);

		const vector<DataType *> *ftypes = 
				static_cast<const DataTypeGroup *>(rtype)->get_data();
		const vector<DataType *> *ttypes = 
				static_cast<const DataTypeGroup *>(ltype)->get_data();

		const vector<OSMExpression *> *lchildren = lexp->get_children();

		assert(lchildren->size() == ftypes->size());
		assert(lchildren->size() == ttypes->size());
		assert(lchildren->size()>=2);
		assert(lchildren->size()<=sizeof(alphanum)/sizeof(char *));

		unsigned int i;
		for (i=0; i<lchildren->size(); i++) {

			assert((*lchildren)[i]->get_type()==OSM_VAR_REF);

			os << indent;
			(*lchildren)[i]->emit_C(os, indent, prefix);
			os << " = ";
			emit_str_with_cast(string(LOCAL_TEMP)+"."+alphanum[i], 
					(*ftypes)[i], (*ttypes)[i], os);
			os << ";" << endl;
		}

		// end a c block
		os << indent << "}" << endl;

		return true;
	}

	// unexpected
	assert(0);
	return false;

}

void osm_assign_expression::emit_C_test(ostream& os, const string& prefix) const
{
	// if the right side is man_ref => allocation
	if (e2->get_type()==OSM_MAN_REF) {
		e2->emit_C_test(os, prefix);
		return;
	}

	// if the left side is var_ref => read 
	if (e1->get_type()==OSM_VAR_REF) {

		assert(e2->get_type()==OSM_DEREF);
		OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);
		const osm_expression *rchild = rexp->get_child();

		if (rchild->get_type()==OSM_MAN_REF) {
			e2->emit_C_test(os, prefix);
		}
		else os << "true";

		return;
	}

	
	// if the left side is deref
	if (e1->get_type()==OSM_DEREF) {
		OSMDeRefExpression *lexp = static_cast<OSMDeRefExpression *>(e1);
		const OSMExpression *lchild = lexp->get_child();

		os << "(";

		// man_ref => allocate
		if (lchild->get_type()==OSM_MAN_REF) {
			lchild->emit_C_test(os, prefix);
		}
		else os << "true";

		if (e2->get_type()==OSM_DEREF) {

			OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);
			const OSMExpression *rchild = rexp->get_child();

			if (rchild->get_type()==OSM_MAN_REF) {
				os << " && ";
				rexp->emit_C_test(os, prefix);
			}
		}

		os << ")";
		return;
	}

	// if the left side is group, right must be deref
	if (e1->get_type()==OSM_GROUP) {
		OSMDeRefExpression *rexp = static_cast<OSMDeRefExpression *>(e2);
		const OSMExpression *rchild = rexp->get_child();
		if (rchild->get_type()==OSM_MAN_REF) {
			rexp->emit_C_test(os, prefix);
		}
		else os << "true";
		return;
	}

	assert(0);
}

int osm_compare_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_COMPARE) return 0;

	const OSMCompareExpression *e2 = 
		static_cast<const OSMCompareExpression *>(o2);

	if (cop != e2->cop) return 0;

	int comp1 = lhs->equal(e2->lhs);
	int comp2 = rhs->equal(e2->rhs);

	if (comp1==0 || comp2==0) return 0;
	else if (comp1==1) return comp2;
	else return -1;
}

int osm_var_ref_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_VAR_REF) return 0;
	const OSMVarRefExpression *e2 = 
		static_cast<const OSMVarRefExpression *>(o2);

	if (vtype!=e2->vtype) return 0;
	if (sym!=e2->sym) return 0;

	if (vtype!=VT_VAR && vtype!=VT_ARG) return 1;
	return -1;
}

int osm_group_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_GROUP) return 0;

	const OSMGroupExpression *e2 = 
		static_cast<const OSMGroupExpression *>(o2);

	if (elements->size()!=e2->elements->size()) return 0;

	int ret = 1;
	for (unsigned ii=0; ii<elements->size(); ii++) {

		int temp = (*elements)[ii]->equal((*(e2->elements))[ii]);
		if (temp==0) return 0;
		else if (temp==-1) ret = -1;
	}

	return ret;
}

int osm_buf_ref_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_BUF_REF) return 0;

	const OSMBufRefExpression *e2 = 
		static_cast<const OSMBufRefExpression *>(o2);

	return sym==e2->sym;
}

int osm_num_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_NUM) return 0;

	const OSMNumExpression *e2 = 
		static_cast<const OSMNumExpression *>(o2);

	return *num==*(e2->num);

}

int osm_str_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_STR) return 0;

	const OSMStrExpression *e2 = 
		static_cast<const OSMStrExpression *>(o2);

	return *str==*(e2->str);
}

int osm_release_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_RELEASE) return 0;

	const OSMReleaseExpression *e2 = 
		static_cast<const OSMReleaseExpression *>(o2);

	return buf->equal(e2->buf);
}


int osm_discard_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_DISCARD) return 0;

	const OSMDiscardExpression *e2 = 
		static_cast<const OSMDiscardExpression *>(o2);

	return buf->equal(e2->buf);
}

int osm_man_ref_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_MAN_REF) return 0;

	const OSMManRefExpression *e2 = 
		static_cast<const OSMManRefExpression *>(o2);

	if (sym!=e2->sym) return 0;
	if (ind==NULL) return e2->ind==NULL;
	return ind->equal(e2->ind);
}

int osm_deref_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_DEREF) return 0;

	const OSMDeRefExpression *e2 = 
		static_cast<const OSMDeRefExpression *>(o2);

	return e->equal(e2->e);
}

int osm_assign_expression::equal(const osm_expression *o2) const
{
	if (o2->get_type()!=OSM_ASSIGN) return 0;

	const OSMAssignExpression *ex2 = 
		static_cast<const OSMAssignExpression *>(o2);

	int comp1 = e1->equal(ex2->e1);
	int comp2 = e2->equal(ex2->e2);

	if (comp1==0 || comp2==0) return 0;
	else if (comp1==1) return comp2;
	else return -1;
}
