
#ifndef __MAD_OSM_EXP_AST_HPP__
#define __MAD_OSM_EXP_AST_HPP__

#include <string>
#include <vector>
#include <iostream>
#include "mad_ast.hpp"
#include "annot_ast.hpp"
#include "data_type_ast.hpp"

namespace mad_ast {

	class buf_context;
	class manager_context;

	/** OSM operators. The order matters here for sorting OSM expressions.
	 *  We want to order OSM expressions on an edge list this:
	 *  0. comparison
	 *  1. buf = manref
	 *  2. var/group = deref
	 *     << Insert normal expressions here.
	 *  3. deref = whatever
	 *  4. discard, release.
	 *  Keeping this order will ensure legal dependency preserved.
	 */
	typedef enum {

		OSM_COMPARE,	/**< a compare b */

		OSM_ASSIGN,		/**< a=b */ 

		OSM_DEREF,		/**< *id, *M[] */
		OSM_MAN_REF, 	/**< m[ind], m[], m[num] */

		OSM_RELEASE,	/**< !buffer */
		OSM_DISCARD,	/**< !!buffer */

		OSM_BUF_REF, 	/**< ident */
		OSM_VAR_REF, 	/**< ident */
		OSM_GROUP,		/**< (a,b,...) */

		OSM_NUM, 		/**< number */
		OSM_STR 		/**< string */

	} EnumOSMOpType;

	typedef enum {

		OSM_CMP_LT,		/**< less than */
		OSM_CMP_GT,		/**< greater than */
		OSM_CMP_LTE,	/**< less than or equal to */
		OSM_CMP_GTE,	/**< greater than or equal to */
		OSM_CMP_EQ,		/**< equal to */
		OSM_CMP_NEQ,	/**< not equal to */
			
	} EnumOSMCompareType;

	/** OSM expression types */
	typedef enum {
		OET_VOID,	/**< for root level expression. */
		OET_TOKEN,	/**< type of token,  become VAL on de-reference. */
		OET_VAL,	/**< value type. */
	} EnumOSMExpType;

	/** Type information of an OSM expression */
	typedef struct osm_exp_type {

		/** Constructorm for void and val.
		 *  @param et    Opcode of the expression.
		 *  @param dt    Data type of manager.
		 */
		osm_exp_type(EnumOSMExpType et, DataType *dt) : 
			exp_type(et), data_type(dt) {
			assert(et!=OET_TOKEN);
			}

		/** Constructorm for token type.
		 *  @param dt    Data type of manager.
		 *  @param mname Class name of manager.
		 */
		osm_exp_type(DataType *dt, Symbol mname) : 
			exp_type(OET_TOKEN), data_type(dt), mname(mname) {}

		/** Destructor. */
		~osm_exp_type() {delete data_type;}

		EnumOSMExpType exp_type;
		DataType *data_type;
		Symbol mname;

	} OSMExpType;


	/* Assignment expression left vs. right
	 * lval\rval  ID    ID_LIST  M[]   *M[]  *ID   NUM/STR 
     * ID          x       x      v     v     v     x
	 * ID_LIST     x       x      x     v     v     x
	 * M[]         x       x      x     x     x     x
	 * *M[]        v       v      x     v     v     v
	 * *ID         v       v      x     v     v     v
	 * NUM/STR     x       x      x     x     x     x
	 */

	/** Base class for osm expressions. */
	typedef class osm_expression : public ast_node, public annot_owner {

		public:

			/** Constructor.
			 *  @param op  The operator.
			 *  @param loc Location of the expression.
			 */
			osm_expression(const EnumOSMOpType &op, const Location *loc) :
				ast_node(loc), op(op), type(NULL), pre_eval(false) {}

			/** Virtual destructor for base class. */
			virtual ~osm_expression() {
				delete type;
			}

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			virtual OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context) = 0;

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for variable names.
			 *  @return True if something is output.
			 */
			virtual bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			virtual void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			/** Get the type of the expression. */
			EnumOSMOpType get_type() const {return op;}

			/** Get the data type of the expression. */
			const DataType *get_data_type() const {
				return type?type->data_type:NULL;
			}

			/** See if the expression should be emitted prior to evals.
			 *  Useful only after infer is called.
			 */
			bool is_pre_eval() const {
				return pre_eval;
			}

			/** See if the expression is the same as another one.
			 *  @param o2 The other osm expression.
			 *  @return 1 if equal, -1 if conditionally equal(local VAR has 
			 *   the same name), 0 if not.
			 */
			virtual int equal(const osm_expression *o2) const = 0;


		protected:
			EnumOSMOpType op;
			OSMExpType *type;

			bool pre_eval;	// true if this osm expression should be
							// emitted before normal eval exp. Set by infer.
	} OSMExpression;

	/** Comparison of OSM expressions.
	 *  @param  x  Left operand.
	 *  @param  y  Right operand.
	 *  @return True if x should be placed before y in execution.
	 */
	bool OSMExpPtrCmp(const OSMExpression *x, const OSMExpression *y);

	/** Class for compare expressions. */
	typedef class osm_compare_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param left  Left side of the comparison.
			 *  @param right Right side of the comparison.
			 *  @param op    The comparison operator.
			 *  @param loc   The location of the expression.
			 */
			osm_compare_expression(OSMExpression *left, OSMExpression *right,
				EnumOSMCompareType op, const Location *loc) :
				osm_expression(OSM_COMPARE, loc),
		   		lhs(left), rhs(right), cop(op), sup_type(NULL) {}

			/** Destructor. */
			~osm_compare_expression() {
				delete lhs;
				delete rhs;
				delete sup_type;
			}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			int equal(const osm_expression *o2) const;

		private:
			OSMExpression *lhs, *rhs;
			EnumOSMCompareType cop;

			DataType *sup_type;

	} OSMCompareExpression;

	/** Class for variable reference OSM expression. */
	typedef class osm_var_ref_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the variable.
			 *  @param loc  Location of the expression.
			 */
			osm_var_ref_expression(Symbol sym, const Location *loc) :
				osm_expression(OSM_VAR_REF, loc), sym(sym) {}

			/** Destructor. */
			~osm_var_ref_expression() {}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** Get the name of the variable. */
			Symbol get_name() const {return sym;}

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			int equal(const osm_expression *o2) const;

		private:

			Symbol sym;
			EnumVarType vtype;

	} OSMVarRefExpression;

	/** Class for group OSM expression. */
	typedef class osm_group_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param elems Array of expressions in the group.
			 *  @param loc   Location of the expession.
			 */
			osm_group_expression(std::vector<OSMExpression *> *elems,
				const Location *loc) :
				osm_expression(OSM_GROUP, loc), elements(elems) {}

			/** Destructor. */
			~osm_group_expression();

			/** Print out the expression. */	
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** Get the children nodes */
			const std::vector<OSMExpression *> *get_children() const
		   			{return elements;}

			int equal(const osm_expression *o2) const;

		private:
			std::vector<OSMExpression *> *elements;

	} OSMGroupExpression;

	/** Class for buffer reference OSM expression. */
	typedef class osm_buf_ref_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the buffer.
			 *  @param loc  Location of the expression.
			 */
			osm_buf_ref_expression(Symbol sym, const Location *loc) :
				osm_expression(OSM_BUF_REF, loc), sym(sym) {}

			/** Destructor. */
			~osm_buf_ref_expression() {}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** Get the name of the buffer. */
			Symbol get_name() const {return sym;}

			int equal(const osm_expression *o2) const;

		private:
			Symbol sym;

	} OSMBufRefExpression;

	/** Class for number OSM expression. */
	typedef class osm_num_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param num  String for the number.
			 *  @param loc  Location of the number.
			 */
			osm_num_expression(num_value *num, const Location *loc) :
				osm_expression(OSM_NUM, loc), num(num) {}

			/** Destructor. */
			~osm_num_expression() {delete num;}

			/** Get the number. */
			const NumValue *get_num() const {return num;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			int equal(const osm_expression *o2) const;

		private:
			num_value *num;

	} OSMNumExpression;

	/** Class for number OSM expression. */
	typedef class osm_str_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param s    The string value.
			 *  @param loc  Location of the number.
			 */
			osm_str_expression(std::string *s, const Location *loc) :
				osm_expression(OSM_STR, loc), str(s) {}

			/** Destructor. */
			~osm_str_expression() {delete str;}

			/** Get the string value. */
			const std::string *get_string() const {return str;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			int equal(const osm_expression *o2) const;

		private:
			std::string *str;

	} OSMStrExpression;

	/** Class for token release OSM expression. */
	typedef class osm_release_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param b   Reference to the buffer.
			 *  @param loc Location of the expression.
			 */
			osm_release_expression(osm_buf_ref_expression *b,
				const Location *loc) :
				osm_expression(OSM_RELEASE, loc), buf(b) {}

			/** Destructor. */
			~osm_release_expression() {delete buf;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			int equal(const osm_expression *o2) const;

			/** Get the buffer */
			const osm_buf_ref_expression *get_buf() const {return buf;}

		private:
			osm_buf_ref_expression *buf;

	} OSMReleaseExpression;

	/** Class for token discard OSM expression. */
	typedef class osm_discard_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param b  Reference to the buffer.
			 *  @param loc Location of the expression.
			 */
			osm_discard_expression(osm_buf_ref_expression *b,
				const Location *loc) :
				osm_expression(OSM_DISCARD, loc), buf(b) {}

			/** Destructor. */
			~osm_discard_expression() {delete buf;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			int equal(const osm_expression *o2) const;

			/** Get the buffer */
			const osm_buf_ref_expression *get_buf() const {return buf;}

		private:
			osm_buf_ref_expression *buf;

	} OSMDiscardExpression;

	/** Class for manager reference OSM expression.
	 *  the index can be empty, a number, an identifier, group of identifier*/
	typedef class osm_man_ref_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param sym  Name of the manager.
			 *  @param b    The index expression, can be NULL.
			 *  @param loc  Location of the expression.
			 */
			osm_man_ref_expression(Symbol sym, 
				osm_expression *b,
				const Location *loc) :
				osm_expression(OSM_MAN_REF, loc), sym(sym),
				ind(b), itype(NULL) {}
 
			/** Destructor. */
			~osm_man_ref_expression() {delete ind; delete itype;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			/** Get the name of the manager. */
			Symbol get_name() const {return sym;}

			/** Get the index. */
			const osm_expression *get_index() const {return ind;}

			/** Get the index type, set after calling infer. */
			const DataType *get_index_type() const {return itype;}

			int equal(const osm_expression *o2) const;

		private:
			Symbol sym;
			osm_expression *ind;

			DataType *itype;	/*index type*/

	} OSMManRefExpression;

	/** Class for dereference OSM expression. */
	typedef class osm_deref_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param e   The expression to dereference.
			 *  @param loc Location of the expression.
			 */
			osm_deref_expression(osm_expression *e,
				const Location *loc) :
				osm_expression(OSM_DEREF, loc), e(e) {}

			/** Destructor. */
			~osm_deref_expression() {delete e;}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			/** Get the children node */
			const osm_expression *get_child() const {return e;}

			int equal(const osm_expression *o2) const;

		private:
			osm_expression *e;

	} OSMDeRefExpression;

	/** Class for assignement OSM expression. */
	typedef class osm_assign_expression : public osm_expression {

		public:
			/** Constructor.
			 *  @param e1  The left hand expression.
			 *  @param e2  The right hand expression.
			 *  @param loc Location of the expression.
			 */
			osm_assign_expression(osm_expression *e1, osm_expression *e2,
				const Location *loc) :
				osm_expression(OSM_ASSIGN, loc), e1(e1), e2(e2) {}

			/** Destructor. */
			~osm_assign_expression() {
				delete e1;
			   	delete e2;
			}

			/** Print out the expression. */
			std::ostream& print(std::ostream &) const;

			/** Infer the type of an osm expression, also type checking.
			 * @param  man_context  Manager definitions.
			 * @param  buf_context  Buffere definitions.
			 * @param  var_context  Variable definitions.
			 * @param  mvr_context  Machine variable definitions.
			 * @param  def_context  Constant definitions.
			 */ 
			OSMExpType *infer(const manager_context *man_context,
				   	const buf_context *buf_context,
					const sym_context *var_context,
					const sym_context *mvr_context,
					const sym_context *def_context);

			/** emit C code.
			 *  @param os The output stream.
			 *  @param indent The indentation string.
			 *  @param prefix The prefix for the variable names.
			 *  @return True if something is output.
			 */
			bool emit_C(std::ostream& os,
					const std::string& indent,
					const std::string& prefix) const;

			/** emit C test code.
			 *  @param os     The output stream.
			 *  @param prefix The prefix to variable names.
			 */
			void emit_C_test(std::ostream& os,
				const std::string& prefix) const;

			/** Get the left hand side expression */
			const OSMExpression *get_lhs() const {return e1;}

			/** Get the right hand side expression */
			const OSMExpression *get_rhs() const {return e2;}

			int equal(const osm_expression *o2) const;

		private:
			osm_expression *e1, *e2;

	} OSMAssignExpression;

	void emit_osm_exp_with_cast(const OSMExpression *exp,
		const DataType *type, const std::string& prefix, std::ostream& os);

}

#endif
