%{

#include <vector>
#include <sstream>
#include <cstdlib>
#include <FlexLexer.h>
#include "annotation.hpp"
#include "program_ast.hpp"

#define YYERROR_VERBOSE
#define YYLTYPE mad_ast::yyltype_t
#define YYLTYPE_IS_TRIVIAL 1

#undef yylloc
#define yylloc annot_loc

unsigned annot_counter;
extern int annot_lex();
extern void annot_error(char *);

using namespace mad_ast;
using mad::Symbol;
using mad::SymNULL;
using mad::symbol_find_or_insert;
using std::vector;
using std::string;
using std::cerr;
using std::endl;

extern Symbol annot_filename;
extern AnnotParagraph *annot_parser_result;

#define new_location(a) new Location(annot_filename, (a))

%}

%union {
	char *strval;

	AnnotParagraph *annot_para_t;

	AnnotStatement *annot_statement_t;

	vector<AnnotValue *> *annot_value_list_t;
	AnnotValue *annot_value_t;

	vector<AnnotType *> *annot_type_list_t;
	AnnotType *annot_type_t;

	vector<AnnotArgument *> *annot_argument_list_t;
	AnnotArgument *annot_argument_t;

	AnnotRelOp annot_op_t;

	NumValue *num_value_t;
}

%token '=' '(' ')' '{' '}' '<' '>' ':' ';' ','

%token TK_EQ "=="
%token TK_IN "<:"
%token TK_VAR "VAR"
%token TK_DEF "DEFINE"
%token TK_INT "int"
%token TK_UINT "uint"
%token TK_STRING "string"

%token <strval> TK_IDENT	"identifier"
%token <strval> TK_NUM		"integer"
%token <strval> TK_STR		"string literal"

%type <annot_para_t> annot_para

%type <annot_statement_t> annot_statement

%type <annot_argument_t> argument
%type <annot_argument_list_t> argument_list

%type <annot_value_list_t> annot_value_list
%type <annot_value_t> annot_value

%type <annot_type_t> annot_type
%type <annot_type_list_t> annot_type_list

%type <annot_op_t> annot_op
%type <num_value_t> num_value


%start annot_para

%%

annot_para: ':' TK_IDENT ':' annot_statement {
		@$.propogate(@1, @3);
		$$ = new AnnotParagraph(symbol_find_or_insert($2), new_location(@$));
		$$->append_statement($4);
		annot_parser_result = $$;
annot_counter++;
		free($2);
	}
	| annot_statement {
		@$.propogate(@1, @1);
		$$ = new AnnotParagraph(SymNULL, new_location(@$));
		$$->append_statement($1);
		annot_parser_result = $$;
annot_counter++;
	}
	| annot_para annot_statement {
		$$ = $1;
		$$->append_statement($2);
	}
	;

annot_statement: TK_IDENT ';' {	// annot_command
		@$.propogate(@1, @2);
		$$ = new AnnotStatementCommand(symbol_find_or_insert($1),
			NULL, new_location(@$));
		free($1);
	}
	| TK_IDENT '(' ')' ';' { // annot_command
		@$.propogate(@1, @4);
		vector<annot_argument *> temp_args;
		$$ = new AnnotStatementCommand(symbol_find_or_insert($1),
			NULL, new_location(@$));
		free($1);
	}
	| TK_IDENT '(' argument_list ')' ';' { // annot_command
		@$.propogate(@1, @4);
		$$ = new AnnotStatementCommand(symbol_find_or_insert($1),
			$3, new_location(@$));
		free($1);
	}
	| annot_value annot_op annot_value ';' {	// annot_relationship
		@$.propogate(@1, @4);
		$$ = new AnnotStatementRelation($1, $3, $2, new_location(@$));
	}
	| TK_VAR TK_IDENT ':' annot_type ';' { // declaration
		@$.propogate(@1, @5);
		$$ = new AnnotStatementVar(symbol_find_or_insert($2),
				$4, new_location(@$));
		free($2);
	}
	| TK_DEF TK_IDENT annot_value ';' { // 
		@$.propogate(@1, @4);
		$$ = new AnnotStatementDef(symbol_find_or_insert($2),
				$3, new_location(@$));
		free($2);
	}
	;

annot_value_list: annot_value {
		$$ = new vector<AnnotValue *>;
		$$->push_back($1);
	}
	| annot_value_list ',' annot_value {
		$$ = $1;
		$$->push_back($3);
	}
	;

annot_value: TK_IDENT {
		$$ = new AnnotValueIdent(symbol_find_or_insert($1), new_location(@1));
		free($1);
	}
	| num_value {
		$$ = new AnnotValueInteger($1, new_location(@1));
		free($1);
	}
	| TK_STR {
		// remove the quotation marks
		$1[strlen($1)-1] = 0;
		$$ = new AnnotValueString($1+1, new_location(@1));
		free($1);
	}
	| '{' '}' { //exmpty set
		@$.propogate(@1, @2);
		$$ = new AnnotValueSet(NULL, new_location(@$));
	}
	| '{' annot_value_list '}' { //set, can be exmpty
		@$.propogate(@1, @3);
		$$ = new AnnotValueSet($2, new_location(@$));
	}
	| '(' annot_value_list ')' { //tuple, cannot be empty
		@$.propogate(@1, @3);
		$$ = new AnnotValueTuple($2, new_location(@$));
	}
	;

argument_list: argument {
		$$ = new vector<AnnotArgument *>;
		$$->push_back($1);
	}
	| argument_list ',' argument {
		$$ = $1;
		$1->push_back($3);
	}
	;

argument: TK_IDENT '=' annot_value {
		@$.propogate(@1, @3);
		$$ = new AnnotArgument(symbol_find_or_insert($1), $3,
				new_location(@$));
		free($1);
	}
	;

annot_type: TK_STRING {
		@$.propogate(@1, @1);
		$$ = new AnnotTypeStr(new_location(@$));
	}
	| TK_INT '<' TK_NUM '>' {
		@$.propogate(@1, @4);
		$$ = new AnnotTypeInt(true, strtol($3, NULL, 0), new_location(@$));
		free($3);
	}
	| TK_UINT '<' TK_NUM '>' {
		@$.propogate(@1, @4);
		$$ = new AnnotTypeInt(false, strtol($3, NULL, 0), new_location(@$));
		free($3);
	}
	| '(' annot_type_list ')' {
		@$.propogate(@1, @3);
		$$ = new AnnotTypeTuple($2, new_location(@$));
	}
	| '{' annot_type '}' {
		@$.propogate(@1, @3);
		$$ = new AnnotTypeSet($2, new_location(@$));
	}
	;

annot_type_list: annot_type {
		$$ = new vector<annot_type *>;
		$$->push_back($1);
	}
	| annot_type_list ',' annot_type {
		$$ = $1;
		$$->push_back($3);
	}
	;

annot_op: TK_EQ {
		$$ = ANNOT_REL_EQ;
	}
	| TK_IN {
		$$ = ANNOT_REL_IN;
	}
	;

num_value: TK_NUM {
		$$ = new num_value(new string($1), false, false);
		free($1);
	}
	| '+' TK_NUM {
		$$ = new num_value(new string($2), false, false);
		free($2);
	}
	| '-' TK_NUM {
		$$ = new num_value(new string($2), true, false);
		free($2);
	}

%%
