%{

#include <vector>
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include "program_ast.hpp"
#include "data_type_ast.hpp"
#include "define_ast.hpp"
#include "manager_ast.hpp"
#include "machine_ast.hpp"
#include "osm_exp_ast.hpp"
#include "exp_ast.hpp"
#include "machine_ast.hpp"
#include "operation_ast.hpp"
#include "cstdlib"
#include "misc.hpp"
#include <FlexLexer.h>

#define YYERROR_VERBOSE
#define YYLTYPE mad_ast::yyltype_t
#define YYLTYPE_IS_TRIVIAL 1

#undef yylloc
#define yylloc madl_loc

using namespace mad_ast;
using mad::Symbol;
using mad::SymNULL;
using mad::symbol_find_or_insert;
using std::vector;
using std::string;
using std::cerr;
using std::cout;
using std::endl;
using std::string;
using std::ifstream;
using std::ofstream;

#define new_location(a) new Location(madl_filename, (a))

/* globals shared across parsers */
extern int madl_lex();
extern void madl_error(char *);

/* some globals for parsing madl program */
extern Symbol madl_filename;
extern Symbol madl_machname;
extern vector<ProgSection *> *madl_parser_result;
extern AnnotOwner *cur_annot_owner;

%}

%union {

	char *strval;

	vector<string *> *string_list_t;

	vector<Symbol> *symbol_list_t;
	Symbol symbol_t;

	vector<ProgSection *> *prog_section_list_t;
	ProgSection *prog_section_t;

	DataType *data_type_t;
	vector<DataType *> *data_type_list_t;

	DataTypeVoid *data_type_void_t;
	DataTypeBit *data_type_bit_t;
	DataTypeString *data_type_string_t;
	DataTypeFloat *data_type_float_t;
	DataTypeGroup *data_type_group_t;
	DataTypeOper *data_type_oper_t;

	MachineSection *machine_section_t;
	vector<MachineParagraph *> *machine_paragraph_list_t;
	MachineParagraph *machine_paragraph_t;

	BufferParagraph *buffer_para_t;
	vector<BufferClause *> *buffer_clause_list_t;
	BufferClause *buffer_clause_t;

	EdgeParagraph *edge_para_t;
	vector<EdgeClause *> *edge_clause_list_t;
	EdgeClause *edge_clause_t;

	StateParagraph *state_para_t;
	StateClause *state_clause_t;
	vector<StateClause *> *state_clause_list_t;

	InitialParagraph *initial_para_t;
	MVarParagraph *mvar_para_t;

	OperationSection *operation_section_t;
	vector<OperationParagraph *> *operation_paragraph_list_t;
	OperationParagraph *operation_paragraph_t;

	FunctionSection *function_section_t;

	ArgClause *arg_clause_t;
	vector<ArgClause *> *arg_clause_list_t;

	VarParagraph *var_para_t;
	VarClause *var_clause_t;
	vector<VarClause *> *var_clause_list_t;

	CodingParagraph *coding_para_t;
	vector<CodingClause *> *coding_clause_list_t;
	CodingClause *coding_clause_t;

	CodingClausePattern *coding_clause_pattern_t;
	CodingClauseVarRef *coding_clause_var_ref_t;
	CodingClauseOrPatterns *coding_clause_or_patterns_t;
	vector<DecodePattern *> *pattern_list_t;

	SyntaxParagraph *syntax_para_t;
	vector<SyntaxClause *> *syntax_clause_list_t;
	SyntaxClause *syntax_clause_t;

	SyntaxClauseLiteral *syntax_clause_literal_t;
	SyntaxClauseVarRef *syntax_clause_var_ref_t;
	SyntaxClauseTabRef *syntax_clause_tab_ref_t;
	SyntaxClauseOrLiterals *syntax_clause_or_literals_t;
	SyntaxClauseOrTabRefs *syntax_clause_or_tab_refs_t;

	vector<SyntaxTableLookup *> *syntax_clause_lookup_list_t;
	SyntaxTableLookup *syntax_clause_lookup_t;

	ManagerSection *manager_section_t;
	vector<ManagerClassClause *> *manager_class_list_t;
	ManagerClassClause *manager_class_clause_t;
	vector<ManagerInstClause *> *manager_inst_list_t;
	ManagerInstClause *manager_inst_clause_t;
	
	ImportSection *import_section_t;
	UsingSection  *using_section_t;
	AnnotationSection *annot_section_t;

	DefineSection *define_section_t;
	vector<DefineClause *> *define_clause_list_t;
	DefineClause *define_clause_t;

	DefineValue *define_value_t;

	vector<NumValue *> *num_value_list_t;
	NumValue *num_value_t;

	DefineValueNum *define_value_num_t;
	DefineValueStr *define_value_str_t;
	DefineValueNumArray *define_value_num_array_t;
	DefineValueStrArray *define_value_str_array_t;

	EvalParagraph *eval_para_t;
	vector<Expression *> *exp_list_t;
	Expression *exp_t;

	TransParagraph *trans_para_t;
	vector<TransClause *> *trans_clause_list_t;
	TransClause *trans_clause_t;
	vector<OSMExpression *> *osm_exp_list_t;
	OSMExpression *osm_exp_t;

	EnumOSMCompareType osm_comp_t;
}

%nonassoc '='
%left  '?' ':'
%left  '|'
%left  '^'
%left  '&'
%left  TK_EQ TK_NEQ 
%left  '<' TK_LTE '>' TK_GTE
%left  TK_RSH TK_LSH
%left  '+' '-'
%left  '*' '/' '%'
%left  TK_CONCAT
%right PS_NEG '~'
%right PS_CAST
%nonassoc '[' '.'

%token TK_EQ	"=="
%token TK_NEQ	"!="
%token TK_LTE	"<="
%token TK_GTE	">="
%token TK_RSH	">>"
%token TK_LSH	"<<"
%token TK_DISCARD	"!!"

%token TK_IMPORT	"IMPORT"
%token TK_USING		"USING"
%token TK_DEFINE	"DEFINE"
%token TK_MANAGER	"MANAGER"
%token TK_CLASS		"CLASS"
%token TK_INSTANCE	"INSTANCE"

%token TK_MACHINE "MACHINE"
%token TK_INITIAL "INITIAL"
%token TK_STATE   "STATE"
%token TK_EDGE    "EDGE"
%token TK_BUFFER  "BUFFER"

%token TK_OPERATION	"OPERATION"
%token TK_CODING	"CODING"
%token TK_SYNTAX	"SYNTAX"
%token TK_VAR		"VAR"
%token TK_EVAL		"EVAL"
%token TK_TRANS		"TRANS"
%token TK_FUNCTION	"FUNCTION"
%token TK_ANNOTATION "ANNOTATION"

%token TK_VOID		"void"
%token TK_UINT		"uint"
%token TK_INT		"int"
%token TK_FLOAT		"float"
%token TK_DOUBLE	"double"
%token TK_STRING	"string"

%token TK_TOWARD	"->"

%token TK_CONCAT	"::"

%token <strval> TK_IDENT	"identifier"
%token <strval> TK_NUM		"integer"
%token <strval> TK_FNUM		"float-point number"
%token <strval> TK_PAT		"boolean pattern"
%token <strval> TK_STR		"string literal"

%type  <data_type_void_t> data_type_void;
%type  <data_type_bit_t> data_type_bit;
%type  <data_type_string_t> data_type_string;
%type  <data_type_float_t> data_type_float;
%type  <data_type_group_t> data_type_group;
%type  <data_type_oper_t> data_type_oper;

%type  <operation_section_t> operation_section operation_head
%type  <operation_paragraph_list_t> operation_paragraph_list
%type  <operation_paragraph_t> operation_paragraph

%type  <function_section_t> function_section function_section_proto 
%type  <function_section_t> function_head

%type  <coding_para_t> coding_para
%type  <coding_clause_t> coding_clause
%type  <coding_clause_list_t> coding_clause_list

%type  <coding_clause_pattern_t> coding_clause_pattern;
%type  <coding_clause_var_ref_t> coding_clause_var_ref;
%type  <coding_clause_or_patterns_t> coding_clause_or_patterns;

%type  <pattern_list_t> coding_disjunct_patterns

%type  <syntax_para_t> syntax_para
%type  <syntax_clause_t> syntax_clause
%type  <syntax_clause_list_t> syntax_clause_list
%type  <string_list_t> syntax_disjunct_strings
%type  <syntax_clause_lookup_list_t> syntax_disjunct_lookups
%type  <syntax_clause_lookup_t> syntax_tab_ref

%type  <syntax_clause_literal_t> syntax_clause_literal;
%type  <syntax_clause_var_ref_t> syntax_clause_var_ref;
%type  <syntax_clause_tab_ref_t> syntax_clause_tab_ref;
%type  <syntax_clause_or_literals_t> syntax_clause_or_literals;
%type  <syntax_clause_or_tab_refs_t> syntax_clause_or_tab_refs;

%type  <arg_clause_list_t> arg_clause_list
%type  <arg_clause_t> arg_clause

%type  <mvar_para_t> mvar_para
%type  <var_para_t> var_para
%type  <var_clause_list_t> var_clause_list
%type  <var_clause_t> var_clause

%type  <data_type_t> data_type arg_type
%type  <data_type_list_t> data_type_list arg_type_list

%type  <symbol_list_t> ident_list

%type  <string_list_t> string_list

%type  <machine_section_t> machine_section machine_head
%type  <machine_paragraph_list_t> machine_paragraph_list
%type  <machine_paragraph_t>  machine_paragraph

%type  <buffer_para_t> buffer_para
%type  <buffer_clause_t> buffer_clause
%type  <buffer_clause_list_t>  buffer_clause_list

%type  <edge_para_t> edge_para
%type  <edge_clause_t> edge_clause 
%type  <edge_clause_list_t> edge_clause_list

%type  <manager_section_t> manager_section 
%type  <manager_class_list_t> manager_class_list manager_class_para
%type  <manager_class_clause_t>  manager_class_clause
%type  <manager_inst_list_t> manager_inst_list manager_inst_para
%type  <manager_inst_clause_t>  manager_inst_clause

%type  <initial_para_t> initial_para

%type  <state_para_t> state_para
%type  <state_clause_t> state_clause
%type  <state_clause_list_t> state_clause_list

%type  <prog_section_list_t> prog prog_section_list
%type  <prog_section_t> prog_section

%type  <import_section_t>  import_section import_head
%type  <using_section_t>   using_section using_head
%type  <annot_section_t>   annot_section annot_head

%type  <num_value_t> num_value
%type  <num_value_list_t> num_value_list
%type  <define_section_t> define_section
%type  <define_clause_list_t> define_clause_list
%type  <define_clause_t> define_clause
%type  <define_value_t> define_value

%type  <define_value_num_t> define_value_num
%type  <define_value_str_t> define_value_str
%type  <define_value_num_array_t> define_value_num_array
%type  <define_value_str_array_t> define_value_str_array

%type  <eval_para_t> eval_para
%type  <exp_list_t> exp_list root_exp_list
%type  <exp_t> exp root_exp tabref_exp

%type  <trans_para_t> trans_para
%type  <trans_clause_list_t> trans_clause_list
%type  <trans_clause_t> trans_clause
%type  <osm_exp_list_t> osm_exp_list osm_exp_clause osm_group_list
%type  <osm_exp_t> osm_exp osm_rval osm_man_ref osm_dref osm_value
%type  <osm_comp_t> osm_comp

%start prog

%%

prog:  { /* empty program is fine. */
		madl_parser_result = new vector<ProgSection *>;
	}
	| prog_section_list {
		madl_parser_result = $$;
	}
	;

prog_section_list: prog_section {
		$$ = new vector<ProgSection *>;
		$$->push_back($1);
	}
	| prog_section_list prog_section {
		$$ = $1;
		$$->push_back($2);
	}
	;

prog_section: import_section {
		$$ = $1;
	}
	| using_section {
		$$ = $1;
	}
	| annot_section {
		$$ = $1;
	}
	| define_section {
		$$ = $1;
	}
	| manager_section {
		$$ = $1;
	}
	| machine_section {
		$$ = $1;
	}
	| operation_section {
		$$ = $1;
	}
	| function_section {
		$$ = $1;
	}
	;

import_section: import_head TK_STR ';' {
		@$.propogate(@1, @3);
		$$ = $1;
		$$->set_file($2);
		$$->set_loc(new_location(@$));
		free($2);
	}
	;

import_head: TK_IMPORT {
		@$.propogate(@1, @1);
		$$ = new ImportSection();
		cur_annot_owner = $$;
	}
	;

using_section: using_head TK_IDENT ';' {
		@$.propogate(@1, @3);
		$$ = $1;
		madl_machname = symbol_find_or_insert($2);
		$$->set_name(madl_machname);
		$$->set_loc(new_location(@$));
		free($2);
	}
	;

using_head: TK_USING {
		@$.propogate(@1, @1);
		$$ = new UsingSection();
		cur_annot_owner = $$;
	}
	;

annot_section: annot_head ';' {
		$$ = $1;
		$$->set_loc(new_location(@$));
	}
	;

annot_head: TK_ANNOTATION {
		@$.propogate(@1, @1);
		$$ = new AnnotationSection();
		cur_annot_owner = $$;
	}
	;

define_section: TK_DEFINE define_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new DefineSection($2, new_location(@$));
	}
	;

define_clause_list: define_clause {
		$$ = new vector<DefineClause *>;
		$$->push_back($1);
	}
	| define_clause_list ';' define_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

define_clause: TK_IDENT ':' data_type_bit '=' define_value {
		@$.propogate(@1, @5);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, $5,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type_string '=' define_value {
		@$.propogate(@1, @5);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, $5,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type_float '=' define_value {
		@$.propogate(@1, @5);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, $5,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type_bit {
		@$.propogate(@1, @3);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, NULL,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type_string {
		@$.propogate(@1, @3);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, NULL,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type_float {
		@$.propogate(@1, @3);
		$$ = new DefineClause(symbol_find_or_insert($1), $3, NULL,
			new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' '(' arg_type_list ')' {
		@$.propogate(@1, @5);
		$$ = new DefineClause(symbol_find_or_insert($1),
			new DataTypeGroup($4, new_location(@4)), NULL, new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	;

arg_type_list: arg_type {
		@$.propogate(@1, @1);
		$$ = new vector<DataType *>;
		$$->push_back($1);
	}
	| arg_type_list ',' arg_type {
		@$.propogate(@1, @3);
		$$ = $1;
		$$->push_back($3);
	}
	;

arg_type: data_type {
		@$.propogate(@1, @1);
		$$ = $1;
		$$->set_readonly(true);
	}
	| data_type '*' {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	;

define_value: define_value_num {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| define_value_str {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| define_value_num_array {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| define_value_str_array {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	;

define_value_num: num_value {
		@$.propogate(@1, @1);
		$$ = new DefineValueNum($1, new_location(@$));
	}
	;

define_value_str: TK_STR {
		@$.propogate(@1, @1);
		$$ = new DefineValueStr(new string($1), new_location(@$));
		free($1);
	}
	;

define_value_num_array: '{' num_value_list '}' {
		@$.propogate(@1, @3);
		$$ = new DefineValueNumArray($2, new_location(@$));
	}
	;

define_value_str_array: '{' string_list '}' {
		@$.propogate(@1, @3);
		$$ = new DefineValueStrArray($2, new_location(@$));
	}
	;

num_value_list: num_value {
		$$ = new vector<NumValue *>;
		$$->push_back($1);
	}
	| num_value_list ',' num_value {
		$$ = $1;
		$$->push_back($3);
	}
	;

num_value: TK_NUM {
		$$ = new num_value(new string($1), false, false);
		free($1);
	}
	| '+' TK_NUM {
		$$ = new num_value(new string($2), false, false);
		free($2);
	}
	| '-' TK_NUM {
		$$ = new num_value(new string($2), true, false);
		free($2);
	}
	| TK_FNUM {
		$$ = new num_value(new string($1), false, true);
		free($1);
	}
	| '+' TK_FNUM {
		$$ = new num_value(new string($2), false, true);
		free($2);
	}
	| '-' TK_FNUM {
		$$ = new num_value(new string($2), true, true);
		free($2);
	}
	;

string_list: TK_STR {	// string list with only ',' as separator
		$$ = new vector<string *>;
		$$->push_back(new string($1));
		free($1);
	}
	| string_list ',' TK_STR {
		$$ = $1;
		$$->push_back(new string($3));
		free($3);
	}
	;

machine_section: machine_head TK_IDENT machine_paragraph_list {
		@$.propogate(@1, @3);
		$$ = $1;
		$$->set_name(symbol_find_or_insert($2));
		$$->set_paragraphs($3);
		$$->set_loc(new_location(@$));
		free($2);
	}
	;

machine_head: TK_MACHINE {
		@$.propogate(@1, @1);
		$$ = new MachineSection();
		cur_annot_owner = $$;
	}
	;

machine_paragraph_list: machine_paragraph {
		@$.propogate(@1, @1);
		$$ = new vector<MachineParagraph *>;
		$$->push_back($1);
	}
	| machine_paragraph_list machine_paragraph {
		@$.propogate(@1, @2);
		$$ = $1;
		$$->push_back($2);
	}
	;

machine_paragraph: initial_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| state_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| edge_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| buffer_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| mvar_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	;

mvar_para: TK_VAR var_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new MVarParagraph($2, new_location(@$));
	}
	;

state_clause: TK_IDENT {
		$$ = new StateClause(symbol_find_or_insert($1), new_location(@1));
		free($1);
		cur_annot_owner = $$;
	}
	;
 
initial_para: TK_INITIAL state_clause ';' {
		@$.propogate(@1, @3);
		$$ = new InitialParagraph($2, new_location(@$));
	}
	;

state_para: TK_STATE state_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new StateParagraph($2, new_location(@$));
	}
	;

state_clause_list: state_clause {
		$$ = new vector<StateClause *>;
		$$->push_back($1);
	}
	| state_clause_list ',' state_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

edge_para: TK_EDGE edge_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new EdgeParagraph($2, new_location(@$));
	}
	;

edge_clause_list: edge_clause {
		$$ = new vector<EdgeClause *>;
		$$->push_back($1);
	}
	| edge_clause_list ';' edge_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

edge_clause: TK_IDENT ':' TK_IDENT TK_TOWARD TK_IDENT {
		@$.propogate(@1, @5);
		$$ = new EdgeClause(symbol_find_or_insert($1),
			 symbol_find_or_insert($3), symbol_find_or_insert($5),
			 new_location(@$));
		free($1);
		free($3);
		free($5);
		cur_annot_owner = $$;
	}
	;

buffer_para: TK_BUFFER buffer_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new BufferParagraph($2, new_location(@$));
	}
	;

buffer_clause_list: buffer_clause {
		$$ = new vector<BufferClause *>;
		$$->push_back($1);
	}
	| buffer_clause_list ';' buffer_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

buffer_clause: TK_IDENT ':' TK_IDENT {
		@$.propogate(@1, @3);
		$$ = new BufferClause(symbol_find_or_insert($1),
				symbol_find_or_insert($3), new_location(@$));
		free($1);
		free($3);
		cur_annot_owner = $$;
	}
	;

manager_section: TK_MANAGER manager_class_para manager_inst_para {
		@$.propogate(@1, @3);
		$$ = new ManagerSection($2, $3, new_location(@$));
	}
	;

manager_class_para: TK_CLASS manager_class_list ';' {
		$$ = $2;
	}
	;

manager_class_list: manager_class_clause {
		$$ = new vector<ManagerClassClause *>;
		$$->push_back($1);
	}
	| manager_class_list ';' manager_class_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

manager_class_clause: TK_IDENT ':' data_type TK_TOWARD data_type {
		@$.propogate(@1, @5);
		$$ = new ManagerClassClause(symbol_find_or_insert($1), $3, $5,
				new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	;

manager_inst_para: TK_INSTANCE manager_inst_list ';' {
		$$ = $2;
	}
	;

manager_inst_list: manager_inst_clause {
		$$ = new vector<ManagerInstClause *>;
		$$->push_back($1);
	}
	| manager_inst_list ';' manager_inst_clause {
		$$ = $1;
		$$->push_back($3);
	}
	;

manager_inst_clause: TK_IDENT ':' TK_IDENT {
		@$.propogate(@1, @3);
		$$ = new ManagerInstClause(symbol_find_or_insert($1), 
				symbol_find_or_insert($3), new_location(@$));
		free($1);
		free($3);
		cur_annot_owner = $$;
	}
	;

operation_section: operation_head TK_IDENT operation_paragraph_list {
		@$.propogate(@1, @3);
		$$ = $1;
		$$->set_name(symbol_find_or_insert($2));
		$$->set_paragraphs($3);
		$$->set_loc(new_location(@$));
		free($2);
	}
	;

operation_head: TK_OPERATION {
		@$.propogate(@1, @1);
		$$ = new OperationSection(madl_machname);
		cur_annot_owner = $$;
	}
	;

operation_paragraph_list: operation_paragraph {
		@$.propogate(@1, @1);
		$$ = new vector<OperationParagraph *>;
		$$->push_back($1);
	}
	| operation_paragraph_list operation_paragraph {
		@$.propogate(@1, @2);
		$$ = $1;
		$$->push_back($2);
	}
	;

operation_paragraph: var_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| coding_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| syntax_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| eval_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| trans_para {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	;

arg_clause_list: arg_clause {
		$$ = new vector<ArgClause *>;
		$$->push_back($1);
	}
	| arg_clause_list ',' arg_clause {
		$$ = $1;
		$$->push_back($3);
	}
	; 

arg_clause: TK_IDENT ':' data_type {
		@$.propogate(@1, @3);
		$$ = new ArgClause(symbol_find_or_insert($1), $3, true,
				new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT ':' data_type '*' {
		@$.propogate(@1, @4);
		$$ = new ArgClause(symbol_find_or_insert($1), $3, false,
				new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	;

function_section: function_section_proto operation_paragraph_list {
		@$.propogate(@1, @2);
		$$ = $1;
		$$->set_paragraphs($2);
		$$->set_loc(new_location(@$));
	}
	;

function_section_proto: function_head TK_IDENT '(' arg_clause_list ')' {
		@$.propogate(@1, @5);
		$$ = $1;
		$$->set_name(symbol_find_or_insert($2));
		$$->set_args($4);
		free($2);
		cur_annot_owner = $$;
	}
	;

function_head: TK_FUNCTION {
		@$.propogate(@1, @1);
		$$ = new FunctionSection();
		cur_annot_owner = $$;
	}
	;

var_para: TK_VAR var_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new VarParagraph($2, new_location(@$));
	}
	;

var_clause_list: var_clause {
		$$ = new vector<VarClause *>;
		$$->push_back($1);
		cur_annot_owner = $1;
	}
	| var_clause_list ';' var_clause {
		$$ = $1;
		$$->push_back($3);
		cur_annot_owner = $3;
	}
	; 

var_clause: TK_IDENT ':' data_type {
		@$.propogate(@1, @3);
		$$ = new VarClause(symbol_find_or_insert($1), $3,
				new_location(@$));
		free($1);
	}
	;

data_type_list: data_type ',' data_type {
		$$ = new vector<DataType *>;
		$$->push_back($1);
		$$->push_back($3);
	}
	| data_type_list ',' data_type {
		$$ = $1;
		$$->push_back($3);
	}
	;

data_type: data_type_void {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	| data_type_bit {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	| data_type_float {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	| data_type_string {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	| data_type_group {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	| data_type_oper {
		@$.propogate(@1, @1);
		$$ =  $1;
	}
	;
	

data_type_void: TK_VOID {
		@$.propogate(@1, @1);
		$$ = new data_type_void(new_location(@$));
	}
	;

data_type_bit: TK_UINT '<' TK_NUM '>' {
		@$.propogate(@1, @4);
		$$ = new data_type_bit(false, strtol($3, NULL, 0), 0,
				new_location(@$));
		free($3);
	}
	| TK_UINT '<' TK_NUM '>' '[' TK_NUM ']' {
		@$.propogate(@1, @7);
		$$ = new data_type_bit(false, strtol($3, NULL, 0),
				strtol($6, NULL, 0), new_location(@$));
		free($3);
		free($6);
	}
	| TK_INT '<' TK_NUM '>' {
		@$.propogate(@1, @4);
		$$ = new data_type_bit(true, strtol($3, NULL, 0), 0,
				new_location(@$));
		free($3);
	}
	| TK_INT '<' TK_NUM '>' '[' TK_NUM ']' {
		@$.propogate(@1, @7);
		$$ = new data_type_bit(true, strtol($3, NULL, 0),
				strtol($6, NULL, 0), new_location(@$));
		free($3);
		free($6);
	}
	;

data_type_float: TK_FLOAT {
		@$.propogate(@1, @1);
		$$ = new data_type_float(false, 0, new_location(@$));
	}
	| TK_FLOAT '[' TK_NUM ']' {
		@$.propogate(@1, @4);
		$$ = new data_type_float(false, strtol($3, NULL, 0), new_location(@$));
		free($3);
	}
	| TK_DOUBLE {
		@$.propogate(@1, @1);
		$$ = new data_type_float(true, 0, new_location(@$));
	}
	| TK_DOUBLE '[' TK_NUM ']' {
		@$.propogate(@1, @4);
		$$ = new data_type_float(true, strtol($3, NULL, 0), new_location(@$));
		free($3);
	}
	;

data_type_string: TK_STRING {
		@$.propogate(@1, @1);
		$$ = new data_type_string(0, new_location(@$));
	}
	| TK_STRING '[' TK_NUM ']' {
		@$.propogate(@1, @4);
		$$ = new data_type_string(strtol($3, NULL, 0), new_location(@$));
		free($3);
	}
	;

data_type_group: '(' data_type_list ')' {
		@$.propogate(@1, @3);
		$$ = new data_type_group($2, new_location(@$));
	}
	;

data_type_oper: '{' ident_list '}' {
		@$.propogate(@1, @3);
		$$ = new data_type_oper($2, SymNULL, new_location(@$));
	}
	| '{' ident_list '}' '(' TK_IDENT ')' {
		@$.propogate(@1, @6);
		$$ = new data_type_oper($2, symbol_find_or_insert($5),
			 new_location(@$));
		free($5);
	}
	;

ident_list: TK_IDENT {
		$$ = new vector<Symbol>;
		$$->push_back(symbol_find_or_insert($1));
		free($1);
	}
	| ident_list ',' TK_IDENT {
		$$ = $1;
		$$->push_back(symbol_find_or_insert($3));
		free($3);
	}
	;

syntax_para: TK_SYNTAX syntax_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new SyntaxParagraph($2, new_location(@$));
	}
	;

syntax_clause_list: syntax_clause {
		$$ = new vector<SyntaxClause *>;
		$1->set_follow(true); /* the first one should follow */
		$1->set_first(true); /* the first one */
		$$->push_back($1);
	}
	| syntax_clause_list syntax_clause {
		$$ = $1;
		$1->push_back($2);
	}
	| syntax_clause_list '^' syntax_clause {
		$$ = $1;
		$3->set_follow(true);
		$1->push_back($3);
	}
	;

syntax_clause: syntax_clause_literal {
		$$ = $1;
	}
	| syntax_clause_var_ref {
		$$ = $1;
	}
	| syntax_clause_tab_ref {
		$$ = $1;
	}
	| syntax_clause_or_literals {
		$$ = $1;
	}
	| syntax_clause_or_tab_refs {
		$$ = $1;
	}
	;

syntax_clause_literal: TK_STR {
		$$ = new syntax_clause_literal(new string($1), new_location(@1));
		free($1);
		cur_annot_owner = $$;
	}
	;

syntax_clause_var_ref: TK_IDENT {
		$$ = new syntax_clause_var_ref(symbol_find_or_insert($1), SymNULL,
			new_location(@1));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT '.' TK_IDENT {
		@$.propogate(@1, @3);
		lower_case($3);
		$$ = new syntax_clause_var_ref(symbol_find_or_insert($1),
			symbol_find_or_insert($3), new_location(@$));
		free($1);
		free($3);
		cur_annot_owner = $$;
	}
	;

syntax_clause_tab_ref: syntax_tab_ref {
		$$ = new syntax_clause_tab_ref($1, new_location(@1));
		cur_annot_owner = $$;
	}
	;

syntax_clause_or_literals: '(' syntax_disjunct_strings ')' {
		@$.propogate(@1, @3);
		$$ = new syntax_clause_or_literals($2, new_location(@$));
		cur_annot_owner = $$;
	}
	;

syntax_clause_or_tab_refs: '(' syntax_disjunct_lookups ')' {
		@$.propogate(@1, @3);
		$$ = new syntax_clause_or_tab_refs($2, new_location(@$));
		cur_annot_owner = $$;
	}
	;

syntax_disjunct_strings: TK_STR '|' TK_STR {
		$$ = new vector<string *>;
		$$->push_back(new string($1));
		$$->push_back(new string($3));
		free($1);
		free($3);
	}
	| syntax_disjunct_strings '|' TK_STR {
		$$ = $1;
		$$->push_back(new string($3));
		free($3);
	}
	;

syntax_disjunct_lookups: syntax_tab_ref '|' syntax_tab_ref {
		$$ = new vector<SyntaxTableLookup *>;
		$$->push_back($1);
		$$->push_back($3);
	}
	| syntax_disjunct_lookups '|' syntax_tab_ref {
		$$ = $1;
		$$->push_back($3);
	}
	;

syntax_tab_ref: TK_IDENT '[' TK_IDENT ']' {
		@$.propogate(@1, @4);
		$$ = new SyntaxTableLookup(symbol_find_or_insert($1),
			symbol_find_or_insert($3), SymNULL, new_location(@$));
		free($1);
		free($3);
	}
	| TK_IDENT '[' TK_IDENT ']' '.' TK_IDENT {
		@$.propogate(@1, @6);
		$$ = new SyntaxTableLookup(symbol_find_or_insert($1),
			symbol_find_or_insert($3), symbol_find_or_insert($6),
			new_location(@$));
		free($1);
		free($3);
		free($6);
	}
	;

coding_para:	TK_CODING coding_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new CodingParagraph($2, new_location(@$));
	}
	;

coding_clause_list: coding_clause {
		$$ = new vector<CodingClause *>;
		$$->push_back($1); 
	}
	| coding_clause_list coding_clause {
		$$ = $1;
		$$->push_back($2); 
	}
	;

coding_clause: coding_clause_pattern {
		$$ = $1;
	}
	| coding_clause_var_ref {
		$$ = $1;
	}
	| coding_clause_or_patterns {
		$$ = $1;
	}
	;

coding_clause_pattern: TK_PAT {
		$$ = new coding_clause_pattern(new DecodePattern($1), new_location(@1));
		free($1);
		cur_annot_owner = $$;
	}
	;

coding_clause_var_ref: TK_IDENT {
		$$ = new coding_clause_var_ref(symbol_find_or_insert($1),
			new_location(@1));
		free($1);
		cur_annot_owner = $$;
	}
	;

coding_clause_or_patterns: '(' coding_disjunct_patterns ')' {
		@$.propogate(@1, @3);
		$$ = new coding_clause_or_patterns($2, new_location(@$));
		cur_annot_owner = $$;
	}
	;

coding_disjunct_patterns: TK_PAT '|' TK_PAT {
		$$=new vector<DecodePattern *>;
		$$->push_back(new DecodePattern($1));
		$$->push_back(new DecodePattern($3));
		free($1);
		free($3);
	}
	| coding_disjunct_patterns '|' TK_PAT {
		$$ = $1;
		$$->push_back(new DecodePattern($3));
		free($3);
	}
	;

eval_para: TK_EVAL root_exp_list ';' {
		@$.propogate(@1, @3);
		$$ = new EvalParagraph($2, new_location(@$));
	}
	;

root_exp_list: root_exp {
		$$ = new vector<Expression *>;
		$$->push_back($1);
	}
	| root_exp_list ';' root_exp {
		$$ = $1;
		$$->push_back($3);
	}
	;

root_exp: TK_IDENT '(' exp_list ')' {
		@$.propogate(@1, @4);
		$$ = new CallExpression(symbol_find_or_insert($1), $3,
			 new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT '=' exp {
		@$.propogate(@1, @3);
		$$ = new AssignExpression( 
				new VarRefExpression(symbol_find_or_insert($1), 
					new_location(@1)),
				$3, new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| tabref_exp '=' exp {
		@$.propogate(@1, @3);
		$$ = new AssignExpression( $1, $3, new_location(@$));
		cur_annot_owner = $$;
	}
	| '+'  TK_IDENT {
		$$ = new DecodeExpression(symbol_find_or_insert($2),
				NULL, new_location(@2));
		free($2);
		cur_annot_owner = $$;
	}
	| '+'  TK_IDENT '=' exp {
		$$ = new DecodeExpression(symbol_find_or_insert($2), $4,
				new_location(@2));
		free($2);
		cur_annot_owner = $$;
	}
	| '@'  TK_IDENT {
		$$ = new ActivateExpression(symbol_find_or_insert($2),
				NULL, new_location(@2));
		free($2);
		cur_annot_owner = $$;
	}
	| '@'  TK_IDENT '=' exp {
		$$ = new ActivateExpression(symbol_find_or_insert($2), $4, 
				new_location(@2));
		free($2);
		cur_annot_owner = $$;
	}
	;

tabref_exp: exp '[' exp ']' {
		@$.propogate(@1, @4);
		$$ = new TabRefExpression($1, $3, NULL, new_location(@$));
	}
	| exp '[' exp ':' exp ']' {
		@$.propogate(@1, @6);
		$$ = new TabRefExpression($1, $3, $5, new_location(@$));
	}
	;

exp_list: exp {
		$$ = new vector<Expression *>;
		$$->push_back($1);
	}
	| exp_list ',' exp {
		$$ = $1;
		$$->push_back($3);
	}
	;

exp:  exp '+' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_ADD, $1, $3, new_location(@$));
	}
	| exp '-' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_SUB, $1, $3, new_location(@$));
	}
	| exp '*' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_MUL, $1, $3, new_location(@$));
	}
	| exp '/' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_DIV, $1, $3, new_location(@$));
	}
	| exp '%' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_MOD, $1, $3, new_location(@$));
	}
	| exp '|' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_OR, $1, $3, new_location(@$));
	}
	| exp '&' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_AND, $1, $3, new_location(@$));
	}
	| exp '^' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_XOR, $1, $3, new_location(@$));
	}
	| exp TK_CONCAT exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_CONCAT, $1, $3, new_location(@$));
	}
	| exp TK_RSH exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_RSH, $1, $3, new_location(@$));
	}
	| exp TK_LSH exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_LSH, $1, $3, new_location(@$));
	}
	| exp '<' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_LT, $1, $3, new_location(@$));
	}
	| exp '>' exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_GT, $1, $3, new_location(@$));
	}
	| exp TK_LTE exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_LTE, $1, $3, new_location(@$));
	}
	| exp TK_GTE exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_GTE, $1, $3, new_location(@$));
	}
	| exp TK_EQ exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_EQ, $1, $3, new_location(@$));
	}
	| exp TK_NEQ exp {
		@$.propogate(@1, @3);
		$$ = new BinaryExpression(mad_ast::OP_NEQ, $1, $3, new_location(@$));
	}
	| '(' TK_UINT '<' TK_NUM '>' ')' exp %prec PS_CAST {
		@$.propogate(@2, @5);
		DataTypeBit *btype = new data_type_bit(false,
				strtol($4, NULL, 0), 0, new_location(@$));
		@$.propogate(@1, @7);
		$$ = new CastExpression(btype, $7, new_location(@$));
		free($4);
	}
	| '(' TK_INT '<' TK_NUM '>' ')' exp %prec PS_CAST {
		@$.propogate(@2, @5);
		DataTypeBit *btype = new data_type_bit(true,
				strtol($4, NULL, 0), 0, new_location(@$));
		@$.propogate(@1, @7);
		$$ = new CastExpression(btype, $7, new_location(@$));
		free($4);
	}
	| '(' TK_FLOAT ')' exp %prec PS_CAST {
		@$.propogate(@1, @4);
		DataTypeFloat *ftype = new data_type_float(false, 0, new_location(@2));
		$$ = new CastExpression(ftype, $4, new_location(@$));
	}
	| '(' TK_DOUBLE ')' exp %prec PS_CAST {
		@$.propogate(@1, @4);
		DataTypeFloat *ftype = new data_type_float(true, 0, new_location(@2));
		$$ = new CastExpression(ftype, $4, new_location(@$));
	}
	| tabref_exp {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| '-' exp %prec PS_NEG {
		@$.propogate(@1, @2);
		$$ = new UnaryExpression(mad_ast::OP_NEG, $2, new_location(@$));
	}
	| '~' exp {
		@$.propogate(@1, @2);
		$$ = new UnaryExpression(mad_ast::OP_COM, $2, new_location(@$));
	}
	| TK_NUM {
		@$.propogate(@1, @1);
		$$ = new NumExpression(new num_value(new string($1), false, false),
				new_location(@1));
		free($1);
	}
	| TK_FNUM {
		@$.propogate(@1, @1);
		$$ = new NumExpression(new num_value(new string($1), false, true),
				new_location(@1));
		free($1);
	}
	| TK_STR {
		@$.propogate(@1, @1);
		$$ = new StrExpression(new string($1), new_location(@1));
		free($1);
	}
	| TK_IDENT {
		@$.propogate(@1, @1);
		$$ = new VarRefExpression(symbol_find_or_insert($1),
				new_location(@1));
		free($1);
	}
	| exp '.' TK_IDENT {
		@$.propogate(@1, @3);
		lower_case($3);
		$$ = new ModExpression($1, symbol_find_or_insert($3), new_location(@$));
		free($3);
	}
	| '(' exp ')' {
		@$.propogate(@1, @3);
		@$.propogate(@1, @2);
		$$ = new UnaryExpression(mad_ast::OP_NOTHING, $2, new_location(@$));
	}
	| exp '?' exp ':' exp {
		@$.propogate(@1, @5);
		$$ = new TernaryExpression($1, $3, $5, new_location(@$));
	}
	;

trans_para: TK_TRANS trans_clause_list ';' {
		@$.propogate(@1, @3);
		$$ = new TransParagraph($2, new_location(@$));
	}
	;

trans_clause_list: trans_clause {
		$$ = new vector<TransClause *>;
		$$->push_back($1);
	}
	| trans_clause_list ';' trans_clause {
		$$ = $1;
		$$->push_back($3);
	}
	| trans_clause_list ';' root_exp {
		$$ = $1;
		$$->back()->append_exp($3);
	}
	;

trans_clause: TK_IDENT ':' osm_exp_clause root_exp {

		vector<Expression *> *v_temp = new vector<Expression *>();
		v_temp->push_back($4);

		@$.propogate(@1, @4);
		$$ = new TransClause(symbol_find_or_insert($1),
				$3, v_temp, new_location(@$));

		free($1);
	}
	| TK_IDENT ':' osm_exp_clause {
		@$.propogate(@1, @3);
		$$ = new TransClause(symbol_find_or_insert($1),
				$3, NULL, new_location(@$));
		free($1);
	}
	| TK_IDENT ':' root_exp {

		vector<Expression *> *v_temp = new vector<Expression *>();
		v_temp->push_back($3);

		@$.propogate(@1, @3);
		$$ = new TransClause(symbol_find_or_insert($1),
				NULL, v_temp, new_location(@$));
		free($1);
	}
	| TK_IDENT ':' {
		@$.propogate(@1, @2);
		$$ = new TransClause(symbol_find_or_insert($1),
				NULL, NULL, new_location(@$));
		free($1);
	}
	;

osm_exp_clause: '{' '}' {
		@$.propogate(@1, @2);
		$$ = new vector<osm_expression *>;
	}
	| '{' osm_exp_list '}' {
		@$.propogate(@1, @3);
		$$ = $2;
	}
	;

osm_exp_list: osm_exp {
		$$ = new vector<osm_expression *>;
		$$->push_back($1);
	}
	| osm_exp_list ',' osm_exp {
		$$ = $1;
		$$->push_back($3);
	}
	;

osm_exp: '!' TK_IDENT { /* release */
		@$.propogate(@1, @2);
		$$ = new OSMReleaseExpression(
				new OSMBufRefExpression(symbol_find_or_insert($2),
					new_location(@2)),
				new_location(@$));
		free($2);
		cur_annot_owner = $$;
	}
	| TK_DISCARD TK_IDENT {	/* discard */
		@$.propogate(@1, @2);
		$$ = new OSMDiscardExpression(
				new OSMBufRefExpression(symbol_find_or_insert($2),
					new_location(@2)),
				new_location(@$));
		free($2);
		cur_annot_owner = $$;
	}
	| TK_IDENT '=' osm_man_ref { /* allocate */
		@$.propogate(@1, @3);
		$$ = new OSMAssignExpression(
				new OSMBufRefExpression(symbol_find_or_insert($1),
					new_location(@1)),
				$3, new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| TK_IDENT '=' osm_dref { /* read */
		@$.propogate(@1, @3);
		$$ = new OSMAssignExpression(
				new OSMVarRefExpression(symbol_find_or_insert($1),
					new_location(@1)),
				$3, new_location(@$));
		free($1);
		cur_annot_owner = $$;
	}
	| osm_dref '=' osm_rval { /* read + write, write */
		@$.propogate(@1, @3);
		$$ = new OSMAssignExpression($1, $3, new_location(@$));
		cur_annot_owner = $$;
	}
	| '(' osm_group_list ')' '=' osm_dref { /* read2 */
		@$.propogate(@1, @3);
		OSMExpression *tmp = new OSMGroupExpression($2, new_location(@$));
		@$.propogate(@1, @5);
		$$ = new OSMAssignExpression(tmp, $5, new_location(@$));
		cur_annot_owner = $$;
	}
	| osm_man_ref { /* allocate_temp */
		@$.propogate(@1, @1);
		$$ = $1;
		cur_annot_owner = $$;
	}
	| '*' osm_man_ref { /* inquire */
		@$.propogate(@1, @2);
		$$ = new OSMDeRefExpression($2, new_location(@$));
		cur_annot_owner = $$;
	}
	| osm_value osm_comp osm_value { /* compare */
		@$.propogate(@1, @3);
		$$ = new OSMCompareExpression($1, $3, $2, new_location(@$));
		cur_annot_owner = $$;
	}
	;

osm_comp: '<' {$$ = OSM_CMP_LT;}
	| '>'     {$$ = OSM_CMP_GT;}
	| TK_EQ   {$$ = OSM_CMP_EQ;}
	| TK_NEQ  {$$ = OSM_CMP_NEQ;}
	| TK_LTE  {$$ = OSM_CMP_LTE;}
	| TK_GTE  {$$ = OSM_CMP_GTE;}
	;

osm_dref: '*' TK_IDENT {
		@$.propogate(@1, @2);
		$$ = new OSMDeRefExpression(
					new OSMBufRefExpression(symbol_find_or_insert($2),
						new_location(@2)),
					new_location(@$));
		free($2);
	}
	| '*' osm_man_ref {
		@$.propogate(@1, @2);
		$$ = new OSMDeRefExpression($2, new_location(@$));
	}
	;

osm_rval: osm_dref {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| osm_value {
		@$.propogate(@1, @1);
		$$ = $1;
	}
	| '(' osm_group_list ')' {
		@$.propogate(@1, @3);
		$$ = new OSMGroupExpression($2, new_location(@$));
	}
	;

osm_man_ref: TK_IDENT '[' '(' osm_group_list ')' ']' {
		@$.propogate(@3, @5);
		OSMExpression *tmp = new OSMGroupExpression($4, new_location(@$));
		@$.propogate(@1, @6);
		$$ = new OSMManRefExpression(symbol_find_or_insert($1), tmp,
				new_location(@$));
		free($1);
	}
	| TK_IDENT '[' TK_IDENT ']' {
		@$.propogate(@1, @4);
		$$ = new OSMManRefExpression(symbol_find_or_insert($1),
				new OSMVarRefExpression(symbol_find_or_insert($3),
					new_location(@3)),
				new_location(@$));
		free($1);
		free($3);
	}
	| TK_IDENT '[' ']' {
		@$.propogate(@1, @3);
		$$ = new OSMManRefExpression(symbol_find_or_insert($1), NULL,
				new_location(@$));
		free($1);
	}
	| TK_IDENT '[' num_value ']' {
		@$.propogate(@1, @4);
		$$ = new OSMManRefExpression(symbol_find_or_insert($1),
				new OSMNumExpression($3, new_location(@3)),
				new_location(@$));
		free($1);
	}
	| TK_IDENT '[' TK_STR ']' {
		@$.propogate(@1, @4);
		$$ = new OSMManRefExpression(symbol_find_or_insert($1),
				new OSMStrExpression(new string($3), new_location(@3)),
				new_location(@$));
		free($1);
		free($3);
	}
	;

osm_group_list: osm_value {
		@$.propogate(@1, @1);
		$$ = new vector<OSMExpression *>;
		$$->push_back($1);
	}
	| osm_group_list ',' osm_value {
		@$.propogate(@1, @3);
		$$ = $1;
		$$->push_back($3);
	}
	;

osm_value: TK_IDENT {
		@$.propogate(@1, @1);
		$$ = new OSMVarRefExpression(symbol_find_or_insert($1),
				new_location(@$));
		free($1);
	}
	| num_value {
		@$.propogate(@1, @1);
		$$ = new OSMNumExpression($1, new_location(@$));
	}
	| TK_STR {
		@$.propogate(@1, @1);
		$$ = new OSMStrExpression(new string($1), new_location(@$));
		free($1);
	}
	;

%%

