
#include "program_ast.hpp"
#include "define_ast.hpp"
#include "manager_ast.hpp"
#include "machine_ast.hpp"
#include "operation_ast.hpp"
#include "emit_c.hpp"
#include "oper_graph.hpp"
#include "opti.hpp"
#include <cassert>
#include <fstream>
#include <string>
#include <list>

using namespace mad_ast;
using namespace mad;
using std::ostream;
using std::ofstream;
using std::endl;
using std::cerr;
using std::vector;
using std::string;
using std::map;
using std::list;

map<OperID, Symbol> oper_name_table;

ostream& import_section::print(ostream& os) const
{
	os << "IMPORT  \"" << file << "\";" << endl;
	return os;	
}

ostream& using_section::print(ostream& os) const
{
	os << "USING  " << symbol_get_string(name) << ";" << endl;
	return os;	
}

ostream& annotation_section::print(ostream& os) const
{
	os << "ANNOTATION  " << endl;
	os << ";" << endl;
	return os;
}

ostream& program::print(ostream& os) const
{
	print_object_list(os, "\n", secs);
	return os;
}

bool import_section::check_annots()
{
	return check_annotations();
}

bool using_section::check_annots()
{
	return check_annotations();
}

bool annotation_section::check_annots()
{
	return check_annotations();
}

program::program() :
	secs(new vector<ProgSection *>),
	oper_cache(new OperContext),
	func_cache(new FunContext),
	coding_ref_graph(NULL),
	decoding_tree_graph(NULL)
{
}

program::~program()
{
	delete_object_list(secs);
	delete oper_cache;
	delete func_cache;
	delete coding_ref_graph;
	delete decoding_tree_graph;
}

void program::append_sections(vector<ProgSection *> *ss)
{
	vector<ProgSection *>::iterator it;
	for (it=ss->begin(); it!=ss->end(); it++) {

		switch ((*it)->get_type()) {

			case SEC_IMPORT: {
				imports.push_back(static_cast<ImportSection *>(*it));
				break;
			}
			case SEC_OPERATION: {
				opers.push_back(static_cast<OperationSection *>(*it));
				break;
			}
			case SEC_FUNCTION: {
				funcs.push_back(static_cast<FunctionSection *>(*it));
				break;
			}
			case SEC_MANAGER: {
				managers.push_back(static_cast<ManagerSection *>(*it));
				break;
			}
			case SEC_MACHINE: {
				machines.push_back(static_cast<MachineSection *>(*it));
				break;
			}
			case SEC_DEFINE: {
				defines.push_back(static_cast<DefineSection *>(*it));
				break;
			}
			case SEC_USING: {
				usings.push_back(static_cast<UsingSection *>(*it));
				break;
			}
			case SEC_ANNOTATION : {
				annots.push_back(static_cast<AnnotationSection *>(*it));
				break;
			}
		}
	}

	secs->insert(secs->end(), ss->begin(), ss->end());
}

bool program::check_annots()
{
	bool ret = true;

	for (vector<AnnotationSection *>::const_iterator ann_it=annots.begin();
		ann_it!=annots.end(); ann_it++) {
		ret = (*ann_it)->check_annots() && ret;
	}

	for (vector<DefineSection *>::const_iterator def_it=defines.begin();
		def_it!=defines.end(); def_it++) {
		ret = (*def_it)->check_annots() && ret;
	}

	for (vector<ManagerSection *>::const_iterator man_it=managers.begin();
		man_it!=managers.end(); man_it++) {
		ret = (*man_it)->check_annots() && ret;
	}

	for (vector<MachineSection *>::const_iterator mit = machines.begin();
		mit!=machines.end(); mit++) {
		ret = (*mit)->check_annots() && ret;
	}

	for (vector<UsingSection *>::const_iterator uit = usings.begin();
		uit!=usings.end(); uit++) {
		ret = (*uit)->check_annots() && ret;
	}

	for (vector<OperationSection *>::const_iterator oit = opers.begin();
		oit != opers.end(); oit++) {
		ret = (*oit)->check_annots() && ret;
	}

	for (vector<FunctionSection *>::const_iterator fit = funcs.begin();
		fit != funcs.end(); fit++) {
		ret = (*fit)->check_annots() && ret;
	}
	return ret;
}

bool program::type_check()
{
	bool ret = true;

	/* check the defines */
	if (defines.size() > 0) {

		/* first we merge them into one */
		for (vector<DefineSection *>::iterator def_it=defines.begin()+1;
			def_it!=defines.end(); def_it++) {
			defines[0]->merge(**def_it);
		}

		/* now we need to worry about only defines[0] */
		ret = defines[0]->type_check() && ret;
	}

	/* define context for convenience of future type-checking */
	SymContext *def_cache = NULL;

	if (defines.size()>0) {
		def_cache = &(defines[0]->def_cache);
	}

	/* More than one manager section */
	if (managers.size()>0) {

		/* first we merge them into one */
		for (vector<ManagerSection *>::iterator man_it=managers.begin()+1;
			man_it!=managers.end(); man_it++) {
			managers[0]->merge(**man_it);
		}

		/* now we only need to worry about the first one */
		ret = managers[0]->type_check() && ret;
	}


	// starting from 1, 0 is null
	MachID mach_id = 1;

	/* check machines */
	vector<MachineSection *>::iterator mit;
	for (mit = machines.begin(); mit!=machines.end(); mit++) {

		Symbol mach_name = (*mit)->get_name();
		if (mach_cache.find(mach_name)!=mach_cache.end()) {
			naming_conflict(mach_name,
				mach_cache[mach_name]->get_loc(), (*mit)->get_loc());
			ret = false;
		}
		else {
			mach_cache[mach_name] = *mit;
		}

		/* type check the machine */
		ManagerContext *mctxt = NULL;
		if (managers.size()>0) {
			mctxt = &(managers[0]->class_cache);
		}

		/* construct an empty context */
		if (!(*mit)->type_check(mctxt, def_cache)) ret = false;
		(*mit)->set_id(mach_id++);
	}

	/* check using sections */
	vector<UsingSection *>::iterator uit;
	for (uit = usings.begin(); uit!=usings.end(); uit++) {

		Symbol uname = (*uit)->get_name();
		if (mach_cache.find(uname)==mach_cache.end()) {
			undefined_ident(uname, (*uit)->get_loc());
			ret = false;
		}
	}

	// starting from 1
	OperID oper_id = 1;

	/* 1. check the naming conflict between operations.
	 * 2. Set up the machine.
	 * 3. Classify paragraphs.
	 */
	vector<OperationSection *>::iterator oit;
	for (oit = opers.begin(); oit != opers.end(); oit++) {

		Symbol oname = (*oit)->get_name();
		const Location *loc = (*oit)->get_loc();

		oper_name_table[oper_id] = oname;
		(*oit)->set_id(oper_id++);

		if (oper_cache->has(oname)) {
			naming_conflict(oname, oper_cache->get_loc(oname), loc);
			ret = false;
		}
		else {
			oper_cache->insert(*oit);
		}

		Symbol mach_name = (*oit)->get_mach_name();

		if (mach_name==SymNULL) {
			generic_error("No machine bound to operation ",
				symbol_get_string((*oit)->get_name()), (*oit)->get_loc());
			ret = false;
		}
		else if (mach_cache.find(mach_name)!=mach_cache.end()) {
			(*oit)->set_machine(mach_cache[mach_name]);
		}
		/* else if the machine name cannot be found, the error
		 * should have been reported by using command */

		/* we also classify the paragraphs. */
		if (!(*oit)->classify_paragraphs()) ret = false;
	}

	vector<FunctionSection *>::iterator fit;
	for (fit = funcs.begin(); fit != funcs.end(); fit++) {

		Symbol oname = (*fit)->get_name();
		const Location *loc = (*fit)->get_loc();

		if (func_cache->has(oname)) {
			naming_conflict(oname, func_cache->get_loc(oname), loc);
			ret = false;
		}
		/* cannot conflict with define names since it would be
		 * ambiguous if functions have the same name */
		else if (defines.size()>0 && defines[0]->def_cache.has(oname)) {
			naming_conflict(oname, defines[0]->def_cache.get_loc(oname), loc);
			ret = false;
		}
		else {
			func_cache->insert(*fit);
		}

		/* classify the function paragraphs. */
		if (!(*fit)->classify_paragraphs()) ret = false;
	}

	/* propogate the dirty flags for the functions.
	 * this must be done prior to type_check since type_check 
	 * needs the dirty information for function call expression.
	 */
	bool change = true;
	while (change) {
		change = false;
		for (fit = funcs.begin(); fit != funcs.end(); fit++) {
			change = (*fit)->propogate_dirty(def_cache, func_cache) || change;
		}
	}

	/* type check the functions */
	for (fit = funcs.begin(); fit != funcs.end(); fit++) {
		(*fit)->type_check(def_cache, func_cache);
	}

	/* stop here since error may cause crashing in further checking,
	 * for instance the operation may not have machine
	 */
	if (!ret) return !summarize_error_warn();

	ManagerContext *manager_cache = NULL;
	if (managers.size()>0)
		manager_cache = &(managers[0]->man_cache);

	/* we check the operations in two steps, first prechecking
	 * will set up all decoding entities, then we build
	 * coding graph and check it, then we sort the operations
	 * and check them one by one. */

	/* precheck */
	for (oit = opers.begin(); oit != opers.end(); oit++) {
		ret = (*oit)->pre_type_check(manager_cache, def_cache,
				oper_cache, func_cache) && ret;
	}

	/* stop here since error may cause crashing in further checking,
	 * for instance the operation may not have machine
	 */
	if (!ret) return !summarize_error_warn();

	/* then build graph */
	coding_ref_graph = new OperGraph(oper_cache, false);
	ret = coding_ref_graph->type_check() && ret;

	/* stop here since error may cause crashing in further checking,
	 * for instance the operation may not have machine
	 */
	if (!ret) return !summarize_error_warn();

	list<OperationSection *> oper_list;
	coding_ref_graph->topological_sort(oper_list);

	/* list sorted, root first */
	list<OperationSection *>::reverse_iterator it;
	for (it=oper_list.rbegin(); it!=oper_list.rend(); it++) {
		ret = (*it)->type_check(manager_cache, def_cache,
				oper_cache, func_cache) && ret;
	}

	/* then build decoding graph */
	decoding_tree_graph = new OperGraph(oper_cache, true);
	ret = decoding_tree_graph->type_check() && ret;

	return !summarize_error_warn();
}

void program::emit_C_defines(const char *prefix)
{
	if (defines.size() > 0) {

		string def_fname = prefix;
		def_fname += DEFN_DEC_FILE;

		ofstream ofs;
		message(def_fname+" ...");
		ofs.open(def_fname.c_str());
		ofs << "\n\n#ifndef __DEFN_DEC_H__\n"
			<< "#define __DEFN_DEC_H__\n";
		ofs << "\n\n#include \"" << prefix << OSM_LIB_INCLUDE << "\"\n";
		defines[0]->emit_C_dec(ofs);
		ofs << "#endif\n";
		ofs.close();

		string def_iname = prefix;
		def_iname += DEFN_IMP_FILE;
		message(def_iname+" ...");
		ofs.open(def_iname.c_str());
		ofs << "\n\n#include \"" << def_fname << "\"\n";
		defines[0]->emit_C(ofs);
		ofs.close();
	}

}

void program::emit_C_funcs(const char *prefix)
{
	if (funcs.size() > 0) {

		string func_fname = prefix;
		func_fname += FUNC_IMP_FILE;

		ofstream ofs;
		message(func_fname+" ...");
		ofs.open(func_fname.c_str());
		vector<FunctionSection *>::iterator fun_it;
		for (fun_it=funcs.begin(); fun_it!=funcs.end(); fun_it++)
			(*fun_it)->emit_C(ofs);
		ofs.close();
	}
}

void program::emit_C_mangs(const char *prefix)
{
	if (managers.size() > 0) {

		string mang_fname = prefix;
		mang_fname += MANG_LIST_FILE;

		ofstream ofs;
		message(mang_fname+" ...");
		ofs.open(mang_fname.c_str());
		ofs << std::dec;
		managers[0]->emit_C_dec(ofs);
		ofs.close();
	}
}

void program::emit_C(const char *prefix)
{

	emit_C_defines(prefix);
	emit_C_funcs(prefix);
	emit_C_mangs(prefix);

	ofstream ofs;

	if (machines.size() > 0) {

		string mach_fname = prefix;
		mach_fname += MACH_DEC_FILE;

		message(mach_fname+" ...");
		ofs.open(mach_fname.c_str());
		ofs << std::dec
			<< "\n\n#ifndef __MACH_DEC_H__\n"
			<< "#define __MACH_DEC_H__\n";
		ofs << "\n\n#include \"" << prefix << OSM_LIB_INCLUDE << "\"\n";
		vector<MachineSection *>::iterator mach_it;
		for (mach_it=machines.begin(); mach_it!=machines.end(); mach_it++)
			(*mach_it)->emit_C_dec(ofs);
		ofs << "\n#endif\n";
		ofs.close();

		string mach_iname = prefix;
		mach_iname += MACH_IMP_FILE;

		message(mach_iname+" ...");
		ofs.open(mach_iname.c_str());
		ofs << std::dec
			<< "\n\n#include \"" << mach_fname << "\"\n";
		for (mach_it=machines.begin(); mach_it!=machines.end(); mach_it++)
			(*mach_it)->emit_C(ofs);
		ofs.close();

		string mach_lname = prefix;
		mach_lname += MACH_LIST_FILE;

		message(mach_lname+" ...");
		ofs.open(mach_lname.c_str());
		ofs << std::dec;
		for (mach_it=machines.begin(); mach_it!=machines.end(); mach_it++)
			ofs << "_MACH_DEF(" << (*mach_it)->get_id() << ", " <<
				symbol_get_string((*mach_it)->get_name()) << ")" << endl;
		ofs << "#undef _MACH_DEF" << endl;
		ofs << "#define _MAX_MACH_ID " << machines.size() << endl;
		ofs.close();
	}

	if (opers.size() > 0) {

		string oper_fname = prefix;
		oper_fname += OPER_DEC_FILE;

		message(oper_fname+" ...");
		ofs.open(oper_fname.c_str());
		ofs << std::dec
			<< "\n\n#ifndef __OPER_DEC_H__\n"
			<< "#define __OPER_DEC_H__\n";
		ofs << "\n\n#include \"" << prefix << OSM_LIB_INCLUDE << "\"\n";
		ofs << "#include \"" << prefix << MACH_DEC_FILE << "\"\n";
		vector<OperationSection *>::iterator oper_it;
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			(*oper_it)->emit_C_dec(ofs);
		ofs << "\n#endif\n";
		ofs.close();

		string oper_iname = prefix;
		oper_iname += OPER_IMP_FILE;
		message(oper_iname+" ...");
		ofs.open(oper_iname.c_str());
		ofs << std::dec << "\n\n#include \"" << oper_fname << "\"\n";
		if (defines.size()) ofs << "#include \""
				<< prefix << DEFN_DEC_FILE << "\"\n";
		if (funcs.size()) ofs << "#include \""
				<< prefix << FUNC_IMP_FILE << "\"\n";

		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++) {

			const std::map<Symbol, DecodeEntity *>& dec_ents =
				(*oper_it)->get_decode_entities();

			map<Symbol, DecodeEntity *>::const_iterator de_it;
			for (de_it=dec_ents.begin(); de_it!=dec_ents.end(); de_it++) {

				DecodeEntity *de = (*de_it).second;

				if (de->need_decoder() && !de->is_single()) {
					ofs << "#include \"" << prefix
						<< DECODER_HDR_FILE(de->get_decoder_name())
					  	<< "\"" << endl;
				}
			}
		}
		ofs << endl;

		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			(*oper_it)->emit_C(ofs);
		ofs.close();

		string oper_lname = prefix;
		oper_lname += OPER_LIST_FILE;
		message(oper_lname+" ...");
		ofs.open(oper_lname.c_str());
		ofs << std::dec;
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			ofs << "_OPER_DEF(" << (*oper_it)->get_id() << ", " <<
					symbol_get_string((*oper_it)->get_name()) << ", " <<
					(*oper_it)->get_mach_id() << "/*" <<
				   	symbol_get_string((*oper_it)->get_mach_name()) << "*/" <<
				   	")" << endl;
		ofs << "#undef _OPER_DEF" << endl;
		ofs << "#define _MAX_OPER_ID " << opers.size() << endl;
		ofs.close();

		/* emit decoder files */
		for (oper_it=opers.begin(); oper_it!=opers.end(); oper_it++)
			(*oper_it)->emit_decode_patterns(prefix);
	}

}

void program::emit_C_ISS(const char *prefix)
{
	/* get all the roots of the decoding tree. */
	vector<const operation_section *> roots =
			decoding_tree_graph->get_root_opers();

	if (roots.size()==0) return;

	/* all optimzed machines here. */
	vector<OptiMachine *> optimized_machines;
	vector<OperContext *> oper_contexts;

	vector<const operation_section *>::const_iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {

		const operation_section *oper = *rit;

		vector<operation_section *> reachables =
			decoding_tree_graph->get_reachables(oper);

		OperContext *reach_ctxt = new OperContext();;
		vector<operation_section *>::iterator oper_it;
		for (oper_it=reachables.begin(); oper_it!=reachables.end(); oper_it++)
			reach_ctxt->insert(*oper_it);

		/* let's try out the optimization stuff here. */
		OptiMachine *omach = new OptiMachine(oper, reach_ctxt);

		/* if fail to optimize, do nothing.*/
		if (!omach->optimize()) return;
		optimized_machines.push_back(omach);
		oper_contexts.push_back(reach_ctxt);
	}

	/* C declaration file. */
	ofstream ofs;

	string msg = prefix;
	msg += MACH_DEC_FILE;

	message(msg+" ...");

    ofs.open(msg.c_str());
    ofs << std::dec
        << "\n\n#ifndef __MACH_DEC_H__\n"
        << "#define __MACH_DEC_H__\n";
    ofs << "\n\n#include \"" << prefix << OSM_LIB_INCLUDE << "\"\n";

	vector<OptiMachine *>::iterator oit;
	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_C_dec_ISS(ofs);

	ofs.close();

	/* C implementation file. */
	msg = prefix;
	msg += MACH_IMP_FILE;
	message(msg+" ...");

	ofs.open(msg.c_str());
	ofs << std::dec
		<< "\n\n#include \"" << prefix << MACH_DEC_FILE << "\"\n";
	if (defines.size())
		ofs << "#include \"" << prefix << DEFN_DEC_FILE << "\"\n";
	if (funcs.size())
		ofs << "#include \"" << prefix << FUNC_IMP_FILE << "\"\n";

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++) {

		const map<Symbol, DecodeEntity *>& dents =
			(*oit)->get_root_oper()->get_decode_entities();

		map<Symbol, DecodeEntity *>::const_iterator dit;
		for (dit=dents.begin(); dit!=dents.end(); dit++) {

			const DecodeEntity *de = (*dit).second;
			if (de->need_decoder() && !de->is_single()) {
				ofs << "#include \"" << prefix
					<< DECODER_HDR_FILE(de->get_decoder_name())
				  	<< "\"" << endl;
			}
		}
	}
	ofs << endl;

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_C_ISS(ofs);

	ofs.close();

	msg = prefix;
	msg += MACH_LIST_FILE;
	message(msg+" ...");

	/* machine list file. */
	ofs.open(msg.c_str());

	unsigned mach_ind=1;
	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++) {

		ofs << "_MACH_DEF(" << mach_ind++ << ", "
			<< symbol_get_string((*oit)->get_name()) << ")" << endl;
		ofs << "#undef _MACH_DEF" << endl;
	}
	ofs << "#define _MAX_MACH_ID " << roots.size() << endl;
	ofs.close();

	emit_C_defines(prefix);
	emit_C_funcs(prefix);
	emit_C_mangs(prefix);

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_decode_patterns(prefix);

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		delete (*oit);

	vector<OperContext *>::iterator cit;
	for (cit=oper_contexts.begin(); cit!=oper_contexts.end(); cit++)
		delete (*cit);
}

void program::emit_C_opt(const char *prefix)
{
	/* get all the roots of the decoding tree. */
	vector<const operation_section *> roots =
			decoding_tree_graph->get_root_opers();

	if (roots.size()==0) return;

	/* all optimzed machines here. */
	vector<OptiMachine *> optimized_machines;
	vector<OperContext *> oper_contexts;

	vector<const operation_section *>::const_iterator rit;
	for (rit=roots.begin(); rit!=roots.end(); rit++) {

		const operation_section *oper = *rit;

		vector<operation_section *> reachables =
			decoding_tree_graph->get_reachables(oper);

		OperContext *reach_ctxt = new OperContext();
		vector<operation_section *>::iterator oper_it;
		for (oper_it=reachables.begin(); oper_it!=reachables.end(); oper_it++)
			reach_ctxt->insert(*oper_it);

		/* let's try out the optimization stuff here. */
		OptiMachine *omach = new OptiMachine(oper, reach_ctxt);

		/* if fail to optimize, do nothing.*/
		if (!omach->optimize()) return;
		optimized_machines.push_back(omach);
		oper_contexts.push_back(reach_ctxt);
	}

	/* C declaration file. */
	ofstream ofs;

	string msg = prefix;
	msg += MACH_DEC_FILE;

	message(msg+" ...");

    ofs.open(msg.c_str());
    ofs << std::dec
        << "\n\n#ifndef __MACH_DEC_H__\n"
        << "#define __MACH_DEC_H__\n";
    ofs << "\n\n#include \"" << prefix << OSM_LIB_INCLUDE << "\"\n";

	vector<OptiMachine *>::iterator oit;
	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_C_dec(ofs);

	ofs.close();

	/* C implementation file. */
	msg = prefix;
	msg += MACH_IMP_FILE;
	message(msg+" ...");

	ofs.open(msg.c_str());
	ofs << std::dec
		<< "\n\n#include \"" << prefix << MACH_DEC_FILE << "\"\n";
	if (defines.size())
		ofs << "#include \"" << prefix << DEFN_DEC_FILE << "\"\n";
	if (funcs.size())
		ofs << "#include \"" << prefix << FUNC_IMP_FILE << "\"\n";

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++) {

		const map<Symbol, DecodeEntity *>& dents =
			(*oit)->get_root_oper()->get_decode_entities();

		map<Symbol, DecodeEntity *>::const_iterator dit;
		for (dit=dents.begin(); dit!=dents.end(); dit++) {

			const DecodeEntity *de = (*dit).second;
			if (de->need_decoder() && !de->is_single()) {
				ofs << "#include \"" << prefix
					<< DECODER_HDR_FILE(de->get_decoder_name())
				  	<< "\"" << endl;
			}
		}
	}
	ofs << endl;

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_C(ofs);

	ofs.close();

	msg = prefix;
	msg += MACH_LIST_FILE;
	message(msg+" ...");

	/* machine list file. */
	ofs.open(msg.c_str());

	unsigned mach_ind=1;
	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++) {

		ofs << "_MACH_DEF(" << mach_ind++ << ", "
			<< symbol_get_string((*oit)->get_name()) << ")" << endl;
		ofs << "#undef _MACH_DEF" << endl;
	}
	ofs << "#define _MAX_MACH_ID " << roots.size() << endl;
	ofs.close();

	emit_C_defines(prefix);
	emit_C_funcs(prefix);
	emit_C_mangs(prefix);

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		(*oit)->emit_decode_patterns(prefix);

	for (oit=optimized_machines.begin(); oit!=optimized_machines.end(); oit++)
		delete (*oit);

	vector<OperContext *>::iterator cit;
	for (cit=oper_contexts.begin(); cit!=oper_contexts.end(); cit++)
		delete (*cit);
}

vector<const ImportSection *> program::get_import_sections() const
{
	return const_vector_transform<ImportSection>(imports);
}

vector<const DefineSection *> program::get_define_sections() const
{
	return const_vector_transform<DefineSection>(defines);
}

vector<const UsingSection *> program::get_using_sections() const
{
	return const_vector_transform<UsingSection>(usings);
}

vector<const AnnotationSection *> program::get_annot_sections() const
{
	return const_vector_transform<AnnotationSection>(annots);
}

vector<const MachineSection *> program::get_machine_sections() const
{
	return const_vector_transform<MachineSection>(machines);
}

vector<const ManagerSection *> program::get_manager_sections() const
{
	return const_vector_transform<ManagerSection>(managers);
}

vector<const OperationSection *> program::get_oper_sections() const
{
	return const_vector_transform<OperationSection>(opers);
}

vector<const FunctionSection *> program::get_func_sections() const
{
	return const_vector_transform<FunctionSection>(funcs);
}

const OperationSection *program::get_operation(Symbol name) const
{
	if (oper_cache->has(name)) return oper_cache->get_object(name);
	return NULL;
}

const FunctionSection *program::get_function(Symbol name) const
{
	if (func_cache->has(name)) return func_cache->get_object(name);
	return NULL;
}

const MachineSection *program::get_machine(Symbol name) const
{
	if (mach_cache.find(name)!=mach_cache.end())
		return (*mach_cache.find(name)).second;
	return NULL;
}
