
#ifndef __MAD_PROG_AST_HPP__
#define __MAD_PROG_AST_HPP__

#include <vector>
#include <iostream>
#include <string>
#include "mad_ast.hpp"
#include "annot_ast.hpp"
#include "data_type_ast.hpp"

using mad::Symbol;

namespace mad_ast {

	/** Enumerate of section types */
	typedef enum {

		SEC_IMPORT,		/**< IMPORT section. */
		SEC_USING,		/**< USING section. */
		SEC_ANNOTATION,	/**< ANNOTATION section. */
		SEC_OPERATION,	/**< OPERATION section. */
		SEC_FUNCTION,	/**< FUNCTION section. */
		SEC_MANAGER,	/**< MANAGER section. */
		SEC_MACHINE,	/**< DIAGRAM section. */
		SEC_DEFINE		/**< DEFINE section. */

	} EnumSecType;

	/* 2^32 operations should be enough */
	typedef unsigned int OperID;

	typedef unsigned int MachID;

	/** Base class of all program sections. */
	typedef class prog_section : public ast_node {

		public:
			/** Constructor.
			 *  @param tp   Type of the section. 
			 *  @param loc  Location of the node. 
			 */
			prog_section(EnumSecType tp, const Location *loc) :
				  ast_node(loc), tp(tp) {}

			/** Destructor. */
			virtual ~prog_section() {};

			/** Get the type of the section. */
			EnumSecType get_type() {return tp;}

		private:
			EnumSecType tp;

	} ProgSection;


	/** Class for import section. */
	typedef class import_section : public prog_section, public annot_owner {

		public:

			/** Constructor. 
			 */
			import_section() : prog_section(SEC_IMPORT, NULL) {}

			/** Destructor. */
			~import_section() {}

			/** Set the file to import.
			 *  @param f  The filename.
			 */
			void set_file(const std::string& f) {file=f;}

			/** Print out section. */
			std::ostream& print(std::ostream &) const;

			/** get the file */
			const std::string& get_file() const {return file;}

			/** check the annotations of the section. */
			bool check_annots();

		private:
			std::string file;

	} ImportSection;


	/** Class for using section. */
	typedef class using_section : public prog_section, public annot_owner {

		public:

			/** Constructor. 
			 */
			using_section() : prog_section(SEC_USING, NULL), name(name) {}

			/** Destructor. */
			~using_section() {}

			/** Set the name of the machine to import.
			 *  @param n Name of the machine.
			 */
			void set_name(Symbol n) {name = n;}
	
			/** get the machine name*/
			Symbol get_name() const {return name;}

			/** Print out the section. */
			std::ostream& print(std::ostream &) const;

			/** check the annotations of the section. */
			bool check_annots();

		private:
			Symbol name;

	} UsingSection;

	typedef class annotation_section : public prog_section, public annot_owner {

		public:

			/** Constructor. 
			 */
			annotation_section() : prog_section(SEC_ANNOTATION, NULL) {}

			/** Destructor. */
			~annotation_section() {}

			/** Print out the section. */
			std::ostream& print(std::ostream &) const;

			/** check the annotations of the section. */
			bool check_annots();

	} AnnotationSection;

	typedef class define_section DefineSection;
	typedef class machine_section MachineSection;
	typedef class manager_section ManagerSection;
	typedef class operation_section OperationSection;
	typedef class function_section FunctionSection;

	/* create a class instead of simply typedef to minimize dependency
	 * between header files. */
	/** class for operation environment */
	typedef class oper_context : public obj_context<operation_section> {
	} OperContext;

	/** class for operation environment */
	typedef class func_context : public obj_context<function_section> {
	} FunContext;

	class oper_graph;

	/** Class for a program. */
	typedef class program {

		public:
			/** Constructor.
			 */
			program();

			/** Destructor. */
			~program();

			/** Append more sections to the program.
			 *  This is used when parsing imported files.
			 *  @param secs Array of sections to append.
			 */  
			void append_sections(std::vector<ProgSection *> *secs);

			/** Print out the program. */
			std::ostream& print(std::ostream &) const;

			/** type check the program */
			bool type_check();

			/** Check annotation syntax */
			bool check_annots();

			/** emit a C program.
			 */
			void emit_C(const char *prefix);

			/** emit an optimized version C program.
			 */
			void emit_C_opt(const char *prefix);

			/** emit an optimized ISS C program.
			 */
			void emit_C_ISS(const char *prefix);

			/** Get the decoding graph. */
			const oper_graph *get_decoding_graph() const {
				return decoding_tree_graph;
			}

			/** Get the import sections. */
			std::vector<const ImportSection *> get_import_sections() const;

			/** Get the define sections. */
			std::vector<const DefineSection *> get_define_sections() const;

			/** Get the using sections. */
			std::vector<const UsingSection *> get_using_sections() const;

			/** Get the annot sections. */
			std::vector<const AnnotationSection *> get_annot_sections() const;

			/** Get the machine sections. */
			std::vector<const MachineSection *> get_machine_sections() const;

			/** Get the manager sections. */
			std::vector<const ManagerSection *> get_manager_sections() const;

			/** Get the operation sections. */
			std::vector<const OperationSection *> get_oper_sections() const;

			/** Get the function sections. */
			std::vector<const FunctionSection *> get_func_sections() const;

			/** Get one operation section. */
			const OperationSection *get_operation(Symbol name) const;

			/** Get one function section. */
			const FunctionSection *get_function(Symbol name) const;

			/** Get one machine section. */
			const MachineSection *get_machine(Symbol name) const;

		private:

			void emit_C_defines(const char *prefix);
			void emit_C_funcs(const char *prefix);
			void emit_C_mangs(const char *prefix);

			std::vector<ProgSection *> *secs;

			std::vector<ImportSection *> imports;
			std::vector<DefineSection *> defines;
			std::vector<UsingSection *> usings;
			std::vector<AnnotationSection *> annots;

			std::vector<MachineSection *> machines;
			std::vector<ManagerSection *> managers;

			std::vector<OperationSection *> opers;
			std::vector<FunctionSection *> funcs;

			std::map<Symbol, MachineSection *> mach_cache;

			OperContext *oper_cache;
			FunContext *func_cache;

			oper_graph *coding_ref_graph;
			oper_graph *decoding_tree_graph;

	} Program;
}


#endif
