
#include <FlexLexer.h>
#include <fstream>
#include <iostream>
#include <set>
#include <string>
#include "project.hpp"

using std::set;
using std::string;
using std::ifstream;
using std::ofstream;
using std::vector;
using namespace mad;
using namespace mad_ast;

vector<ProgSection *> *madl_parse_wrapper(ifstream *infile, const char *fname);

project_manager::project_manager(const char *mfile,
	const std::vector<const char *>& ipath, 
	bool syntax, bool print, bool annot, bool opt, bool iss,
	const char *prefix) :
	main_file(mfile), inc_paths(ipath),
   	syntax_only(syntax), pretty_print(print), check_annot(annot),
	to_optimize(opt), to_iss(iss), prog(NULL), status(true), prefix(prefix)
{
	process();
}

project_manager::~project_manager()
{
	delete prog;
}

void project_manager::process()
{
	if (!open_project()) {
		status = false;
		return;
	}

	if (!syntax_only) {
		message("Type checking ...");
		if (prog->type_check()) {
			message("Emitting C library ...");
			if (to_iss)
				prog->emit_C_ISS(prefix);
			else if (to_optimize) 
				prog->emit_C_opt(prefix);
			else
				prog->emit_C(prefix);
		}
		else status = false;
	}
	if (check_annot) {
		message("Checking annotation ...");
		status = prog->check_annots();
	}

}


/** Open a file. 
 *  @param fname Name of the file.
 *  @return Array of sections in the file.
 */
vector<ProgSection *> *project_manager::open_file(const char *fname)
{
	ifstream infile;
	string full_name(fname);

	/* first check the current directory */
	infile.open(fname);

	/* no such file. */
	if (!infile.is_open()) {

		/* check the include path. */
		vector<const char *>::const_iterator path_it;
		for (path_it=inc_paths.begin();
				path_it!=inc_paths.end(); path_it++) {
	
			full_name = *path_it;
			full_name += "/";
			full_name += fname;

			infile.clear();
			infile.open(full_name.c_str());
			if (infile.is_open()) break;
		}
	}

	if (!infile.is_open()) {
		string error_msg = "Cannot open file ";
		error_msg += full_name;
		error(error_msg);
		return NULL;
	}

	string msg = "Parsing ";
	msg += full_name;
	msg += " ...";
	message(msg);

	return madl_parse_wrapper(&infile, full_name.c_str());
}

bool project_manager::open_project()
{
	set<string> openned_files;
	set<string> new_files;

	prog = new Program();

	new_files.insert(main_file);
	while (!new_files.empty()) {

		string fname = *new_files.begin();

		new_files.erase(new_files.begin());

		/* see if the file has already been openned */
		if (openned_files.find(fname)!=openned_files.end())
			continue;

		/* open the file. */
		vector<ProgSection *> *sections = open_file(fname.c_str());

		/* error */
		if (sections==NULL) return false;

		/* the file already successfully openned */
		openned_files.insert(fname);

		/* check the import section */
		vector<ProgSection *>::iterator it;
		for (it=sections->begin(); it!=sections->end(); it++)
		{
			if ((*it)->get_type()==SEC_IMPORT) {

				const string& newf =
				   static_cast<ImportSection *>(*it)->get_file();

				new_files.insert(newf);
			}
		}

		/* echo the input to output. */
		if (pretty_print) {

			string ofname;
			const char *cp = fname.c_str();
			/* strip the existing path so the output is in current dir */
			if ((cp = rindex(cp, '/'))!=NULL) {
				ofname = cp+1;
			} else {
				ofname = fname;
			}

		   	ofname += ".out";
			ofstream ofs;

			string msg = "Printing ";
			msg += ofname;
			msg += " ...";
			message(msg);

			ofs.open(ofname.c_str());
			print_object_list(ofs, "\n", sections);
			ofs.close();
		}

		/* append the content to the program */
		prog->append_sections(sections);

		delete sections;
	}

	return true;
}
