%{


#include "program_ast.hpp"
#include "data_type_ast.hpp"
#include "define_ast.hpp"
#include "manager_ast.hpp"
#include "machine_ast.hpp"
#include "osm_exp_ast.hpp"
#include "exp_ast.hpp"
#include "operation_ast.hpp"
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

using namespace mad_ast;
using namespace mad;
using std::istream;
using std::ifstream;
using std::string;

#define YYLTYPE mad_ast::yyltype_t
#include "parse_mad.hpp"

#define YY_NO_UNPUT
#define YY_USER_ACTION  madl_loc.advance_column(YYLeng());

/* tracking parsing location */
Symbol madl_filename;
string madl_annot_buf;
Symbol madl_machname;
vector<ProgSection *> *madl_parser_result;

YYLTYPE annot_begin_loc;
extern YYLTYPE madl_loc;
extern int madl_parse();

madl_FlexLexer *madl_lexer;
annot_owner *cur_annot_owner;


static void madl_lexerror(const mad_ast::yyltype_t *loc, char c) 
{
	std::stringstream sstr;
	sstr << "Unknown character " << c << " at "
		 << symbol_get_string(madl_filename) << ":" << *loc;
	error(sstr.str());
}

static void process_annotation()
{
	if (cur_annot_owner==NULL) {
		std::stringstream sstr;
		sstr << "Annotation without an anchor " << " at "
			 << symbol_get_string(madl_filename) << ":" << annot_begin_loc;
		error(sstr.str());
	}
	else {
		 cur_annot_owner->append_annotation
			(new Annotation(strdup(madl_annot_buf.c_str()),
				new Location(madl_filename, annot_begin_loc)));
	}
}


%}

%option c++
%option yylineno
%option noyywrap

NUM		[0-9]+
HEXNUM  0[xX][0-9a-fA-F]+
BINNUM  0[bB][01]+

DIG     [0-9]
EXP     [Ee][+-]?{DIG}+

PAT     [01-]+
STR		\"(\\.|[^\\"])*\"

WHITES  [ \t]*
IDENT   [a-zA-Z_][a-zA-Z_0-9]*


%s coding
%x comment
%x annot

%%

%{
	/* At each yylex invocation, mark the current position as the
	start of the next token.  */
	(madl_loc.step());
%}

"SYNTAX"	return TK_SYNTAX;
{STR}		madl_lval.strval = strdup(YYText()+1); madl_lval.strval[YYLeng()-2]=0; return TK_STR;

"void"		return TK_VOID;
"uint"		return TK_UINT;
"int"		return TK_INT;
"float"		return TK_FLOAT;
"double"	return TK_DOUBLE;
"string"	return TK_STRING;

"VAR"		return TK_VAR;
"IMPORT"	return TK_IMPORT;
"USING"		return TK_USING;
"DEFINE"	return TK_DEFINE;
"MANAGER"	return TK_MANAGER;
"MACHINE"	return TK_MACHINE;

"CLASS"		return TK_CLASS;
"INSTANCE"	return TK_INSTANCE;
"INITIAL"	return TK_INITIAL;
"BUFFER"	return TK_BUFFER;
"STATE"		return TK_STATE;
"EDGE"		return TK_EDGE;

"OPERATION"		return TK_OPERATION;
"FUNCTION"		return TK_FUNCTION;
"ANNOTATION"	return TK_ANNOTATION;

"EVAL"		return TK_EVAL;
"TRANS"		return TK_TRANS;

"::"		return TK_CONCAT;
"->"		return TK_TOWARD;
"=="		return TK_EQ;
"!="		return TK_NEQ;
"<="		return TK_LTE;
">="		return TK_GTE;
"<<"		return TK_LSH;
">>"		return TK_RSH;
"!!"		return TK_DISCARD;

"CODING"		BEGIN(coding); return TK_CODING;
<coding>{PAT}	madl_lval.strval = strdup(YYText()); return TK_PAT;
<coding>";"		BEGIN(INITIAL); return ';';

{IDENT}		madl_lval.strval = strdup(YYText()); return TK_IDENT;
{NUM}       madl_lval.strval = strdup(YYText()); return TK_NUM;
{HEXNUM}    madl_lval.strval = strdup(YYText()); return TK_NUM;
{BINNUM}    madl_lval.strval = strdup(YYText()); return TK_NUM;

{DIG}+{EXP}[fF]?              madl_lval.strval = strdup(YYText()); return TK_FNUM;
{DIG}*"."{DIG}+({EXP})?[fF]?  madl_lval.strval = strdup(YYText()); return TK_FNUM;
{DIG}+"."{DIG}*({EXP})?[fF]?  madl_lval.strval = strdup(YYText()); return TK_FNUM;

"##"			BEGIN(comment);
<comment>"\n"+	madl_loc.advance_line(YYLeng()); madl_loc.step();
<comment>.		/*ignore*/
<comment>"##"	BEGIN(INITIAL);

"#"		{int c;
		 while ((c=yyinput()) != 0)
			if (c=='\n') break;
		 madl_loc.advance_line(1); madl_loc.step();
		}

"$$"			{
				BEGIN(annot); madl_annot_buf.erase(); 
				annot_begin_loc = madl_loc;
				}

<annot>"\n"+	{ 
				madl_loc.advance_line(YYLeng()); madl_loc.step();
				madl_annot_buf += YYText();
				}

<annot>"$$"		{BEGIN(INITIAL);
				 process_annotation();
				}

<annot>.		madl_annot_buf.append(1, *(YYText()));

"$"		{char c;
		 annot_begin_loc = madl_loc;
		 madl_annot_buf.erase();
		 while ((c=yyinput()) != 0)
			if (c=='\n') break;
			else madl_annot_buf.append(1, c);
		 madl_loc.advance_line(1); madl_loc.step();
		 process_annotation();
		}

"("		return '(';
")"		return ')';
"["		return '[';
"]"		return ']';
"{"		return '{';
"}"		return '}';
"<"		return '<';
">"		return '>';
";"		return ';';
","		return ',';
":"		return ':';
"|"		return '|';
"&"		return '&';
"/"		return '/';
"%"		return '%';
"^"		return '^';
"="		return '=';
"+"		return '+';
"-"		return '-';
"~"		return '~';
"#"		return '#';
"!"		return '!';
"*"		return '*';
"?"		return '?';
"."		return '.';
"@"		return '@';


{WHITES}	madl_loc.step();
"\n"+		madl_loc.advance_line(YYLeng()); madl_loc.step();
<<EOF>>		return 0;

.			{
			 madl_lexerror (&madl_loc, *YYText());
			 madl_loc.step();
			}

%%

void madl_error (char *s)
{
	stringstream sstr;
	sstr << s << " near " << symbol_get_string(madl_filename)
		 << ":" << madl_lexer->lineno();
	fatal(sstr.str());
}

int madl_lex()
{
    return madl_lexer->yylex();                                    
}

vector<ProgSection *> *madl_parse_wrapper(ifstream *infile, const char *fname)
{
	madl_filename = symbol_find_or_insert(fname);
	madl_machname = SymNULL;
	madl_lexer = new madl_FlexLexer(infile);
	cur_annot_owner = NULL;

	madl_loc.reset();
	madl_parse();

	delete madl_lexer;
	return madl_parser_result;
}
